#ifndef __MATRIX3_HPP
#define __MATRIX3_HPP

#include "Foundation/Matrix3.h"

#include "Foundation/vec3.h"

MATRIX3_INLINE Matrix3::Matrix3()
{
  for(int i=0;i<3;i++)
  {
    for(int j=0;j<3;j++)
    {
      m[i][j]=0;
    }
  }
}

MATRIX3_INLINE Matrix3::Matrix3(const Vec3& V1, const Vec3& V2,const Vec3& V3)
{
  m[0][0]=V1.data[0];
  m[0][1]=V2.data[0];
  m[0][2]=V3.data[0];
  m[1][0]=V1.data[1];
  m[1][1]=V2.data[1];
  m[1][2]=V3.data[1];
  m[2][0]=V1.data[2];
  m[2][1]=V2.data[2];
  m[2][2]=V3.data[2];
  
}

MATRIX3_INLINE Matrix3::Matrix3(const double a[3][3])
{
  m[0][0]=a[0][0];
  m[0][1]=a[0][1];
  m[0][2]=a[0][2];
  m[1][0]=a[1][0];
  m[1][1]=a[1][1];
  m[1][2]=a[1][2];
  m[2][0]=a[2][0];
  m[2][1]=a[2][1];
  m[2][2]=a[2][2];
}

MATRIX3_INLINE Matrix3::Matrix3(const Matrix3& rhs)
{
  for(int i=0;i<3;i++)
  {
    for(int j=0;j<3;j++)
    {
      m[i][j]=rhs.m[i][j];
    }
  }
}

MATRIX3_INLINE Matrix3::~Matrix3()
{
}

/*!
  Determinant
*/
MATRIX3_INLINE double Matrix3::det()
{
  return m[0][0]*(m[1][1]*m[2][2]-m[1][2]*m[2][1])+
    m[0][1]*(m[1][2]*m[2][0]-m[1][0]*m[2][2])+
    m[0][2]*(m[1][0]*m[2][1]-m[1][1]*m[2][0]);
}


// Solve 3x3 equation system via determinants
// throws exception MatSingularError() in case of a singular matrix
// needs 59 Flops ( 36 mult , 20 add/sub , 3 div )  
MATRIX3_INLINE Vec3 Matrix3::solve(const Vec3& rhs) const
{
  double det1,detx,dety,detz;

  det1=m[0][0]*(m[1][1]*m[2][2]-m[1][2]*m[2][1])+m[0][1]*(m[1][2]*m[2][0]-m[1][0]*m[2][2])+m[0][2]*(m[1][0]*m[2][1]-m[1][1]*m[2][0]);
  if(det1 == 0){
    throw  MatSingularError();
  }
  detx=rhs.data[0]*(m[1][1]*m[2][2]-m[1][2]*m[2][1])+m[0][1]*(m[1][2]*rhs.data[2]-rhs.data[1]*m[2][2])+m[0][2]*(rhs.data[1]*m[2][1]-m[1][1]*rhs.data[2]);
  dety=m[0][0]*(rhs.data[1]*m[2][2]-m[1][2]*rhs.data[2])+rhs.data[0]*(m[1][2]*m[2][0]-m[1][0]*m[2][2])+m[0][2]*(m[1][0]*rhs.data[2]-rhs.data[1]*m[2][0]);
  detz=m[0][0]*(m[1][1]*rhs.data[2]-rhs.data[1]*m[2][1])+m[0][1]*(rhs.data[1]*m[2][0]-m[1][0]*rhs.data[2])+rhs.data[0]*(m[1][0]*m[2][1]-m[1][1]*m[2][0]);
  double x=detx/det1;
  double y=dety/det1;
  double z=detz/det1;
  return Vec3(x,y,z);
}    

// in situ inversion
// based on a`[ij]=det(A[ij])/det(M)
// where A[ij] is the adjoined submatrix to a[ij]
// 50 Flops (27 mult, 14 add/sub, 9 div)
MATRIX3_INLINE void Matrix3::invert()
{
  double det1,temp[3][3];

  det1=m[0][0]*(m[1][1]*m[2][2]-m[1][2]*m[2][1])+m[0][1]*(m[1][2]*m[2][0]-m[1][0]*m[2][2])+m[0][2]*(m[1][0]*m[2][1]-m[1][1]*m[2][0]);
  if(det1 == 0){ // det(M)==0 -> singular matrix
    throw  MatSingularError();
  } 
  temp[0][0]=(m[1][1]*m[2][2]-m[1][2]*m[2][1])/det1;
  temp[0][1]=(m[1][2]*m[2][0]-m[1][0]*m[2][2])/det1;
  temp[0][2]=(m[1][0]*m[2][1]-m[1][1]*m[2][0])/det1;
  temp[1][0]=(m[0][2]*m[2][1]-m[0][1]*m[2][2])/det1;
  temp[1][1]=(m[0][0]*m[2][2]-m[0][2]*m[2][0])/det1;
  temp[1][2]=(m[0][1]*m[2][0]-m[0][0]*m[2][1])/det1;
  temp[2][0]=(m[1][2]*m[0][1]-m[1][1]*m[0][2])/det1;
  temp[2][1]=(m[1][0]*m[0][2]-m[1][2]*m[0][0])/det1;
  temp[2][2]=(m[1][1]*m[0][0]-m[1][0]*m[0][1])/det1;
  for(int i=0;i<3;i++){
    for(int j=0;j<3;j++){
      m[i][j]=temp[i][j];
    }
  }
}

MATRIX3_INLINE Matrix3 Matrix3::inv()
{
  Matrix3 res=*this;

  res.invert();
  
  return res;
}

MATRIX3_INLINE void Matrix3::transpose()
{
  double h;

  h=m[1][0];
  m[1][0]=m[0][1];
  m[0][1]=h;
  h=m[2][0];
  m[2][0]=m[0][2];
  m[0][2]=h;
  h=m[2][1];
  m[2][1]=m[1][2];
  m[1][2]=h;
}

MATRIX3_INLINE Matrix3 Matrix3::trans() const
{
  Matrix3 res;
  
  res.m[0][0]=m[0][0];
  res.m[0][1]=m[1][0];
  res.m[0][2]=m[2][0];
  res.m[1][0]=m[0][1];
  res.m[1][1]=m[1][1];
  res.m[1][2]=m[2][1];
  res.m[2][0]=m[0][2];
  res.m[2][1]=m[1][2];
  res.m[2][2]=m[2][2];

  return res;
}

// 15 Flops ( 9 mult,6 add)
MATRIX3_INLINE Vec3 Matrix3::operator *(const Vec3& V) const
{
  double x=m[0][0]*V.data[0]+m[0][1]*V.data[1]+m[0][2]*V.data[2];
  double y=m[1][0]*V.data[0]+m[1][1]*V.data[1]+m[1][2]*V.data[2];
  double z=m[2][0]*V.data[0]+m[2][1]*V.data[1]+m[2][2]*V.data[2];
  
  return Vec3(x,y,z);
}

// 9 Flops
MATRIX3_INLINE Matrix3 Matrix3::operator *(double d) const
{
  Matrix3 res;

  res.m[0][0]=m[0][0]*d;
  res.m[0][1]=m[0][1]*d;
  res.m[0][2]=m[0][2]*d;
  res.m[1][0]=m[1][0]*d;
  res.m[1][1]=m[1][1]*d;
  res.m[1][2]=m[1][2]*d;
  res.m[2][0]=m[2][0]*d;
  res.m[2][1]=m[2][1]*d;
  res.m[2][2]=m[2][2]*d;

  return res;
}

// 45 Flops (27mult, 18add)
MATRIX3_INLINE Matrix3 Matrix3::operator *(const Matrix3& rhs) const
{
  Matrix3 res;

  for(int i=0;i<3;i++){
    for(int j=0;j<3;j++){
      res.m[i][j]=m[i][0]*rhs.m[0][j]+m[i][1]*rhs.m[1][j]+m[i][2]*rhs.m[2][j];
    }
  }
  return res;
}

// 9 Flops
MATRIX3_INLINE Matrix3& Matrix3::operator+=(const Matrix3& rhs)
{
  for(int i=0;i<3;i++){
    for(int j=0;j<3;j++){
      m[i][j]=m[i][j]+rhs.m[i][j];
    }
  }
  return *this;
}

/*!
  add two matrices
*/
MATRIX3_INLINE Matrix3 Matrix3::operator +(const Matrix3& rhs) const
{
  Matrix3 res;
  
  for(int i=0;i<3;i++){
    for(int j=0;j<3;j++){
      res.m[i][j]=m[i][j]+rhs.m[i][j];
    }
  }

  return res;
}

/*!
  subtract two matrices
*/
MATRIX3_INLINE Matrix3 Matrix3::operator-(const Matrix3& rhs) const
{
  Matrix3 res;
  
  for(int i=0;i<3;i++) {
    for(int j=0;j<3;j++) {
      res.m[i][j]=m[i][j]-rhs.m[i][j];
    }
  }

  return res;
}

/*!
  calculate the trace of a matrix
*/
MATRIX3_INLINE double Matrix3::trace() const
{
  return m[0][0]+m[1][1]+m[2][2];
}

/*!
  calculate the euclidian norm of a matrix
*/ 
MATRIX3_INLINE double Matrix3::norm() const
{
  double res=0.0;

  for(int i=0;i<3;i++){
    for(int j=0;j<3;j++){
      res+=m[i][j]*m[i][j];
    }
  }
  return res;
}

// matrix from vector
MATRIX3_INLINE Matrix3 star(const Vec3& V)
{
  Matrix3 res;

  res.m[0][1]=-V.Z();
  res.m[0][2]=V.Y();
  res.m[1][0]=V.Z();
  res.m[1][2]=-V.X();
  res.m[2][0]=-V.Y();
  res.m[2][1]=V.X();

  return res;
}

// unit matrix
MATRIX3_INLINE Matrix3 Matrix3::Unit()
{
  Matrix3 res;

  res.m[0][0]=1.0;
  res.m[1][1]=1.0;
  res.m[2][2]=1.0;

  return res;
}

/*!
  scalar * matrix 
*/ 
MATRIX3_INLINE Matrix3 operator*(double d,const Matrix3& M)
{
  Matrix3 res;

  for(int i=0;i<3;i++){
    for(int j=0;j<3;j++){
      res.m[i][j]=d*M.m[i][j];
    }
  }

  return res;
}

/*!
  calculate eigenvectors & eigenvalues

  \returns v1 eigenvector with largest eigenvalue
  \returns v2 eigenvector with middle eigenvalue
  \returns v3 eigenvector with smalles eigenvalue
  \returns e1 largest eigenvalue
  \returns e2 middle eigenvalue
  \returns e3 smalles eigenvalue
*/
MATRIX3_INLINE void Matrix3::eigen(Vec3& v1,Vec3& v2,Vec3& v3,double& e1,double& e2,double& e3)
{
  Vec3 tv1,tv2,tv3;
  double te1,te2,te3;
  // --- eigenvalues first ---
  // get determinant
  double I1=m[0][0]+m[1][1]+m[2][2];
  double I2=m[0][0]*m[1][1]-m[0][1]*m[1][0]+
    m[0][0]*m[2][2]-m[0][2]*m[2][0]+
    m[1][1]*m[2][2]-m[1][2]*m[2][1];
  double I3=det();
  if(I3==0){ // simple case, matrix singluar -> 1 eigenvalue 0
    // solve quadratic eqn.
    double p=0.5*I1;
    double q=sqrt(p*p-I2);
    te1=p+q;
    te2=p-q;
    // calc eigenvectors
    Matrix3 m=*this;
    m+=Matrix3::Unit()*-1.0*te1;
    tv1=(m.solve(Vec3(0.0,0.0,0.0))).unit();
    m=*this;
    m+=Matrix3::Unit()*-1.0*te2;
    tv2=(m.solve(Vec3(0.0,0.0,0.0))).unit();
    tv3=cross(tv1,tv2); 
  }  
  // sort the stuff
  if((te1>0) && (te1>te2)){ 
    e1=te1;
    if(te2>0){
      e2=te2;
      e3=0.0;
    } else {
      e2=0.0;
      e3=te2;
    }
  } else if ((te1<0) && (te1<te2)) {
    e3=te1;
    if(te2>0){
      e1=te2;
      e2=0.0;
    } else {
      e1=0.0;
      e2=te2;
    }
  } else {
    e2=te1;
    if(te2>0){
      e1=te2;
      e3=0.0;
    } else {
      e1=0.0;
      e3=te2;
    }
  } 
}

// output
MATRIX3_INLINE ostream& operator<<(ostream& ostr, const Matrix3& M)
{
  ostr << "[" ;
  ostr << M.m[0][0] << ", " << M.m[0][1] << ", " << M.m[0][2] <<" |";
  ostr << M.m[1][0] << ", " << M.m[1][1] << ", " << M.m[1][2] <<" |";
  ostr << M.m[2][0] << ", " << M.m[2][1] << ", " << M.m[2][2] <<"]";
  return ostr;
}

#endif // __MATRIX3_HPP
