/*=============================================================================

******************************************************************************
*                                               *
* COPYRIGHT ??????? - All Rights Reserved *
*  *
* This software is the property of ??????????????. No part of this code   *
* may be copied in any form or by any means without the expressed written   *
* consent of ???????????. Copying, use or modification of this software     *
* by any unauthorised person is illegal unless that    *
* person has a software license agreement with ?????????????.            *
*                                                          *
******************************************************************************

*********************************************************************************/

#include "Geometry/BlockGenerator.h"
#include "Geometry/GridIterator.h"
#include "Geometry/SimpleParticle.h"

#include <float.h>

namespace esys
{
  namespace lsm
  {
    BlockGenerator::BlockGenerator(
      NTable            &nTable,
      ParticlePool      &particlePool,
      const BoundingBox &bBox,
      const BoolVector  &periodicDimensions,
      double            tolerence
    ) : ParticleGenerator(nTable, particlePool),
        m_bBox(bBox),
        m_periodicDimensions(periodicDimensions),
        m_tolerence(tolerence)
    {
    }

    BlockGenerator::~BlockGenerator()
    {
    }
    
    const BoundingBox &BlockGenerator::getBBox() const
    {
      return m_bBox;
    }

    size_t BlockGenerator::getNumParticles() const
    {
      return m_idSet.size();
    }

    int BlockGenerator::getNextId()
    {
      return static_cast<int>(getNTable().getNumParticles());
    }

    double BlockGenerator::getTolerence() const
    {
      return m_tolerence;
    }

    bool BlockGenerator::is2d() const
    {
      return ((getBBox().getMaxPt().Z() - getBBox().getMinPt().Z()) <= 0);
    }

    bool BlockGenerator::particleFitsInBBox(const SimpleParticle &particle) const
    {
      return
        (
          (
            m_periodicDimensions[0]
            ||
            (
              m_bBox.contains(particle.getPos() - Vec3(particle.getRad(), 0, 0), getTolerence())
              &&
              m_bBox.contains(particle.getPos() + Vec3(particle.getRad(), 0, 0), getTolerence())
            )
          )
          &&
          (
            m_periodicDimensions[1]
            ||
            (
              m_bBox.contains(particle.getPos() - Vec3(0, particle.getRad(), 0), getTolerence())
              &&
              m_bBox.contains(particle.getPos() + Vec3(0, particle.getRad(), 0), getTolerence())
            )
          )
          &&
          (
            is2d() || m_periodicDimensions[2]
            ||
            (
              m_bBox.contains(particle.getPos() - Vec3(0, 0, particle.getRad()), getTolerence())
              &&
              m_bBox.contains(particle.getPos() + Vec3(0, 0, particle.getRad()), getTolerence())
            )
          )
        );
    }

    bool BlockGenerator::particleFitsWithNeighbours(const SimpleParticle &particle) const
    {
      const ParticleVector neighbours = 
        getNTable().getNeighbourVector(
          particle.getPos(),
          particle.getRad() + getTolerence()
        );
      ParticleVector::const_iterator iter = neighbours.begin();
      for (; iter != neighbours.end(); iter++) {
        const double dist = (particle.getPos() - (*iter)->getPos()).norm();
        if (dist < ((particle.getRad() + (*iter)->getRad()) - getTolerence())) {
          return false;
        }
      }
      return true;
    }

    bool BlockGenerator::particleFits(const SimpleParticle &particle) const
    {
      return (particleFitsInBBox(particle) && particleFitsWithNeighbours(particle));
    }

    void BlockGenerator::insertParticle(const SimpleParticle &particle)
    {
      SimpleParticle *pParticle = getParticlePool().construct(particle);
      m_particleVector.push_back(pParticle);
      m_idSet.insert(pParticle->getID());
      getNTable().insert(pParticle);
    }

    bool BlockGenerator::contains(const SimpleParticle &particle) const
    {
      return (m_idSet.find(particle.getID()) != m_idSet.end());
    }

    SimpleParticle BlockGenerator::generateParticle(const Vec3 &point)
    {
      return SimpleParticle(point, getRadius(), getNextId());
    }

    void BlockGenerator::generateSeedParticles()
    {
      GridIterator pointIt = GridIterator(getBBox(), getGridRadius());
      while (pointIt.hasNext()) {
        SimpleParticle particle = generateParticle(pointIt.next());
        if (particleFits(particle)) {
          insertParticle(particle);
        }
      }
    }

  };
};
