/*=============================================================================
??
??******************************************************************************
??*                                               *
??* COPYRIGHT ??????? - All Rights Reserved *
??*  *
??* This software is the property of ??????????????. No part of this code   *
??* may be copied in any form or by any means without the expressed written   *
??* consent of ???????????. Copying, use or modification of this software     *
??* by any unauthorised person is illegal unless that    *
??* person has a software license agreement with ?????????????.            *
??*                                                          *
??******************************************************************************
??
*********************************************************************************/

#ifndef ESYS_LSMCIRCULARNEIGHBOURTABLE_H
#define ESYS_LSMCIRCULARNEIGHBOURTABLE_H

#include "Geometry/NeighbourTable.h"
#include <boost/pool/object_pool.hpp>

#include <sstream>
#include <stdexcept>

namespace esys
{
  namespace lsm
  {
    /**
     *
     */
    template <class TmplParticle>
    class CircularNeighbourTable : public NeighbourTable<TmplParticle>
    {
    public:
      typedef TmplParticle                 Particle;
      typedef boost::object_pool<Particle> ParticlePool;
      typedef std::vector<bool>            BoolVector;

      CircularNeighbourTable(
        const BoundingBox  &bBox,
        double             gridSpacing,
        const BoolVector   &periodicDimensions,
        ParticlePool       &particlePool,
        double             circBorderWidth = 0.0
      )
        : NeighbourTable<TmplParticle>(bBox, gridSpacing),
          m_periodicDimensions(periodicDimensions),
          m_pParticlePool(&particlePool),
          m_numClonedParticles(0),
          m_circGridWidth(1)
      {
        checkPeriodicDimensions();
        if (circBorderWidth <= 0.0) {
          circBorderWidth = getGridSpacing();
        }
        m_circGridWidth = int(ceil(circBorderWidth/getGridSpacing()));
      }

      void checkPeriodicDimensions()
      {
        if (m_periodicDimensions.size() != 3) {
          std::stringstream msg;
          msg
            << "CircularNeighbourTable::CircularNeighbourTable -"
            << " size of periodic dimensions argument ("
            << m_periodicDimensions.size()
            << ") is not equal to 3";
          throw std::runtime_error(msg.str());
        }
        int numPeriodic = 0;
        for (int i = 0; i < 3; i++) {
          numPeriodic += (m_periodicDimensions[i] ? 1 : 0);
        }
        if (numPeriodic > 1) {
          std::stringstream msg;
          msg
            << "CircularNeighbourTable::CircularNeighbourTable -"
            << " only a single dimension may be periodic.";
          throw std::runtime_error(msg.str());
        }
      }

      virtual ~CircularNeighbourTable()
      {
      }

      void insertClone(Particle *pParticle, const Vec3 &newPosition)
      {
        Particle *pNewParticle = m_pParticlePool->construct(*pParticle);              
        pNewParticle->moveTo(newPosition);
        NeighbourTable<TmplParticle>::insert(pNewParticle);
        m_numClonedParticles++;
      }

      void insert(Particle *pParticle)
      {
        const Vec3L minIdx = getVecIndex(pParticle->getPos() - pParticle->getRad());
        const Vec3L maxIdx = getVecIndex(pParticle->getPos() + pParticle->getRad());
        insertInTable(pParticle, minIdx, maxIdx);
        addInserted(pParticle);
        for (int dimIdx = 0; dimIdx < 3; dimIdx++) {
          if (m_periodicDimensions[dimIdx]) {
            if (minIdx[dimIdx] < (getMinVecIndex()[dimIdx] + m_circGridWidth)) {
              Vec3 shift = Vec3::ZERO;
              shift[dimIdx] = getBBox().getSizes()[dimIdx];
              insertClone(pParticle, pParticle->getPos() + shift);
            }
            if (maxIdx[dimIdx] > (getMaxVecIndex()[dimIdx] - m_circGridWidth)) {
              Vec3 shift = Vec3::ZERO;
              shift[dimIdx] = getBBox().getSizes()[dimIdx];
              insertClone(pParticle, pParticle->getPos() - shift);
            }
          }
        }
      }

      void insert(Particle &particle)
      {
        insert(&particle);
      }

      size_t getNumClonedParticles() const
      {
        return m_numClonedParticles;
      }

      size_t getNumParticles() const
      {
        return size() - getNumClonedParticles();
      }

      private:
        BoolVector   m_periodicDimensions;
        ParticlePool *m_pParticlePool;
        int          m_numClonedParticles;
        int          m_circGridWidth;
    };
  };
};

#endif
