/*=============================================================================
??
??******************************************************************************
??*                                               *
??* COPYRIGHT ??????? - All Rights Reserved *
??*  *
??* This software is the property of ??????????????. No part of this code   *
??* may be copied in any form or by any means without the expressed written   *
??* consent of ???????????. Copying, use or modification of this software     *
??* by any unauthorised person is illegal unless that    *
??* person has a software license agreement with ?????????????.            *
??*                                                          *
??******************************************************************************
??
*********************************************************************************/

#ifndef ESYS_LSMNEIGHBOURTABLE_H
#define ESYS_LSMNEIGHBOURTABLE_H

#include <Foundation/BoundingBox.h>
#include <Geometry/Vec3L.h>
#include <vector>
#include <algorithm>

namespace esys
{
  namespace lsm
  {
    /**
     *
     */
    template <class TmplParticle>
    class NeighbourTable
    {
    public:
      typedef TmplParticle            Particle; 
      typedef std::vector<Particle *> ParticleVector;

      NeighbourTable(const BoundingBox &bBox, double gridSpacing)
        : m_dimensions(),
          m_minIndex(),
          m_maxIndex(Vec3L(-1, -1, -1)),
          m_gridSpacing(gridSpacing),
          m_bBox(bBox),
          m_insertedParticles(),
          m_pTable(NULL)
      {
        resize(bBox, gridSpacing);
      }

      virtual ~NeighbourTable()
      {
        delete [] m_pTable;
      }
      
      void clear()
      {
        for (int i = getMinVecIndex().X(); i <= getMaxVecIndex().X(); i++) {
          for (int j = getMinVecIndex().Y(); j <= getMaxVecIndex().Y(); j++) {
            for (int k = getMinVecIndex().Z(); k <= getMaxVecIndex().Z(); k++) {
              m_pTable[getScalarIndex(i, j, k)].clear();
            }
          }
        }
        m_insertedParticles.clear();
      }
      
      double getGridSpacing() const
      {
        return m_gridSpacing;
      }

      void resize(const BoundingBox &bBox, double gridSpacing)
      {
        ParticleVector particles = m_insertedParticles;
        clear();
        m_bBox = bBox;
        m_gridSpacing = gridSpacing;

        const Vec3 dims  = m_bBox.getSizes()/gridSpacing;
        m_dimensions = Vec3L(int(floor(dims[0])), int(floor(dims[1])), int(floor(dims[2])));
        m_dimensions = m_dimensions.max(Vec3L(1, 1, 1));

        delete [] m_pTable;
        m_pTable = NULL;
        m_pTable = new ParticleVector[m_dimensions.X()*m_dimensions.Y()*m_dimensions.Z()];
        m_minIndex = Vec3L(0, 0, 0);
        m_maxIndex = (m_dimensions - 1);
        for (
          typename ParticleVector::iterator it = particles.begin();
          it != particles.end();
          it++
        )
        {
          insert(*it);
        }
      }

      const Vec3L &getDimensions() const
      {
        return m_dimensions;
      }

      const BoundingBox &getBBox() const
      {
        return m_bBox;
      }

      const Vec3 &getMinPt() const
      {
        return getBBox().getMinPt();
      }

      /**
       * Return the number of particles inserted into this table.
       */
      size_t size() const
      {
        return m_insertedParticles.size();
      }

      int getScalarIndex(int xIdx, int yIdx, int zIdx) const
      {
        return 
          xIdx*m_dimensions.Z()*m_dimensions.Y()
          +
          yIdx*m_dimensions.Z()
          +
          zIdx;
      }

      int getScalarIndex(const Vec3L &index) const
      {
        return getScalarIndex(index.X(), index.Y(), index.Z()); 
      }

      int getScalarIndex(const Vec3 &pt) const
      {
        return getScalarIndex(getVecIndex(pt));
      }

      const Vec3L &getMinVecIndex() const
      {
        return m_minIndex;
      }

      const Vec3L &getMaxVecIndex() const
      {
        return m_maxIndex;
      }

      Vec3L getVecIndex(const Vec3 &pt) const
      {
        const Vec3 relPos = Vec3((pt - getMinPt())/m_gridSpacing);
        const Vec3L index = Vec3L(int(floor(relPos.X())), int(floor(relPos.Y())), int(floor(relPos.Z())));
        return getMinVecIndex().max(getMaxVecIndex().min(index));
      }

      ParticleVector getNeighbourVector(const Vec3 &pt, double radius) const
      {
        ParticleVector neighbours;
        neighbours.reserve(128);
        const Vec3L min = getVecIndex(pt - radius);
        const Vec3L max = getVecIndex(pt + radius);
        for (int i = min.X(); i <= max.X(); i++) {
          for (int j = min.Y(); j <= max.Y(); j++) {
            for (int k = min.Z(); k <= max.Z(); k++) {
              neighbours.insert(
                neighbours.end(),
                m_pTable[getScalarIndex(i, j, k)].begin(),
                m_pTable[getScalarIndex(i, j, k)].end()
              );
            }
          }
        }
        return neighbours;
      }

      ParticleVector getUniqueNeighbourVector(const Vec3 &pt, double radius) const
      {
        ParticleVector neighbours = getNeighbourVector(pt, radius);
        std::sort(neighbours.begin(), neighbours.end());
        typename ParticleVector::iterator uniqueEnd =
          std::unique(neighbours.begin(), neighbours.end());
        neighbours.erase(
          uniqueEnd,
          neighbours.end()
        );

        return neighbours;
      }

      ParticleVector getNeighbourVector(const Vec3 &pt) const
      {
        return m_pTable[getScalarIndex(pt)];
      }

      void insert(Particle *pParticle)
      {
        const Vec3L minIdx = getVecIndex(pParticle->getPos() - pParticle->getRad());
        const Vec3L maxIdx = getVecIndex(pParticle->getPos() + pParticle->getRad());
        insertInTable(pParticle, minIdx, maxIdx);
        addInserted(pParticle);
      }

      void insert(Particle &particle)
      {
        insert(&particle);
      }

      class ParticleIterator
      {
      public:
        ParticleIterator(ParticleVector &particleVector)
         : m_it(particleVector.begin()),
           m_end(particleVector.end())
        {
        }

        bool hasNext()
        {
          return (m_it != m_end);
        }

        typename ParticleVector::reference next()
        {
          typename ParticleVector::reference p = *m_it;
          m_it++;
          return p;
        }

        private:
          typename ParticleVector::iterator m_it;
          typename ParticleVector::iterator m_end;
      };
      
      ParticleIterator getParticleIterator()
      {
        return ParticleIterator(m_insertedParticles);
      }

    protected:

      void insertInTable(Particle *pParticle, const Vec3L &minIdx, const Vec3L &maxIdx)
      {
        for (int i = minIdx.X(); i <= maxIdx.X(); i++) {
          for (int j = minIdx.Y(); j <= maxIdx.Y(); j++) {
            for (int k = minIdx.Z(); k <= maxIdx.Z(); k++) {
              m_pTable[getScalarIndex(i, j, k)].push_back(pParticle);
            }
          }
        }
      }

      void addInserted(Particle *pParticle)
      {
        m_insertedParticles.push_back(pParticle);      
      }

    private:
      Vec3L          m_dimensions;
      Vec3L          m_minIndex;
      Vec3L          m_maxIndex;
      double         m_gridSpacing;
      BoundingBox    m_bBox;
      ParticleVector m_insertedParticles;
      ParticleVector *m_pTable;
    };
  };
};

#endif
