/*=============================================================================

******************************************************************************
*                                               *
* COPYRIGHT ??????? - All Rights Reserved *
*  *
* This software is the property of ??????????????. No part of this code   *
* may be copied in any form or by any means without the expressed written   *
* consent of ???????????. Copying, use or modification of this software     *
* by any unauthorised person is illegal unless that    *
* person has a software license agreement with ?????????????.            *
*                                                          *
******************************************************************************

*********************************************************************************/

#include "Geometry/RandomBlockGenerator.h"
#include "Geometry/SimpleParticle.h"
#include "Geometry/GridIterator.h"
#include "Geometry/SimpleNTable.h"
#include "Geometry/Sphere3d.h"
#include "Geometry/Plane.h"
#include "Geometry/ParticleFitter.h"

#include <algorithm>
#include <stdexcept>
#include <float.h>

namespace esys
{
  namespace lsm
  {
    RandomBlockGenerator::RandomBlockGenerator(
      NTable            &nTable,
      ParticlePool      &particlePool,
      const BoundingBox &bBox,
      const BoolVector  &periodicDimensions,
      double            tolerence,
      double            minSphereRadius,
      double            maxSphereRadius,
      const PlaneVector &fitPlaneVector,
      int               maxInsertionFailures
    )
      : BlockGenerator(nTable, particlePool, bBox, periodicDimensions, tolerence),
        m_minRadius(minSphereRadius),
        m_maxRadius(maxSphereRadius),
        m_fitPlaneVector(fitPlaneVector),
        m_maxInsertionFailures(maxInsertionFailures)
    {
    }

    RandomBlockGenerator::~RandomBlockGenerator()
    {
    }

    double RandomBlockGenerator::getRandom(double minVal, double maxVal) const
    {
      return 
        minVal
        +
        ((maxVal-minVal)*(static_cast<double>(rand()))/(static_cast<double>(RAND_MAX)));
    }

    double RandomBlockGenerator::getGridRadius() const
    {
      return m_maxRadius;
    }

    double RandomBlockGenerator::getRadius() const
    {
      return getRandom(m_minRadius, m_maxRadius);
    }

    class PlaneComparer
    {
    public:
      PlaneComparer(const SimpleParticle &particle)
        : m_pParticle(&particle)
      {
      }
      
      inline bool operator()(const Plane &plane1, const Plane &plane2) const
      {
        return 
          (
            plane1.sep(m_pParticle->getPos())
            <
            plane2.sep(m_pParticle->getPos())
          );
      }
    private:
      const SimpleParticle *m_pParticle;
    };

    const PlaneVector &RandomBlockGenerator::getFitPlaneVector() const
    {
      return m_fitPlaneVector;
    }

    Plane RandomBlockGenerator::getClosestFitPlane(const SimpleParticle &particle) const
    {
      PlaneVector fitPlanes = getFitPlaneVector();
      if (fitPlanes.size() > 0) {
        std::sort(fitPlanes.begin(), fitPlanes.end(), PlaneComparer(particle));

        return fitPlanes[0];
      }
      return Plane(Vec3(-1.0, 0, 0), Vec3(DBL_MAX/2, DBL_MAX/2, DBL_MAX/2));
    }

    Vec3 RandomBlockGenerator::getRandomPoint() const
    {
      return 
        Vec3(
          getRandom(getBBox().getMinPt().X(), getBBox().getMaxPt().X()),
          getRandom(getBBox().getMinPt().Y(), getBBox().getMaxPt().Y()),
          getRandom(getBBox().getMinPt().Z(), getBBox().getMaxPt().Z())
        );
    }

    class ParticleComparer
    {
    public:
      ParticleComparer(const SimpleParticle &particle) : m_pParticle(&particle)
      {
      }

      inline bool operator()(const SimpleParticle &p1, const SimpleParticle &p2) const
      {
        return 
          (
            ((p1.getPos() - m_pParticle->getPos()).norm() - p1.getRad())
            <
            ((p2.getPos() - m_pParticle->getPos()).norm() - p2.getRad())
          );
      }

      inline bool operator()(const SimpleParticle *p1, const SimpleParticle *p2) const
      {
        return (*this)(*p1, *p2);
      }

    private:
      const SimpleParticle *m_pParticle;
    };

    RandomBlockGenerator::ParticleVector RandomBlockGenerator::getClosestNeighbors(
      const SimpleParticle& particle,
      int numClosest
    )
    {
      ParticleVector neighbourVector =
        getNTable().getUniqueNeighbourVector(particle.getPos(), m_maxRadius + getTolerence());
      std::sort(neighbourVector.begin(), neighbourVector.end(), ParticleComparer(particle));

      if (neighbourVector.size() > numClosest) {
        neighbourVector.erase(neighbourVector.begin() + numClosest, neighbourVector.end());
      }

      return neighbourVector;
    }

    bool RandomBlockGenerator::particleFits(const SimpleParticle &particle) const
    {
      return 
        (
          (particle.getRad() >= m_minRadius)
          &&
          (particle.getRad() <= m_maxRadius)
          &&
          BlockGenerator::particleFits(particle)
        );
    }

    int RandomBlockGenerator::getMaxInsertionFailures() const
    {
      return m_maxInsertionFailures;
    }

    FitterPtrVector RandomBlockGenerator::getFitterPtrVector()
    {
      FitterPtrVector fitters;
      fitters.push_back(FitterPtr(new MoveToSurfaceFitter(*this)));
      if (is2d())
      {
        fitters.push_back(FitterPtr(new TwoDParticleFitter(*this)));
        if (getFitPlaneVector().size() > 0) {
          fitters.push_back(FitterPtr(new TwoDPlaneParticleFitter(*this)));
        }
      }
      else
      {
        fitters.push_back(FitterPtr(new ThreeDParticleFitter(*this)));
        if (getFitPlaneVector().size() > 0) {
          fitters.push_back(FitterPtr(new ThreeDPlaneParticleFitter(*this)));
        }
      }

      return fitters;
    }

    void RandomBlockGenerator::generateFillParticles()
    {
      int numFails    = 0;
      int insertCount = 0;
      FitterPtrVector fitters = getFitterPtrVector();
      SimpleParticle particle = ParticleFitter::INVALID;
      SimpleParticle fitParticle = ParticleFitter::INVALID;
      Plane plane;
      ParticleVector neighbours;
      while (numFails < getMaxInsertionFailures()) {
        fitParticle = ParticleFitter::INVALID;
        particle = generateParticle(getRandomPoint());
        neighbours = getClosestNeighbors(particle, 4);
        plane = getClosestFitPlane(particle);
        
        for (
          FitterPtrVector::iterator it = fitters.begin();
          (
            (it != fitters.end())
            &&
            (!fitParticle.isValid())
          );
          it++
        )
        {
          fitParticle = (*it)->getFitParticle(particle, neighbours, plane);
        }
        if (fitParticle.isValid()) {
          numFails = 0;
          insertCount++;
          insertParticle(fitParticle);
        } else {
          numFails++;
        }
      }
      std::cout
        << "BoundingBox: minPt = " << getBBox().getMinPt()
        << ", maxPt = " << getBBox().getMaxPt() << std::endl
        << "BoundingBox: sizes = " << getBBox().getSizes() << std::endl;
      for (
          FitterPtrVector::iterator it = fitters.begin();
          (it != fitters.end());
          it++      
      ) {
        std::cout << *(*it) << std::endl;
      }
      std::cout << "Total inserted = " << insertCount << std::endl;
    }

    void RandomBlockGenerator::generate()
    {
      generateSeedParticles();
      generateFillParticles();
    }
  };
};
