#include "EdgeTest.h"

// -- Project includes --
#include "Geometry/Edge.h"
#include "Geometry/Triangle.h"
#include "Foundation/vec3.h"

namespace LsmTest
{
  CPPUNIT_TEST_SUITE_REGISTRATION( EdgeTest );

  /*! 
    constructor - do nothing
  */
  EdgeTest::EdgeTest() : CppUnit::TestCase()
  {}

  
  /*!
    destructor - do nothing
  */
  EdgeTest::~EdgeTest(){}

  /*!
   */
  void EdgeTest::setUp()
  {
    CppUnit::TestCase::setUp();
  }
  
  /*!
   */
  void EdgeTest::tearDown()
  {
    CppUnit::TestCase::tearDown();
  }

  /*!
    test Edge::dist()
  */
  void EdgeTest::testDist()
  {
    Vec3 p1=Vec3(1.0,2.0,3.0);
    Vec3 p2=Vec3(3.0,2.0,1.0);
    Vec3 pt1=Vec3(3.0,2.0,3.0);
    Vec3 pt2=Vec3(2.0,3.0,2.0);
    Vec3 pt3=Vec3(1.0,2.0,4.0);

    Edge E(0,1,p1,p2);
    pair<bool,double> res;

    // test 1
    res=E.dist(pt1);
    CPPUNIT_ASSERT_EQUAL(true,res.first);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(sqrt(2.0),res.second,1e-10);
    
    // test 2
    res=E.dist(pt2);
    CPPUNIT_ASSERT_EQUAL(true,res.first);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0,res.second,1e-10);
    
    // test 3 
    res=E.dist(pt3);
    CPPUNIT_ASSERT_EQUAL(false,res.first);
  }

  /*!
    test Edge::sep()
  */
  void EdgeTest::testSep()
  {
    Vec3 p1=Vec3(1.0,2.0,3.0);
    Vec3 p2=Vec3(3.0,2.0,1.0);
    Vec3 pt1=Vec3(3.0,2.0,3.0);
    Vec3 pt2=Vec3(2.0,3.0,2.0);
    Vec3 pt3=Vec3(1.0,2.0,4.0);

    Edge E(0,1,p1,p2);
 
    // test 1
    CPPUNIT_ASSERT_DOUBLES_EQUAL(sqrt(2.0),E.sep(pt1),1e-10);
    
    // test 2
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0,E.sep(pt2),1e-10);
    
    // test 3 
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0,E.sep(pt3),1e-10);
  }

  /*!
    test Edge::getBoundingBoxMin
  */
  void EdgeTest::testBBXMin()
  {
    Vec3 p1=Vec3(1.0,2.0,3.0);
    Vec3 p2=Vec3(3.0,2.0,1.0);

    Edge E(0,1,p1,p2);

    Vec3 bbmin=E.getBoundingBoxMin();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0,bbmin.X(),1e-10);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0,bbmin.Y(),1e-10);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0,bbmin.Z(),1e-10);
  }

  /*!
    test Edge::getBoundingBoxMax
  */
  void EdgeTest::testBBXMax()
  {
    Vec3 p1=Vec3(1.0,2.0,3.0);
    Vec3 p2=Vec3(3.0,2.0,1.0);

    Edge E(0,1,p1,p2);

    Vec3 bbmax=E.getBoundingBoxMax();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0,bbmax.X(),1e-10);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0,bbmax.Y(),1e-10);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0,bbmax.Z(),1e-10);
  }

  /*!
    test Edge::getDirectionFromPoint
  */
  void EdgeTest::testDir()
  {
    Vec3 p1=Vec3(1.0,2.0,3.0);
    Vec3 p2=Vec3(3.0,2.0,1.0);
    Vec3 pt1=Vec3(3.0,2.0,3.0);
    Vec3 pt2=Vec3(2.0,3.0,2.0);

    Edge E(0,1,p1,p2);
    Vec3 dir;

    // test 1
    dir=E.getDirectionFromPoint(pt1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(sqrt(2.0)/2.0,dir.X(),1e-10);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0,dir.Y(),1e-10);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(sqrt(2.0)/2.0,dir.Z(),1e-10);
    // test 2
    dir=E.getDirectionFromPoint(pt2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0,dir.X(),1e-10);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0,dir.Y(),1e-10);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0,dir.Z(),1e-10);
  }

  /*!
    test Edge::isValidContact
  */
  void EdgeTest::testValid()
  {
    // corners for 2 triangles
    Vec3 p1=Vec3(0.0,0.0,0.0);
    Vec3 p2=Vec3(0.0,1.0,0.0);
    Vec3 p3=Vec3(1.0,0.0,0.0);
    Vec3 p4=Vec3(0.0,0.0,-1.0);
    // the triangles
    Triangle Tr1=Triangle(1,2,3,p1,p2,p3,0,0);
    Triangle Tr2=Triangle(1,2,4,p1,p2,p4,0,0);
    // edge 
    Edge E1=Edge(0,1,p1,p2); // no tri
    Edge E2=Edge(0,1,p1,p2,&Tr1); // one tri
    Edge E3=Edge(0,1,p1,p2,&Tr1,&Tr2); // two tris
    // test points
    Vec3 test_p1=Vec3(0.25,0.25,1.0);
    Vec3 test_p2=Vec3(-1.0,0.25,-0.25);
    Vec3 test_p3=Vec3(-1.0,0.25,-1.0);

    // tests
    CPPUNIT_ASSERT(E1.isValidContact(test_p1));
    CPPUNIT_ASSERT(E1.isValidContact(test_p2));
    CPPUNIT_ASSERT(E1.isValidContact(test_p3));
    
    CPPUNIT_ASSERT(!E2.isValidContact(test_p1));
    CPPUNIT_ASSERT(E2.isValidContact(test_p2));
    CPPUNIT_ASSERT(E2.isValidContact(test_p3));
    
    CPPUNIT_ASSERT(!E3.isValidContact(test_p1));
    CPPUNIT_ASSERT(!E3.isValidContact(test_p2));
    CPPUNIT_ASSERT(E3.isValidContact(test_p3));   
   }

};
