/*=============================================================================
??
??******************************************************************************
??*                                               *
??* COPYRIGHT ??????? - All Rights Reserved *
??*  *
??* This software is the property of ??????????????. No part of this code   *
??* may be copied in any form or by any means without the expressed written   *
??* consent of ???????????. Copying, use or modification of this software     *
??* by any unauthorised person is illegal unless that    *
??* person has a software license agreement with ?????????????.            *
??*                                                          *
??******************************************************************************
??
*********************************************************************************/

#include "GeometryTest/GougeBlock3DTest.h"
#include "Geometry/GougeBlock3D.h"
#include "UnitTest/Macros.h"

using namespace esys::lsm;

namespace LsmTest
{
  class GougeBlock3DTest::Impl
  {
  public:
    Impl(const GougeBlockPrms &prms, const GougeBlockPrms &twoDPrms)
      : m_prms(prms),
        m_twoDPrms(twoDPrms)
    {
    }

    GougeBlockPrms m_prms;
    GougeBlockPrms m_twoDPrms;
  };

  CPPUNIT_TEST_SUITE_REGISTRATION( GougeBlock3DTest );

  GougeBlock3DTest::GougeBlock3DTest()
      : CppUnit::TestCase()
  {
  }

  GougeBlock3DTest::~GougeBlock3DTest()
  {
  }

  void GougeBlock3DTest::setUp()
  {
    CppUnit::TestCase::setUp();
    srand(12345);
    m_implPtr = 
      ImplPtr(
        new Impl(
          GougeBlockPrms(
              BoundingBox(Vec3(0, 0, 0), Vec3(10, 80, 15)),
              1.0,
              XZ,
              ParticleBlockPrms(20.0, 0.45, 0.90),
              ParticleBlockPrms(10.0, 0.25, 0.50),
              BoolVector(3, false),
              32,
              0.001
          ),
          GougeBlockPrms(
              BoundingBox(Vec3(0, 0, 0), Vec3(10, 80, 0)),
              1.0,
              XZ,
              ParticleBlockPrms(0.0, 0.45, 0.90),
              ParticleBlockPrms(0.0, 0.25, 0.50),
              BoolVector(3, false),
              32,
              0.001
          )
        )
      );
  }

  void GougeBlock3DTest::tearDown()
  {
    CppUnit::TestCase::tearDown();
  }

  void GougeBlock3DTest::test2dPrmsBlockSize()
  {
    BoolVector circDimensions = BoolVector(3, false);
    circDimensions[0] = true;
    GougeBlockPrms prms = 
      GougeBlockPrms(
          BoundingBox(Vec3(0, 0, 0), Vec3(10, 80, 0)),
          1.0,
          XZ,
          ParticleBlockPrms(10.0, 0.45, 0.90),
          ParticleBlockPrms(20.0, 0.25, 0.50),
          circDimensions,
          32,
          0.001
      );
    LSMUNIT_ASSERT_DOUBLES_EQUAL(prms.getBBox().getMinPt().X(),  0.0);
    LSMUNIT_ASSERT_DOUBLES_EQUAL(prms.getBBox().getMaxPt().X(), 11.0);
    LSMUNIT_ASSERT_DOUBLES_EQUAL(
      prms.getBBox().getMinPt().X(),
      prms.getFaultPackingInfoVector()[0].getBBox().getMinPt().X()
    );
    prms.getFaultPackingInfoVector();
    LSMUNIT_ASSERT_DOUBLES_EQUAL(
      prms.getBBox().getMaxPt().X(),
      prms.getFaultPackingInfoVector()[0].getBBox().getMaxPt().X()
    );
  }

  void GougeBlock3DTest::test3dPrmsBlockSize()
  {
    BoolVector circDimensions = BoolVector(3, false);
    circDimensions[0] = true;
    GougeBlockPrms prms = 
      GougeBlockPrms(
          BoundingBox(Vec3(0, 0, 0), Vec3(36, 66, 4.30940107676)),
          1.0,
          XZ,
          ParticleBlockPrms( 3.0, 0.3, 1.00),
          ParticleBlockPrms(10.0, 0.3, 1.00),
          circDimensions,
          32,
          0.001
      );
    LSMUNIT_ASSERT_DOUBLES_EQUAL(prms.getBBox().getMinPt().X(),  0.0);
    LSMUNIT_ASSERT_DOUBLES_EQUAL(prms.getBBox().getMaxPt().X(), 36.0);
    LSMUNIT_ASSERT_DOUBLES_EQUAL(
      prms.getBBox().getMinPt().X(),
      prms.getFaultPackingInfoVector()[0].getBBox().getMinPt().X()
    );
    prms.getFaultPackingInfoVector();
    LSMUNIT_ASSERT_DOUBLES_EQUAL(
      prms.getBBox().getMaxPt().X(),
      prms.getFaultPackingInfoVector()[0].getBBox().getMaxPt().X()
    );
    GougeBlock3D blk(prms);
  }

  void GougeBlock3DTest::testGougeBlockPrms()
  {
    GougeBlockPrms prms = m_implPtr->m_prms;

    CPPUNIT_ASSERT_EQUAL(static_cast<size_t>(2), prms.getFaultPackingInfoVector().size());
    CPPUNIT_ASSERT_EQUAL(
      prms.getBBox().getMinPt().X(),
      prms.getFaultPackingInfoVector()[0].getBBox().getMinPt().X()
    );
    CPPUNIT_ASSERT_EQUAL(
      prms.getBBox().getMinPt().Z(),
      prms.getFaultPackingInfoVector()[0].getBBox().getMinPt().Z()
    );

    CPPUNIT_ASSERT_EQUAL(
      prms.getBBox().getMaxPt().X(),
      prms.getFaultPackingInfoVector()[0].getBBox().getMaxPt().X()
    );
    CPPUNIT_ASSERT_EQUAL(
      prms.getBBox().getMaxPt().Z(),
      prms.getFaultPackingInfoVector()[0].getBBox().getMaxPt().Z()
    );

    CPPUNIT_ASSERT_EQUAL(static_cast<size_t>(2), prms.getRegularBBoxVector().size());
    CPPUNIT_ASSERT_EQUAL(static_cast<size_t>(1), prms.getGougePackingInfoVector().size());
  }

  class SizeCheckerVisitor
  {
  public:
    SizeCheckerVisitor(double minRadius, double maxRadius)
      : m_minRadius(minRadius),
        m_maxRadius(maxRadius)
    {
    }

    void visitSimpleParticle(const SimpleParticle &particle)
    {
      if (
        (particle.getRad() > m_maxRadius)
        ||
        (particle.getRad() < m_minRadius)
      )
      {
        std::stringstream msg;
        msg 
          << "Particle radius " << particle.getRad()
          << " does not lie in range ["
          << m_minRadius << ", " << m_maxRadius
          << "]";
        throw std::runtime_error(msg.str());
      }
    }
  private:
    double m_minRadius;
    double m_maxRadius;
  };

  void GougeBlock3DTest::testGougeBlock()
  {
    GougeBlock3D gougeBlock(m_implPtr->m_prms);

    gougeBlock.generate();

    SizeCheckerVisitor 
      sizeChecker(
        m_implPtr->m_prms.getMinRadius(),
        m_implPtr->m_prms.getMaxRadius()
      );
    gougeBlock.visitParticles(sizeChecker);

    CPPUNIT_ASSERT(gougeBlock.getNumParticles() > 4000);
  }

  class ZDimCheckerVisitor
  {
  public:
    void visitSimpleParticle(const SimpleParticle &particle) const
    {
      if (particle.getPos().Z() != 0.0) {
        std::stringstream msg;
        msg
          << "Particle with id " << particle.getID() << " does not"
          << " have zero z-dimension, pos = (" << particle.getPos() << ")";
        throw std::runtime_error(msg.str());
      }
    }
  };

  void GougeBlock3DTest::test2d()
  {
    GougeBlock3D gougeBlock(m_implPtr->m_twoDPrms);

    gougeBlock.generate();

    gougeBlock.visitParticles(ZDimCheckerVisitor());
    CPPUNIT_ASSERT_EQUAL(225, gougeBlock.getNumParticles());
  }
};
