/*=============================================================================
??
??******************************************************************************
??*                                               *
??* COPYRIGHT ??????? - All Rights Reserved *
??*  *
??* This software is the property of ??????????????. No part of this code   *
??* may be copied in any form or by any means without the expressed written   *
??* consent of ???????????. Copying, use or modification of this software     *
??* by any unauthorised person is illegal unless that    *
??* person has a software license agreement with ?????????????.            *
??*                                                          *
??******************************************************************************
??
*********************************************************************************/

#include "GeometryTest/GridIteratorTest.h"
#include "UnitTest/Macros.h"
#include "Geometry/GridIterator.h"
#include "Foundation/BoundingBox.h"
#include "Foundation/vec3.h"

#include <float.h>

using namespace esys::lsm;

namespace LsmTest
{
  CPPUNIT_TEST_SUITE_REGISTRATION( GridIteratorTest );

  GridIteratorTest::GridIteratorTest()
      : CppUnit::TestCase()
  {
  }

  GridIteratorTest::~GridIteratorTest()
  {
  }

  void GridIteratorTest::setUp()
  {
    CppUnit::TestCase::setUp();
  }

  void GridIteratorTest::tearDown()
  {
    CppUnit::TestCase::tearDown();
  }

  int GridIteratorTest::count(GridIterator &it)
  {
    int i = 0;
    for (; it.hasNext(); i++)
    {
      Vec3 pt = it.next();
    }
    return i;
  }

  BoundingBox boundingBox(GridIterator &it)
  {
    BoundingBox bBox(Vec3(DBL_MAX, DBL_MAX, DBL_MAX), Vec3(-DBL_MAX, -DBL_MAX, -DBL_MAX));
    for (; it.hasNext();)
    {
      Vec3 pt = it.next();
      bBox = 
        BoundingBox(comp_min(bBox.getMinPt(), pt), comp_max(bBox.getMaxPt(), pt));
    }
    return bBox;
  }

  void GridIteratorTest::testIterator()
  {
    GridIterator it1 = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(2, 2, 2)), 1.0);
    CPPUNIT_ASSERT_EQUAL(1, count(it1));

    GridIterator it2 = GridIterator(BoundingBox(Vec3(0, 1, 3), Vec3(10, 7, 9)), 1.0);
    CPPUNIT_ASSERT_EQUAL(48, count(it2));
  }

  void GridIteratorTest::testZeroDim()
  {
    GridIterator itXy = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(10, 2, 0)), 1.0);
    CPPUNIT_ASSERT_EQUAL(5, count(itXy));
    itXy = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(10, 2, 0)), 1.0);
    CPPUNIT_ASSERT_EQUAL(boundingBox(itXy), itXy.getBoundingBox());

    itXy = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(10, 4, 0)), 1.0);
    CPPUNIT_ASSERT_EQUAL(10, count(itXy));
    itXy = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(10, 4, 0)), 1.0);
    CPPUNIT_ASSERT_EQUAL(boundingBox(itXy), itXy.getBoundingBox());        
    
    GridIterator itXz = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(10, 0, 7)), 1.0);
    CPPUNIT_ASSERT_EQUAL(20, count(itXz));

    GridIterator itYz = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(0, 10, 7)), 1.0);
    CPPUNIT_ASSERT_EQUAL(0, count(itYz));
  }

  void GridIteratorTest::testSphereBoundingBox()
  {
    GridIterator it = GridIterator(BoundingBox(Vec3(1, 2, 3), Vec3(32, 15, 7)), 1.0);
    BoundingBox sphereBBox = it.getSphereBBox();
    CPPUNIT_ASSERT_EQUAL(sphereBBox, GridIterator(sphereBBox, 1.0).getSphereBBox());
    
    it = GridIterator(BoundingBox(Vec3(1, 2, 0), Vec3(32, 15, 0)), 1.0);
    sphereBBox = it.getSphereBBox();
    CPPUNIT_ASSERT_EQUAL(sphereBBox, GridIterator(sphereBBox, 1.0).getSphereBBox());
    
    it = GridIterator(BoundingBox(Vec3(1, 0, 0), Vec3(33, 2.0, 10)), 1.0);
    sphereBBox = it.getSphereBBox();
    CPPUNIT_ASSERT_EQUAL(sphereBBox, GridIterator(sphereBBox, 1.0).getSphereBBox());
    
    it = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(36, 66.0, 3.6)), 1.0);
    sphereBBox = it.getSphereBBox();
    CPPUNIT_ASSERT_EQUAL(sphereBBox, GridIterator(sphereBBox, 1.0).getSphereBBox());

    it = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(36, 65.6867, 4.3094)), 1.0);
    sphereBBox = it.getSphereBBox();
    CPPUNIT_ASSERT_EQUAL(sphereBBox, GridIterator(sphereBBox, 1.0).getSphereBBox());
  }

  void GridIteratorTest::test2dSphereBoundingBox()
  {
    GridIterator it = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(18, 20, 0)), 1.0);
    LSMUNIT_ASSERT_DOUBLES_EQUAL( 0, it.getSphereBBox().getMinPt().X());
    LSMUNIT_ASSERT_DOUBLES_EQUAL(19, it.getSphereBBox().getMaxPt().X());
    
    BoundingBox sphereBBox = it.getSphereBBox();
    CPPUNIT_ASSERT_EQUAL(sphereBBox, GridIterator(sphereBBox, 1.0).getSphereBBox());    
  }

  void GridIteratorTest::testCentrePointBoundingBox()
  {
    GridIterator itX = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(10, 2.0, 2.0)), 1.0);
    CPPUNIT_ASSERT_EQUAL(5, count(itX));
    itX = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(10, 2.0, 2.0)), 1.0);
    CPPUNIT_ASSERT_EQUAL(boundingBox(itX), itX.getBoundingBox());

    GridIterator itXz = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(10, 2.0, 4.0)), 1.0);
    CPPUNIT_ASSERT_EQUAL(10, count(itXz));
    itXz = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(10, 2.0, 4.0)), 1.0);
    CPPUNIT_ASSERT_EQUAL(boundingBox(itXz), itXz.getBoundingBox());

    GridIterator itXz2 = GridIterator(BoundingBox(Vec3(1, 1, 1), Vec3(11, 3.0, 7.0)), 1.0);
    CPPUNIT_ASSERT_EQUAL(15, count(itXz2));
    itXz2 = GridIterator(BoundingBox(Vec3(1, 1, 1), Vec3(11, 3.0, 7.0)), 1.0);
    CPPUNIT_ASSERT_EQUAL(boundingBox(itXz2), itXz2.getBoundingBox());

    GridIterator itXyz = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(10, 4.0, 4.0)), 1.0);
    CPPUNIT_ASSERT_EQUAL(16, count(itXyz));
    itXyz = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(10, 4.0, 4.0)), 1.0);
    CPPUNIT_ASSERT_EQUAL(boundingBox(itXyz), itXyz.getBoundingBox());

    GridIterator itXyz2 = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(10, 6.0, 4.0)), 1.0);
    CPPUNIT_ASSERT_EQUAL(32, count(itXyz2));
    itXyz2 = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(10, 6.0, 4.0)), 1.0);
    CPPUNIT_ASSERT_EQUAL(boundingBox(itXyz2), itXyz2.getBoundingBox());

    GridIterator itXyz3 = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(10, 8.0, 8.0)), 1.0);
    CPPUNIT_ASSERT_EQUAL(100, count(itXyz3));
    itXyz3 = GridIterator(BoundingBox(Vec3(0, 0, 0), Vec3(10, 8.0, 8.0)), 1.0);
    BoundingBox bBox = boundingBox(itXyz3);
    CPPUNIT_ASSERT((bBox.getMinPt() - itXyz3.getBoundingBox().getMinPt()).norm() < 0.00001);
    CPPUNIT_ASSERT((bBox.getMaxPt() - itXyz3.getBoundingBox().getMaxPt()).norm() < 0.00001);
  }
};
