/*=============================================================================
??
??******************************************************************************
??*                                               *
??* COPYRIGHT ??????? - All Rights Reserved *
??*  *
??* This software is the property of ??????????????. No part of this code   *
??* may be copied in any form or by any means without the expressed written   *
??* consent of ???????????. Copying, use or modification of this software     *
??* by any unauthorised person is illegal unless that    *
??* person has a software license agreement with ?????????????.            *
??*                                                          *
??******************************************************************************
??
*********************************************************************************/

#include "GeometryTest/NeighbourTableTest.h"
#include "GeometryTest/TestParticle.h"

#include "Geometry/NeighbourTable.h"

using namespace esys::lsm;

namespace LsmTest
{
  CPPUNIT_TEST_SUITE_REGISTRATION( NeighbourTableTest );

  typedef esys::lsm::NeighbourTable<TestParticle> TestTable;

  NeighbourTableTest::NeighbourTableTest()
      : CppUnit::TestCase()
  {
  }

  NeighbourTableTest::~NeighbourTableTest()
  {
  }

  void NeighbourTableTest::setUp()
  {
    CppUnit::TestCase::setUp();
  }

  void NeighbourTableTest::tearDown()
  {
    CppUnit::TestCase::tearDown();
  }

  void NeighbourTableTest::testConstructor()
  {
    TestTable table1 = TestTable(BoundingBox(Vec3(0, 0, 0), Vec3(10, 20, 30)), 1);
    
    CPPUNIT_ASSERT_EQUAL(Vec3L(10, 20, 30), table1.getDimensions());
    
    TestTable table2 = TestTable(BoundingBox(Vec3(1, 1, 1), Vec3(10, 19, 22)), 3);
    CPPUNIT_ASSERT_EQUAL(Vec3L(3, 6, 7), table2.getDimensions());
    
    TestTable table3 = TestTable(BoundingBox(Vec3(1, 1, 1), Vec3(10, 19, 22)), 3.012321);
    CPPUNIT_ASSERT_EQUAL(Vec3L(2, 5, 6), table3.getDimensions());
  }

  void NeighbourTableTest::testInsert()
  {
    TestTable table = TestTable(BoundingBox(Vec3(0, 0, 0), Vec3(10, 20, 30)), 2);

    // Insert a particle in "first" cell
    TestParticle p1 = TestParticle(Vec3(1, 1, 1), 0.5);
    CPPUNIT_ASSERT_EQUAL((size_t)0, table.size());
    table.insert(p1);
    CPPUNIT_ASSERT_EQUAL((size_t)1, table.size());

    CPPUNIT_ASSERT_EQUAL((size_t)1, table.getNeighbourVector(Vec3(1, 1, 1)).size());
    CPPUNIT_ASSERT_EQUAL(&p1, table.getNeighbourVector(Vec3(1, 1, 1))[0]);

    // Insert a particle in "last" cell
    TestParticle p2 = TestParticle(Vec3(9, 19, 29), 0.5);
    table.insert(p2);
    CPPUNIT_ASSERT_EQUAL((size_t)2, table.size());
    CPPUNIT_ASSERT_EQUAL((size_t)1, table.getNeighbourVector(Vec3(9.5, 18.5, 29.25)).size());
    CPPUNIT_ASSERT_EQUAL(&p2, table.getNeighbourVector(Vec3(9.5, 18.5, 29.25))[0]);

    // Insert a particle which overlaps two table cells.
    TestParticle p3 = TestParticle(Vec3(1.75, 1, 1), 0.75);
    table.insert(p3);
    CPPUNIT_ASSERT_EQUAL((size_t)3, table.size());
    TestTable::ParticleVector neighbours = table.getNeighbourVector(Vec3(1, 1, 1));
    CPPUNIT_ASSERT_EQUAL((size_t)2, neighbours.size());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p1) != neighbours.end());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p3) != neighbours.end());

    neighbours = table.getNeighbourVector(Vec3(3, 1, 1));
    CPPUNIT_ASSERT_EQUAL((size_t)1, neighbours.size());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p3) != neighbours.end());

    // Insert a particle which overlaps eight table cells.
    TestParticle p4 = TestParticle(Vec3(2, 2, 2), 1.0);
    table.insert(p4);
    CPPUNIT_ASSERT_EQUAL((size_t)4, table.size());
    neighbours = table.getNeighbourVector(Vec3(1, 1, 1));
    CPPUNIT_ASSERT_EQUAL((size_t)3, neighbours.size());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p1) != neighbours.end());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p3) != neighbours.end());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p4) != neighbours.end());

    neighbours = table.getNeighbourVector(Vec3(3, 1, 1));
    CPPUNIT_ASSERT_EQUAL((size_t)2, neighbours.size());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p3) != neighbours.end());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p4) != neighbours.end());

    neighbours = table.getNeighbourVector(Vec3(1, 3, 1));
    CPPUNIT_ASSERT_EQUAL((size_t)1, neighbours.size());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p4) != neighbours.end());

    neighbours = table.getNeighbourVector(Vec3(3, 3, 1));
    CPPUNIT_ASSERT_EQUAL((size_t)1, neighbours.size());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p4) != neighbours.end());    

    neighbours = table.getNeighbourVector(Vec3(1, 1, 3));
    CPPUNIT_ASSERT_EQUAL((size_t)1, neighbours.size());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p4) != neighbours.end());

    neighbours = table.getNeighbourVector(Vec3(3, 1, 3));
    CPPUNIT_ASSERT_EQUAL((size_t)1, neighbours.size());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p4) != neighbours.end());

    neighbours = table.getNeighbourVector(Vec3(1, 3, 3));
    CPPUNIT_ASSERT_EQUAL((size_t)1, neighbours.size());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p4) != neighbours.end());

    neighbours = table.getNeighbourVector(Vec3(3, 3, 3));
    CPPUNIT_ASSERT_EQUAL((size_t)1, neighbours.size());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p4) != neighbours.end());    
    
  }
  
  void NeighbourTableTest::testOutsideBBoxInsert()
  {
    TestTable table = TestTable(BoundingBox(Vec3(6, 4, 2), Vec3(26, 24, 22)), 2);

    TestParticle p[] =
      {
        TestParticle(Vec3(5, 3, 1), 1.0),
        TestParticle(Vec3(3, 7, 7), 0.5),
        TestParticle(Vec3(0, 28, 7), 0.5),
        TestParticle(Vec3(29, 3, 45), 0.5)
      };
    const int numP = (sizeof(p)/sizeof(TestParticle));

    table.insert(p[0]);
    TestTable::ParticleVector neighbours = table.getNeighbourVector(Vec3(6, 4, 2));
    CPPUNIT_ASSERT_EQUAL((size_t)1, neighbours.size());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p[0]) != neighbours.end());

    table.insert(p[1]);
    neighbours = table.getNeighbourVector(Vec3(6, 7, 7));
    CPPUNIT_ASSERT_EQUAL((size_t)1, neighbours.size());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p[1]) != neighbours.end());

    table.insert(p[2]);
    neighbours = table.getNeighbourVector(Vec3(6, 24, 7));
    CPPUNIT_ASSERT_EQUAL((size_t)1, neighbours.size());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p[2]) != neighbours.end());
    
    table.insert(p[3]);
    neighbours = table.getNeighbourVector(Vec3(26, 4, 22));
    CPPUNIT_ASSERT_EQUAL((size_t)1, neighbours.size());
    CPPUNIT_ASSERT(std::find(neighbours.begin(), neighbours.end(), &p[3]) != neighbours.end());    
  }

  void NeighbourTableTest::testGetNeighbours()
  {
    TestTable table = TestTable(BoundingBox(Vec3(0, 0, 0), Vec3(10, 10, 10)), 2);

    typedef std::vector<TestParticle> TestParticleVector;
    TestParticleVector particles;
    particles.reserve(1024);
    for (int x = 3; x <= 7; x+=2) {
      for (int y = 3; y <= 7; y+=2) {
        for (int z = 3; z <= 7; z+=2) {
          particles.push_back(TestParticle(Vec3(x, y, z), 0.5));
          table.insert(*(particles.rbegin()));          
        }
      }
    }

    CPPUNIT_ASSERT_EQUAL((size_t)27, table.size());
    TestTable::ParticleVector neighbours = table.getNeighbourVector(Vec3(5, 5, 5));
    CPPUNIT_ASSERT_EQUAL((size_t)1, neighbours.size());
    
    neighbours = table.getNeighbourVector(Vec3(6, 6, 6));
    CPPUNIT_ASSERT_EQUAL((size_t)1, neighbours.size());
    
    neighbours = table.getNeighbourVector(Vec3(6, 6, 6), 1);
    CPPUNIT_ASSERT_EQUAL((size_t)8, neighbours.size());

    neighbours = table.getUniqueNeighbourVector(Vec3(6, 6, 6), 1);
    CPPUNIT_ASSERT_EQUAL((size_t)8, neighbours.size());

    // insert a particle which occurs in more than one table cell.
    particles.push_back(TestParticle(Vec3(5, 6, 5), 0.5));
    table.insert(*(particles.rbegin()));
    neighbours = table.getNeighbourVector(Vec3(6, 6, 6), 1);
    CPPUNIT_ASSERT_EQUAL((size_t)10, neighbours.size());
    neighbours = table.getUniqueNeighbourVector(Vec3(6, 6, 6), 1);
    CPPUNIT_ASSERT_EQUAL((size_t)9, neighbours.size());
  }
};
