#include "CPartRender.h"
#include "console.h"
#include "Matrix3.h"

SPRenderParticle::SPRenderParticle()
{
  m_col=Colour(0.0,0.0,0.0);
}

void SPRenderParticle::addPatch(const Vec3& c,double r,const Vec3& n,double a)
{
  rspatch p;

  p.center=c;
  p.radius=r;
  p.normal=n;
  p.angle=a;

  m_patches.push_back(p);
}

void SPRenderParticle::addColor(const Colour& C)
{
  m_col+=C;
  console.XDebug() << "new Color: " << m_col.red << " " << m_col.green << " " << m_col.blue << "\n";
}

void SPRenderParticle::draw()
{
  vector<rspatch>::iterator iter;
  if(m_is2D){
    glDisable(GL_LIGHTING);
    for(iter=m_patches.begin();iter!=m_patches.end();iter++){
      drawPatch2D(iter->center,iter->radius,iter->normal,iter->angle,m_col);
    }
  } else {
    for(iter=m_patches.begin();iter!=m_patches.end();iter++){
      drawPatch(iter->center,iter->radius,iter->normal,iter->angle,m_col);
    }
  }
}

/*!
  return estimate of minimum position
*/
Vec3 SPRenderParticle::minPos() const
{
  Vec3 min;
  vector<rspatch>::const_iterator iter=m_patches.begin();

  min=iter->center-Vec3(iter->radius,iter->radius,iter->radius);
  for(iter=m_patches.begin();iter!=m_patches.end();iter++){
    min.X()=(min.X()<iter->center.X()-iter->radius) ? min.X() : iter->center.X()-iter->radius;
    min.Y()=(min.Y()<iter->center.Y()-iter->radius) ? min.Y() : iter->center.Y()-iter->radius;
    min.Z()=(min.Z()<iter->center.Z()-iter->radius) ? min.Z() : iter->center.Z()-iter->radius;
  }

  return min;
}


/*!
  return estimate of maximum position
*/
Vec3 SPRenderParticle::maxPos() const
{
  Vec3 max;
  vector<rspatch>::const_iterator iter=m_patches.begin();

  max=iter->center+Vec3(iter->radius,iter->radius,iter->radius);
  for(iter=m_patches.begin();iter!=m_patches.end();iter++){
    max.X()=(max.X()>iter->center.X()+iter->radius) ? max.X() : iter->center.X()+iter->radius;
    max.Y()=(max.Y()>iter->center.Y()+iter->radius) ? max.Y() : iter->center.Y()+iter->radius;
    max.Z()=(max.Z()>iter->center.Z()+iter->radius) ? max.Z() : iter->center.Z()+iter->radius;
  }

  return max;
}

/*!
  scale whole particle
  
  \param scl the scale factor
*/
void SPRenderParticle::scale(double scl)
{
  vector<rspatch>::iterator iter;
  m_center=scl*m_center;
  for(iter=m_patches.begin();iter!=m_patches.end();iter++){
    iter->center=scl*iter->center;
    iter->radius*=scl;
  }
}
  
void SPRenderParticle::extractFromPacket(Cxdr& cx)
{
  tconsole.XDebug() << "SPRenderParticle::extractFromPacket()\n";
  rspatch rsp;
  Matrix3 rot;
  Vec3 pos,center,normal,h1,h2,h3;
  double rad;
  unsigned long is2D,np; 

  cx >> is2D;
  m_is2D=(is2D==1);
  cx >> pos.X();
  cx >> pos.Y();
  cx >> pos.Z();
  m_center=pos;
  cx >> rad;
  // rot matrix
  cx >> h1.X();
  cx >> h1.Y();
  cx >> h1.Z();
  cx >> h2.X();
  cx >> h2.Y();
  cx >> h2.Z();
  cx >> h3.X();
  cx >> h3.Y();
  cx >> h3.Z();
  rot=Matrix3(h1,h2,h3);
  console.XDebug() << "extracted rot matrix : " << rot << "\n";
  cx >> np;
  console.XDebug() << "np : " << np << "\n";
  for(unsigned int i=0;i<np;i++){
    cx >> center.X();
    cx >> center.Y();
    cx >> center.Z();
    rsp.center=pos+rot*center;
    cx >> rsp.radius;
    cx >> normal.X();
    cx >> normal.Y();
    cx >> normal.Z();
    rsp.normal=rot*normal;
    cx >> rsp.angle;
    m_patches.push_back(rsp);
  }
}

/*!
  render a patch

  \todo remove hardwired resolution
*/
void SPRenderParticle::drawPatch(const Vec3& center,double rad,const Vec3& normal,double angle,const Colour& C)
{
  GLfloat Cols[4] ;

  console.XDebug() << "SPRenderParticle::drawPatch( " << center << " " << rad << " " << normal << " " << angle <<")\n";
  
  glEnable(GL_NORMALIZE);

  Cols[0] = (float)C.red ;
  Cols[1] = (float)C.green ;
  Cols[2] = (float)C.blue ;
  Cols[3] = (float)C.alpha ;

  glEnable(GL_LIGHTING) ; 
  glPushMatrix() ; 
 
  int  i,j;
  double alpha,beta, alpha2,beta2,dalpha,dbeta;
  Vec3 P1,P2,P3,P4;
  
  double betamax=acos(angle);
  int nbeta=int(ceil(betamax/0.25));

  dbeta=betamax/nbeta;
  dalpha=2.0*M_PI/12.0;
 
  double phi=atan2(normal.Z(),normal.X())*180.0/M_PI;
  double theta=acos(normal.Y())*180.0/M_PI;
    
  glMatrixMode(GL_MODELVIEW);
    
  glTranslatef(center.X(),center.Y(),center.Z());
  glRotatef(phi,0.0,1.0,0.0); // rot y
  glRotatef(theta,0.0,0.0,-1.0); // rot x
    
  P1=rad*Vec3(0.0,1.0,0.0);
  for(j=0;j<12;j++){ // inner part
    P2.X()=rad*cos(j*dalpha)*sin(dbeta);
    P2.Z()=rad*sin(j*dalpha)*sin(dbeta);
    P2.Y()=rad*cos(dbeta);
    P3.X()=rad*cos((j+1)*dalpha)*sin(dbeta);
    P3.Z()=rad*sin((j+1)*dalpha)*sin(dbeta);
    P3.Y()=rad*cos(dbeta);
    glBegin(GL_TRIANGLES);
    glColor4fv(Cols);
    glNormal3d(0.0,1.0,0.0) ;
    glVertex3d(P1.X(),P1.Y(),P1.Z()) ;
    glNormal3d(P3.X(),P3.Y(),P3.Z()) ;
    glVertex3d(P3.X(),P3.Y(),P3.Z()) ;
    glNormal3d(P2.X(),P2.Y(),P2.Z()) ;
    glVertex3d(P2.X(),P2.Y(),P2.Z()) ;
    glEnd();
  }
  for(i=1;i<nbeta;i++){ //  
    beta=double(i)*dbeta;
    beta2=double(i+1)*dbeta;
    for(j=0;j<12;j++){ // 
      alpha=double(j)*dalpha;
      alpha2=double(j+1)*dalpha;
      P1.X()=rad*cos(alpha)*sin(beta);
      P1.Z()=rad*sin(alpha)*sin(beta);
      P1.Y()=rad*cos(beta);
      P2.X()=rad*cos(alpha2)*sin(beta);
      P2.Z()=rad*sin(alpha2)*sin(beta);
      P2.Y()=rad*cos(beta);
      P3.X()=rad*cos(alpha2)*sin(beta2);
      P3.Z()=rad*sin(alpha2)*sin(beta2);
      P3.Y()=rad*cos(beta2);
      P4.X()=rad*cos(alpha)*sin(beta2);
      P4.Z()=rad*sin(alpha)*sin(beta2);
      P4.Y()=rad*cos(beta2);
      glBegin(GL_QUADS);
      glColor4fv(Cols);
      glNormal3d(P1.X(),P1.Y(),P1.Z()) ;
      glVertex3d(P1.X(),P1.Y(),P1.Z()) ;
      glNormal3d(P2.X(),P2.Y(),P2.Z()) ;
      glVertex3d(P2.X(),P2.Y(),P2.Z()) ;
      glNormal3d(P3.X(),P3.Y(),P3.Z()) ;
      glVertex3d(P3.X(),P3.Y(),P3.Z()) ;
      glNormal3d(P4.X(),P4.Y(),P4.Z()) ;
      glVertex3d(P4.X(),P4.Y(),P4.Z()) ;
      glEnd();
    }
  }    
  glPopMatrix() ;
  glDisable(GL_NORMALIZE);
}

/*!
  render a patch in 2D

  \todo remove hardwired resolution
*/
void SPRenderParticle::drawPatch2D(const Vec3& center,double rad,const Vec3& normal,double angle,const Colour& C)
{
  console.XDebug() << "SPRenderParticle::drawPatch2D( " << m_center << " " << center << " " << rad << " " << normal << " " << angle <<")\n";
  
  glPushMatrix() ; 

  double betamax=acos(angle);
 
  int nbeta=int(ceil(betamax/0.25));
  double dbeta=betamax/nbeta;  
  
  glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  
  glBegin(GL_POLYGON);
  glColor3d(C.red,C.green,C.blue);

  double theta=atan2(normal.Y(),normal.X());

  glVertex3d(m_center.X(),m_center.Y(),0.0);
  for(int j=-nbeta;j<=nbeta;j++){
    double p1x=center.X()+rad*cos(theta+j*dbeta);
    double p1y=center.Y()+rad*sin(theta+j*dbeta);
    glVertex3d(p1x,p1y,0.0);
  }
 
  glEnd();
  glPopMatrix();
}
