//--- Project includes ---
#include "LinearColorMap.h"
#include "console.h"

// -- IO includes --
#include <iostream>
#include <fstream>

using std::ofstream;
using std::ifstream;
using std::ios_base;
using std::ios;
using std::endl;

// -- STL includes
#include <utility>

using std::pair;

/*!
  Default Constructor. The color map is initialized as red-blue ramp ranging from 0 to 1.
*/
CLinearColorMap::CLinearColorMap() : AColorMap()
{
  addNode(0.0,Colour(1.0,0.0,0.0));
  addNode(0.5,Colour(0.0,1.0,0.0));
  addNode(1.0,Colour(0.0,0.0,1.0));
}

/*!
  Constructor. The color map is initialized as red-blue ramp ranging from 0 to 1.

  \param name the name of the colormap
*/  
CLinearColorMap::CLinearColorMap(const string& name) : AColorMap(name)
{
  addNode(0.0,Colour(1.0,0.0,0.0));
  addNode(0.5,Colour(0.0,1.0,0.0));
  addNode(1.0,Colour(0.0,0.0,1.0));
}

/*!
  The copy constructor

  \param CM the color map to be copied
*/
CLinearColorMap::CLinearColorMap(const CLinearColorMap& CM) : AColorMap(CM)
{
  m_data=CM.m_data;
}

/*!
  Destructor
*/
CLinearColorMap::~CLinearColorMap()
{}

/*!
  Assignment operator
*/
CLinearColorMap& CLinearColorMap::operator=(const CLinearColorMap& CLM)
{
  AColorMap::operator=(CLM);

  m_data=CLM.m_data;

  return *this;
}

/*!
  Get color for a given scalar value. For a value between two node points the color is determined by linear interpolation. For a value outside the range it depends on the value of m_circular.

  \param d the value
  \todo implement periodic boundary conditions
*/ 
Colour CLinearColorMap::getColor(double d) const
{
  Colour C;

  map<double,Colour>::const_iterator lower=m_data.lower_bound(d);
  map<double,Colour>::const_iterator upper;

  if(lower->first==d){ // exactly at node point
    C=lower->second;
  } else if(((lower!=m_data.end()) || (d<lower->first)) && (lower!=m_data.begin())){ // between two nodes
    upper=lower;
    lower--;
    double dist=upper->first-lower->first;
    double w2=(d-lower->first)/dist;
    double w1=(upper->first-d)/dist;
    C=Mix(lower->second,upper->second,w1,w2);
  } else if (m_data.size()!=0){
    switch (m_boundary){
    case BConst : C=lower->second; break;
    case BTrans : C=Colour(0.0,0.0,0.0,0.0) ; break;
    }
  } else { // 0-sized colormap - > return 0-color, output warning
    C=Colour(0.0,0.0,0.0);
    console.Warning() << "using empty color map\n";
  }

  return C;

}

/*!
  Save a linear color map to a file. 

  \param filename the name of the file
*/
void CLinearColorMap::SaveAs(const string& filename) const
{
  console.Debug() << "Saving color map to file " << filename.c_str() << "\n";

  AColorMap::SaveAs(filename);
  ofstream outfile(filename.c_str(),ios::app);
  outfile << "LINEAR\n";
  outfile << m_data.size() << endl;
  for(map<double,Colour>::const_iterator iter=m_data.begin();iter!=m_data.end();iter++){
    outfile << iter->first << endl;        // node location
    outfile << iter->second.red << endl;
    outfile << iter->second.green << endl;
    outfile << iter->second.blue << endl;
    outfile << iter->second.alpha << endl;
  }
  outfile.close();
}
 
/*!
  Add a node point.

  \param p the node position
  \param c the node color
*/
void CLinearColorMap::addNode(double p,const Colour& c)
{
  m_data.insert(pair<double,Colour>(p,c));
}

/*!
  Remove a node point. If an illegal index is used, nothing is done and a warning is raised.

  \param idx the index of the node.
*/
void CLinearColorMap::deleteNode(int idx)
{
  if((idx>=0) && (idx<m_data.size())){
    int count=0;
    bool done=false;
    map<double,Colour>::iterator iter=m_data.begin();
    while((iter!=m_data.end())&&(!done)){
      if(count<idx){
	count++;
	iter++;
      } else {
	done=true;
      } 
     }
     m_data.erase(iter);
  }else{
    console.Warning() << "CLinearColorMap::deleteNode: illegal index: " << idx << "\n";
  }
}

void CLinearColorMap::setNodePos(int,double)
{}

void CLinearColorMap::setNodeColor(int,const Colour&)
{}

double CLinearColorMap::getLow() const
{
  map<double,Colour>::const_iterator iter=m_data.begin();
  return iter->first;
}

double CLinearColorMap::getHigh() const
{
  map<double,Colour>::const_reverse_iterator iter=m_data.rbegin();
  return iter->first;
}
