#include "RemoteData.h"
#include "console.h"
#include "shake_debug.h"
#include "ShakeRenderer.h"
#include "RDParticleField.h"


#include <algorithm>
using std::min;
using std::max;

//--------------------------------------
//  CFieldDescription functions
//--------------------------------------

CFieldDescription::CFieldDescription() 
{}

CFieldDescription::~CFieldDescription() 
{}

void CFieldDescription::setName(string name) 
{
  console.Debug()<< "Name : " << name.c_str() << "\n";
  m_name = name ;
}

vector<ADataField::AFieldShape> CFieldDescription::getAvailableShapes()
{
  return m_available_shapes;
}

/*!
  Add a shape to the list of the available shapes of a field. If s is the first shape added, it is made the default shape.

  \param s the shape
*/
void CFieldDescription::addShape(const ADataField::AFieldShape& s)
{
  if(m_available_shapes.size()==0){
    m_shape=s;
  }
  m_available_shapes.push_back(s);
}

string CFieldDescription::getName() 
{
  return m_name ;
}

ADataField::ADataType CFieldDescription::getType() 
{
  return m_type ;
}

void CFieldDescription::setType(ADataField::ADataType type) {
  m_type = type ;
}

//--------------------------------------
//  CDataParticle functions
//--------------------------------------

CDataParticle::CDataParticle()
{
  m_valid=false;
}

CDataParticle::CDataParticle(const Vec3&  pos, double rad)
{
  m_Pos=pos;
  m_rad=rad;
  m_valid=true;
}

//--------------------------------------
//  CRemoteData functions
//--------------------------------------


CRemoteData::CRemoteData() 
{
  m_backend=NULL;
  cur_Id = 1 ;
  scale_calculated=false ;
}

CRemoteData::~CRemoteData() 
{
}

/*! 
  Add a Field, return the Id of the field
  \param f pointer to field to be added
*/
int CRemoteData::addField(ADataField *f) 
{
  m_activefields.push_back(f) ;
  f->m_Id = cur_Id ;
  cur_Id++ ;
  return f->m_Id ;
}

/*! 
  Add an available Field for download. If field with the same name already exists, ignore the new one.
  \param f pointer to field to be added
*/
void CRemoteData::addFieldDescription(CFieldDescription f) 
{
  bool have=false;
  for(vector<CFieldDescription>::iterator iter=m_fieldsAvailable.begin();
      iter!=m_fieldsAvailable.end();
      iter++){
    if(iter->getName()==f.getName()){
      have=true;
    }
  }
  if(!have){
    m_fieldsAvailable.push_back(f) ;
  }
}

/*!
  Get the names of the active fields
*/ 
vector<string> CRemoteData::getActiveFieldNames()
{
  vector<string> res;

  for(list<ADataField *>::iterator siter=m_activefields.begin();siter!=m_activefields.end();siter++){
    res.push_back((*siter)->getName());
  }
  return res;
}

/*!
  check if a field is already active

  \param name the name of the field
*/
bool CRemoteData::isActive(const string& name)
{
  bool active=false;
  
  list<ADataField *>::iterator siter=m_activefields.begin();
  while((siter!=m_activefields.end()) && !active){
    if((*siter)->getName()==name) {
      active=true;
    }
    siter++;
  }

  console.XDebug()<< "CRemoteData::isActive " << name.c_str() << "  " << active << "\n";
  return active; 
}

/*!
  make available field active, i.e. visible

  \param  name the name of the field
*/
void CRemoteData::activateField(const string& name)
{
  console.XDebug() << "CRemoteData::activateField( " << name.c_str() << " )\n";
  // get the field
  vector<CFieldDescription>::iterator diter=m_fieldsAvailable.begin();
  bool found=false;
  CFieldDescription cfd;

  while((diter!=m_fieldsAvailable.end()) && !found){
    if(name==diter->getName()){
      found=true;
      cfd=*diter;
    }
    diter++;
  }
  if(found){
    if(m_backend!=NULL){
      ADataField *adfield =(ADataField*)NULL;
      if(isActive(name)){ //field is already active
      
      } else { //new field 
	if (cfd.getType()==ADataField::FLD_SCALAR){ // scalar field
	  if(cfd.getShape()==ADataField::FLD_PARTICLE){ // per particle
	    adfield=new CScalarParticleField(name);
	  }else if(cfd.getShape()==ADataField::FLD_CONN){ // per connection
	    console.Critical() << "per connection field not implemented\n";
	  } else {
	    console.Critical() << "other field not implemented\n";
	  }
	}else{ // vector field
	  console.Critical() << "vector field not implemented\n";
	}
      }
      if(adfield!=NULL){
	m_backend->getFieldData(adfield);
	addField(adfield);
      }
      console.Debug()<<"activating field " << name.c_str()  << "\n";
    } else {
      console.Error()<<"trying to activate field while backend==NULL\n";
    }
  }else{
    console.Error()<<"trying to activate unknown field\n";
  }
}

/*!
  reload current values of active fields
*/
void CRemoteData::reloadActiveFields()
{
  for(list<ADataField*>::iterator iter=m_activefields.begin();iter!=m_activefields.end();iter++){
    (*iter)->clear();
    m_backend->getFieldData(*iter);
  }
}

/*!
  Delete a field from the list of active fields. If a non-existing or inactive field is attempted to be removed, nothing happens.
 
  \param name field name
*/
void CRemoteData::deleteField(const string& name) 
{
  console.XDebug() << "CRemoteData::deleteField  " << name.c_str() << "\n";

  list<ADataField *>::iterator f=m_activefields.begin();
  bool done=false;

  while((f!=m_activefields.end()) && (!done)) {
    console.XDebug() << "name: " <<  (*f)->getName().c_str() << "\n";
    if ((*f)->getName() == name) {
      m_activefields.erase(f) ;
      console.XDebug() << "erased field " << name.c_str() << "nfields : " << m_activefields.size() << "\n";
      done=true;
    }
    f++;
  }
  return ;
}


/*!
  return field associated with a given Id, return NULL if error
  
  \param n Field Id
  \todo deuglify
*/
ADataField * CRemoteData::getActiveField(int n) const 
{
  console.XDebug()<<"CRemoteData::getActiveField( " << n << " )\n";

  list<ADataField *>::const_iterator f ;
  ADataField *adf=(ADataField *)(NULL) ;
  int i=0;

  for(f=m_activefields.begin();f!=m_activefields.end();f++) {
    if (i == n) {
      adf=(ADataField *)*f ; 
    }
    i++;
  }
  console.XDebug()<< (void*)(*f) << "\n";
  return adf;
}

/*!
  return particle data of a given particle
  \param i Particle Id
*/
CDataParticle & CRemoteData::getParticle(int i) 
{
  return m_particles[i] ;
}

/*!
  return particle data of a given particle
  \param i Particle Id
*/
SPRenderParticle & CRemoteData::getCParticle(int i) 
{
  return m_sp_particles[i] ;
}



/*!
  add a particle to the data set

  \param pos the position of the particle
  \param rad the radius of the particle
  \param i the global id of the particle
*/
void CRemoteData::addParticle(const Vec3& pos,real rad,int i)
{
  m_particles[i]=CDataParticle(pos,rad);
}

/*!
  add a spherical patch particle to the data set

  \param sp the particle
  \param i the global id of the particle
*/
void CRemoteData::addParticle(const SPRenderParticle& sp,int i)
{
  m_sp_particles[i]=sp;
}

/*!
  get number of particles
*/
int CRemoteData::getNbParticle() {
  return m_particles.size();
}


void CRemoteData::calcScale() 
{
  real minX,minY,minZ,maxX,maxY,maxZ ;
  int i;
  
  // spherical particles
  if ((m_particles.size()> 0)&&(!scale_calculated)) {
    console.Debug() << "calculating scale for " << m_particles.size() << " particles\n" ;
    
    int i=0;

    minX = m_particles[i].Pos().X() - m_particles[i].Rad()/2.0 ;
    minY = m_particles[i].Pos().Y() - m_particles[i].Rad()/2.0 ;
    minZ = m_particles[i].Pos().Z() - m_particles[i].Rad()/2.0 ;
    maxX = m_particles[i].Pos().X() + m_particles[i].Rad()/2.0 ;
    maxY = m_particles[i].Pos().Y() + m_particles[i].Rad()/2.0 ;
    maxZ = m_particles[i].Pos().Z() + m_particles[i].Rad()/2.0 ;
    for (i=0; i<m_particles.size(); i++) {
      minX = min(minX,m_particles[i].Pos().X() - m_particles[i].Rad()/2.0) ;
      minY = min(minY,m_particles[i].Pos().Y() - m_particles[i].Rad()/2.0) ;
      minZ = min(minZ,m_particles[i].Pos().Z() - m_particles[i].Rad()/2.0) ;
      maxX = max(maxX,m_particles[i].Pos().X() + m_particles[i].Rad()/2.0) ;
      maxY = max(maxY,m_particles[i].Pos().Y() + m_particles[i].Rad()/2.0) ;
      maxZ = max(maxZ,m_particles[i].Pos().Z() + m_particles[i].Rad()/2.0) ;
    }
    m_centre = (Vec3(maxX,maxY,maxZ)-Vec3(minX,minY,minZ))/2.0 ;
    m_scale = max(max(maxX-minX,maxY-minY),maxZ-minZ) /2.0 ;
    scale_calculated = true ;
  }
  // complex particles
  if ((m_sp_particles.size()> 0)&&(!scale_calculated)) {
    console.Debug() << "calculating scale for " << m_sp_particles.size() << "sp-particles\n" ;
    Vec3 minp,maxp;

    int i=0;

    minp=m_sp_particles[i].minPos();
    maxp=m_sp_particles[i].maxPos();
    for (i=0; i<m_sp_particles.size(); i++) {
      minp=comp_min(minp,m_sp_particles[i].minPos());
      maxp=comp_max(maxp,m_sp_particles[i].maxPos());
    }
    m_centre = (minp-maxp)/2.0 ;
    console.Debug() << "min,max " << minp << " " << maxp << "\n"; 
    m_scale = max(max(maxp.X()-minp.X(),maxp.Y()-minp.Y()),maxp.Z()-minp.Z()) /2.0 ;
    scale_calculated = true ;
  } 
  

  if (!scale_calculated) {
    m_scale = 1.0 ;
  }
}

real CRemoteData::scale() {
  return m_scale ;
}

Vec3 & CRemoteData::centre() {
  return m_centre ;
}
