#include "BondedInteraction.h"
#include "console.h"
#include "tml/message/packed_message_interface.h"

CBondedInteraction::CBondedInteraction():APairInteraction()
{
  m_k=0;
  m_break=0;
  m_r0=0;
  m_dist=0;
  m_force=Vec3(0.0,0.0,0.0);
  m_tag = -1;
}

CBondedInteraction::CBondedInteraction(
  CParticle* p1,
  CParticle* p2,
  const CBondedIGP& param
)
  : APairInteraction(p1, p2)
{
  m_k=param.k;
  m_break=(m_p1->getRad()+m_p2->getRad())*param.rbreak;
  m_r0=p1->getRad()+p2->getRad();
  // setup initial distance in case we want to break before first calcForces()
  Vec3 D=p1->getPos()-p2->getPos();
  m_dist=sqrt(D*D);
  m_force=Vec3(0.0,0.0,0.0);
  m_tag = param.tag;
}

CBondedInteraction::~CBondedInteraction()
{
  //
  // if (m_p1!=NULL) m_p1->setFlag();
  // if (m_p2!=NULL) m_p2->setFlag();
}

bool CBondedInteraction::broken()
{
  if(m_dist>m_break){
    console.Debug() << "bond broken" << "\n" ;
    console.Debug() << "ids : " << m_p1->getID() <<  " " << m_p2->getID() << "\n" ;
    console.Debug() << "positions : " << m_p1->getPos() <<  m_p2->getPos() << "\n";
    console.Debug() << "dist : " << m_dist << "\n" ;
    console.Debug() << "break : " << m_break << "\n" ;
  }
  return(m_dist>m_break);
}

/*!
  Set breaking distance to sum of the radii multiplied with given "relative breaking distance"

  \param rel_break the relative breaking distance
*/
void CBondedInteraction::setBreak(double rel_break)
{
  m_break=(m_p1->getRad()+m_p2->getRad())*rel_break;
}

/*!
  Calculate bonded elastic forces. 21 Flops
*/
void CBondedInteraction::calcForces()
{
  //  console.XDebug() 
  //    << "bonded interaction: [" << m_p1->getID() << " - "
  //    << m_p2->getID() << "]" << m_p1->getPos()
  //    << m_p2->getPos() << "\n"; 
  const Vec3 D=m_p1->getPos()-m_p2->getPos();
  m_dist=sqrt(D*D);

  m_force = D*(m_k*(m_dist-m_r0)/m_dist);
  const Vec3 pos=m_p2->getPos()+(m_p2->getRad()/(m_p1->getRad()+m_p2->getRad()))*D;

  m_p2->applyForce(m_force,pos);
  m_p1->applyForce(-1.0*m_force,pos);
  m_cpos=pos;
}

/*!
  get the potential energy stored in the interaction
*/
double CBondedInteraction::getPotentialEnergy() const
{
  double e_pot_norm=0.5*m_force*m_force/m_k;

  return e_pot_norm;
}

/*!
  Get the particle member function which returns a scalar field of a given name.

  \param name the name of the field 
*/
CBondedInteraction::ScalarFieldFunction CBondedInteraction::getScalarFieldFunction(const string& name)
{
  CBondedInteraction::ScalarFieldFunction sf;

  if (name=="potential_energy"){
    sf=&CBondedInteraction::getPotentialEnergy;
  } else if (name=="count"){
    sf=&CBondedInteraction::Count;
  } else {
    sf=NULL;
    cerr << "ERROR - invalid name for interaction scalar  access function" << endl; 
  }
  
  return sf;
}

/*!
  Get the particle member function which returns a vector field of a given name.

  \param name the name of the field 
*/
CBondedInteraction::VectorFieldFunction CBondedInteraction::getVectorFieldFunction(const string& name)
{
  CBondedInteraction::VectorFieldFunction sf;

  sf=NULL;
  cerr << "ERROR - invalid name for interaction vector  access function" << endl; 
  
  return sf;
}

/*!
  Get the particle member function which returns a checked scalar field of a given name.

  \param name the name of the field
*/
CBondedInteraction::CheckedScalarFieldFunction CBondedInteraction::getCheckedScalarFieldFunction(const string& name)
{
	CBondedInteraction::CheckedScalarFieldFunction sf;

	sf=NULL;
	cerr << "ERROR - invalid name for interaction scalar  access function" << endl;

	return sf;
}

/*!
  Pack a CBondedInteraction into a TML packed message

  \param I the interaction
*/
template<>
void TML_PackedMessageInterface::pack<CBondedInteraction>(const CBondedInteraction& I)
{
  append(I.m_k);
  append(I.m_r0);
  append(I.m_dist);
  append(I.m_break);
  append(I.m_id[0]);
  append(I.m_id[1]);
  append(I.getTag());
}

/*!
  Unpack a CBondedInteraction from a TML packed message

  \param I the interaction
*/
template<>
void TML_PackedMessageInterface::unpack<CBondedInteraction>(CBondedInteraction& I)
{
  I.m_k=pop_double();
  I.m_r0=pop_double();
  I.m_dist=pop_double();
  I.m_break=pop_double();
  I.m_id.erase(I.m_id.begin(),I.m_id.end());
  I.m_id.push_back(pop_int());
  I.m_id.push_back(pop_int());
  I.setTag(pop_int());
}

ostream& operator<<(ostream& ost,const CBondedInteraction& BI)
{
  ost << "[" << BI.m_p1->getID() << " - " << BI.m_p2->getID() << "]";
  return ost;
}
