#ifndef __BONDEDINTERACTION_H
#define __BONDEDINTERACTION_H

// -- project includes --
#include "Model/Interaction.h"
#include "Model/Particle.h"
#include "Model/InteractionParam.h"
#include "Foundation/vec3.h"
#include "Parallel/CheckPointable.h"

// -- I/O includes --
#include <iostream>
using std::ostream;

// -- STL includes --
#include <utility>

using std::pair;

/*!
  \brief Interaction parameters for bonded interaction
  \author Steffen Abe

  $Revision$
  $Date$
*/
struct CBondedIGP
{
  /**
   * Spring constant. 
   */
  double k;
  
  /**
   * Breaking distance/radius?
   */
  double rbreak;
  
  int tag;
};

/*!
   \brief Elastic interaction between bonded particles
   \author Steffen Abe

   $Revision$
   $Date$
*/
class CBondedInteraction : public APairInteraction
{
 public: // types
  typedef CBondedIGP ParameterType;

  typedef double (CBondedInteraction::* ScalarFieldFunction)() const;
  typedef pair<bool,double> (CBondedInteraction::* CheckedScalarFieldFunction)() const;
  typedef Vec3 (CBondedInteraction::* VectorFieldFunction)() const;

 private:
  double m_k;     //!< spring constant
  double m_r0;    //!< equilibrium distance
  double m_dist;  //!< current distance, cached from last calcForces()
  double m_break; //!< breaking distance
  Vec3 m_force;   //!< current force, cached for E_pot calculation
  Vec3 m_cpos;
  int  m_tag;     //!< Interaction tag;

 public:
  CBondedInteraction();
  CBondedInteraction(
    CParticle *particle1,
    CParticle *particle2,
    const CBondedIGP &params
  );

  virtual ~CBondedInteraction();

  static ScalarFieldFunction getScalarFieldFunction(const string&);
  static CheckedScalarFieldFunction getCheckedScalarFieldFunction(const string&);
  static VectorFieldFunction getVectorFieldFunction(const string&);

  virtual void calcForces();
  void setBreak(double);
  bool broken();
  
  
  inline int getTag() const {return m_tag;}
  inline void setTag(int tag) {m_tag = tag;}
  
  double getPotentialEnergy() const;

  virtual Vec3 getPos(){return m_cpos;};

  friend ostream& operator<<(ostream&,const CBondedInteraction&);
  friend class TML_PackedMessageInterface;
};

/**
 * Helper class for checkpointing BondedInteraction data.
 */
class BondedInteractionCpData : public esys::lsm::CheckPointable
{
public:
  typedef int ParticleId;
  typedef int InteractionTag;
  
  BondedInteractionCpData() : m_p1Id(-1), m_p2Id(-1), m_tag(-1)
  {
  }

  BondedInteractionCpData(const CBondedInteraction &bondedInteraction)
    : m_p1Id(bondedInteraction.getPairID().first),
      m_p2Id(bondedInteraction.getPairID().second),
      m_tag(bondedInteraction.getTag())
  {
  }

  BondedInteractionCpData(
    ParticleId particle1Id,
    ParticleId particle2Id,
    InteractionTag interactionTag
  )
    : m_p1Id(particle1Id),
      m_p2Id(particle2Id),
      m_tag(interactionTag)
  {
  }

  void set(const CBondedInteraction &bondedInteraction)
  {
    m_p1Id = bondedInteraction.getPairID().first;
    m_p2Id = bondedInteraction.getPairID().second;
    m_tag  = bondedInteraction.getTag();
  }

  void set(ParticleId particle1Id, ParticleId particle2Id, InteractionTag interactionTag)
  {
    m_p1Id = particle1Id;
    m_p2Id = particle2Id;
    m_tag  = interactionTag;
  }
  
  ParticleId getP1Id() const
  {
    return m_p1Id;
  }
  
  ParticleId getP2Id() const
  {
    return m_p2Id;
  }
  
  InteractionTag getTag() const
  {
    return m_tag;
  }
  
  virtual void saveCheckPointData(std::ostream &oStream)
  {
    const char delim = ' ';
    oStream
      << m_p1Id << delim
      << m_p2Id << delim
      << m_tag;
  }

  virtual void loadCheckPointData(std::istream &iStream)
  {
    iStream
      >> m_p1Id
      >> m_p2Id
      >> m_tag;
  }

private:
  ParticleId     m_p1Id;
  ParticleId     m_p2Id;
  InteractionTag m_tag;
};

#endif //__BONDEDINTERACTION_H
