#ifndef __DAMPING_H
#define __DAMPING_H

// -- project includes --
#include "Model/InteractionGroup.h"
#include "Foundation/vec3.h"

class CVarMPIBuffer;
class AMPIBuffer;
/*!
  \class CDampingIGP
  \brief Interaction group parameters for CDampingGroup

  \author Steffen Abe
  $Revision$
  $Date$
*/ 
class CDampingIGP : public AIGParam
{
protected:
  Vec3 m_vref; //!< reference velocity
  double m_visc; //!< artificial viscosity
  double m_dt;   //!< time step
  int m_max_iter; //!< max nr. of iterations

public:
  virtual void  packInto(CVarMPIBuffer*) const;
  void setVRef(const Vec3 V){m_vref=V;};
  Vec3 getVRef()const{return m_vref;};
  void setVisc(double v){m_visc=v;};
  double getVisc()const{return m_visc;};
  void setTimeStep(double t){m_dt=t;};
  double getTimeStep()const{return m_dt;};
  void setMaxIter(int mi){m_max_iter=mi;};
  int getMaxIter()const {return m_max_iter;};
};

CDampingIGP* extractDampingIGP(AMPIBuffer*);

/*!
  \class CDamping
  \brief Damping of the particle motion by an artificial viscosity

  \author Steffen Abe
  $Revision$
  $Date$
*/

template <class T>
class CDamping
{
protected:
  T *m_p; //!< the particle
  Vec3 m_vref; //!< reference velocity
  double m_visc; //!< artificial viscosity
  double m_dt;   //!< time step
  int m_maxiter;   //!< iteration limit
  double m_E_diss; //!< dissipaed energy

  static double s_limit2; //!< square error limit for iteration 
  static int s_flops;       

public:
  typedef CDampingIGP ParameterType;

  typedef double (CDamping::* ScalarFieldFunction)() const;
  typedef pair<bool,double> (CDamping::* CheckedScalarFieldFunction)() const;
  typedef Vec3 (CDamping::* VectorFieldFunction)() const;

  static ScalarFieldFunction getScalarFieldFunction(const string&);
  static CheckedScalarFieldFunction getCheckedScalarFieldFunction(const string&);
  static VectorFieldFunction getVectorFieldFunction(const string&);

  CDamping(T*,const Vec3&,double,double,int); // to be obsoleted
  CDamping(T*,const CDampingIGP&);
  virtual ~CDamping();

  inline void setLimit(double limit){s_limit2=limit*limit;};
  virtual void calcForces();
  virtual bool hasTag(int,int) const;
  virtual Vec3 getPosFirst() const {return m_p->getPos();};
  virtual Vec3 getPosSecond() const {return Vec3(0.0,0.0,0.0);};
  virtual Vec3 getPos(){return m_p->getPos();};

  static void zeroFlops(){s_flops=0;};
  static int Flops(){return s_flops;};

  double getDissipatedEnergy() const;
};

#include "Model/Damping.hpp"

#endif //__DAMPING_H
