//----------------------------------------------
//       CEWallInteractionGroup functions
//----------------------------------------------

#include "Fields/WallForceField.h"
#include "Fields/WallPosField.h"
#include "Foundation/console.h"

template<class T>
CEWallInteractionGroup<T>::CEWallInteractionGroup(TML_Comm* comm):AWallInteractionGroup<T>(comm)
{}

template<class T>
CEWallInteractionGroup<T>::CEWallInteractionGroup(TML_Comm* comm,const CEWallIGP* I):AWallInteractionGroup<T>(comm)
{
  console.XDebug() << "making CEWallInteractionGroup pos : " << I->getWallPos() << " norm : " << I->getWallNormal() << "\n";

  m_k=I->getSpringConst();
  m_wall=new CWall(I->getWallPos(),I->getWallNormal());
}

template<class T>
void CEWallInteractionGroup<T>::calcForces()
{

  console.XDebug() << "calculating " << m_interactions.size() << " elastic wall forces\n" ;

  for(
    typename vector<CElasticWallInteraction<T> >::iterator it=m_interactions.begin();
    it != m_interactions.end();
    it++
  ){
    it->calcForces();
  }
}

template<class T>
void CEWallInteractionGroup<T>::Update(ParallelParticleArray<T>* PPA)
{

  console.XDebug() << "CEWallInteractionGroup::Update()\n" ;

  // empty particle list first
  m_interactions.erase(m_interactions.begin(),m_interactions.end());
  m_inner_count=0;
  // build new particle list
  typename ParallelParticleArray<T>::ParticleListHandle plh=
    PPA->getParticlesAtPlane(m_wall->getOrigin(),m_wall->getNormal());
  for(typename ParallelParticleArray<T>::ParticleListIterator iter=plh->begin();
      iter!=plh->end();
      iter++){
    bool iflag=PPA->isInInner((*iter)->getPos());
    m_interactions.push_back(CElasticWallInteraction<T>(*iter,m_wall,m_k,iflag));
    m_inner_count+=(iflag ? 1 : 0);
  }
  console.XDebug() << "end CEWallInteractionGroup::Update()\n";
}


/*!
  Apply a given force to the wall. Only forces in the direction of the given force are 
  considered, free movement is assumed in perpendicular directions.

  \param F the force
  \warning Forces not perpendicular to the wall make no sense here, but this is not checked!
*/
template<class T>
void CEWallInteractionGroup<T>::applyForce(const Vec3& F)
{
  // calculate local K
  double K=m_inner_count*m_k;
  // get global K
  double K_global=m_comm->sum_all(K);

  int it=0;
  double d;
  Vec3 O_f=F.unit(); // direction of the applied force
  do{
    // calculate local F
    Vec3 F_local=Vec3(0.0,0.0,0.0);
    for (
      typename vector<CElasticWallInteraction<T> >::iterator iter=m_interactions.begin();
      iter!=m_interactions.end();
      iter++
    ){
      if(iter->isInner()){
        Vec3 f_i=iter->getForce();
        F_local+=(f_i*O_f)*O_f; // add component of f_i in O_f direction
      }
    }
    // get global F
    // by component (hack - fix later,i.e. sum_all for Vec3)
    double fgx=m_comm->sum_all(F_local.X());
    double fgy=m_comm->sum_all(F_local.Y());
    double fgz=m_comm->sum_all(F_local.Z());
    Vec3 F_global=Vec3(fgx,fgy,fgz);

    // calc necessary wall movement
    d=((F+F_global)*O_f)/K_global;
    // move the wall
    m_wall->moveBy(d*O_f);
    it++;
  } while((it<10)&&(fabs(d)>10e-6)); // check for convergence
}


template<class T>
void CEWallInteractionGroup<T>::addToFields(map<string,AField*>* fm)
{
  console.XDebug()<<"CEWallInteractionGroup<T>::addToFields\n";
  map<string,AField*>::iterator iter=fm->find("WallForces");
  if(iter==fm->end()){
    console.XDebug()<<"No wall force field yet, generating it\n";
    CWallForceField *wallf=new CWallForceField;
    wallf->addWall(m_wall);
    fm->insert(pair<string,AField*>("WallForces",wallf));
  }else{
    console.XDebug()<<"Have wall force field, adding wall\n";
    //CWallForceField *wallf=dynamic_cast<CWallForceField*>(iter->second);
    CWallForceField *wallf=(CWallForceField*)(iter->second);
    if(wallf!=NULL){
      wallf->addWall(m_wall);
    } else { // can't happen
      console.Error() << "dynamic cast failed\n";
    }
  }
  // position
  iter=fm->find("WallPos");
  if(iter==fm->end()){
    console.XDebug()<<"No wall position field yet, generating it\n";
    CWallPosField *wallp=new CWallPosField;
    wallp->addWall(m_wall);
    fm->insert(pair<string,AField*>("WallPos",wallp));
  } else {
    console.XDebug()<<"Have wall position field, adding wall\n";
    //CWallForceField *wallf=dynamic_cast<CWallForceField*>(iter->second);
    CWallPosField *wallp=(CWallPosField*)(iter->second);
    if(wallp!=NULL){
      wallp->addWall(m_wall);
    } else { // can't happen
      console.Error() << "dynamic cast failed\n";
    }
  }
  console.XDebug()<<"end  CEWallInteractionGroup<T>::addToFields\n";
}

template<class T>
ostream& operator<<(ostream& ost,const CEWallInteractionGroup<T>& IG)
{
  ost << "CEWallInteractionGroup" << endl << flush;
  ost << *(IG.m_wall) << endl << flush;
 
  return ost;
}
