#ifndef __FRICTIONINTERACTION_H
#define __FRICTIONINTERACTION_H

// -- project includes --
#include "Interaction.h"
#include "Particle.h"
#include "InteractionParam.h"
#include "vec3.h"

// -- I/O includes --
#include <iostream>
using std::ostream;

// -- STL includes --
#include <utility>

using std::pair;


/*!
  \struct CFrictionIGP
  \brief Interaction parameters for frictional interaction
  \author Steffen Abe
  $Revision$
  $Date$
*/
struct CFrictionIGP
{
  double k;
  double mu;
  double k_s;
  double dt;
};

/*!
   \class CFrictionInteraction
   \brief Frictional+Elastic interaction between particles
   \author Steffen Abe
   $Revision$
   $Date$
*/
class CFrictionInteraction : public APairInteraction
{
 public: // types
  typedef CFrictionIGP ParameterType;

  typedef double (CFrictionInteraction::* ScalarFieldFunction)() const;
  typedef pair<bool,double> (CFrictionInteraction::* CheckedScalarFieldFunction)() const;
  typedef Vec3 (CFrictionInteraction::* VectorFieldFunction)() const;

  static ScalarFieldFunction getScalarFieldFunction(const string&);
  static CheckedScalarFieldFunction getCheckedScalarFieldFunction(const string&);
  static VectorFieldFunction getVectorFieldFunction(const string&);

 protected:
  double m_k;     //!< spring constant
  double m_r0;    //!< equilibrium distance
  double m_mu;    //!< coefficient of friction
  double m_ks;    //!< shear stiffness (Cundall)
  double m_dt;    //!< time step
  Vec3 m_Ffric;   //!< current frictional force
  Vec3 m_force_deficit; //!< difference between fric. force & force necessary for slip
  Vec3 m_cpos; //!< contact position
  Vec3 m_normal_force; //!< current normal force
  bool m_is_slipping; //!< static/dynamic status of the interaction
  bool m_is_touching; //!< contact status of the interaction
  double m_E_diss; //!< dissipaed energy

 public:
  CFrictionInteraction();
  CFrictionInteraction(CParticle*,CParticle*);
  CFrictionInteraction(CParticle*,CParticle*,const CFrictionIGP&);
  virtual ~CFrictionInteraction();

  virtual void calcForces();
  virtual bool isPersistent();

  pair<bool,double> getAbsFrictionalForce() const;
  pair<bool,double> getAbsFN() const;
  pair<bool,double> getAbsMuFN() const;
  pair<bool,double> getSlipVelocity() const;
  double getPotentialEnergy() const;
  double getSlipping()const;
  double getTouching()const;
  double getSticking()const;
  double getDissipatedEnergy() const;
  Vec3 getForce() const;
  virtual Vec3 getPos(){return m_cpos;};

  pair<bool,double> getMuEff(const Vec3&,const Vec3&) const;
  pair<bool,double> getMuEffXY() const {return getMuEff(Vec3(1.0,0.0,0.0),Vec3(0.0,1.0,0.0));};
  pair<bool,double> getMuEffXZ() const {return getMuEff(Vec3(1.0,0.0,0.0),Vec3(0.0,0.0,1.0));};

  friend ostream& operator<<(ostream&,const CFrictionInteraction&);
  friend class TML_PackedMessageInterface;
};

#endif //__FRICTIONINTERACTION_H
