
#include "Model/Particle.h"
#include "Model/SimpleParticleData.h"
#include "Foundation/console.h"
#include "tml/message/packed_message_interface.h"
#include "Parallel/mpibuf.h"
#include "Parallel/mpisgbuf.h"

// --- field includes ---
#include "Fields/StdScalarField.h"
#include "Fields/StdComponentField.h"

const Vec3 zeroVec3(0.0,0.0,0.0);

bool CParticle::s_do2Calculations = false;

CParticle::CParticle()
  : CBasicParticle(zeroVec3,0.0),
    m_sigma(), 
    m_vel(),
    m_force(),
    m_oldpos(),
    m_initpos(),
    m_circular_shift(), //!< shift vector if particle is circular image
    m_mass(0.0),
    m_div_mass(0.0),
    flag(false)
{
}

CParticle::CParticle(const esys::lsm::SimpleParticleData &data)
  : CBasicParticle(data),
    m_sigma(),
    m_vel(),
    m_force(),
    m_oldpos(data.getPosition()),
    m_initpos(data.getPosition()),
    m_circular_shift(), //!< shift vector if particle is circular image
    m_mass(data.getMass()),
    m_div_mass((data.getMass() != 0) ? 1.0/data.getMass() : 0.0),
    flag(false)
{
}

/*!
  Construct particle. Old and initial position are assumed to be identical to current position. 

  \param rad radius
  \param mass mass
  \param pos current position
  \param vel current velocity
  \param force currently applied force
  \param id particle id
*/
CParticle::CParticle(double rad,double mass,const Vec3& pos,const Vec3& vel,const Vec3& force,int id) : CBasicParticle(pos,rad)
{
  m_oldpos=pos;
  m_initpos=pos;
  m_vel=vel;
  m_force=force;
  m_mass=mass;
  if(m_mass!=0.0){
    m_div_mass=1.0/m_mass;
  } else {
    m_div_mass=0.0;
  }
  m_global_id=id;
  m_circular_shift = zeroVec3;
  flag=false;
}

CParticle::CParticle(double rad,double mass,const Vec3& pos,const Vec3& oldpos,const Vec3& initpos,const Vec3& vel,const Vec3& force,int id) : CBasicParticle(pos,rad)
{
  m_oldpos=oldpos;
  m_initpos=initpos;
  m_vel=vel;
  m_force=force;
  m_mass=mass;
  if(m_mass!=0.0){
    m_div_mass=1.0/m_mass;
  } else {
    m_div_mass=0.0;
  }
  m_global_id=id;
  m_circular_shift = zeroVec3;
  flag=false;
}

/*!
  Pack the particle into a transport buffer

  \param buffer a pointer to the transport buffer
  \todo Basicparticle data should be handled by BasicParticle::packInto
*/
void CParticle::packInto(AMPIBuffer* buffer) const
{
  buffer->append(m_tag);
  buffer->append(m_pos) ;
  buffer->append(m_oldpos);
  buffer->append(m_initpos);
  buffer->append(m_circular_shift);
  buffer->append(m_vel);
  buffer->append(m_force);
  buffer->append(m_rad);
  buffer->append(m_mass);
  buffer->append(m_global_id);  
}

/*!
  extract particle from a scatter/gather buffer

  \param buffer a pointer to the buffer
  \param slice from which slice of the buffer to extract
  \warning no typechecking, i.e. the packed data is extracted from the buffer and interpreted as CParticle, no matter what actually has be put into the buffer (change?)
*/  
CParticle& CParticle::extractFrom(AMPISGBufferRoot* buffer,int slice)
{
  double db[6*3+2] ;
  
  m_tag=buffer->pop_int(slice);
  buffer->pop_doubles(slice,db,20);
  m_pos=Vec3(db[0],db[1],db[2]) ;
  m_oldpos=Vec3(db[3],db[4],db[5]) ;
  m_initpos=Vec3(db[6],db[7],db[8]) ;
  m_circular_shift=Vec3(db[9],db[10],db[11]) ;
  m_vel=Vec3(db[12],db[13],db[14]) ;
  m_force=Vec3(db[15],db[16],db[17]) ;
  m_rad = db[18];
  m_mass = db[19];
  if(m_mass!=0.0){
    m_div_mass=1.0/m_mass;
  } else {
    m_div_mass=0.0;
  }
  m_global_id=buffer->pop_int(slice);

  return *this ; 
}

/*!
  extract data from buffer and generate new particle

  \param buffer the buffer
*/
CParticle* CParticle::generateFromBuffer(AMPIBuffer* buffer)
{
  CParticle* res=new CParticle;
  double db[6*3+2] ;
  
  res->m_tag=buffer->pop_int();
  buffer->pop_doubles(db,20);
  res->m_pos=Vec3(db[0],db[1],db[2]) ;
  res->m_oldpos=Vec3(db[3],db[4],db[5]) ;
  res->m_initpos=Vec3(db[6],db[7],db[8]) ;
  res->m_circular_shift=Vec3(db[9],db[10],db[11]) ;
  res->m_vel=Vec3(db[12],db[13],db[14]) ;
  res->m_force=Vec3(db[15],db[16],db[17]) ;
  res->m_rad = db[18];
  res->m_mass = db[19];
  if(res->m_mass!=0.0){
    res->m_div_mass=1.0/res->m_mass;
  } else {
    res->m_div_mass=0.0;
  }
  res->m_global_id=buffer->pop_int();

  return res;
}

int CParticle::getPackSize()
{
  return 200; // replace by MPI_Pack_size()
}

/*!
  Get the particle member function which returns a scalar field of a given name.

  \param name the name of the field 
*/
CParticle::ScalarFieldFunction CParticle::getScalarFieldFunction(const string& name)
{
  CParticle::ScalarFieldFunction sf;

  if(name=="id"){
    sf=&CParticle::getIDField;
  } else if(name=="tag"){
    sf=&CParticle::getTagField;    
  } else if(name=="sigma_xx_2d"){
    sf=&CParticle::sigma_xx_2D;
  } else if(name=="sigma_xy_2d"){
    sf=&CParticle::sigma_xy_2D;
  } else if(name=="sigma_yy_2d"){
    sf=&CParticle::sigma_yy_2D;
  } else if(name=="sigma_d"){
    sf=&CParticle::sigma_d;
  } else if(name=="e_kin"){
    sf=&CParticle::getKineticEnergy;
  } else if(name=="radius"){
    sf=&CParticle::getRad;
  } else if(name=="v_abs"){
    sf=&CParticle::getAbsVel;
  } else {
    sf=NULL;
    cerr << "ERROR - invalid name for particle scalar  access function" << endl; 
  }

  return sf;
}

/*!
  Get the particle member function which returns a vector field of a given name.

  \param name the name of the field 
*/
CParticle::VectorFieldFunction CParticle::getVectorFieldFunction(const string& name)
{
  CParticle::VectorFieldFunction sf;

  if(name=="displacement"){
    sf=&CParticle::getTotalDisplacement;
  } else if(name=="velocity"){
    sf=&CParticle::getVel;
  } else if(name=="position"){
    sf=&CParticle::getPos;
  } else if(name=="force"){
    sf=&CParticle::getForce;
  } else {
    sf=NULL;
    cerr << "ERROR - invalid name for particle vector access function" << endl; 
  }

  return sf;
}

/*!
  Pack a CParticle into a TML packed message

  \param p the particle
  \todo BasicParticle data should be handled by pack<Basicparticle>
*/
template<>
void TML_PackedMessageInterface::pack<CParticle>(const CParticle& p)
{
  append(p.m_tag);
  append(p.m_pos.X()) ;
  append(p.m_pos.Y()) ;
  append(p.m_pos.Z()) ;
  append(p.m_oldpos.X());
  append(p.m_oldpos.Y());
  append(p.m_oldpos.Z());
  append(p.m_initpos.X());
  append(p.m_initpos.Y());
  append(p.m_initpos.Z());
  append(p.m_circular_shift.X());
  append(p.m_circular_shift.Y());
  append(p.m_circular_shift.Z());
  append(p.m_vel.X());
  append(p.m_vel.Y());
  append(p.m_vel.Z());
  append(p.m_force.X());
  append(p.m_force.Y());
  append(p.m_force.Z());
  append(p.m_rad);
  append(p.m_mass);
  append(p.m_global_id); 
}

/*!
  Unpack a CParticle from a TML packed message

  \param p the particle
*/
template<>
void TML_PackedMessageInterface::unpack<CParticle>(CParticle& p)
{
  double db[6*3+2] ;
 
  p.m_tag=pop_int();
  pop_doubles(db,20);
  p.m_pos=Vec3(db[0],db[1],db[2]) ;
  p.m_oldpos=Vec3(db[3],db[4],db[5]) ;
  p.m_initpos=Vec3(db[6],db[7],db[8]) ;
  p.m_circular_shift=Vec3(db[9],db[10],db[11]) ;
  p.m_vel=Vec3(db[12],db[13],db[14]) ;
  p.m_force=Vec3(db[15],db[16],db[17]) ;
  p.m_rad = db[18];
  p.m_mass = db[19];
  if(p.m_mass!=0.0){
    p.m_div_mass=1.0/p.m_mass;
  } else {
    p.m_div_mass=0.0;
  }
  p.m_global_id=pop_int();
}

/*!
  Do the time integration for the particle by 1st order method. 13 Flops

  \param dt the time step
*/
void CParticle::integrate(double dt)
{
  if (getDo2dCalculations()) {
    m_force = Vec3(m_force.X(), m_force.Y(), 0);
  }
#if 1
  // 1st order scheme
  m_vel+=m_force*m_div_mass*dt;
  m_pos+=m_vel*dt;
#else
  // 2nd order scheme.
  const Vec3  Acc = m_force*m_div_mass ;
  m_pos +=  dt*m_vel + 0.5*dt*dt*Acc ;
  m_vel +=  dt*Acc;
#endif
}

/*!
  zero forces on particle
*/
void CParticle::zeroForce()
{
  m_force=Vec3(0.0,0.0,0.0);
  m_sigma=Matrix3();
}

/*!
  write particle data as a line in openDX format into a stream(file)
  
  \param ost the output stream
  \param slid from which sublattice the particle is coming
*/
void CParticle::writeAsDXLine(ostream& ost,int slid)
{
  ost << m_pos.X() << " " << m_pos.Y() << " " << m_pos.Z() << " " ; //position
  ost << slid << " " ;
  ost << m_rad << " " << m_mass << " " ; // radius,mass
  Vec3 disp=m_pos-m_initpos; 
  ost << disp.X() << " "  << disp.Y() << " "  << disp.Z() << " " ; //displacement
  ost << m_vel.X() << " "  << m_vel.Y() << " "  << m_vel.Z() << " " ; // velocity
  ost << endl;
}

/*!
  get values to be exchanged, i.e. pos and vel
*/
CParticle::exchangeType CParticle::getExchangeValues()
{
  return 
    exchangeType(
      m_pos     - m_circular_shift,
      m_initpos - m_circular_shift,
      m_oldpos  - m_circular_shift,
      m_vel
    );
}

/*!
  set pos and vel from exchangeType

  \param E the exchanged values, i.e. a pair of pos and vel 
*/
void CParticle::setExchangeValues(const exchangeType& E)
{
  m_pos     = E.m_pos     + m_circular_shift;
  m_initpos = E.m_initPos + m_circular_shift;
  m_oldpos  = E.m_oldPos  + m_circular_shift;
  m_vel     = E.m_vel;
}

/*!
  set circular shift vector

  \param cv the circular shift vector
*/
void CParticle::setCircular(const Vec3& cv)
{
  m_circular_shift += cv;
  m_pos            += cv;
  m_initpos        += cv;
  m_oldpos         += cv;
}

/*!
  Pack an exchangeType into a TML packed message

  \param p the exchangeType
*/
template<>
void TML_PackedMessageInterface::pack<CParticle::exchangeType>(const CParticle::exchangeType& p)
{
  append(p.m_pos.X());
  append(p.m_pos.Y());
  append(p.m_pos.Z());
  append(p.m_initPos.X());
  append(p.m_initPos.Y());
  append(p.m_initPos.Z());
  append(p.m_oldPos.X());
  append(p.m_oldPos.Y());
  append(p.m_oldPos.Z());
  append(p.m_vel.X());
  append(p.m_vel.Y());
  append(p.m_vel.Z());
}

/*!
  Unpack an exchangeType from a TML packed message
*/
template<>
void TML_PackedMessageInterface::unpack<CParticle::exchangeType>(CParticle::exchangeType& p)
{
  const int numDoubles = 12;
  double db[numDoubles];

  pop_doubles(db, numDoubles);
  p.m_pos     = Vec3(db[0],  db[1],  db[2]);
  p.m_initPos = Vec3(db[3],  db[4],  db[5]);
  p.m_oldPos  = Vec3(db[6],  db[7],  db[8]);
  p.m_vel     = Vec3(db[9], db[10], db[11]);
}

/*!
  extract particle from transport buffer

  \param buffer a pointer to the transport buffer
  \warning no typechecking, i.e. the packed data is extracted from the buffer and interpreted as CParticle, no matter what actually has be put into the buffer (change?)
*/  
CParticle extractCParticleFrom(AMPIBuffer* buffer)
{
  Vec3 pos=buffer->pop_vector() ;
  Vec3 oldpos=buffer->pop_vector() ;
  Vec3 initpos=buffer->pop_vector() ;
  Vec3 vel=buffer->pop_vector() ;
  Vec3 force=buffer->pop_vector() ;
  double rad=buffer->pop_double();
  double mass=buffer->pop_double();
  int id=buffer->pop_int();
  return CParticle(rad,mass,pos,oldpos,initpos,vel,force,id);
}


/*!
  extract particle from a scatter/gather buffer

  \param buffer a pointer to the buffer
  \param slice from which slice of the buffer to extract
  \warning no typechecking, i.e. the packed data is extracted from the buffer and interpreted as CParticle, no matter what actually has be put into the buffer (change?)
*/  
CParticle extractCParticleFrom(AMPISGBufferRoot* buffer,int slice)
{
  Vec3 pos=buffer->pop_vector(slice);
  Vec3 oldpos=buffer->pop_vector(slice);
  Vec3 initpos=buffer->pop_vector(slice);
  Vec3 vel=buffer->pop_vector(slice);
  Vec3 force=buffer->pop_vector(slice);
  double rad=buffer->pop_double(slice);
  double mass=buffer->pop_double(slice);
  int id=buffer->pop_int(slice);
  return CParticle(rad,mass,pos,oldpos,initpos,vel,force,id);
}

/*!
  Apply a force to a particle at a given position. Also updates particle stress tensor. 

  \param force the force
  \param pos the position at which the force acts on the particle
*/
void CParticle::applyForce(const Vec3& force,const Vec3& pos)
{
  m_force+=force;
  Vec3 rel_pos=pos-m_pos;
  for(int i=0;i<3;i++){
    for(int j=0;j<3;j++){
      m_sigma(i,j)-=force[i]*rel_pos[j];
    }
  }
}

/*!
  get deviatoric stress
*/
double CParticle::sigma_d() const
{
  double scale=1.0/(M_PI*m_rad*m_rad);
  
  double sig_d=(m_sigma-m_sigma.trace()*Matrix3::Unit()).norm();

  return sig_d;
}


map<string,AField*> CParticle::generateFields(ParallelParticleArray<CParticle>* particles)
{
  map<string,AField*> res;

  return res;
}

ostream& operator<<(ostream& ost, const CParticle& CP)
{
  ost << "--CParticle " << CP.m_global_id << "  --\n";
  ost << "Radius : " << CP.m_rad << " Mass : " << CP.m_mass << "\n"; 
  ost << "Position : " << CP.m_pos << "\n";
  ost << "Velocity : " << CP.m_vel << "\n";
  ost << "Force    : " << CP.m_force << "\n";
  return ost;
}

void CParticle::saveCheckPointData( std::ostream& oStream){
  const char delim = ' ';
  oStream 
    << getPos()                        << delim
    << getRad()                        << delim
    << getID()                         << delim
    << getTag()                        << delim
    << getMass()                       << delim
    << getInitPos() - m_circular_shift << delim
    << getOldPos()                     << delim
    << getVel()                        << delim
    << getForce();
}

void CParticle::loadCheckPointData(std::istream &iStream)
{
  Vec3 pos;
  Vec3 oldpos;
  Vec3 initpos;
  Vec3 vel;
  Vec3 force;
  Vec3 circular_shift;
  double rad;
  double mass;
  int id;
  int tag;
  
  iStream
    >> pos
    >> rad
    >> id
    >> tag
    >> mass
    >> initpos
    >> oldpos
    >> vel
    >> force;
  
  CParticle 
    particle
    (
      rad,
      mass,
      pos,
      oldpos,
      initpos + circular_shift,
      vel,
      force,
      id
    );
  particle.setTag(tag);
  
  *this = particle;
}
