#include "RotElasticInteraction.h"
#include "console.h"

CRotElasticInteraction::CRotElasticInteraction(CRotParticle* p1,CRotParticle* p2,double k):ARotPairInteraction(p1,p2)
{
  m_kr=k;
}

CRotElasticInteraction::CRotElasticInteraction(CRotParticle* p1,CRotParticle* p2,const CRotElasticIGP& param):ARotPairInteraction(p1,p2)
{
  m_kr=param.m_kr;
}

Vec3 CRotElasticInteraction::getForce() const
{
  Vec3 force = Vec3(0.0, 0.0, 0.0);
  Vec3 D = m_p1->getPos()-m_p2->getPos();
  double dist = D*D;
  double eq_dist = m_p1->getRad()+m_p2->getRad();
  if (dist < (eq_dist*eq_dist)) {
    dist = sqrt(dist);
    force = D*(m_kr*(dist - eq_dist)/dist);
  }
  return force;
}

/*!
  Calculate free elastic forces. 23 Flops if in contact, 10 Flops if not
 */
void CRotElasticInteraction::calcForces()
{ 
  //  console.XDebug() << "elastic interaction: [" << m_p1->getID() << " - " << m_p2->getID() << "]" << m_p1->getPos() << m_p2->getPos() << "\n"; 
  Vec3 D=m_p1->getPos()-m_p2->getPos();
  double dist=D*D;
  double eq_dist=m_p1->getRad()+m_p2->getRad();
  if(dist<(eq_dist*eq_dist)){
    dist=sqrt(dist);
    Vec3 force=D*(m_kr*(dist-eq_dist)/dist);
    Vec3 pos=m_p2->getPos()+(m_p2->getRad()/eq_dist)*D;
    m_p2->applyForce(force,pos);
    m_p1->applyForce(-1.0*force,pos);

  }
}

/*!
  get the potential energy stored in the interaction
*/
double CRotElasticInteraction::getPotentialEnergy() const
{
  const Vec3 normalForce = getForce();
  return (0.5*normalForce*normalForce/m_kr);
}

ostream& operator<<(ostream& ost,const CRotElasticInteraction& BI)
{
  ost << "[" << BI.m_p1->getID() << " - " << BI.m_p2->getID() << "]";
  return ost;
}


/*!
  Get the particle member function which returns a scalar field of a given name.
 
  \param name the name of the field
*/
CRotElasticInteraction::ScalarFieldFunction CRotElasticInteraction::getScalarFieldFunction(const string& name)
{
  CRotElasticInteraction::ScalarFieldFunction sf;
  
  if (name=="potential_energy")
  {
    sf=&CRotElasticInteraction::getPotentialEnergy;
  }
  
  return sf;
}

/*!
  Get the particle member function which returns a checked scalar field of a given name.

  \param name the name of the field
*/
CRotElasticInteraction::CheckedScalarFieldFunction CRotElasticInteraction::getCheckedScalarFieldFunction(const string& name)
{
  CRotElasticInteraction::CheckedScalarFieldFunction sf;

  sf=NULL;
  cerr << "ERROR - invalid name for interaction vector access function" << endl;

  return sf;
}


/*!
  Get the particle member function which returns a vector field of a given name.
 
  \param name the name of the field
*/
CRotElasticInteraction::VectorFieldFunction CRotElasticInteraction::getVectorFieldFunction(const string& name)
{
  CRotElasticInteraction::VectorFieldFunction vf;
  
  vf=NULL;
  cerr << "ERROR - invalid name for interaction vector access function" << endl;
  
  return vf;
}
