
#include "Model/RotFricInteraction.h"
#include "Foundation/console.h"

// --- TML includes ---
#include "tml/message/packed_message_interface.h"

extern double calc_angle(double , double);

CRotFrictionInteraction::CRotFrictionInteraction():ARotPairInteraction() 
{
  m_k=0.0;
  m_mu_d=0.0;
  m_mu_s=0.0;
  m_r0=0.0;
  m_ks=0.0;
  m_dt=0.0;
  m_is_slipping=false;
  m_is_touching=false;
 
  m_p1=NULL;
  m_p2=NULL;
  m_id.push_back(-1);
  m_id.push_back(-1);

}

/*!
  constructor for CFrictionInteraction without friction parameters, only calls
  the constructor of APairInteraction with the 2 particle pointers
*/
CRotFrictionInteraction::CRotFrictionInteraction(CRotParticle* p1,CRotParticle* p2):ARotPairInteraction(p1,p2)
{
  m_is_slipping=false;
  m_is_touching=false;
 
  m_Ffric = Vec3(0.0,0.0,0.0);
}


CRotFrictionInteraction::CRotFrictionInteraction(CRotParticle* p1,CRotParticle* p2,const CRotFrictionIGP& param):ARotPairInteraction(p1,p2)
{
  m_k=param.k;
  m_mu_d = param.mu_d;
  m_mu_s = param.mu_s;
  m_ks=param.k_s;
  m_r0=p1->getRad()+p2->getRad();
  m_dt=param.dt;
  m_cpos=p1->getPos()+((p2->getPos()-p1->getPos())*p1->getRad()/m_r0);
  m_is_slipping=false;
  m_is_touching=false;
  
  m_Ffric = Vec3(0.0,0.0,0.0);
}

CRotFrictionInteraction::~CRotFrictionInteraction()
{
}


/*!
  Calculate elastic and frictional forces. 
*/
void CRotFrictionInteraction::calcForces()
{
  // calculate distance
  const Vec3 D=m_p1->getPos()-m_p2->getPos();
  double dist=D*D;
  const double eq_dist=m_p1->getRad()+m_p2->getRad();
  // check if there is contact
  if(dist<(eq_dist*eq_dist)){ // contact -> calculate forces
    //--- elastic force ---
    dist=sqrt(dist);
    const Vec3 force(D*(m_k*(dist-eq_dist)/dist));
    m_normal_force=force;
    const Vec3 pos(m_p2->getPos()+(m_p2->getRad()/eq_dist)*D);

    // apply elastic force 
    m_p2->applyForce(force,pos);
    m_p1->applyForce(-1.0*force,pos); 

    //--- frictional force ---
    // Calculate the relative linear-displacement of point
    // pos on each of the particles.
    const Vec3 vp1 = m_p1->getVel() + cross(m_p1->getAngVel(), pos-m_p1->getPos());
    const Vec3 vp2 = m_p2->getVel() + cross(m_p2->getAngVel(), pos-m_p2->getPos());
    Vec3 ds((vp2-vp1)*m_dt);

    // Get the tangential component by subtracting off the normal component.
    ds -= ((ds*D)/(D.norm2()))*D;

    m_Ffric += m_ks*ds;
    const double FfricNorm = m_Ffric.norm();    
    const double forceNorm = force.norm();
    if (FfricNorm > forceNorm*m_mu_s) {
      m_is_slipping = true;
      m_Ffric = m_Ffric*((m_mu_d*forceNorm)/FfricNorm);
    } else {
      m_is_slipping = false;
    }

    const Vec3 Moment1(cross(pos-m_p1->getPos(),  m_Ffric));
    const Vec3 Moment2(cross(pos-m_p2->getPos(), -m_Ffric));
    m_p1->applyMoment(Moment1);
    m_p2->applyMoment(Moment2);

    m_p1->applyForce(m_Ffric,pos);
    m_p2->applyForce(-1.0*m_Ffric,pos);
    m_cpos=pos;
    m_is_touching=true;
  } else { // no contact -> all forces are 0
    m_Ffric=Vec3(0.0,0.0,0.0);
    m_force_deficit=Vec3(0.0,0.0,0.0);
    m_normal_force=Vec3(0.0,0.0,0.0);
    m_is_slipping=false;
    m_is_touching=false;
  }
}

bool CRotFrictionInteraction::isPersistent()
{
  Vec3 D=m_p1->getPos()-m_p2->getPos();
  double dist=D*D;
  double eq_dist=m_p1->getRad()+m_p2->getRad();
  return dist<=(eq_dist*eq_dist);
}

/*!
  get the force needed to overcome friction and make the interaction slip
*/
double CRotFrictionInteraction::getAbsForceDeficit()const
{
  return m_force_deficit.norm();
}

/*!
 Calculate the normal force.
*/
void CRotFrictionInteraction::calcNormalForce()
{
  Vec3 pos;
  // calculate distance
  const Vec3 D=m_p1->getPos()-m_p2->getPos();
  double dist=D*D;
  const double eq_dist=m_p1->getRad()+m_p2->getRad();
  // check if there is contact
  if(dist<(eq_dist*eq_dist)){ // contact -> calculate forces
    //--- elastic force ---
    dist=sqrt(dist);
    m_normal_force=D*(m_k*(dist-eq_dist)/dist);
  }
}

/*!
  get the potential energy stored in the interaction
*/
double CRotFrictionInteraction::getPotentialEnergy() const
{
  double e_pot_norm=0.5*m_normal_force*m_normal_force/m_k;

  return e_pot_norm;
}

/*!
  Get the static/dynamic status of the interaction. Returns 1 for a contact in dynamic 
  friction, 0 for static or no contact
*/
double CRotFrictionInteraction::getSlipping() const
{
  double res=m_is_slipping ? 1.0 : 0.0;
  return res;
}

/*!
  Get the contact status of the interaction. Returns 1 for an actual contact, 
  0 for no contact
*/
double CRotFrictionInteraction::getTouching() const
{
  double res=m_is_touching ? 1.0 : 0.0;
  return res;
}

Vec3 CRotFrictionInteraction::getForce() const
{
  const Vec3 f=m_is_touching ? m_Ffric+m_normal_force : Vec3(0.0,0.0,0.0);
  return f; 
}

/*!
  Get the particle member function which returns a scalar field of a given name.

  \param name the name of the field 
*/
CRotFrictionInteraction::ScalarFieldFunction CRotFrictionInteraction::getScalarFieldFunction(const string& name)
{
  CRotFrictionInteraction::ScalarFieldFunction sf;

  if(name=="force_deficit"){
    sf=&CRotFrictionInteraction::getAbsForceDeficit;
  } else if (name=="potential_energy"){
    sf=&CRotFrictionInteraction::getPotentialEnergy;
  } else if (name=="slipping"){
    sf=&CRotFrictionInteraction::getSlipping;
  } else if (name=="touching"){
    sf=&CRotFrictionInteraction::getTouching;
  } else if (name=="count"){
    sf=&CRotFrictionInteraction::Count;
  } else {
    sf=NULL;
    cerr << "ERROR - invalid name for interaction scalar  access function" << endl; 
  }
  
  return sf;
}

CRotFrictionInteraction::CheckedScalarFieldFunction CRotFrictionInteraction::getCheckedScalarFieldFunction(const string& name)
{
  CRotFrictionInteraction::CheckedScalarFieldFunction sf = NULL;
  cerr << "ERROR - invalid name for interaction scalar  access function" << endl; 
  return sf;
}

/*!
  Get the particle member function which returns a vector field of a given name.

  \param name the name of the field 
*/
CRotFrictionInteraction::VectorFieldFunction CRotFrictionInteraction::getVectorFieldFunction(const string& name)
{
  CRotFrictionInteraction::VectorFieldFunction vf=NULL;

  if (name=="total_force") {
    vf = &CRotFrictionInteraction::getForce;    
  }
  else {
    cerr << "ERROR - invalid name for interaction vector access function" << endl;
  }

  return vf;
}

/*!
  Pack a CFrictionInteraction into a TML packed message

  \param I the interaction
*/
template<>
void TML_PackedMessageInterface::pack<CRotFrictionInteraction>(const CRotFrictionInteraction& I)
{
  append(I.m_k);
  append(I.m_r0);
  append(I.m_mu_d);
  append(I.m_mu_s);
  append(I.m_ks);
  append(I.m_dt);
  append(I.m_id[0]);
  append(I.m_id[1]);
  append(static_cast<int>(I.m_is_slipping));
  append(I.m_Ffric.X());
  append(I.m_Ffric.Y());
  append(I.m_Ffric.Z());
}

/*!
  Unpack a CFrictionInteraction from a TML packed message

  \param I the interaction
*/
template<>
void TML_PackedMessageInterface::unpack<CRotFrictionInteraction>(CRotFrictionInteraction& I)
{
  I.m_k=pop_double();
  I.m_r0=pop_double();
  I.m_mu_d=pop_double();
  I.m_mu_s=pop_double();
  I.m_ks=pop_double();
  I.m_dt=pop_double();
  I.m_id.clear();
  I.m_id.push_back(pop_int());
  I.m_id.push_back(pop_int());
  I.m_is_slipping = static_cast<bool>(pop_int());
  I.m_Ffric.X() = pop_double();
  I.m_Ffric.Y() = pop_double();
  I.m_Ffric.Z() = pop_double();
}

ostream& operator<<(ostream& ost,const CRotFrictionInteraction& BI)
{
  ost << "[" << BI.m_p1->getID() << " - " << BI.m_p2->getID() << "]";
  return ost;
}
