#ifndef __ROTPARTICLE_H
#define __ROTPARTICLE_H

// -- project includes --
#include "Foundation/vec3.h"
#include "Foundation/Matrix3.h"
#include "Model/Particle.h"
#include "Foundation/Quaternion.h"

template <class T> class ParallelParticleArray;
class AMPISGBufferRoot;
class AMPIBuffer;
class AField;

//--- MPIincludes ---
#include <mpi.h>

//--- STL includes ---
#include <map>
#include <vector>
#include <utility>
#include <string>

using std::map;
using std::vector;
using std::pair;
using std::string;

namespace esys
{
   namespace lsm
   {
     class SimpleParticleData;
   }
}

/*!
  \class CRotParticle
  \brief Class for a rotational particle

  \author  
  $Revision$
  $Date$ 
*/
class CRotParticle : public CParticle
{
 public: // types

  class exchangeType
  {
  public:
    exchangeType()
      : m_pos(),
        m_initPos(),
        m_vel(),
        m_angVel(),
        m_quat()
    {
    }

    exchangeType(
      const Vec3 &pos,
      const Vec3 &initPos,
      const Vec3 &vel,
      const Vec3 &currAngVel,
      const Quaternion &quat
    )
      : m_pos(pos),
        m_initPos(initPos),
        m_vel(vel),
        m_angVel(currAngVel),
        m_quat(quat)
    {
    }
  public:
    Vec3       m_pos;
    Vec3       m_initPos;
    Vec3       m_vel;
    Vec3       m_angVel;
    Quaternion m_quat;

    friend class TML_PackedMessageInterface;    
  };
  typedef double (CRotParticle::* ScalarFieldFunction)() const;
  typedef Vec3 (CRotParticle::* VectorFieldFunction)() const;

 protected:
 
  Quaternion  m_quat;
  Quaternion  m_initquat;
  Vec3        m_angVel; //! Angular velocity at time t
  Vec3        m_moment;
  double      m_inertRot;
  double      m_div_inertRot;

 public:
  CRotParticle();
  CRotParticle(const esys::lsm::SimpleParticleData &particleData);
  CRotParticle(
    double      rad,
    double      mass,
    const Vec3& pos,
    const Vec3& vel,
    const Vec3& force,
    int         id
  );
  CRotParticle(
    double      rad,
    double      mass,
    const Vec3& pos,
    const Vec3& vel,
    const Vec3& force,
    int         id,
    Quaternion& quat,
    double      inertRot,
    const Vec3& moment,
    const Vec3& angvel  
  );
  CRotParticle(
    double            rad,
    double            mass,
    const Vec3&       pos,
    const Vec3&       oldpos,
    const Vec3&       initpos,
    const Vec3&       vel,
    const Vec3&       force,
    int               id,
    const Quaternion& quat,
    const Quaternion& initquat,
    double            inertRot,
    const Vec3&       moment,
    const Vec3&       angvel  
  );
  
  virtual ~CRotParticle(){};

  static int getPackSize();
  static ScalarFieldFunction getScalarFieldFunction(const string&);
  static VectorFieldFunction getVectorFieldFunction(const string&);

  // Need to define this for template class using forAllParticles call in Parallel/SubLattice.hpp.
  Vec3 getDisplacement() const {return CParticle::getDisplacement();};
  void resetDisplacement() {CParticle::resetDisplacement();}

  inline const Vec3 &getAngVel () const { return m_angVel;}
  inline void setAngVel(const Vec3 &V) { m_angVel = V;}  
  inline Quaternion getInitQuat() const { return m_initquat;}
  inline Quaternion getQuat() const { return m_quat;}
  inline double getInertRot () const { return m_inertRot; }
  inline Vec3 getMoment()  const {return m_moment;}
  void applyMoment( const Vec3&);
  void integrate(double);
  void zeroForce();
  void rescale();
  void setCircular(const Vec3& cv);
  double getAngularKineticEnergy() const {return 0.2*m_mass*m_rad*m_rad*(m_angVel*m_angVel);}
  double getLinearKineticEnergy() const {return 0.5*m_mass*m_vel*m_vel;}
  double getKineticEnergy() const {return getLinearKineticEnergy() + getAngularKineticEnergy();}
  void writeAsDXLine(ostream&,int slid=0);

  static map<string,AField*> generateFields(ParallelParticleArray<CRotParticle>*);

  friend ostream& operator<<(ostream&, const CRotParticle&);
  void print(){cout << *this << endl << flush;};

  virtual void saveCheckPointData(std::ostream& oStream);
  virtual void loadCheckPointData(std::istream& iStream);

  CRotParticle::exchangeType getExchangeValues();
  void setExchangeValues(const CRotParticle::exchangeType &e);

  // stress 
  inline double sigma_xx_2D() const {return m_sigma(0,0)/(M_PI*m_rad*m_rad);};
  inline double sigma_xy_2D() const {return m_sigma(0,1)/(M_PI*m_rad*m_rad);};
  inline double sigma_yy_2D() const {return m_sigma(1,1)/(M_PI*m_rad*m_rad);};
//  inline double sigma_d() const;
  static void get_type() {cout <<" CRotParticel" ;};
  friend class TML_PackedMessageInterface;
};

#endif //__ROTPARTICLE_H
