
#include "ModelTest/BondedInteractionTest.h"
#include "ModelTest/Integrator.h"

// -- test includes --
#include "Model/Particle.h"
#include "Model/BondedInteraction.h"
#include "Model/SimpleParticleData.h"

#include "UnitTest/Macros.h"
#include <boost/limits.hpp>

using namespace esys::lsm;

namespace LsmTest
{
  CPPUNIT_TEST_SUITE_REGISTRATION( BondedInteractionTest );

  typedef LsmTest::Integrator<CBondedInteraction> BondIntegrator;
  
  class BondedInteractionTest::Impl
  {
  public:
    Impl()
      : m_dt(0.0025),
        m_p1(),
        m_p2(),
        m_prms()
    {
    }

    double       m_dt;
    CParticle    m_p1;
    CParticle    m_p2;
    CBondedIGP   m_prms;
  };

  /*!
    constructor - do nothing
  */
  BondedInteractionTest::BondedInteractionTest()
    : CppUnit::TestCase(),
      m_impl(new Impl())
  {
  }

  /*!
    destructor - do nothing
  */
  BondedInteractionTest::~BondedInteractionTest()
  {
  }

  /*!
   */
  void BondedInteractionTest::setUp()
  {
    CppUnit::TestCase::setUp();
    m_impl->m_prms.k      =  1.00;
    m_impl->m_prms.rbreak = 10.00;
    m_impl->m_prms.tag    = 0;

    SimpleParticleData sp1(0, 0, Vec3(0, 0, 0), 1.0);
    SimpleParticleData sp2(1, 0, Vec3(2, 0, 0), 1.0);
    sp1.setMass(1.0);
    sp2.setMass(1.0);

    m_impl->m_p1 = CParticle(sp1);
    m_impl->m_p2 = CParticle(sp2);
  }

  /*!
   */ 
  void BondedInteractionTest::tearDown()
  {
    CppUnit::TestCase::tearDown();
  }

  void BondedInteractionTest::testZeroForce()
  {
    const Vec3 zeroVec(0.0, 0.0, 0.0);
    const double dt = m_impl->m_dt;
    // setup two particles
    CParticle &p1 = m_impl->m_p1;
    CParticle &p2 = m_impl->m_p2;

    Vec3 p1OrigPos = Vec3(0, 0, 0);
    Vec3 p2OrigPos = Vec3(2, 0, 0);
    p1.moveTo(p1OrigPos);
    p2.moveTo(p2OrigPos);

    CBondedInteraction interaction(&p1, &p2, m_impl->m_prms);

    interaction.calcForces();

    CPPUNIT_ASSERT_EQUAL(zeroVec, p1.getForce());
    CPPUNIT_ASSERT_EQUAL(zeroVec, p2.getForce());

    p1.integrate(dt);
    p2.integrate(dt);
    CPPUNIT_ASSERT_EQUAL(p1OrigPos, p1.getPos());
    CPPUNIT_ASSERT_EQUAL(p2OrigPos, p2.getPos());
  }

  void BondedInteractionTest::testElasticOnlyForce()
  {
    const Vec3 zeroVec(0.0, 0.0, 0.0);

    // setup two particles
    CParticle &p1 = m_impl->m_p1;
    CParticle &p2 = m_impl->m_p2;

    Vec3 p1OrigPos = Vec3(0.0, 0.0, 0.0);
    Vec3 p2OrigPos = Vec3(1.8, 0.0, 0.0);
    p1.moveTo(p1OrigPos);
    p2.moveTo(p2OrigPos);

    CBondedInteraction interaction(&p1, &p2, m_impl->m_prms);

    interaction.calcForces();

    LSMUNIT_ASSERT_VEC3_EQUAL(Vec3(-0.2, 0.0, 0.0), p1.getForce());
    LSMUNIT_ASSERT_VEC3_EQUAL(Vec3( 0.2, 0.0, 0.0), p2.getForce());

    p1.integrate(m_impl->m_dt);
    p2.integrate(m_impl->m_dt);
    CPPUNIT_ASSERT(p1OrigPos.X() > p1.getPos().X());
    CPPUNIT_ASSERT(p2OrigPos.X() < p2.getPos().X());
  }

  void BondedInteractionTest::testElasticEnergyConservation()
  {
    // setup two particles
    CParticle &p1 = m_impl->m_p1;
    CParticle &p2 = m_impl->m_p2;

    Vec3 p1OrigPos = Vec3(0.00, 0.0, 0.0);
    Vec3 p2OrigPos = Vec3(1.95, 0.0, 0.0);
    p1.moveTo(p1OrigPos);
    p2.moveTo(p2OrigPos);

    CBondedInteraction interaction(&p1, &p2, m_impl->m_prms);

    BondIntegrator integrator(interaction);

    const double initialEnergy = integrator.getTotalEnergy();

    integrator.integrate(2000, m_impl->m_dt);

    const double energy = integrator.getTotalEnergy();

    // Make sure we are no more than 3% of the initial energy
    LSMUNIT_ASSERT_ENERGY_DISIPATED(initialEnergy, energy, initialEnergy*0.03);

    //
    // Check positions, linear and angular velocity as a regression test
    //
    LSMUNIT_ASSERT_VEC3_EQUAL(Vec3(-0.0073976953111071176, 0, 0), p1.getPos());
    LSMUNIT_ASSERT_VEC3_EQUAL(Vec3(0.025062162320759734, 0, 0), p2.getVel());
  }

  void BondedInteractionTest::testElasticAndLinearEnergyConservation()
  {
    // setup two particles
    CParticle &p1 = m_impl->m_p1;
    CParticle &p2 = m_impl->m_p2;

    Vec3 p1OrigPos = Vec3(0.00, 0.0, 0.0);
    Vec3 p2OrigPos = Vec3(1.95, 0.0, 0.0);
    p1.moveTo(p1OrigPos);
    p2.moveTo(p2OrigPos);
    p2.setVel(Vec3(-0.01, 0.01, -0.01));

    CBondedInteraction interaction(&p1, &p2, m_impl->m_prms);

    BondIntegrator integrator(interaction);

    const double initialEnergy = integrator.getTotalEnergy();

    integrator.integrate(2000, m_impl->m_dt);

    const double energy = integrator.getTotalEnergy();
    
    // Make sure we are no more than 3% of the initial energy
    LSMUNIT_ASSERT_ENERGY_DISIPATED(initialEnergy, energy, initialEnergy*0.03);
    
  }
};
