
#include "ModelTest/RotBondedInteractionTest.h"
#include "ModelTest/Integrator.h"

// -- test includes --
#include "Model/RotParticle.h"
#include "Model/RotBondedInteraction.h"
#include "Model/SimpleParticleData.h"

#include "UnitTest/Macros.h"
#include <boost/limits.hpp>

using namespace esys::lsm;

namespace LsmTest
{
  CPPUNIT_TEST_SUITE_REGISTRATION( RotBondedInteractionTest );

  typedef LsmTest::Integrator<CRotBondedInteraction> BondIntegrator;
  
  class RotBondedInteractionTest::Impl
  {
  public:
    Impl()
      : m_dt(0.005),
        m_p1(),
        m_p2(),
        m_prms()
    {
    }

    double          m_dt;
    CRotParticle    m_p1;
    CRotParticle    m_p2;
    CRotBondedIGP m_prms;
  };

  /*!
    constructor - do nothing
  */
  RotBondedInteractionTest::RotBondedInteractionTest()
    : CppUnit::TestCase(),
      m_impl(new Impl())
  {
  }

  /*!
    destructor - do nothing
  */
  RotBondedInteractionTest::~RotBondedInteractionTest()
  {
  }

  /*!
   */
  void RotBondedInteractionTest::setUp()
  {
    CppUnit::TestCase::setUp();
    m_impl->m_prms.kr = 1.00;
    m_impl->m_prms.ks = 1.00;
    m_impl->m_prms.kt = 1.00;
    m_impl->m_prms.kb = 1.00;
    m_impl->m_prms.max_nForce =  1000000.0;
    m_impl->m_prms.max_shFroce = 1000000.0;
    m_impl->m_prms.max_tMoment = 1000000.0;
    m_impl->m_prms.max_bMoment = 1000000.0;

    SimpleParticleData sp1(0, 0, Vec3(0, 0, 0), 1.0);
    SimpleParticleData sp2(1, 0, Vec3(2, 0, 0), 1.0);
    sp1.setMass(1.0);
    sp2.setMass(1.0);

    m_impl->m_p1 = CRotParticle(sp1);
    m_impl->m_p2 = CRotParticle(sp2);
  }

  /*!
   */ 
  void RotBondedInteractionTest::tearDown()
  {
    CppUnit::TestCase::tearDown();
  }

  void RotBondedInteractionTest::testZeroForce()
  {
    const Vec3 zeroVec(0.0, 0.0, 0.0);
    const double dt = m_impl->m_dt;
    // setup two particles
    CRotParticle &p1 = m_impl->m_p1;
    CRotParticle &p2 = m_impl->m_p2;

    Vec3 p1OrigPos = Vec3(0, 0, 0);
    Vec3 p2OrigPos = Vec3(2, 0, 0);
    p1.moveTo(p1OrigPos);
    p2.moveTo(p2OrigPos);

    CRotBondedInteraction interaction(&p1, &p2, m_impl->m_prms);

    interaction.calcForces();

    CPPUNIT_ASSERT_EQUAL(zeroVec, p1.getForce());
    CPPUNIT_ASSERT_EQUAL(zeroVec, p2.getForce());
    CPPUNIT_ASSERT_EQUAL(zeroVec, p1.getMoment());
    CPPUNIT_ASSERT_EQUAL(zeroVec, p2.getMoment());

    p1.integrate(dt);
    p2.integrate(dt);
    CPPUNIT_ASSERT_EQUAL(p1OrigPos, p1.getPos());
    CPPUNIT_ASSERT_EQUAL(p2OrigPos, p2.getPos());
    CPPUNIT_ASSERT_EQUAL(zeroVec, p1.getAngVel());
    CPPUNIT_ASSERT_EQUAL(zeroVec, p2.getAngVel());
  }

  void RotBondedInteractionTest::testElasticOnlyForce()
  {
    const Vec3 zeroVec(0.0, 0.0, 0.0);

    // setup two particles
    CRotParticle &p1 = m_impl->m_p1;
    CRotParticle &p2 = m_impl->m_p2;

    Vec3 p1OrigPos = Vec3(0.0, 0.0, 0.0);
    Vec3 p2OrigPos = Vec3(1.8, 0.0, 0.0);
    p1.moveTo(p1OrigPos);
    p2.moveTo(p2OrigPos);

    CRotBondedInteraction interaction(&p1, &p2, m_impl->m_prms);

    interaction.calcForces();

    LSMUNIT_ASSERT_VEC3_EQUAL(Vec3(-0.2, 0.0, 0.0), p1.getForce());
    LSMUNIT_ASSERT_VEC3_EQUAL(Vec3( 0.2, 0.0, 0.0), p2.getForce());
    CPPUNIT_ASSERT_EQUAL(zeroVec, p1.getMoment());
    CPPUNIT_ASSERT_EQUAL(zeroVec, p2.getMoment());

    p1.integrate(m_impl->m_dt);
    p2.integrate(m_impl->m_dt);
    CPPUNIT_ASSERT_EQUAL(zeroVec, p1.getAngVel());
    CPPUNIT_ASSERT_EQUAL(zeroVec, p2.getAngVel());
    CPPUNIT_ASSERT(p1OrigPos.X() > p1.getPos().X());
    CPPUNIT_ASSERT(p2OrigPos.X() < p2.getPos().X());
  }
  
  void RotBondedInteractionTest::testElasticEnergyConservation()
  {
    // setup two particles
    CRotParticle &p1 = m_impl->m_p1;
    CRotParticle &p2 = m_impl->m_p2;

    Vec3 p1OrigPos = Vec3(0.00, 0.0, 0.0);
    Vec3 p2OrigPos = Vec3(1.95, 0.0, 0.0);
    p1.moveTo(p1OrigPos);
    p2.moveTo(p2OrigPos);

    CRotBondedInteraction interaction(&p1, &p2, m_impl->m_prms);

    BondIntegrator integrator(interaction);

    const double initialEnergy = integrator.getTotalEnergy();

    integrator.integrate(int(nearbyint(1.0/m_impl->m_dt)), m_impl->m_dt);

    const double energy = integrator.getTotalEnergy();
    
    // Make sure error is no more than 2% of the initial energy
    CPPUNIT_ASSERT_DOUBLES_EQUAL(initialEnergy, energy, initialEnergy*0.02);
  }

  void RotBondedInteractionTest::testElasticAndAngularEnergyConservation()
  {
    // setup two particles
    CRotParticle &p1 = m_impl->m_p1;
    CRotParticle &p2 = m_impl->m_p2;

    Vec3 p1OrigPos = Vec3(0.00, 0.0, 0.0);
    Vec3 p2OrigPos = Vec3(1.95, 0.0, 0.0);
    p1.moveTo(p1OrigPos);
    p2.moveTo(p2OrigPos);
    p2.setAngVel(Vec3(0, 0, 0.05));

    CRotBondedInteraction interaction(&p1, &p2, m_impl->m_prms);

    BondIntegrator integrator(interaction);

    const double initialEnergy = integrator.getTotalEnergy();

    integrator.integrate(int(nearbyint(1.0/m_impl->m_dt)), m_impl->m_dt);

    const double energy = integrator.getTotalEnergy();
    
    // Make sure error is no more than 2% of the initial energy
    CPPUNIT_ASSERT_DOUBLES_EQUAL(initialEnergy, energy, initialEnergy*0.02);
  }

  void RotBondedInteractionTest::testElasticAngularAndLinearEnergyConservation()
  {
    // setup two particles
    CRotParticle &p1 = m_impl->m_p1;
    CRotParticle &p2 = m_impl->m_p2;

    Vec3 p1OrigPos = Vec3(0.00, 0.0, 0.0);
    Vec3 p2OrigPos = Vec3(1.95, 0.0, 0.0);
    p1.moveTo(p1OrigPos);
    p2.moveTo(p2OrigPos);
    p2.setAngVel(Vec3(0, 0, 0.05));
    p2.setVel(Vec3(-0.01, 0.01, -0.01));

    CRotBondedInteraction interaction(&p1, &p2, m_impl->m_prms);

    BondIntegrator integrator(interaction);

    const double initialEnergy = integrator.getTotalEnergy();

    integrator.integrate(int(nearbyint(1.0/m_impl->m_dt)), m_impl->m_dt);

    const double energy = integrator.getTotalEnergy();
    
    // Make sure error is no more than 2% of the initial energy
    CPPUNIT_ASSERT_DOUBLES_EQUAL(initialEnergy, energy, initialEnergy*0.02);
  }
};
