
#include "ModelTest/RotFrictionInteractionTest.h"
#include "ModelTest/Integrator.h"

// -- test includes --
#include "Model/RotParticle.h"
#include "Model/RotFricInteraction.h"
#include "Model/SimpleParticleData.h"

#include "UnitTest/Macros.h"
#include <boost/limits.hpp>

using namespace esys::lsm;

namespace LsmTest
{
  const double TOL = numeric_limits<double>::epsilon()*1024;
  
  typedef Integrator<CRotFrictionInteraction> InteractionIntegrator;
  
  CPPUNIT_TEST_SUITE_REGISTRATION( RotFrictionInteractionTest );

  class RotFrictionInteractionTest::Impl
  {
  public:
    Impl()
      : m_dt(0.01),
        m_p1(),
        m_p2(),
        m_prms()
    {
    }

    double          m_dt;
    CRotParticle    m_p1;
    CRotParticle    m_p2;
    CRotFrictionIGP m_prms;
  };

  /*!
    constructor - do nothing
  */
  RotFrictionInteractionTest::RotFrictionInteractionTest()
    : CppUnit::TestCase(),
      m_impl(new Impl())
  {
  }

  /*!
    destructor - do nothing
  */
  RotFrictionInteractionTest::~RotFrictionInteractionTest()
  {
  }

  /*!
   */
  void RotFrictionInteractionTest::setUp()
  {
    CppUnit::TestCase::setUp();
    m_impl->m_prms.k    = 1.0;
    m_impl->m_prms.mu_d = 0.2;
    m_impl->m_prms.mu_s = 0.2;
    m_impl->m_prms.k_s  = 0.5;
    m_impl->m_prms.dt   = 0.01;

    SimpleParticleData sp1(0, 0, Vec3(0, 0, 0), 1.0);
    SimpleParticleData sp2(1, 0, Vec3(2, 0, 0), 1.0);
    sp1.setMass(1.0);
    sp2.setMass(1.0);

    m_impl->m_p1 = CRotParticle(sp1);
    m_impl->m_p2 = CRotParticle(sp2);
  }

  /*!
   */ 
  void RotFrictionInteractionTest::tearDown()
  {
    CppUnit::TestCase::tearDown();
  }

  void RotFrictionInteractionTest::testZeroForce()
  {
    const Vec3 zeroVec(0.0, 0.0, 0.0);
    const double dt = m_impl->m_dt;
    // setup two particles
    CRotParticle &p1 = m_impl->m_p1;
    CRotParticle &p2 = m_impl->m_p2;

    Vec3 p1OrigPos = Vec3(0, 0, 0);
    Vec3 p2OrigPos = Vec3(2, 0, 0);
    p1.moveTo(p1OrigPos);
    p2.moveTo(p2OrigPos);

    CRotFrictionInteraction interaction(&p1, &p2, m_impl->m_prms);

    interaction.calcForces();

    CPPUNIT_ASSERT_EQUAL(zeroVec, p1.getForce());
    CPPUNIT_ASSERT_EQUAL(zeroVec, p2.getForce());
    CPPUNIT_ASSERT_EQUAL(zeroVec, p1.getMoment());
    CPPUNIT_ASSERT_EQUAL(zeroVec, p2.getMoment());

    p1.integrate(dt);
    p2.integrate(dt);
    CPPUNIT_ASSERT_EQUAL(p1OrigPos, p1.getPos());
    CPPUNIT_ASSERT_EQUAL(p2OrigPos, p2.getPos());
    CPPUNIT_ASSERT_EQUAL(zeroVec, p1.getAngVel());
    CPPUNIT_ASSERT_EQUAL(zeroVec, p2.getAngVel());
  }

  void RotFrictionInteractionTest::testElasticOnlyForce()
  {
    const Vec3 zeroVec(0.0, 0.0, 0.0);

    // setup two particles
    CRotParticle &p1 = m_impl->m_p1;
    CRotParticle &p2 = m_impl->m_p2;

    Vec3 p1OrigPos = Vec3(0.0, 0.0, 0.0);
    Vec3 p2OrigPos = Vec3(1.8, 0.0, 0.0);
    p1.moveTo(p1OrigPos);
    p2.moveTo(p2OrigPos);

    CRotFrictionInteraction interaction(&p1, &p2, m_impl->m_prms);

    interaction.calcForces();

    LSMUNIT_ASSERT_VEC3_EQUAL(Vec3(-0.2, 0.0, 0.0), p1.getForce());
    LSMUNIT_ASSERT_VEC3_EQUAL(Vec3( 0.2, 0.0, 0.0), p2.getForce());
    CPPUNIT_ASSERT_EQUAL(zeroVec, p1.getMoment());
    CPPUNIT_ASSERT_EQUAL(zeroVec, p2.getMoment());

    p1.integrate(m_impl->m_dt);
    p2.integrate(m_impl->m_dt);
    CPPUNIT_ASSERT_EQUAL(zeroVec, p1.getAngVel());
    CPPUNIT_ASSERT_EQUAL(zeroVec, p2.getAngVel());
    CPPUNIT_ASSERT(p1OrigPos.X() > p1.getPos().X());
    CPPUNIT_ASSERT(p2OrigPos.X() < p2.getPos().X());
  }

  void RotFrictionInteractionTest::testShearForce()
  {
    const Vec3 zeroVec(0.0, 0.0, 0.0);

    // setup two particles
    CRotParticle &p1 = m_impl->m_p1;
    CRotParticle &p2 = m_impl->m_p2;

    Vec3 p1OrigPos = Vec3(0.0, 0.0, 0.0);
    Vec3 p2OrigPos = Vec3(1.8, 0.2, 0.0);
    p1.moveTo(p1OrigPos);
    p2.moveTo(p2OrigPos);
    p2.setVel(0.01*(p2.getPPos() - p2.getInitPos()));

    CRotFrictionInteraction interaction(&p1, &p2, m_impl->m_prms);

    interaction.calcForces();

    CPPUNIT_ASSERT(p1.getForce().X() <  0.0);
    CPPUNIT_ASSERT(p1.getForce().X() > -0.2);
    CPPUNIT_ASSERT(p1.getForce().Y() <  0.0);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(p1.getForce().Z(), 0.0, TOL);
    CPPUNIT_ASSERT(p2.getForce().X() >  0.0);
    CPPUNIT_ASSERT(p2.getForce().X() <  0.2);
    CPPUNIT_ASSERT(p2.getForce().Y() >  0.0);
    CPPUNIT_ASSERT(p2.getForce().Z() == 0.0);

    CPPUNIT_ASSERT(p1.getMoment().X() == 0.0);
    CPPUNIT_ASSERT(p1.getMoment().Y() == 0.0);
    CPPUNIT_ASSERT(p1.getMoment().Z() >  0.0);

    p1.integrate(m_impl->m_dt);
    p2.integrate(m_impl->m_dt);
    CPPUNIT_ASSERT_EQUAL(0.0, p1.getAngVel().X());
    CPPUNIT_ASSERT_EQUAL(0.0, p1.getAngVel().Y());
    CPPUNIT_ASSERT(0.0 < p1.getAngVel().Z());
    CPPUNIT_ASSERT_EQUAL(0.0, p2.getAngVel().X());
    CPPUNIT_ASSERT_EQUAL(0.0, p2.getAngVel().Y());
    CPPUNIT_ASSERT(0.0 < p2.getAngVel().Z());
  }

  void RotFrictionInteractionTest::testShearForceWithVelocity()
  {
    const Vec3 zeroVec(0.0, 0.0, 0.0);

    // setup two particles
    CRotParticle &p1 = m_impl->m_p1;
    CRotParticle &p2 = m_impl->m_p2;

    Vec3 p1OrigPos = Vec3(0.0, 0.0, 0.0);
    Vec3 p2OrigPos = Vec3(2.0, 0.0, 0.0);
    p1.moveTo(p1OrigPos);
    p2.moveTo(p2OrigPos);
    p2.setVel(0.1*(Vec3(1.8, 0.2, 0.0)-p2.getPos()));
    
    CRotFrictionInteraction interaction(&p1, &p2, m_impl->m_prms);

    InteractionIntegrator integrator(interaction);

    const double initialEnergy = integrator.getTotalEnergy();
    integrator.integrate(100, m_impl->m_dt);
    const double energy = integrator.getTotalEnergy();

    LSMUNIT_ASSERT_ENERGY_DISIPATED(initialEnergy, energy, 0);
    interaction.calcForces();

    CPPUNIT_ASSERT(p1.getForce().X() <  0.0);
    CPPUNIT_ASSERT(p1.getForce().X() > -0.2);
    CPPUNIT_ASSERT(p1.getForce().Y() >  0.0);
    CPPUNIT_ASSERT(p1.getForce().Z() == 0.0);
    CPPUNIT_ASSERT(p2.getForce().X() >  0.0);
    CPPUNIT_ASSERT(p2.getForce().X() <  0.2);
    CPPUNIT_ASSERT(p2.getForce().Y() <  0.0);
    CPPUNIT_ASSERT(p2.getForce().Z() == 0.0);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(p1.getMoment().X(), 0.0, TOL);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(p1.getMoment().Y(), 0.0, TOL);
    CPPUNIT_ASSERT(p1.getMoment().Z() >  0.0);

    p1.integrate(m_impl->m_dt);
    p2.integrate(m_impl->m_dt);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, p1.getAngVel().X(), TOL);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, p1.getAngVel().Y(), TOL);
    CPPUNIT_ASSERT(0.0 < p1.getAngVel().Z());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, p2.getAngVel().X(), TOL);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, p2.getAngVel().Y(), TOL);
    CPPUNIT_ASSERT(0.0 < p2.getAngVel().Z());
  }

  void RotFrictionInteractionTest::testElasticPlusMoment()
  {
    const Vec3 zeroVec(0.0, 0.0, 0.0);

    // setup two particles
    CRotParticle &p1 = m_impl->m_p1;
    CRotParticle &p2 = m_impl->m_p2;

    Vec3 p1OrigPos = Vec3(0.00, 0.00, 0.0);
    Vec3 p2OrigPos = Vec3(1.95, 0.00, 0.0);
    p1.moveTo(p1OrigPos);
    p2.moveTo(p2OrigPos);
    p1.setAngVel(Vec3(0.0, 0.0, -0.0101));
    p2.setAngVel(Vec3(0.0, 0.0,  0.0099));

    CRotFrictionInteraction interaction(&p1, &p2, m_impl->m_prms);
    InteractionIntegrator integrator(interaction);
    
    const double initialEnergy = integrator.getTotalEnergy();
    
    integrator.integrate(100, m_impl->m_dt);
    
    const double energy = integrator.getTotalEnergy();
    LSMUNIT_ASSERT_ENERGY_DISIPATED(initialEnergy, energy, 0);

    interaction.calcForces();
    CPPUNIT_ASSERT(-0.2 <  p1.getForce().X());
    CPPUNIT_ASSERT( 0.0 <  p1.getForce().Y());
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, p1.getForce().Z(), TOL);
    CPPUNIT_ASSERT( 0.2 >  p2.getForce().X());
    CPPUNIT_ASSERT( 0.0 >  p2.getForce().Y());
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, p2.getForce().Z(), TOL);
    CPPUNIT_ASSERT(0.0 < p1.getMoment().Z());
    CPPUNIT_ASSERT(0.0 < p2.getMoment().Z());

    p1.integrate(m_impl->m_dt);
    p2.integrate(m_impl->m_dt);
    CPPUNIT_ASSERT(0.0 > p1.getAngVel().Z());
    CPPUNIT_ASSERT(0.0 < p2.getAngVel().Z());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, p2.getAngVel().X(), TOL);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, p2.getAngVel().Y(), TOL);
    CPPUNIT_ASSERT(p1OrigPos.X() > p1.getPos().X());
    CPPUNIT_ASSERT(p2OrigPos.X() < p2.getPos().X());
  }
};
