#include "ModelTest/RotParticleTest.h"
#include "UnitTest/Macros.h"
#include "Model/RotParticle.h"
#include "Model/SimpleParticleData.h"
#include "Foundation/vec3.h"

using namespace esys::lsm;
namespace LsmTest
{
  CPPUNIT_TEST_SUITE_REGISTRATION( RotParticleTest );

  class RotParticleTest::Impl
  {
  public:
    Impl()
      : m_p(),
        m_dt(0.001)
    {
      const double mass   = 1;
      const double radius = 1;
      const Vec3   pos    = Vec3(0, 0, 0);
      const int    id     = 0;
      const int    tag    = 0;
      
      SimpleParticleData particleData(id, tag, pos, radius);
      particleData.setMass(particleData.get3dMass());
      
      m_p = CRotParticle(particleData);
    }
  
    CRotParticle m_p;
    double       m_dt;
  };
  
  RotParticleTest::RotParticleTest()
    : CppUnit::TestCase(),
      m_implPtr()
  {
  }

  RotParticleTest::~RotParticleTest()
  {
  }

  void RotParticleTest::setUp()
  {
    CppUnit::TestCase::setUp();
    m_implPtr = ImplPtr(new Impl());
  }

  void RotParticleTest::tearDown()
  {
    CppUnit::TestCase::tearDown();
  }

  void RotParticleTest::testIntegrationConstantAngVel()
  {
    CRotParticle &p = m_implPtr->m_p;
    const Vec3 angVel(0.0, 0.0, 0.1);
    p.setAngVel(angVel);

    const double initialEnergy = p.getKineticEnergy();
    const int numTimeSteps = int(nearbyint(1.0/m_implPtr->m_dt));
    for (int i = 0; i < numTimeSteps; i++) {
      p.integrate(m_implPtr->m_dt);
      p.zeroForce();
      p.rescale();
    }
    const double energy = p.getKineticEnergy();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(initialEnergy, energy, 0.0);
    LSMUNIT_ASSERT_VEC3_EQUAL(angVel, p.getAngVel());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(angVel.norm(), 2*acos(p.getQuat().return_sca()), m_implPtr->m_dt);
  }

  void RotParticleTest::testIntegrationInitialMoment()
  {
    CRotParticle &p = m_implPtr->m_p;
    const Vec3 moment(0.0, 0.0, 0.01);
    p.applyMoment(moment);

    for (int i = 0; i < 2000; i++) {
      p.integrate(m_implPtr->m_dt);
      p.zeroForce();
      p.rescale();      
    }
    const double energy = p.getKineticEnergy();

    CPPUNIT_ASSERT(energy > 0.0);
    CPPUNIT_ASSERT_EQUAL(p.getAngVel().X(), 0.0);
    CPPUNIT_ASSERT_EQUAL(p.getAngVel().Y(), 0.0);
    CPPUNIT_ASSERT(p.getAngVel().Z() > 0.0);
  }
};
