/*=============================================================================
??
??******************************************************************************
??*                                               *
??* COPYRIGHT ??????? - All Rights Reserved *
??*  *
??* This software is the property of ??????????????. No part of this code   *
??* may be copied in any form or by any means without the expressed written   *
??* consent of ???????????. Copying, use or modification of this software     *
??* by any unauthorised person is illegal unless that    *
??* person has a software license agreement with ?????????????.            *
??*                                                          *
??******************************************************************************
??
*********************************************************************************/

#include "Parallel/CheckPointController.h"
#include "Parallel/mpicmdbuf.h"
#include "Parallel/mpibarrier.h"
#include "Parallel/sublattice_cmd.h"
#include "Parallel/MpiInfo.h"
#include "Parallel/CheckPointParams.h"
#include "Parallel/mpivbuf.h"
#include "Parallel/CheckPointInfo.h"

#include <fstream>

CheckPointController::CheckPointController()
  : m_spatialDomain(Vec3(0, 0, 0), Vec3(0, 0, 0)),
    m_fileNamePrefix(""),
    m_beginTime(-1),
    m_endTime(-1),
    m_timeInterval(-1),
    m_geoInfo(),
    m_numTimeSteps(0),
    m_timeStepSize(0.0)
{
}

CheckPointController::CheckPointController(
  const std::string &fileNamePrefix,
  int beginTime,
  int endTime,
  int timeInterval
) : m_spatialDomain(Vec3(0, 0, 0), Vec3(0, 0, 0)),
    m_fileNamePrefix(fileNamePrefix),
    m_beginTime(beginTime),
    m_endTime(endTime),
    m_timeInterval(timeInterval),
    m_geoInfo(),
    m_numTimeSteps(0),
    m_timeStepSize(0.0)
{
}

CheckPointController::~CheckPointController()
{
}

void CheckPointController::performCheckPoint(int currentTime)
{
  if (isCheckPoint(currentTime))
  {
    issueCheckPointCmd(currentTime);
  }
}

void CheckPointController::issueCheckPointCmd(int currentTime)
{
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD, MpiInfo(MPI_COMM_WORLD).rank());
  CMPIBarrier barrier(MPI_COMM_WORLD);

  cmd_buffer.broadcast(CMD_SAVECHECKPOINT);
  
  CVarMPIBuffer buffer(MPI_COMM_WORLD);

  // send check-point parameters
  CheckPointParams checkPointParams(m_fileNamePrefix, currentTime, MpiInfo().rank());
  checkPointParams.packInto(&buffer);
  buffer.broadcast(0);
  buffer.clear();

  std::ofstream oStream(checkPointParams.getFileName().c_str());
  esys::lsm::CheckPointInfo cpInfo;

  cpInfo.setLatticeDataFiles(getLatticeDataFiles(currentTime));
  cpInfo.setGeometryInfo(getGeometryInfo());
  cpInfo.setTimeStep(currentTime);
  cpInfo.setTimeStepSize(getTimeStepSize());
  cpInfo.setNumTimeSteps(getNumTimeSteps());
  cpInfo.write(oStream);

  oStream.close();

  barrier.wait("CheckPoint");
}

std::string CheckPointController::getLatticeDataFileName(const std::string &fileNamePrefix, int timeStep, int rank)
{
  return CheckPointParams(fileNamePrefix, timeStep, rank).getFileName();
}

esys::lsm::StringVector CheckPointController::getLatticeDataFiles(int timeStep, int size)
{
  esys::lsm::StringVector fileNames;
  for (int rank = 1; rank < size; rank++)
  {
    fileNames.push_back(getLatticeDataFileName(m_fileNamePrefix, timeStep, rank));
  }
  return fileNames;
}

esys::lsm::StringVector CheckPointController::getLatticeDataFiles(int timeStep)
{
  MpiInfo mpiInfo;
  
  return getLatticeDataFiles(timeStep, MpiInfo().size());
}

bool CheckPointController::isCheckPoint(int time)
{
  return
    (
      (
        (time >= m_beginTime)
        &&
        (time <= m_endTime)
        &&
        (((time - m_beginTime) % m_timeInterval) == 0)
      )
      ||
      (time == m_endTime)
    );
}

void CheckPointController::setCheckPointParams(
  const std::string &fileNamePrefix,
  int beginTime,
  int endTime,
  int timeInterval
)
{
  m_fileNamePrefix = fileNamePrefix;
  m_beginTime      = beginTime;
  m_endTime        = endTime;
  m_timeInterval   = timeInterval;
}

const esys::lsm::BoundingBox &CheckPointController::getSpatialDomain() const
{
  return m_spatialDomain;
}

void CheckPointController::setSpatialDomain(const esys::lsm::BoundingBox &bbox)
{
  m_spatialDomain = bbox;
}

void CheckPointController::setGeometryInfo(const esys::lsm::GeometryInfo &geoInfo)
{
  m_geoInfo = geoInfo;
}

esys::lsm::GeometryInfo CheckPointController::getGeometryInfo() const
{
  return m_geoInfo;
}

int CheckPointController::getNumTimeSteps() const
{
  return m_numTimeSteps;
}

void CheckPointController::setNumTimeSteps(int numTimeSteps)
{
  m_numTimeSteps = numTimeSteps;
}

double CheckPointController::getTimeStepSize() const
{
  return m_timeStepSize;
}

void CheckPointController::setTimeStepSize(double timeStepSize)
{
  m_timeStepSize = timeStepSize;
}

