/*=============================================================================
??
??******************************************************************************
??*                                               *
??* COPYRIGHT ??????? - All Rights Reserved *
??*  *
??* This software is the property of ??????????????. No part of this code   *
??* may be copied in any form or by any means without the expressed written   *
??* consent of ???????????. Copying, use or modification of this software     *
??* by any unauthorised person is illegal unless that    *
??* person has a software license agreement with ?????????????.            *
??*                                                          *
??******************************************************************************
??
*********************************************************************************/

#ifndef CHECKPOINTCONTROLLER_H
#define CHECKPOINTCONTROLLER_H

#include "Geometry/GeometryInfo.h"
#include "Foundation/StringUtil.h"
#include "Foundation/BoundingBox.h"

#include <string>

/**
 * Controls the issue of check-pointing commands to slave processes.
 */
class CheckPointController
{
public:
  /**
   * Default constructor.
   *
   * Default parameters cause isCheckPoint to return false.
   */
  CheckPointController();

  /**
   * Instantiates and sets the parameters which determine when
   * a check-point should occur.
   *
   * @see setCheckpointParams
   */
  CheckPointController(
    const std::string &fileNamePrefix,
    int beginTime,
    int endTime,
    int timeInterval
  );

  /**
   *
   */
  virtual ~CheckPointController();

  /**
   * Determines whether specified argument is a check-point
   * time (see isCheckPoint). If currentTime is a check-point,
   * issues commands to slave processes to perform check-point.
   *
   * @param currentTime Parameter used to determine whether
   *                    a check point should occur.
   */
  virtual void performCheckPoint(int currentTime);

  /**
   * Issues the check-point command to slave processes.
   *
   * @param currentTime The check-point time.
   */
  virtual void issueCheckPointCmd(int currentTime);

  /**
   * Returns whether specified time is a check-point time.
   *
   * @param time This value is checked against the checkpoint
   *             parameters.
   */
  bool isCheckPoint(int time);

  /**
   * Sets the parameters which determine when a check-point should occur.
   */
  void setCheckPointParams(
    const std::string &fileNamePrefix,
    int beginTime,
    int endTime,
    int timeInterval
  );

  std::string getLatticeDataFileName(const std::string &fileNamePrefix, int timeStep, int rank);

  esys::lsm::StringVector getLatticeDataFiles(int timeStep, int size);

  esys::lsm::StringVector getLatticeDataFiles(int timeStep);

  /**
   * Sets the spatial extent in which particles are tracked.
   */
  void setGeometryInfo(const esys::lsm::GeometryInfo &geoInfo);

  /**
   * Returns the spatial extent in which particles are tracked.
   */
  const esys::lsm::BoundingBox &getSpatialDomain() const;

  /**
   * Sets geometry info.
   */
  void setSpatialDomain(const esys::lsm::BoundingBox &bBox);

  /**
   * Returns geometry info.
   */
  esys::lsm::GeometryInfo getGeometryInfo() const;

  /**
   *
   */
  int getNumTimeSteps() const;
  
  /**
   * Sets the number of time steps.
   */
  void setNumTimeSteps(int numTimeSteps);

  /**
   *
   */
  double getTimeStepSize() const;
  
  /**
   * Sets the time step size.
   */
  void setTimeStepSize(double timeStepSize);

private:
  /**
   *
   */
  esys::lsm::BoundingBox m_spatialDomain;

  /**
   * Prefix of check-point files.
   */
  std::string m_fileNamePrefix;

  /**
   *
   */
  int m_beginTime;

  /**
   *
   */
  int m_endTime;

  /**
   *
   */
  int m_timeInterval;

  /**
   *
   */
  esys::lsm::GeometryInfo m_geoInfo;

  /**
   *
   */
  int m_numTimeSteps;

  /**
   *
   */
  double m_timeStepSize;
};

#endif
