
#include "Parallel/LatticeMaster.h"
#include "Parser/ModelFnc.h"

//--- IO includes --
#include <iostream>
#include <iomanip>
#include <fstream>
#include <stdexcept>

using std::ifstream;
using std::ofstream;
using std::fstream;
using std::ios;

// -- STL includes --
#include <map>
#include <algorithm>

using std::map;

// -- project includes --
#include "Model/TempPartStore.h"
#include "Parallel/mpi_tag_defs.h"
#include "Parallel/mpicmdbuf.h"
#include "Parallel/mpisgvbuf.h"
#include "Parallel/sublattice_cmd.h"
#include "Parallel/mpibarrier.h"
#include "Model/Damping.h"
//#include "Model/RotDampingGroup.h"
#include "Model/ElasticInteractionGroup.h"
//#include "Model/RotElasticInteractionGroup.h"
#include "Model/BWallInteractionGroup.h"
#include "Model/ViscWallIG.h"
#include "Model/SoftBWallInteractionGroup.h"
#include "Model/BodyForceGroup.h"
#include "Parallel/master_main.h"
#include "Foundation/Timer.h"

#include "Fields/ParticleFieldMaster.h"
#include "Fields/ScalarParticleDistributionMaster.h"
#include "Fields/InteractionFieldMaster.h"
#include "Fields/VectorInteractionFieldMaster.h"
#include "Fields/VectorTriangleFieldMaster.h"
#include "Parallel/CheckPointInfo.h"
#include "Parallel/CheckPointLoader.h"
#include "Parallel/CheckPointController.h"
#include "Parallel/GeometryReader.h"
#include "Foundation/BoundingBox.h"
#include "Parallel/MeshReader.h"
#include "Parallel/MpiInfo.h"

using namespace esys::lsm;

CLatticeMaster::CLatticeMaster()
  : m_timingFileName(),
    m_pTimers(new MpiWTimers()),
    m_pCheckPointController(new CheckPointController()),
    m_processDims(3, 0)
{
  m_isInitialized = false ;
  m_is_dx=false;
  m_dt = 0.0;
}

CLatticeMaster::~CLatticeMaster()
{
  console.Debug() << "master finalizing\n";
  delete m_pTimers;
  delete m_pCheckPointController;
  for (vector<AFieldMaster*>::iterator it = m_save_fields.begin(); it != m_save_fields.end(); it++)
  {
    delete (*it);
  }
  m_save_fields.clear();
  MPI_Finalize();
}

void CLatticeMaster::init()
{
  MPI_Comm_size(MPI_COMM_WORLD, &m_global_size);
  MPI_Comm_rank(MPI_COMM_WORLD, &m_global_rank);

  console.Debug() << "Master has rank " << m_global_rank << "\n" ;
  console.Debug() << "Global size is  " << m_global_size << "\n" ;
}

void CLatticeMaster::setNumSteps(int s)
{
  m_max_ts=s;
  m_pCheckPointController->setNumTimeSteps(s);
}

void CLatticeMaster::saveTimingDataToFile(const std::string &fileNamePrefix)
{
  std::stringstream sStream;
  sStream << fileNamePrefix << m_global_rank << ".csv";
  setTimingFileName(sStream.str());

  CMPILCmdBuffer cmdBuffer(MPI_COMM_WORLD, MpiInfo(MPI_COMM_WORLD).rank());
  CMPIBarrier barrier(MPI_COMM_WORLD);
  cmdBuffer.broadcast(CMD_PERFORMTIMING);

  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  // send timing filename prefix
  buffer.append(fileNamePrefix.c_str());
  buffer.broadcast(0);
  buffer.clear();
  barrier.wait("performTiming");
}

void CLatticeMaster::setTimingFileName(const std::string &fileName)
{
  m_timingFileName = fileName;
}

const std::string &CLatticeMaster::getTimingFileName() const
{
  return m_timingFileName;
}

void CLatticeMaster::do2dCalculations()
{
  CMPILCmdBuffer cmdBuffer(MPI_COMM_WORLD, MpiInfo(MPI_COMM_WORLD).rank());
  CMPIBarrier barrier(MPI_COMM_WORLD);
  cmdBuffer.broadcast(CMD_DO2DCALCULATIONS);
  barrier.wait("do2dCalculations cmd");
}

/*!
  make a lattice from particles of the given type

  \param ptype the type of particle
  \param nrange search range
  \param alpha pair search cutoff
  \param dt time step
*/
void CLatticeMaster::makeLattice(const char* ptype, double nrange, double alpha, double dt)
{
  console.XDebug() << "CLatticeMaster::makeLattice( " << ptype << " , " << dt << " )\n";

  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  m_pCheckPointController->setTimeStepSize(dt);
  m_particle_type = string(ptype);

  //send command to slave
  cmd_buffer.broadcast(CMD_MAKELATTICE);
  
  // send lattice parameters
  m_dt = dt;

  CLatticeParam clp(ptype, nrange, alpha, dt, getProcessDims());
  clp.packInto(&buffer);
  buffer.broadcast(0);
  buffer.clear();

  makeWorkerComm();

  barrier.wait("CLatticeMaster::makeLattice");
  console.XDebug() << "end CLatticeMaster::makeLattice\n";
}

void CLatticeMaster::setProcessDims(const CLatticeParam::ProcessDims &dims)
{
  m_processDims = dims;
}

const CLatticeParam::ProcessDims &CLatticeMaster::getProcessDims() const
{
  return m_processDims;
}

/*!
  Generate a communicator excluding the master. Has to be done in the master
  too because the new communicator is derived from MPI_COMM_WORLD and
  MPI_Comm_create has to be executed by all members of the original communicator
*/
void CLatticeMaster::makeWorkerComm()
{
  MPI_Group group_world,group_workers;
  MPI_Comm comm_workers;
  int ranks[1];
  ranks[0]=0;
  MPI_Comm_group(MPI_COMM_WORLD,&group_world);
  MPI_Group_excl(group_world,1,ranks,&group_workers);
  MPI_Comm_create(MPI_COMM_WORLD,group_workers,&comm_workers);

}

/*!
  Complex init from file. Read file and send particle data to sublattices 

  \param filename the nameof the file to be read
  \param nx blocks in x-direction
  \param ny blocks in y-direction
  \param nz blocks in z-direction
*/
void CLatticeMaster::initComplexFromFile(const string& filename, int nx, int ny, int nz)
{
  console.XDebug()<<"CLatticeMaster::initComplexFromFile\n";
   
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CMPIVarSGBufferRoot sgbuffer(MPI_COMM_WORLD,1024);
  CMPIVarSGBufferRoot nbuffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);
  ifstream infile;
  double xmin,xmax,ymin,ymax,zmin,zmax;
  int *np; // number of particles per block
  int is2D;

  if(m_global_size-1!=nx*ny*nz){
    console.Critical() << "ERROR !! Wrong nr. of slaves/sublattices ! SHUTTING DOWN !!\n\n";
    MPI_Finalize();
    exit(1);
  }

  np=new int[m_global_size];
  // zero particle numbers
  for(int i=0;i<m_global_size;i++){
    np[i]=0;
  }

   // open file
  infile.open(filename.c_str());

  // check if file is open
  if(!infile){
    console.Critical() << "can not open model description file " << filename << "\n";
  } else {
    // read type parameter and pack it into the buffer
    string type;
    infile >> type;
    for(int i=1;i<m_global_size;i++){ 
      sgbuffer.append(type.c_str(),i);
      string dim;
      infile >> dim;
      if (dim=="2d"){
        is2D=1;
        sgbuffer.append(int(1),i);
      } else {
        is2D=0;
        sgbuffer.append(int(0),i);
      }
    }
    // read model area (xmin/max, ymin/max, zmin/max)
    infile >> xmin >> xmax;
    infile >> ymin >> ymax;
    infile >> zmin >> zmax;

    // calc variables for distribution
    double dx=(xmax-xmin)/double(nx);
    double dy=(ymax-ymin)/double(ny);
    double dz=(zmax-zmin)/double(nz);
    
    // read file and put date into buffer
    int nparts;
    infile >> nparts; // number of particles 
    
    for(int part=0;part<nparts;part++){
      Vec3 pos,vel;
      double rad,mass;
      int npatches;
      
      // get particle params (radius,mass,position,velocity)
      infile >> rad;
      infile >> mass;
      infile >> pos.X();
      infile >> pos.Y();
      infile >> pos.Z();
      infile >> vel.X();
      infile >> vel.Y();
      infile >> vel.Z();

      // from pos, decide which block it goes to
      double rx=pos.X()-xmin;
      int ix=int(floor(rx/dx));
      double ry=pos.Y()-ymin;
      int iy=int(floor(ry/dy));
      double rz=pos.Z()-zmin;
      int iz=int(floor(rz/dz));
      int idx=ix*ny*nz+iy*nz+iz+1; // +1 because 0->master

      np[idx]++; 
      
      console.XDebug() 
        << "id,rad,mass,pos,vel,idx "
        << part << " "
        << rad << " , "
        << mass << " , "
        << pos << " , "
        << vel << " , "
        << idx << "\n";

      sgbuffer.append(is2D,idx);
      sgbuffer.append(rad,idx);
      sgbuffer.append(mass,idx);
      sgbuffer.append(pos.X(),idx);
      sgbuffer.append(pos.Y(),idx);
      sgbuffer.append(pos.Z(),idx);
      sgbuffer.append(vel.X(),idx);
      sgbuffer.append(vel.Y(),idx);
      sgbuffer.append(vel.Z(),idx);
      sgbuffer.append(part,idx); // particle id

      infile >> npatches;
      sgbuffer.append(npatches,idx); // nr. of patches
      console.XDebug() << "npatches " << npatches << "\n";
      for(int patch=0;patch<npatches;patch++){
      Vec3 center,normal;
      double rad2,ang;
      
      infile >> center.X();
      infile >> center.Y();
      infile >> center.Z();
      infile >> rad2;
      infile >> normal.X();
      infile >> normal.Y();
      infile >> normal.Z();
      infile >> ang;
      sgbuffer.append(center.X(),idx);
      sgbuffer.append(center.Y(),idx);
      sgbuffer.append(center.Z(),idx);
      sgbuffer.append(rad2,idx);
      sgbuffer.append(normal.X(),idx);
      sgbuffer.append(normal.Y(),idx);
      sgbuffer.append(normal.Z(),idx);
      sgbuffer.append(ang,idx);
      }
    }
    // close file
    infile.close();

    // put per-block number into buffer 
    for(int i=1;i<m_global_size;i++){
      nbuffer.append(np[i],i);
    }

    // send command to workers
    cmd_buffer.broadcast(CMD_INITCOMPLEX);

    // send particle numbers
    nbuffer.scatter();
    barrier.wait("initComplexFromFile:np");

    // distribute
    sgbuffer.scatter();

    // wait for the workers to do the rest
    barrier.wait("initComplexFromFile");

    // call neighborsearch
    searchNeighbors(true);  
    barrier.wait("initComplexFromFile:ns");
  }
  console.XDebug()<<"end CLatticeMaster::initComplex\n";
}

bool CLatticeMaster::haveSetSpatialDomain() const
{
  return (m_pCheckPointController->getSpatialDomain().getVolume() > 0.0);
}

void CLatticeMaster::setSpatialDomain(const Vec3 &minBBoxPt, const Vec3 &maxBBoxPt)
{
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);

  m_pCheckPointController->setSpatialDomain(BoundingBox(minBBoxPt, maxBBoxPt));

  // none of the boundaries circular -> simple initLattice
  cmd_buffer.broadcast(CMD_INITLATTICE);
  esys::lsm::Vec3Vector corners;
  corners.push_back(minBBoxPt);
  corners.push_back(maxBBoxPt);
  m_world_comm.broadcast_cont_packed(corners);
  getSlaveSpatialDomains();
}

void  CLatticeMaster::initialiseGeometry(const GeometryInfo &geoInfo)
{
  CMPIBarrier barrier(MPI_COMM_WORLD);
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);

  m_pCheckPointController->setGeometryInfo(geoInfo);

  // send command to slaves to start initLattice fnc there
  if (geoInfo.hasAnyPeriodicDimensions()){
    // at least one boundary circular -> initLatticeCirc
    console.XDebug()<< "Circular Lattice Init\n";
    cmd_buffer.broadcast(CMD_INITLATTICECIRC);

    console.XDebug()<< "Broadcasting bonding-box corners\n";
    m_world_comm.broadcast_cont_packed(geoInfo.getBBoxCorners());

    // send circ. markers
    m_world_comm.broadcast_cont(geoInfo.getPeriodicDimensions());

    getSlaveSpatialDomains();

  } else {
    if (!haveSetSpatialDomain()) {
      setSpatialDomain(geoInfo.getBBoxCorners()[0], geoInfo.getBBoxCorners()[1]);
    }
  }
}

void CLatticeMaster::getSlaveSpatialDomains()
{
  // get slave dimensions and coordinates 
  multimap<int,int> slavedims;
  multimap<int,int> slavecoords;
  m_world_comm.gather(slavedims);
  m_world_comm.gather(slavecoords);

  console.XDebug()<<"got dims & coords\n";

  // extract slave dimensions from received data 
  int dim_x,dim_y,dim_z; // dimensions in x,y,z direction
  multimap<int,int>::iterator iter=slavedims.find(1);
  dim_x=iter->second;
  iter++;
  dim_y=iter->second;
  iter++;
  dim_z=iter->second;
  // setup coord->rank mapping for slaves
  int nslaves=dim_x*dim_y*dim_z; // number of slaves
  for(int i=0;i<nslaves;i++){
    // extract slave coordinates from received data
    int cx,cy,cz; // x-,y-,z- coord.
    
    multimap<int,int>::iterator coord_iter=slavecoords.find(i+1); // slaves are 1..n
    cx=coord_iter->second;
    coord_iter++;
    cy=coord_iter->second;
    coord_iter++;
    cz=coord_iter->second;
  }
}

template <class TmplIterator, class TmplParticle>
void CLatticeMaster::addParticles(TmplIterator &particleIt)
{
  CMPIBarrier barrier(MPI_COMM_WORLD);
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);

  // cast storage pointer to correct type 
  vector<TmplParticle> particleVector;
  console.XDebug()<< "Reserving vector for " << particleIt.getNumRemaining() << " particles..." << "\n";

  const int numBroadcastParticles = 250000;
  particleVector.reserve(numBroadcastParticles);
  console.XDebug()<< "Reading particles..." << "\n";
  barrier.wait("INIT thingie");
  for (int i = 0; particleIt.hasNext(); i++) {
    const typename TmplIterator::ValueType &particleData = particleIt.next();

    if (((i+1) % 50000) == 0) {
      console.XDebug() << "Adding particle with id " << particleData.getID() << "\n";
    }

    particleVector.push_back(TmplParticle(particleData));

    if (((i+1) % numBroadcastParticles) == 0)
    {
      console.XDebug() << "Broadcasting receive cmd...." << "\n";
      cmd_buffer.broadcast(CMD_RECEIVEPARTICLES);
      console.XDebug() << "Broadcasting particles...." << "\n";
      m_world_comm.broadcast_cont_packed(particleVector);
      barrier.wait("Post particle broadcast.");
      barrier.wait("Post Command.");
      particleVector.clear();
      particleVector.reserve(numBroadcastParticles);
    }
  }
  console.XDebug() << "Done reading particles...." << "\n";
  console.XDebug() << "Broadcasting final particle-receive cmd...." << "\n";
  cmd_buffer.broadcast(CMD_RECEIVEPARTICLES);
  console.XDebug() << "Broadcasting final set of particles...." << "\n";
  m_world_comm.broadcast_cont_packed(particleVector);
  barrier.wait("Post final particle broadcast.");
  barrier.wait("Post final particle-broadcast command.");
  // -- end read particles
}

template <class TmplIterator>
void CLatticeMaster::addConnections(TmplIterator &connectionIt)
{
  console.XDebug()<< "Reserving " << connectionIt.getNumRemaining() << "particle id's." << "\n";
  m_temp_conn[0].reserve(m_temp_conn[0].size() + 2*connectionIt.getNumRemaining());
  console.XDebug()
    << "Reading "
    << connectionIt.getNumRemaining()
    << "connections..." << "\n";

  for (int i = 0; connectionIt.hasNext(); i++)
  {
    const typename TmplIterator::ValueType &data = connectionIt.next();
    if (i % 50000 == 0)
    {
      console.XDebug() << "Adding connection number " << i << "\n";
    }
    m_temp_conn[data.getTag()].push_back(data.getP1Id());
    m_temp_conn[data.getTag()].push_back(data.getP2Id());
  }
  console.XDebug()<< "Done reading connections..." << "\n";
  // -- end read connections
}

template <class TmplParticle>
void  CLatticeMaster::readGeometry(const std::string &fileName)
{
  GeometryReader geoReader(fileName);
  GeometryInfo geoInfo = geoReader.getGeometryInfo();
  initialiseGeometry(geoInfo);

  GeometryReader::ParticleIterator &particleIt = geoReader.getParticleIterator();
  console.XDebug()<< "Number of particles: " << particleIt.getNumRemaining() << "\n";
  if (geoReader.getParticleType() != "Simple") {
    throw 
      std::runtime_error(
        (std::string("Unknown particle type ") + geoReader.getParticleType()).c_str()
      );
  }

  addParticles<GeometryReader::ParticleIterator,TmplParticle>(particleIt);

  // -- build ntable
  console.XDebug() << "Building ntable (searchNeighbours)..." << "\n";
  searchNeighbors(true);

  GeometryReader::ConnectionIterator &connectionIt =
    geoReader.getConnectionIterator();
  console.XDebug()<< "Number of connections: " << connectionIt.getNumRemaining() << "\n";

  addConnections(connectionIt);
  // -- build ntable
  searchNeighbors(true);
  console.Debug()<<"end CLatticeMaster::readGeometryFile\n";
}

void  CLatticeMaster::readGeometryFile(const string& fileName)
{
  console.Debug()<<"CLatticeMaster::readGeometryFile( " << fileName << ")\n";

  if (m_particle_type == "Basic"){
    readGeometry<CParticle>(fileName);
  } else if (m_particle_type == "Rot") {
    readGeometry<CRotParticle>(fileName);
  }
}

void CLatticeMaster::loadCheckPointData(const string &checkPointFileName)
{
  CheckPointInfo cpInfo;

  std::ifstream iStream(checkPointFileName.c_str());
  if (!iStream) {
    throw std::runtime_error(
      std::string("Could not open file ") + checkPointFileName
    );
  }

  cpInfo.read(iStream);

  initialiseGeometry(cpInfo.getGeometryInfo());

  CheckPointLoader cpLoader(cpInfo.getLatticeDataFiles());
  cpLoader.loadInto(*this);
  
  searchNeighbors(true);
}

/*!
  add a wall to the sublattice

  \param ipos initial position of the wall
  \param inorm initial normal of the wall
  \param k spring constant of the wall
*/

void CLatticeMaster::initWall(const Vec3& ipos,const Vec3& inorm,double k)
{
  console.Debug() << "CLatticeMaster::initWall("<< ipos << " | " << inorm << ")\n" ;

  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CEWallIGP wigp;
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  //send command to slave
  cmd_buffer.broadcast(CMD_GETWALL);
  wigp.setName("wall");
  wigp.setSpringConst(k);
  wigp.setWall(ipos,inorm);
  wigp.packInto(&buffer);
  buffer.broadcast(m_global_rank);
  barrier.wait("CLatticeMaster::initWall");
  console.Debug() << "end CLatticeMaster::initWall() \n";
}

/*!
  add a bonded wall to the sublattice 

  \param ipos initial position of the wall
  \param inorm initial normal of the wall
  \param k spring constant of the wall
*/

void CLatticeMaster::initBondedWall(const Vec3& ipos,const Vec3& inorm,double k,int tag)
{
  console.Debug() << "CLatticeMaster::initBondedWall("<< ipos << " | " << inorm << ")\n" ;

  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CBWallIGP wigp;
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  //send command to slave
  cmd_buffer.broadcast(CMD_GETBONDEDWALL);
  wigp.setName("bondedwall");
  wigp.setSpringConst(k);
  wigp.setTag(tag);
  wigp.setWall(ipos,inorm);
  wigp.packInto(&buffer);
  buffer.broadcast(m_global_rank);
  barrier.wait("CLatticeMaster::initBondedWall");
  console.Debug() << "end CLatticeMaster::initBondedWall() \n";
}

/*!
  add a wall with viscous drag to the sublattice 

  \param ipos initial position of the wall
  \param inorm initial normal of the wall
  \param k spring constant of the wall
  \param nu viscosity
  \param tag 
*/

void CLatticeMaster::initViscWall(const Vec3& ipos,const Vec3& inorm,double k,double nu,int tag)
{
  console.Debug() << "CLatticeMaster::initViscWall("<< ipos << " | " << inorm << ")\n" ;

  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CVWallIGP wigp;
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  //send command to slave
  cmd_buffer.broadcast(CMD_GETVISCWALL);
  wigp.setName("bondedwall");
  wigp.setSpringConst(k);
  wigp.setTag(tag);
  wigp.setWall(ipos,inorm);
  wigp.packInto(&buffer);
  wigp.setNu(nu);
  buffer.broadcast(m_global_rank);
  barrier.wait("CLatticeMaster::initViscWall");
  console.Debug() << "end CLatticeMaster::initBondedWall() \n";
}

/*!
 */
void CLatticeMaster::initSoftBondedWall(const Vec3& ipos,const Vec3& inorm,double kx, double ky, double kz, int tag)
{
  console.Debug() << "CLatticeMaster::initSoftBondedWall("<< ipos << " | " << inorm << ")\n" ;

  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CSoftBWallIGP wigp;
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  //send command to slave
  cmd_buffer.broadcast(CMD_GETSOFTBONDEDWALL);
  wigp.setName("softbondedwall");
  wigp.setAllK(kx,ky,kz);
  wigp.setTag(tag);
  wigp.setWall(ipos,inorm);
  wigp.packInto(&buffer);
  buffer.broadcast(m_global_rank);
  barrier.wait("CLatticeMaster::initSoftBondedWall");
  console.Debug() << "end CLatticeMaster::initSoftBondedWall() \n";
}

/*!
  Move all particles with a given tag to a given displacement relative to their original position. 

  \param id the tag of the particles to be moved
  \param d the displacement
*/
void CLatticeMaster::moveParticleTo(int id,const Vec3& d)
{
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  cmd_buffer.broadcast(CMD_PMOVE);
  buffer.append(id);
  buffer.append(d) ; 
  buffer.broadcast(m_global_rank);
  barrier.wait("moveParticleTo");
}

/*!
  Move a node in a TriMeshIG by a given amount.

  \param tm_name the name of the TriMeshIG
  \param id the node id
  \param d the displacement
*/
void CLatticeMaster::moveSingleNodeBy(const string& name,int id,const Vec3& d)
{  
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  cmd_buffer.broadcast(CMD_MOVENODE);
  buffer.append(name.c_str());
  buffer.append(id) ;
  buffer.append(d);
  buffer.broadcast(m_global_rank);
  barrier.wait("moveSingleNodeBy");
}

/*!
  Move all nodes with a given tag in a TriMeshIG by a given amount.

  \param tm_name the name of the TriMeshIG
  \param tag the tag
  \param d the displacement
*/
void CLatticeMaster::moveTaggedNodesBy(const string& name,int tag,const Vec3& d)
{
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  cmd_buffer.broadcast(CMD_MOVETAGGEDNODES);
  buffer.append(name.c_str());
  buffer.append(tag) ;
  buffer.append(d);
  buffer.broadcast(m_global_rank);
  barrier.wait("moveTaggedNodesBy");
}

/*!
  Add a given tag to the particle closest to a given position. Only the bits
  set in the mask will be influenced, i.e. new_tag=(old_tag & !mask) | (tag & mask).

  \param tag the tag
  \param mask the mask
  \param pos the position
*/
void CLatticeMaster::tagParticleNearestTo(int tag,int mask,const Vec3& pos)
{
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  cmd_buffer.broadcast(CMD_PTAG);
  buffer.append(tag);
  buffer.append(mask) ;
  buffer.append(pos);
  buffer.broadcast(m_global_rank);
  barrier.wait("tagParticleNearestTo");
}

/*!
  Set all particles with a given tag to be non-dynamic, i.e. to have infinite mass

  \param tag the tag
*/
void CLatticeMaster::setParticleNonDynamic(int tag)
{
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  cmd_buffer.broadcast(CMD_PSETND);
  buffer.append(tag);
  buffer.broadcast(m_global_rank);
  barrier.wait("setParticleNonDynamic");
}

/*! 
  set the velocity of a particle 

  \param id  the id of the particle to be moved
  \param V the velocity
*/
void CLatticeMaster::setParticleVel(int id,const Vec3& V)
{
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  cmd_buffer.broadcast(CMD_PVEL);
  buffer.append(id);
  buffer.append(V) ; 
  buffer.broadcast(m_global_rank);
  barrier.wait("setParticleVel");
}
 
/*!
  move a wall to by given vector

  \param id the nr of the wall to be moved
  \param d the movement
*/
void CLatticeMaster::moveWallBy(int id,const Vec3& d)
{
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  cmd_buffer.broadcast(CMD_WMOVE);
  buffer.append(id);
  buffer.append(d); 
  buffer.broadcast(m_global_rank);
  barrier.wait("moveWallBy");
}

/*!
  apply a given force to a wall

  \param id the nr of the wall 
  \param f the force
*/
void CLatticeMaster::applyForceToWall(int id,const Vec3& f)
{
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  cmd_buffer.broadcast(CMD_WFORCE);
  buffer.append(id);
  buffer.append(f); 
  buffer.broadcast(m_global_rank);
  barrier.wait("applyForceToWall");
}

/*!
  set velocity of a wall. Only affects ViscWall, i.e. wall position
  doesn't get updatet !!

  \param id the nr of the wall 
  \param v the velocity
*/
void CLatticeMaster::setVelocityOfWall(int id,const Vec3& v)
{
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  cmd_buffer.broadcast(CMD_WVEL);
  buffer.append(id);
  buffer.append(v); 
  buffer.broadcast(m_global_rank);
  barrier.wait("setVelocityOfWall");
}

/*!
  perform a single time step
 */
void CLatticeMaster::oneStep()
{
  m_pTimers->start("OneStepForceAndExchange");
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  m_pTimers->start("ForceCalculation");
  cmd_buffer.broadcast(CMD_CALC); // local calculations
  barrier.wait("oneStep.1");
  m_pTimers->stop("ForceCalculation");

  m_pTimers->start("BoundaryDataExchange");
  cmd_buffer.broadcast(CMD_XCHANGE); // exchange boundary data
  barrier.wait("oneStep.2");
  m_pTimers->stop("BoundaryDataExchange");

  console.Debug() << "=== TIME STEP ===\n"; 
  //-- times --
  console.Timing() << "Force calculation  " << m_pTimers->getTiming("ForceCalculation") << "\n";
  console.Timing() << "Boundary exchange  " << m_pTimers->getTiming("BoundaryDataExchange") << "\n";
  m_pTimers->stop("OneStepForceAndExchange");
  console.Timing() << "one step           " << m_pTimers->getTiming("OneStepForceAndExchange") << "\n";
  //---
}

/*!
  neighbor search. A check if the search is necessary is performed first

  \param force if true, force neighborsearch even if displacement is below threshold
*/
void CLatticeMaster::searchNeighbors(bool force)
{
  console.Debug() << "CLatticeMaster::searchNeighbors\n" ; 
  
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CMPIBarrier barrier(MPI_COMM_WORLD);
  
  bool need_search=true; 

  if(!force){
    need_search=checkNeighbors();
  }
  if(need_search){
    cmd_buffer.broadcast(CMD_NSEARCH);
    barrier.wait("searchNeighbors");
  }
  console.Debug() << "end CLatticeMaster::searchNeighbors\n" ; 
}

/*!
  Test if neighbor search is necessary.

*/
bool CLatticeMaster::checkNeighbors()
{
  console.Debug() << "CLatticeMaster::checkNeighbors()\n";
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CMPIBarrier barrier(MPI_COMM_WORLD);
  CMPISGBufferRoot buffer(MPI_COMM_WORLD,8);
  
  bool res=false;
  
  cmd_buffer.broadcast(CMD_CHECKNEIGHBORS);
  buffer.gather();
  int i=1;
  while((i<m_global_size)&&(!res)){
    int b=buffer.pop_int(i);
    res=(b==1); // replace with pop_bool -> to be implemented    
    i++;
  } 
  barrier.wait("checkNeighbors");
  console.Debug() << "end CLatticeMaster::checkNeighbors(), res= " << res << "\n";

  return res;
}

/*!
 */
void CLatticeMaster::updateInteractions()
{
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  console.XDebug() << "CLatticeMaster::updateInteractions()\n";
  // send the command
  cmd_buffer.broadcast(CMD_UPDATE);
 
  barrier.wait("updateInteractions");
  console.XDebug() << "end CLatticeMaster::updateInteractions()\n";
}

/*!
  Add a field to the list to be saved

  \param fieldname the name of the field to be saved
  \param filename the name of the file
  \param dt save every dt time steps

  \warning This is a preliminary implementation, i.e. will most likely change
*/
void CLatticeMaster::addSaveField(const string& fieldname, const string& filename, int dt, const string& type)
{
  SSaveParams ssp;
  
  ssp.fieldname=fieldname;
  ssp.filename=filename;
  ssp.dt=dt;
  if (type=="scalar"){
    ssp.type=SSaveParams::FLD_SCALAR;
  } else if(type=="vector"){
    ssp.type=SSaveParams::FLD_VECTOR;
  } else {
    console.Error() <<"Trying to add field with unknown type\n";
  }

  m_save_parameters.push_back(ssp);
}

/*!
  Get a given field from the sublattices and save it in an ASCII-File

  \param fieldname the name of the field to be saved
  \param filename the name of the file
  \param append append to an existing file or open new one?

  \warning This is a preliminary implementation, i.e. will most likely change
*/
void CLatticeMaster::saveSingleField(const string& fieldname, const string& filename, bool append, int type)
{
  console.XDebug() << "CLatticeMaster::saveSingleField( " << fieldname.c_str() << " " << filename.c_str() << " " << append << " )\n";

  fstream outfile;
  
  int i,j,idx ;
  unsigned long n; 
  Vec3 vdata;
  double data;

  CMPIBarrier barrier(MPI_COMM_WORLD);
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIVarSGBufferRoot sgbuffer(MPI_COMM_WORLD);

  // open file with correct mode
  if(append){
    outfile.open(filename.c_str(),ios::app | ios::out );
  } else {
    outfile.open(filename.c_str(),ios::out);
  }
  //if(!outfile.is_open()){
  //  console.Error() << "couldn't open file\n";
  //}

   //== data ==
  // send command
  cmd_buffer.broadcast(CMD_FIELDCOLLECT);

  // send field name
  buffer.append(fieldname.c_str());
  buffer.broadcast(0);
  
  // get data
  sgbuffer.clear();
  sgbuffer.gather();

  unsigned long sumparts = 0 ;
  vector<unsigned long> nparts;
  for(i=1;i<m_global_size;i++){
    nparts.push_back(sgbuffer.pop_int(i));
    sumparts+=nparts[i-1];
  }

  // extract data and save
  switch(type){
  case SSaveParams::FLD_SCALAR:
    for(i=1;i<m_global_size;i++){
      for(j=0;j<nparts[i-1];j++){
	data=sgbuffer.pop_double(i);
	idx=sgbuffer.pop_int(i);
	outfile << idx << " " << data << " ";
	console.Debug() << "writing " << idx << " , " << data << "\n";
      }
    }
    outfile << "\n";
    break;
  case SSaveParams::FLD_VECTOR:
    {
      map<int,Vec3> data_map;
      for(i=1;i<m_global_size;i++){
	for(j=0;j<nparts[i-1];j++){
	  vdata=sgbuffer.pop_vector(i);
	  idx=sgbuffer.pop_int(i);
	  data_map[idx]+=vdata;
	}
      }
      for(map<int,Vec3>::iterator iter=data_map.begin();
	  iter!=data_map.end();
	  iter++){
	outfile << iter->first << " " << iter->second << " ";
      }
      outfile << endl;
    }
  break;
  default:
    console.Error()<< "trying to collect field of unknown type\n";
  }

  barrier.wait("sendFieldData.2");

  // close file
  outfile.close();

  console.XDebug() << "end CLatticeMaster::saveSingleField\n";
}

/*!
  add a scalar particle field to the list of fields to be saved

  \param filename the name of the file the field is saved into 
  \param fieldname the name of the field
  \param savetype output file format
  \param t_0 first timestep to be saved
  \param t_end last timestep to be saved
  \param dt timesteps between saves
*/
void CLatticeMaster::addScalarParticleSaveField(
  const string& filename,
  const string& fieldname,
  const string& savetype,
  int t_0,
  int t_end,
  int dt
)
{
  console.Debug()
    << "CLatticeMaster::addScalarParticleSaveField("
    << filename
    << ","
    << fieldname
    << ","
    << t_0
    << ","
    << t_end
    << ","
    << dt
    << ")\n";

  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  //send cmd to slave
  cmd_buffer.broadcast(CMD_ADD_SPF);

  AFieldMaster* new_fm = 
    new ScalarParticleFieldMaster(
      &m_world_comm,
      fieldname,
      filename,
      savetype,
      t_0,
      t_end,
      dt
    );
  m_save_fields.push_back(new_fm);

  barrier.wait("CLatticeMaster::addScalarParticleSaveField");

  console.Debug() << "end CLatticeMaster::addScalarParticleSaveField()\n";
} 

/*!
  add a scalar particle field to the list of fields to be saved

  \param filename the name of the file the field is saved into 
  \param fieldname the name of the field
  \param savetype output file format
  \param t_0 first timestep to be saved
  \param t_end last timestep to be saved
  \param dt timesteps between saves
  \param tag tag of the particles for which the field is saved
  \param mask the mask used in tag comparisons
*/
void CLatticeMaster::addTaggedScalarParticleSaveField(const string& filename, const string& fieldname,const string& savetype, int t_0,int t_end,int dt,int tag,int mask)
{
  console.Debug() << "CLatticeMaster::addTaggedScalarParticleSaveField(" << filename << ","<< fieldname << "," << t_0 << "," << t_end << "," << dt << ")\n";

  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  //send cmd to slave
  cmd_buffer.broadcast(CMD_ADD_SPF);

  AFieldMaster* new_fm=new ScalarParticleFieldMaster(&m_world_comm,fieldname,filename,savetype,t_0,t_end,dt,tag,mask);
  m_save_fields.push_back(new_fm);

  barrier.wait("CLatticeMaster::addTaggedScalarParticleSaveField");

  console.Debug() << "end CLatticeMaster::addTaggedScalarParticleSaveField()\n";
} 
 
/*!
  save the distribution/histogram of a scalar field on tagged particles

  \param filename the name of the file the field is saved into 
  \param fieldname the name of the field
  \param savetype output file format (WINDOW or GLOBAL)
  \param t_0 first timestep to be saved
  \param t_end last timestep to be saved
  \param dt timesteps between adding samples into the distribution
  \param t_snap time between snapshot saves 
  \param tag tag of the particles for which the field is saved
  \param mask the mask used in tag comparisons
  \param x_0 minimum data size in distribution
  \param x_max maximum data size in distribution
  \param nx number of bins
*/
void CLatticeMaster::addTaggedScalarParticleDistributionSaver(const string& filename,const string& fieldname,const string& savetype,int t_0,int t_end,int dt,int t_snap,int tag,int mask,double x_0,double x_max,int nx)
{
  console.Debug() << "CLatticeMaster::addTaggedScalarParticleDistributionSaver(" << filename << ","<< fieldname << "," << t_0 << "," << t_end << "," << dt << ")\n";

  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  //send cmd to slave
  cmd_buffer.broadcast(CMD_ADD_SPF);

  AFieldMaster* new_fm=new ScalarParticleDistributionMaster(&m_world_comm,fieldname,filename,savetype,t_0,t_end,dt,t_snap,x_0,x_max,nx,tag,mask);
  m_save_fields.push_back(new_fm);

  barrier.wait("CLatticeMaster::addTaggedScalarParticleDistributionSaver");

  console.Debug() << "end CLatticeMaster::addTaggedScalarParticleSaveField()\n";
}

/*!
  add a vector particle field to the list of fields to be saved

  \param filename the name of the file the field is saved into 
  \param fieldname the name of the field
  \param savetype output file format
  \param t_0 first timestep to be saved
  \param t_end last timestep to be saved
  \param dt timesteps between saves
*/
void CLatticeMaster::addVectorParticleSaveField(const string& filename,const string& fieldname,const string& savetype,int t_0,int t_end,int dt)
{
  console.Debug() << "CLatticeMaster::addVectorParticleSaveField(" << filename << ","<< fieldname << "," << t_0 << "," << t_end << "," << dt << ")\n";

  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  //send cmd to slave
  cmd_buffer.broadcast(CMD_ADD_VPF);

  AFieldMaster* new_fm=new VectorParticleFieldMaster(&m_world_comm,fieldname,filename,savetype,t_0,t_end,dt);
  m_save_fields.push_back(new_fm);

  barrier.wait("CLatticeMaster::addVectorParticleSaveField");

  console.Debug() << "end CLatticeMaster::addVectorParticleSaveField()\n";
} 

/*!
  add a vector particle field to the list of fields to be saved

  \param filename the name of the file the field is saved into 
  \param fieldname the name of the field
  \param savetype output file format
  \param t_0 first timestep to be saved
  \param t_end last timestep to be saved
  \param dt timesteps between saves
  \param tag tag of the particles for which the field is saved
  \param mask the mask used in tag comparisons
*/
void CLatticeMaster::addTaggedVectorParticleSaveField(const string& filename,const string& fieldname,const string& savetype,int t_0,int t_end,int dt,int tag, int mask)
{
  console.Debug() << "CLatticeMaster::addTaggedVectorParticleSaveField(" << filename << ","<< fieldname << "," << t_0 << "," << t_end << "," << dt << "," << tag << "," << mask << ")\n";

  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  //send cmd to slave
  cmd_buffer.broadcast(CMD_ADD_VPF);

  AFieldMaster* new_fm=new VectorParticleFieldMaster(&m_world_comm,fieldname,filename,savetype,t_0,t_end,dt,tag,mask);
  m_save_fields.push_back(new_fm);

  barrier.wait("CLatticeMaster::addTaggedVectorParticleSaveField");

  console.Debug() << "end CLatticeMaster::addTaggedVectorParticleSaveField()\n";
} 

void CLatticeMaster::performCheckPoints(
  const string &fileNamePrefix,
  int beginTime,
  int endTime,
  int timeInterval
)
{
  m_pCheckPointController->setCheckPointParams(
    fileNamePrefix,
    beginTime,
    endTime,
    timeInterval
  );
}

/*!
  add a scalar interaction field to the list of fields to be saved

  \param filename the name of the file the field is saved into 
  \param fieldname the name of the field
  \param igname the name of the interaction group from which the field is taken
  \param savetype output file format
  \param t_0 first timestep to be saved
  \param t_end last timestep to be saved
  \param dt timesteps between saves
  \param checked choose between normal and checked field (defaults to false)
*/
void CLatticeMaster::addScalarInteractionSaveField(
  const string& filename,
  const string& fieldname,
  const string& igtype,
  const string& igname,
  const string& savetype,
  int t_0,
  int t_end,
  int dt,
  bool checked
)
{
  console.Debug()
     << "CLatticeMaster::addScalarInteractionSaveField("
     << filename
     << ","
     << fieldname
     << ","
     << igname
     << ","
     << t_0
     << ","
     << t_end
     << ","
     << dt
     << ")\n";

  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  //send cmd to slave
  cmd_buffer.broadcast(CMD_ADD_SIF);

  AFieldMaster* new_fm = 
    new ScalarInteractionFieldMaster(
      &m_world_comm,
      fieldname,
      igtype,
      igname,
      filename,
      savetype,
      t_0,
      t_end,
      dt,
      checked
    );
  m_save_fields.push_back(new_fm);

  barrier.wait("CLatticeMaster::addScalarInteractionSaveField");

  console.Debug() << "end CLatticeMaster::addScalarInteractionSaveField()\n";
} 

/*!
  add a vector field on the triangles of a given trimesh to the list of fields to be saved

  \param filename the name of the file the field is saved into 
  \param fieldname the name of the field
  \param meshname the name of the mesh from which the field is taken
  \param savetype the format in which the data is to be saved
  \param t_0 first timestep to be saved
  \param t_end last timestep to be saved
  \param dt timesteps between saves
*/
void CLatticeMaster::addVectorTriangleSaveField(const string& filename,
						const string& fieldname,
						const string& meshname,
						const string& savetype,
						int t_0,int t_end,int dt)
{
  console.Debug() << "CLatticeMaster::addVectorTriangleSaveField\n";
  
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  //send cmd to slave
  cmd_buffer.broadcast(CMD_ADD_VTF);

  AFieldMaster* new_fm = new VectorTriangleFieldMaster(&m_world_comm,fieldname,meshname,filename,savetype,t_0,t_end,dt);
  m_save_fields.push_back(new_fm);

  barrier.wait("CLatticeMaster::addVectorTriangleSaveField");

  console.Debug() << "end CLatticeMaster::addVectorTriangleSaveField\n";
}

/*!
  add a vector interaction field to the list of fields to be saved

  \param filename the name of the file the field is saved into 
  \param fieldname the name of the field
  \param igname the name of the interaction group from which the field is taken
  \param savetype output file format
  \param t_0 first timestep to be saved
  \param t_end last timestep to be saved
  \param dt timesteps between saves
  \param checked choose between normal and checked field (defaults to false)
*/
void CLatticeMaster::addVectorInteractionSaveField(const string& filename,const string& fieldname,const string& igtype,
						   const string& igname,const string& savetype,int t_0,int t_end,int dt,bool checked)
{
  console.Debug() << "CLatticeMaster::addVectorInteractionSaveField()\n";
  
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  //send cmd to slave
  cmd_buffer.broadcast(CMD_ADD_VIF);

  AFieldMaster* new_fm = new VectorInteractionFieldMaster(&m_world_comm,fieldname,igtype,igname,filename,savetype,t_0,t_end,dt,checked);
  m_save_fields.push_back(new_fm);

  barrier.wait("CLatticeMaster::addVectorInteractionSaveField");

  console.Debug() << "end CLatticeMaster::addVectorInteractionSaveField()\n";
} 


/*!
  add a scalar interaction field to the list of fields to be saved

  \param filename the name of the file the field is saved into 
  \param fieldname the name of the field
  \param igname the name of the interaction group from which the field is taken
  \param savetype output file format
  \param t_0 first timestep to be saved
  \param t_end last timestep to be saved
  \param dt timesteps between saves
  \param tag the particle tag
  \param mask the mask used for tag comparisons
	\param checked choice between "full" and "checked" fields
*/
void CLatticeMaster::addTaggedScalarInteractionSaveField(const string& filename, const string& fieldname,const string& igtype,const string& igname,const string& savetype, int t_0,int t_end,int dt,int tag,int mask, bool checked)
{
  console.Debug() << "CLatticeMaster::addTaggedScalarInteractionSaveField(" << filename << ","<< fieldname << "," << igname << "," << t_0 << "," << t_end << "," << dt << ")\n";

  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  //send cmd to slave
  cmd_buffer.broadcast(CMD_ADD_SIF);

  AFieldMaster* new_fm=new ScalarInteractionFieldMaster(&m_world_comm,fieldname,igtype,igname,filename,savetype,t_0,t_end,dt,tag,mask,checked);
  m_save_fields.push_back(new_fm);

  barrier.wait("CLatticeMaster::addTaggedScalarInteractionSaveField");

  console.Debug() << "end CLatticeMaster::addTaggedScalarInteractionSaveField()\n";
} 

/*!
  Collect particle data from the sublattices and save it as openDX file
  the header is saved as fnameXXX.general, the data as fnameXXX.data, where
  XXX=t

  \param fname the filename
  \param t running number of saved files
*/
void CLatticeMaster::saveDataAsDX(const char* fname,int t)
{
  int fnlen=strlen(fname)+1;
  int sumparts=0;
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CMPIBarrier barrier(MPI_COMM_WORLD);
  
  cmd_buffer.broadcast(CMD_COLLECT);
  //-- setup file names ---
  char *hname=new char[fnlen+20];
  char *dname=new char[fnlen+20];
  char tnbr[10];
  sprintf(tnbr,"%d",t);
  strncpy(hname,fname,fnlen);
  strcat(hname,tnbr);
  strcat(hname,".general");
  strncpy(dname,fname,fnlen);
  strcat(dname,tnbr);
  strcat(dname,".data");
  //-- collect particle data ---
  CMPIVarSGBufferRoot buffer(MPI_COMM_WORLD,36000);
  buffer.gather();
  //-- extract and save data ---
  ofstream data_file(dname);
  for(int i=1;i<m_global_size;i++){
    int nparts=buffer.pop_int(i);
    sumparts+=nparts;
    for(int j=0;j<nparts;j++){
      CParticle temp_part=extractCParticleFrom(&buffer,i);
      temp_part.writeAsDXLine(data_file,i);
    }
  }
  data_file.close();
  //-- write header ---
  ofstream header_file(hname);
  header_file << "file = " << dname << endl;
  header_file << "points = " << sumparts << endl;
  header_file << "format = text" << endl;
  header_file << "interleaving = field" << endl;
  header_file << "field = locations, sublatticeID, radius, mass, displacement, velocity" << endl;
  header_file << "structure = 3-vector, scalar, scalar, scalar, 3-vector, 3-vector" << endl;
  header_file << "type = float, float, float, float, float, float" << endl;
  header_file << endl << "end" << endl;
  header_file.close();
  //-- clean up ---
  barrier.wait("saveDataAsDX");
  delete hname,dname;
} 

void CLatticeMaster::setDXSaving(const string& filename,int dt)
{
  m_dxfilename=filename;
  m_dxt=dt;
  m_is_dx=true;
}

/*!
  run the simulation
*/
void CLatticeMaster::run()
{
  runInit() ;

  m_pTimers->start("TotalRunTime");
	
  while (m_t<m_max_ts){
    runOneStep() ;
  }
  m_pTimers->clear();
  runEnd() ;
}

/*!
  Initialisation to run the simulation
*/
void CLatticeMaster::runInit()
{
  console.Debug() << "timer resolution : " << MPI_Wtick() << "\n" ;
  m_isInitialized = true ;
  m_t=0 ; 
  m_pTimers->clear();
}

/*!
  Finalize after running the simulation
*/
void CLatticeMaster::runEnd()
{
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  console.Debug() << "total time: " << m_pTimers->getTiming("TotalRunTime") << " seconds for " << m_max_ts-1 << " time steps\n" ;
  cmd_buffer.broadcast(CMD_FINISH);
  barrier.wait("runEnd");
  barrier.wait("runEnd2");
}

void CLatticeMaster::saveTimingData()
{
  if (!getTimingFileName().empty()) {

    m_pTimers->start("WorkerTimingDataSave");
    CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
    CMPIBarrier barrier(MPI_COMM_WORLD);
    cmd_buffer.broadcast(CMD_SAVETIMINGDATA);
    barrier.wait("saveTimingData");
    m_pTimers->stop("WorkerTimingDataSave");

    m_pTimers->appendData(getTimingFileName());    
  }
}

/**
 *  Perform a single time step of the simulation.
 */
void CLatticeMaster::runOneStep()
{
  m_pTimers->start("RunOneStep");

  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  m_t++ ; 
  if ((m_t % 10)==0) {
    console.Info() << "Time step: " << m_t << "\n" ;
  } else {
    console.Debug() << "Time step: " << m_t << "\n" ;
  }

  m_pTimers->start("LoadingMechanism");
  SParserAccess::Parser()->Eval("loadingMechanism");
  m_pTimers->stop("LoadingMechanism");

  m_pTimers->start("NeighbourSearch");
  if(m_t==1){  
    searchNeighbors(true);
  } else {
    searchNeighbors(false);
  }
  m_pTimers->stop("NeighbourSearch");

  //  barrier.wait("runOneStep.2");
  m_pTimers->start("UpdateInteractions");
  updateInteractions();
  m_pTimers->stop("UpdateInteractions");

  if (m_t == 1) {
    m_pCheckPointController->performCheckPoint(0);
  }

  m_pTimers->start("CalcForcesAndExchange");
  oneStep();
  m_pTimers->stop("CalcForcesAndExchange");

  m_pTimers->start("SavingData");
  // saving fields
  bool app;
  if(m_t==1){ 
    app=false; // if 1st time step -> new file
  } else {
    app=true; // else append
  }
  for(vector<SSaveParams>::iterator iter=m_save_parameters.begin();iter!=m_save_parameters.end();iter++){
    if((m_t % iter->dt)==0){
      saveSingleField(iter->fieldname,iter->filename,app,iter->type);
    }
  }
  if(m_is_dx){
    if((m_t % m_dxt)==0){
      saveDataAsDX(m_dxfilename.c_str(),m_t);
    }
  }

  // new field saving
  m_pTimers->start("FieldMasterCollect");
  m_pTimers->pause("FieldMasterCollect");
  m_pTimers->start("FieldMasterWrite");
  m_pTimers->pause("FieldMasterWrite");
  for(vector<AFieldMaster*>::iterator iter=m_save_fields.begin();
      iter!=m_save_fields.end();
      iter++){
    if((*iter)->needSave(m_t)){
      m_pTimers->resume("FieldMasterCollect");    
      cmd_buffer.broadcast(CMD_SEND_FIELDS);
      (*iter)->collect();
      barrier.wait("runOneStep.collect");
      m_pTimers->pause("FieldMasterCollect");      
      m_pTimers->resume("FieldMasterWrite");
      (*iter)->write();
      m_pTimers->pause("FieldMasterWrite");
    }
  }
  m_pTimers->stop("FieldMasterCollect");  
  m_pTimers->stop("FieldMasterWrite");

  // Save a check-point (if required)
  m_pTimers->start("CheckPoint");
  m_pCheckPointController->performCheckPoint(m_t);
  m_pTimers->stop("CheckPoint");

  m_pTimers->stop("SavingData");

  m_pTimers->stop("RunOneStep");

  m_pTimers->stop("TotalRunTime", true);

  saveTimingData();

  console.Timing() << "updateInteractions : " << m_pTimers->getTiming("UpdateInteractions") << " seconds\n" ;
  console.Timing() << "searchNeighbors    : " <<  m_pTimers->getTiming("NeighbourSearch") << " seconds\n" ;
  console.Timing() << "one step           : " <<  m_pTimers->getTiming("CalcForcesAndExchange") << " seconds\n" ;
  console.Timing() << "saving data        : " <<  m_pTimers->getTiming("SavingData") << " seconds\n" ;
}


/*!
  Create and add a new bonded IG

  \param name the name of the bonded IG
  \param k spring constant of the created bonds
  \param dist maximum distance between two particles to create bond
  \param break_dist (relative) breaking distance of bonds
  
*/
void CLatticeMaster::addBondedIG(int tag,const string& name,double k,double break_dist)
{
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CVarMPIBuffer pbuffer(MPI_COMM_WORLD,20);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  console.XDebug() << "CLatticeMaster::addBondedIG()\n ";

  // send the command
  cmd_buffer.broadcast(CMD_ADDBONDEDIG);
  // send parameters
  pbuffer.append(tag);
  pbuffer.append(name.c_str());
  pbuffer.append(k);
  pbuffer.append(break_dist);

  pbuffer.broadcast(0);

  // broadcast connection data
  m_world_comm.broadcast_cont(m_temp_conn[tag]);

  barrier.wait("addBondedIG()");
  console.XDebug() << "end CLatticeMaster::addBondedIG()" << "\n";
} 

void CLatticeMaster::addRotBondedIG(
  int tag,
  const string& name,
  double kr,
  double ks,
  double kt,
  double kb,
  double max_nForce,
  double max_shFroce,
  double max_tMoment,
  double max_bMoment
)
{
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CVarMPIBuffer pbuffer(MPI_COMM_WORLD, 256);  // 20 need change ?
  CMPIBarrier barrier(MPI_COMM_WORLD);

  console.XDebug() << "CLatticeMaster::addRotBondedIG()\n ";

  // send the command
  cmd_buffer.broadcast(CMD_ADDROTBONDEDIG);
  // send parameters
  pbuffer.append(tag);
  pbuffer.append(name.c_str());
  pbuffer.append(kr);
  pbuffer.append(ks);
  pbuffer.append(kt);
  pbuffer.append(kb);
  pbuffer.append(max_nForce);
  pbuffer.append(max_shFroce);
  pbuffer.append(max_tMoment);
  pbuffer.append(max_bMoment);

  pbuffer.broadcast(0);

  // broadcast connection data
  m_world_comm.broadcast_cont(m_temp_conn[tag]);

  barrier.wait("addRotBondedIG()");
  console.XDebug() << "end CLatticeMaster::addRotBondedIG()" << "\n";
}

/*!
  add group of particle-particle interactions

  \param type
  \param name
  \param params
*/
void CLatticeMaster::addPairIG(const string& type, const string& name, list<OpValue> params)
{
  console.XDebug()<<"CLatticeMaster::addPairIG\n";
   
  CVarMPIBuffer pbuffer(MPI_COMM_WORLD,2000);
  CMPIBarrier barrier(MPI_COMM_WORLD);
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);

  //send the command
  cmd_buffer.broadcast(CMD_ADDPIG);
  if(type=="Elastic"){
    console.Debug()<<"adding elastic PIG\n";
    console.XDebug()<< "name of the PIG is : " << name.c_str() << "\n";
    pbuffer.append("Elastic");
    list<OpValue>::iterator iter=params.begin();
    pbuffer.append(name.c_str());
    pbuffer.append(iter->Content.d); // k
    pbuffer.broadcast(0);
  } else if(type=="Friction"){
    console.Debug()<<"adding friction PIG\n";
    console.XDebug()<< "name of the PIG is : " << name.c_str() << "\n";
    pbuffer.append("Friction");
    list<OpValue>::iterator iter=params.begin();
    pbuffer.append(name.c_str());
    pbuffer.append(iter->Content.d); iter++; // k
    pbuffer.append(iter->Content.d); iter++; // mu
    pbuffer.append(iter->Content.d); iter++; // k_s
    pbuffer.append(iter->Content.d); // dt
    pbuffer.broadcast(0);
  } else if(type=="AdhesiveFriction"){
    console.Debug()<<"adding adhesive friction PIG\n";
    console.XDebug()<< "name of the PIG is : " << name.c_str() << "\n";
    pbuffer.append("AdhesiveFriction");
    list<OpValue>::iterator iter=params.begin();
    pbuffer.append(name.c_str());
    pbuffer.append(iter->Content.d); iter++; // k
    pbuffer.append(iter->Content.d); iter++; // mu
    pbuffer.append(iter->Content.d); iter++; // k_s
    pbuffer.append(iter->Content.d); iter++; // dt
    pbuffer.append(iter->Content.d); // r_cut
    pbuffer.broadcast(0);
  } else if (type=="FractalFriction") {
    console.Debug()<<"adding fractal friction PIG\n";
    console.XDebug()<< "name of the PIG is : " << name.c_str() << "\n";
    pbuffer.append(type.c_str());
    list<OpValue>::iterator iter=params.begin();
    pbuffer.append(name.c_str());
    //string ffname=string(iter->Content.str); iter++;
    string ffname=string("Data/test1.frac"); //iter++;
    console.XDebug() << "frac file: " << ffname << "\n";
    pbuffer.append(iter->Content.d); iter++; // k
    pbuffer.append(iter->Content.d); iter++; // mu_0
    pbuffer.append(iter->Content.d); iter++; // k_s
    pbuffer.append(iter->Content.d); iter++; // dt
    pbuffer.append(iter->Content.d); iter++; // x0
    pbuffer.append(iter->Content.d); iter++; // y0
    pbuffer.append(iter->Content.d); iter++; // dx
    pbuffer.append(iter->Content.d); iter++; // dy
    int nx=iter->Content.l; iter++;
    pbuffer.append(nx);
    int ny=iter->Content.l; 
    pbuffer.append(ny);
    // read coeff file and put it into buffer
    ifstream infile(ffname.c_str());
    for(int i=0;i<nx*ny;i++){
      double mu;
      infile >> mu;
      console.XDebug() << i << " : " << mu << "\n";
      pbuffer.append(mu);
    }
    infile.close();
    pbuffer.broadcast(0);    
  } else {
    console.Error()<<"trying to add PIG of unknown type\n";
  }
  barrier.wait("addPairIG()");
  console.XDebug()<<"end CLatticeMaster::addPairIG\n";
}

/*!
  add a group of triangle mesh (surface) to particle interactions

  \param type the type of the interaction (elastic...)
  \param name the name of the itneraction group
  \param meshname the name of the mesh 
  \param params the actual interaction parameters (k,...)
*/
void CLatticeMaster::addTriMeshIG(const std::string& type,
				  const std::string& name,
				  const std::string& meshname,
				  std::list<OpValue> params)
{
  console.XDebug() << "begin CLatticeMaster::addTriMeshIG\n";
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CVarMPIBuffer pbuffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  // send command 
  cmd_buffer.broadcast(CMD_ADDTRIMESHIG);

  // send data
  pbuffer.append(type.c_str());
  pbuffer.append(name.c_str());
  pbuffer.append(meshname.c_str());
  if(type=="Elastic"){
    console.XDebug() << "TriMeshIG is elastic\n";
    // send parameters
    list<OpValue>::iterator iter=params.begin();
    pbuffer.append(iter->Content.d); // k
    pbuffer.broadcast(0);
  } else {
    throw std::runtime_error(std::string("Unknown TriMesh interaction type: ")+type);
  }
  barrier.wait("addTriMeshIG");
  
  console.XDebug() << "end CLatticeMaster::addTriMeshIG\n";
}

/*!
  add a group of bonded triangle mesh (surface) to particle interactions

    \param name the name of the itneraction group
    \param meshname the name of the mesh
    \param k spring constant
    \param r_break (relative) breaking distance
    \param buildtype the way to initially build the IG
    \param buildparam the parameters for building the IG
*/
void CLatticeMaster::addBondedTriMeshIG(const std::string& name,
					const std::string& meshname,
					double k,
					double r_break,
					const std::string& buildtype,
					std::list<OpValue> buildparams)
{
  console.XDebug() << "begin CLatticeMaster::addBondedTriMeshIG\n";
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CVarMPIBuffer pbuffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  // send command 
  cmd_buffer.broadcast(CMD_ADDBONDEDTRIMESHIG);

  // send data
  pbuffer.append(name.c_str());
  pbuffer.append(meshname.c_str());
  pbuffer.append(k);
  pbuffer.append(r_break);
  pbuffer.append(buildtype.c_str());
  if(buildtype=="BuildByTag"){
    console.XDebug() << "BuildByTag\n";
    // extract build parameters
    list<OpValue>::iterator iter=buildparams.begin();
    pbuffer.append(int(iter->Content.l)); iter++; // tag
    pbuffer.append(int(iter->Content.l)); 
    // send params
    pbuffer.broadcast(0);
  } else if (buildtype=="BuildByGap"){
    list<OpValue>::iterator iter=buildparams.begin();
    pbuffer.append(iter->Content.d); // max. gap 
    // send params
    pbuffer.broadcast(0);
  } else {
    throw std::runtime_error(std::string("Unknown BTriMeshInteraction build type: ")+buildtype);
  }
  barrier.wait("readBondedTriMeshIG");
  
  console.XDebug() << "end CLatticeMaster::addBondedTriMeshIG\n";
}

/*!
  add a triangle mesh

  \param name the name of the mesh
  \param filename the name of the mesh file
*/
void CLatticeMaster::addTriMesh(const string& name, const string& filename)
{
  console.XDebug() << "begin CLatticeMaster::addTriMesh\n";
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  CVarMPIBuffer pbuffer(MPI_COMM_WORLD);

  // send command 
  cmd_buffer.broadcast(CMD_ADDTRIMESH);

  pbuffer.append(name.c_str());
  pbuffer.broadcast(0);
  readAndDistributeTriMesh(filename);

  console.XDebug() << "end CLatticeMaster::addTriMesh\n";
}

/*!
  read a triangle mesh from a file and distribute the data to the workers

  \param meshfilename the filename
*/
void CLatticeMaster::readAndDistributeTriMesh(const std::string& meshfilename)
{
  console.XDebug() << "CLatticeMaster::readAndDistributeTriMesh(" << meshfilename << ")\n";
  // buffers
  CMPIBarrier barrier(MPI_COMM_WORLD);
  vector<MeshNodeData> node_send_buffer;
  vector<MeshTriData> tri_send_buffer;

  // mesh reader
  MeshReader reader(meshfilename);


  // --- Nodes ---
  MeshReader::NodeIterator &niter=reader.getNodeIterator();
  // read nodes into vector
  while(niter.hasNext()){
    node_send_buffer.push_back(niter.next());
  }
 
  // --- Triangles ---
  MeshReader::TriIterator &titer=reader.getTriIterator();
  // read triangles into vector
  while(titer.hasNext()){
    tri_send_buffer.push_back(titer.next());
  }

  // distribute the node vector
  m_world_comm.broadcast_cont_packed(node_send_buffer);
  // distribute the triangle vector
  m_world_comm.broadcast_cont_packed(tri_send_buffer);

  barrier.wait("readAndDistributeTriMesh");
  console.XDebug() << "end CLatticeMaster::readAndDistributeTriMesh\n";
}

/*!
  add body force

  \param type the type of force, i.e. Gravity,...
  \param prms the interaction parameters
*/
void CLatticeMaster::addBodyForce(const string& type, const esys::lsm::BodyForceIGP &prms)
{
  console.XDebug() << "begin CLatticeMaster::addBodyForce\n";  
  if (type == "Gravity") {
    console.XDebug()<<"CLatticeMaster::addSingleIG\n";
     
    CVarMPIBuffer pbuffer(MPI_COMM_WORLD,20);
    CMPIBarrier barrier(MPI_COMM_WORLD);
    CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
  
    //send the command
    cmd_buffer.broadcast(CMD_ADDSIG);
    // console.Debug()<<"adding Gravity IG\n";
    pbuffer.append(type.c_str());
    prms.packInto(&pbuffer);
    pbuffer.broadcast(0);

    barrier.wait("addSingleIG()");
  }
  else {
    throw std::runtime_error(
      std::string("Unknown body force type: ")
      +
      type
    );
  }
  console.XDebug() << "end CLatticeMaster::addBodyForce\n";  
}

void CLatticeMaster::addRotPairIG(const string& type, const string& name, list<OpValue> params)
{
  console.XDebug()<<"CLatticeMaster::addRotPairIG\n";
                                                                                
  CVarMPIBuffer pbuffer(MPI_COMM_WORLD,2000);
  CMPIBarrier barrier(MPI_COMM_WORLD);
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);
                                                                                
  //send the command
  cmd_buffer.broadcast(CMD_ADDROTPIG);
  if(type=="RotElastic"){
    console.Debug()<<"adding Rotelastic PIG\n";
    console.XDebug()<< "name of the PIG is : " << name.c_str() << "\n";
    pbuffer.append("RotElastic");
    list<OpValue>::iterator iter=params.begin();
    pbuffer.append(name.c_str());
    pbuffer.append(iter->Content.d); // k
    pbuffer.broadcast(0);
  
  } else if(type=="RotFriction"){
    console.Debug()<<"adding friction PIG\n";
    console.XDebug()<< "name of the PIG is : " << name.c_str() << "\n";
    pbuffer.append("RotFriction");
    list<OpValue>::iterator iter=params.begin();
    pbuffer.append(name.c_str());
    pbuffer.append(iter->Content.d); iter++; // k
    pbuffer.append(iter->Content.d); iter++; // mu_s
    pbuffer.append(iter->Content.d); iter++; // mu_d
    pbuffer.append(iter->Content.d); iter++; // k_s
    pbuffer.append(iter->Content.d); // dt
    pbuffer.broadcast(0);
  } else {
    console.Error()<<"trying to add PIG of unknown type\n";
 
  }

  barrier.wait("addRotPairIG()");
  console.XDebug()<<"end CLatticeMaster::addRotPairIG\n";
}

void CLatticeMaster::addSingleIG(const string& type, const string& name, list<OpValue> params)
{
  console.XDebug()<<"CLatticeMaster::addSingleIG\n";

  if(type=="Gravity"){

    list<OpValue>::iterator iter=params.begin();
    double x = iter->Content.d;
    iter++;
    double y = iter->Content.d;
    iter++;
    double z = iter->Content.d;
    
    addBodyForce(type, esys::lsm::BodyForceIGP(name, Vec3(x, y, z)));
  }else{
    throw std::runtime_error(
      std::string("Unknown body force type: ")
      +
      type
    );
  }
  console.XDebug()<<"end CLatticeMaster::addSingleIG\n";
}

/*!
  Add damping to a lattice. Send parameters to slaves
*/
void CLatticeMaster::addDamping(const string& type,list<OpValue> params)
{
  console.XDebug()<<"CLatticeMaster::addDamping\n";
   
  CVarMPIBuffer pbuffer(MPI_COMM_WORLD,20);
  CMPIBarrier barrier(MPI_COMM_WORLD);
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);

  //send the command
  cmd_buffer.broadcast(CMD_ADDDAMP);
  if(type=="Damping"){
    // console.Debug()<<"adding Damping\n";
    pbuffer.append("Damping");
    CDampingIGP igp;
    list<OpValue>::iterator iter=params.begin();
    igp.setName("damp");
    igp.setVRef(Vec3(0.0,0.0,0.0));
    igp.setVisc(iter->Content.d);
    console.XDebug()<< "visc : " << iter->Content.d;
    iter++;
    igp.setTimeStep(iter->Content.d);
    console.XDebug()<< "dt : " << iter->Content.d;
    iter++;
    igp.setMaxIter(iter->Content.l);
    igp.packInto(&pbuffer);
    pbuffer.broadcast(0);
  }else{
    console.Error()<<"trying to add Damping of unknown type\n";
  }
  barrier.wait("addDamping()");
  console.XDebug()<<"end CLatticeMaster::addDamping\n";
}

/*!
  set IG s1 as excluding IG in s2, i.e. an two particles interacting in s1 
  can't interact in s2
*/
void CLatticeMaster::addExIG(const string& s1,const string& s2)
{
  console.XDebug()<<"CLatticeMaster::addExIG\n";
   
  CVarMPIBuffer pbuffer(MPI_COMM_WORLD,20);
  CMPIBarrier barrier(MPI_COMM_WORLD);
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,m_global_rank);

  //send the command
  cmd_buffer.broadcast(CMD_EXIG);
  pbuffer.append(s1.c_str());
  pbuffer.append(s2.c_str());
  pbuffer.broadcast(0);
  barrier.wait("addExIG()");
  console.XDebug()<<"end CLatticeMaster::addExIG\n";
}  
