#ifndef __LATTICEMASTER_H
#define __LATTICEMASTER_H

//--- Project includes ---

#include "Parallel/mpibuf.h"
#include "Parallel/mpivbuf.h"
#include "Parallel/LatticeParam.h"
#include "Model/Particle.h"
#include "Model/RotParticle.h"
#include "Parser/mtypes.h"
#include "SocketServer.h"
#include "Fields/FieldMaster.h"

//--- MPI includes ---
#include <mpi.h>

//--- TML includes ---
#include "tml/comm/comm_world.h"

// -- STL includes --
#include <vector>
#include <list>
#include <map>
#include <utility>

using std::vector;
using std::map;
using std::pair;

class CSocketSvc;
class CheckPointController;
namespace esys
{
  namespace lsm
  {
    class GeometryInfo;
    class BodyForceIGP;
  }
}

/*!
  \struct SSaveParams
  \brief Struct containing parameters for saving a field to a file
 
  \warning Preliminary, will change
*/
struct SSaveParams {
    std::string fieldname; //!< name of the field
    std::string filename;  //!< filename
    int dt;           //!< save every dt timesteps
    int type;         //!< field type, i.e. scalar/vector

    enum DataType {FLD_SCALAR, FLD_VECTOR};
};

/*!
  \class CLatticeMaster
  \brief Class for the initialisation and control of a group of sublattices 
 
  \author Steffen Abe, David Place, Shane Latham
  $Revision$
  $Date$
*/
class MpiWTimers;

class CLatticeMaster {
private:
    std::string                           m_timingFileName;
    MpiWTimers                            *m_pTimers;
    CheckPointController                  *m_pCheckPointController;
    esys::lsm::CLatticeParam::ProcessDims m_processDims;

protected:
    map<int,vector<int> > m_temp_conn;
    vector<SSaveParams>   m_save_parameters;
    vector<AFieldMaster*> m_save_fields;

    double m_x_max,m_x_min,m_y_max,m_y_min,m_z_max,m_z_min;
    int m_global_rank;
    int m_global_size;
    int m_max_ts;
    int m_center_id;
    double m_total_time;
    int m_t ;
    double m_dt;
    int m_dxt; // obsolete
    std::string m_dxfilename; // obsolete
    bool m_is_dx; // obsolete
    bool m_isInitialized ;
    std::string m_particle_type;

    void makeWorkerComm();

    void saveTimingData();
    void readAndDistributeTriMesh(const std::string&);

    TML_CommWorld m_world_comm;

public:
    CLatticeMaster();
    ~CLatticeMaster();

    int getTimeStep() const {return m_t;}
    double getTimeStepSize() const {return m_dt;}
    
    void init();
    void run();
    void runInit();
    void runOneStep();
    void runEnd();
    void oneStep();
    void searchNeighbors(bool);
    bool checkNeighbors();
    void updateInteractions();
    void addBondedIG(int,const std::string&,double,double);
    void addPairIG(const std::string&,const std::string&, std::list<OpValue>);
    void addTriMesh(const std::string&,const std::string&);
    void addTriMeshIG(const std::string&,const std::string&,const std::string&,std::list<OpValue>);
    void addBondedTriMeshIG(const std::string&,const std::string&,double,double,const std::string&,std::list<OpValue>);
    void addDamping(const std::string&,std::list<OpValue>);
    void addBodyForce(const std::string& type, const esys::lsm::BodyForceIGP &prms);
    void addSingleIG(const std::string&,const std::string&,std::list<OpValue>);
    void addExIG(const std::string&,const std::string&);
    void setNumSteps(int s);
    int getNumSteps() {return m_max_ts;};
    int getSteps() {return m_t;};
    void addRotPairIG(const string&,const string&,list<OpValue>);  
    void addRotBondedIG(int,const string&,double,double,double,double,double,double,double,double);  
    //void addRotDamping(const string&,list<OpValue>);
    
    void initWall(const Vec3&,const Vec3&,double);
    void initBondedWall(const Vec3&,const Vec3&,double,int);
    void initViscWall(const Vec3&,const Vec3&,double,double,int);
    void initSoftBondedWall(const Vec3&,const Vec3&,double,double,double,int);
    void moveParticleTo(int,const Vec3&);
    void setParticleNonDynamic(int);
    void setParticleVel(int,const Vec3&);
    void moveWallBy(int,const Vec3&);
    void setVelocityOfWall(int,const Vec3&);
    void tagParticleNearestTo(int,int,const Vec3&);
    void applyForceToWall(int,const Vec3&);
    void moveSingleNodeBy(const string&,int,const Vec3&);
    void moveTaggedNodesBy(const string&,int,const Vec3&);

    void saveDataAsDX(const char*,int); // ??
    void setDXSaving(const std::string&,int); // ??
    void saveTimingDataToFile(const std::string &fileNamePrefix);

    /**
     * Enforces particles to be restricted to motion in the x-y plane.
     */
    void do2dCalculations();

    /**
     * Sets explicit process partitioning info for MPI_Dims_create.
     *
     * @param dims Dimension process partition vector.
     *             dims[0] x-partitioning,
     *             dims[1] y-partitioning and
     *             dims[2] z-partitioning. A zero value for a dimension
     *             is converted to a non-zero value using MPI_Dims_create
     */
    void setProcessDims(const esys::lsm::CLatticeParam::ProcessDims &dims);

    const esys::lsm::CLatticeParam::ProcessDims &getProcessDims() const;

    /*!
     * Specifies a file in which timing results are written.
     *
     * \param fileName Name of file, overwritten if it exists.
     */
    void setTimingFileName(const std::string &fileName);

    /*!
     * Returns the name of the file to which timing results are written.
     *
     * \return Name of file, empty std::string if no timing results are being
     * recorded.
     */
    const std::string &getTimingFileName() const;

    //! functions to send data to the FE via a CSocketSvc
    bool sendParticleData(CSocketSvc *socket) ; // remove ??
    //bool queryFieldAvailable(CSocketSvc *socket) ;
    //bool sendFieldData(CSocketSvc *,CShakePacket&);

    //! functions to save data to file
    void addSaveField(const std::string&, const std::string&,int,const std::string&);
    void saveSingleField(const std::string&, const std::string&, bool,int);

    //! "new" field saving functions
    void addScalarParticleSaveField(const std::string&,const std::string&,const std::string&,int,int,int);
    void addTaggedScalarParticleSaveField(const std::string&,const std::string&,const std::string&,int,int,int,int,int);
    void addVectorParticleSaveField(const std::string&,const std::string&,const std::string&,int,int,int);
    void addTaggedVectorParticleSaveField(const std::string&,const std::string&,const std::string&,int,int,int,int,int);
    void addScalarInteractionSaveField(const std::string&,const std::string&,const std::string&,const std::string&,const std::string&,int,int,int,bool checked=false);
    void addVectorInteractionSaveField(const std::string&,const std::string&,const std::string&,const std::string&,const std::string&,int,int,int,bool checked=false);
    void addTaggedScalarInteractionSaveField(const std::string&,const std::string&,const std::string&,const std::string&,const std::string&,int,int,int,int,int,bool);
    void addTaggedScalarParticleDistributionSaver(const std::string&,const std::string&,const std::string&,int,int,int,int,int,int,double,double,int);
    void addVectorTriangleSaveField(const string&,const string&,const string&,const string&,int,int,int);

    /**
     * Initialises parameters for performing model checkpointing.
     *
     * @param fileNamePrefix Path prefix for checkpoint files. Multiple checkpoint
     *                       files may be generated for a single timestep snapshot.
     * @param beginTime      Time to begin checkpointing. Time of first checkpoint.
     * @param endTime        End time for checkpointing. Time of last checkpoint.
     * @param timeInterval   Time interval between checkpoint file generation.
     */
    void performCheckPoints(
        const std::string &fileNamePrefix,
        int beginTime,
        int endTime,
        int timeInterval
    );

    //! initialization functions
    void makeLattice(const char*,double,double,double);
    void initComplexFromFile(const std::string&,int,int,int);// to be removed

    /**
     * Defines the bounding box in which particles may roam.
     */
    void setSpatialDomain(const Vec3 &minBBoxPt, const Vec3 &maxBBoxPt);
    
    /**
     * Returns whether the setSpacialDomain method has been called with a
     * non-zero volume bounding box.
     */
    bool haveSetSpatialDomain() const;

    /**
     * Gathers slave dimensions and coordinates.
     */
    void getSlaveSpatialDomains();

    /**
     * Initialises the bounding-box and circular boundary conditions
     * of the lattice geometry.
     *
     * @param geoInfo Object containing bounding box of all particles.
     */
    void initialiseGeometry(const esys::lsm::GeometryInfo &geoInfo);

    /**
     * Template method for reading geometry from a specified file.
     * The template parameter is the type of particles created from
     * the particle data contained in the geometry file.
     *
     * @param fileName the name of the file containing geometry info.
     */
    template <class TmplParticle>
    void  readGeometry(const std::string &fileName);

    /**
     * Reads an initial particle configuration from a geometry file.
     *
     * @param fileName Name of file containing geometry information.
     */
    void readGeometryFile(const std::string &fileName);

    /**
     * Reads particle and interaction configuration from a check-point
     * summary file.
     *
     * @param fileName Name of file containing summary of check-point
     *                 data.
     */
    void loadCheckPointData(const std::string &checkPointFileName);

    /**
     * Adds enumeration of particles to the lattice.
     *
     * @param it Iterator object with next() and hasNext() methods.
     *           The TmplIterator::next() method is required to
     *           return an object which can accepted as a constructor
     *           argument for the CParticle class.
     */
    template <class TmplIterator, class TmplParticle>
    void addParticles(TmplIterator &it);

    /**
     * Adds enumeration of connections/bonds to the lattice.
     *
     * @param it Iterator object with next() and hasNext() methods.
     *           The TmplIterator::next() method is required to
     *           return an object which can accepted as a constructor
     *           argument for the SimpleConnectionData class.
     */
    template <class TmplIterator>
    void addConnections(TmplIterator &it);
};

#endif
