#ifndef __SUBLATTICE_H
#define __SUBLATTICE_H

// -- project includes --

#include "Parallel/mpibuf.h"
#include "Parallel/mpivbuf.h"
#include "Parallel/ASubLattice.h"
#include "Parallel/LatticeParam.h"

#include "Foundation/shake_math.h"
#include "Foundation/Timer.h"

#include "TriMesh.h"

// -- interactions --
#include "Model/Interaction.h"
#include "Model/InteractionParam.h"
#include "Model/EWallInteractionGroup.h"
#include "Model/BWallInteractionGroup.h"
#include "Model/ViscWallIG.h"
#include "Model/SoftBWallInteractionGroup.h"

#include "Model/RotDampingGroup.h"
#include "Model/RotBondedInteraction.h"
#include "Model/RotElasticInteraction.h"
#include "Model/RotFricInteraction.h"

// -- fields --
#include "Fields/Field.h"

#include "Fields/FieldSlave.h"
#include "Fields/ScalarParticleFieldSlave.h"
#include "Fields/VectorParticleFieldSlave.h"
#include "Fields/ScalarInteractionFieldSlave.h"

//--- TML includes ---
#include "tml/comm/comm.h"
#include "tml/comm/comm_world.h"

//--- MPI includes ---
#include <mpi.h>

//--- STL includes ---
#include <vector>
#include <map>
#include <string>

using std::vector;
using std::map;
using std::string;

template <class T> class ParallelParticleArray;
class AParallelInteractionStorage;

/*!
  \class TSubLattice
  \brief Class for a Sublattice

  \author Steffen Abe
  $Revision$
  $Date$
*/
template <typename T>
class TSubLattice : public ASubLattice
{
 protected:
  // -- particles storage--
  ParallelParticleArray<T> *m_ppa;

  // -- interaction storage --
  map<string,AParallelInteractionStorage*> m_dpis; //! dynamic interactions
  map<string,AParallelInteractionStorage*> m_bpis; //! persistent interactions
  
  typedef map<string,AInteractionGroup<T>*> NameIGroupMap;
  NameIGroupMap m_singleParticleInteractions; //! Body/gravitational force interactions

  // -- special interaction groups --
  map<string,AParallelInteractionStorage*> m_damping;
  vector<AWallInteractionGroup<T>*> m_WIG;
  map<string,TriMesh*> m_mesh; 

  // -- parameters --
  double m_dt; // time step for integration
  double m_nrange; // search range for neighbor table
  //  double m_rmax; // max. particle radius
  double m_alpha; // search range padding

  int m_last_ns; // timestamp of last neighborsearch
 
  // -- MPI --
  int m_rank; //!<global rank
  MPI_Comm m_comm;
  MPI_Comm m_worker_comm; //!< MPI communicator between workers (excl. master)
  TML_Comm m_tml_worker_comm; //!< TML version of the communicator between workers (excl. master)
  int m_num_workers;
  TML_CommWorld m_world_comm;
  /**
   * Process partitioning, non-zero entries are taken literally,
   * zero entries are computed via MPI_Dims_create.
   */
  esys::lsm::CLatticeParam::ProcessDims m_dims;

  // -- times --
  double packtime; 
  double unpacktime;
  double commtime;
  double forcetime;

  // -- fields --
  map<string,AField*> m_fields; //!< the fields available in the SubLattice

  // -- "new" fields --
  map<int,AFieldSlave*> m_field_slaves;
  
  MpiWTimers *m_pTimers;

  // -- private functions --
  void calcForces();
  void zeroForces();
  void integrate(double);
 
 public:
  TSubLattice(const esys::lsm::CLatticeParam &prm, int rank, MPI_Comm comm);
  virtual ~TSubLattice();

  virtual vector<int> getCommCoords() const {return m_ppa->getCommCoords();};
  virtual vector<int> getCommDims() const {return m_ppa->getCommDims();};

  virtual void receiveParticles(TML_Comm*);

  virtual void addOwnedParticleByParam(double,double,const Vec3&,int); // deprecated ?
  void addOwnedParticle(T*); // deprecated ?

  virtual const MPI_Comm &getWorkerComm() const {return m_worker_comm;}
  
  virtual void initNeighborTable(const Vec3&,const Vec3&);
  virtual void initNeighborTable(const Vec3&,const Vec3&,const vector<bool>&);
  virtual void addWall(const CEWallIGP*);
  virtual void addBondedWall(const CBWallIGP*);
  virtual void addViscWall(const CVWallIGP*);
  virtual void addSoftBondedWall(const CSoftBWallIGP*);
  virtual void addTriMesh();
  virtual void addTriMeshIG();
  virtual void addBondedTriMeshIG();
  virtual void addPairIG();
  virtual void addRotPairIG();
  virtual void addSingleIG();
  virtual void addBondedIG();
  virtual void addRotBondedIG();
  virtual void addDamping();
 // virtual void addRotDamping();
  virtual void setExIG();

  virtual void oneStep();
  virtual void exchangePos();
  void setTimeStep(double dt){m_dt=dt;};

  virtual void rebuildParticleArray();
  virtual void rebuildInteractions();
  virtual void searchNeighbors();
  virtual void updateInteractions();
  virtual void checkNeighbors();
  void resetDisplacements();
  
  virtual void do2dCalculations();

  virtual void moveParticleTo();
  virtual void moveWallBy();
  virtual void applyForceToWall();
  virtual void setParticleVelocity();
  virtual void setVelocityOfWall();
  virtual void setParticleNonDynamic();
  virtual void tagParticleNearestTo();
  virtual void moveSingleNode();
  virtual void moveTaggedNodes();

  virtual void sendDataToMaster();
  virtual void countParticles();

  virtual void printStruct();
  virtual void printData();
  virtual void printTimes();

  virtual void makeFields();
  virtual void countFieldData();
  virtual void collectFieldData();

  //-- "new" field functions --
  virtual void addScalarParticleField();
  virtual void addVectorParticleField();
  virtual void addScalarInteractionField();
  virtual void addVectorInteractionField();
  virtual void addVectorTriangleField();
  virtual void sendFieldData();

  virtual void printField(const string&); //!< field debugging
  virtual void printAllFields(); 
  
  virtual void saveCheckPointData(std::ostream &oStream);
  virtual void loadCheckPointData(std::istream &iStream);
  
  virtual void setTimer(MpiWTimers &timers) {m_pTimers = &timers;}
};

#include "SubLattice.hpp"

#endif
