// -- project includes --

#include "Parallel/SubLattice.h"
#include "Parallel/MpiInfo.h"
#include "Parallel/mpivbuf.h"
#include "Parallel/mpisgvbuf.h"
#include "Parallel/mpibarrier.h"
#include "Parallel/mpia2abuf.h"
#include "Model/BondedInteraction.h"
//#include "Model/DampingGroup.h"
#include "Model/Damping.h"
#include "Model/FrictionInteraction.h"
#include "Model/FractalFriction.h"
#include "Model/AdhesiveFriction.h"
#include "Model/VWFrictionInteraction.h"
#include "Model/RotFricInteraction.h"
#include "Model/RotElasticInteraction.h"
#include "Model/MeshData.h"
#include "Model/ETriMeshInteraction.h"
#include "Model/BTriMeshInteraction.h"

// --- parallel storage includes ---
#include "ppa/src/pp_array.h"
#include "pis/pi_storage_eb.h"
#include "pis/pi_storage_ed.h"
#include "pis/pi_storage_ne.h"
#include "pis/pi_storage_single.h"
#include "pis/trimesh_pis.h"
#include "pis/trimesh_pis_ne.h"
#include "pis/trimesh_pis_eb.h"

// --- field includes ---
#include "Fields/ScalarParticleFieldSlaveTagged.h"
#include "Fields/VectorParticleFieldSlaveTagged.h"
#include "Fields/ScalarInteractionFieldSlaveTagged.h"
#include "Fields/ScalarParticleFieldSlaveTagged.h"
#include "Fields/ScalarInteractionFieldSlaveTagged.h"
#include "Fields/CheckedScalarInteractionFieldSlave.h"
#include "Fields/CheckedScalarInteractionFieldSlaveTagged.h"
#include "Fields/VectorTriangleFieldSlave.h"

#include "Model/BodyForceGroup.h"

#include <mpi.h>
#include <stdlib.h>
#include <stdexcept>

// -- STL includes --
#include <algorithm>
#include <stdexcept>

using std::runtime_error;

// -- IO includes --
#include <iostream>
using std::cerr;
using std::flush;
using std::endl;
using esys::lsm::CLatticeParam;

//----------------------------------------------
//   TSubLattice functions
//----------------------------------------------

/*!
  construct SubLattice
 
  \param param Lattice parameters
  \param rank the MPI rank
  \param comm the MPI communicator
*/
template <class T>
TSubLattice<T>::TSubLattice(const CLatticeParam &param, int rank, MPI_Comm comm)
  : m_dims(3, 0),
    m_pTimers(NULL)
{
  // -- MPI stuff  --
  m_rank=rank;

  m_comm=comm;
  m_dims = param.processDims();
  
  // -- setup communicator for workers only --
  MPI_Group group_world,group_workers;
  int ranks[1];
  ranks[0]=0;
  MPI_Comm_group(MPI_COMM_WORLD,&group_world);
  MPI_Group_excl(group_world,1,ranks,&group_workers);
  MPI_Comm_create(MPI_COMM_WORLD,group_workers,&m_worker_comm);
  MPI_Comm_size(m_worker_comm,&m_num_workers);
  m_tml_worker_comm=TML_Comm(m_worker_comm);


  // -- set parameters
  m_dt=param.dt();
  m_alpha=param.alpha();
  m_nrange=param.nrange();
  console.Debug()<< "dt,nrange,alpha : " << m_dt << " , " << m_nrange << " , " << m_alpha << "\n";
  
  commtime=0.0;
  packtime=0.0;
  unpacktime=0.0;
  forcetime=0.0;
  
  m_last_ns=-1;
}

template <class T>
TSubLattice<T>::~TSubLattice()
{
	for(typename vector<AWallInteractionGroup<T>*>::iterator vit=m_WIG.begin();vit!=m_WIG.end();vit++)
	{
		delete *vit;
	}
	if(m_ppa!=NULL) delete m_ppa;
}

/*!
  Initialize particle storage. The dimensions are global.
 
  \param min minimum corner
  \param max maximum corner
*/
template <class T>
void TSubLattice<T>::initNeighborTable(const Vec3& min,const Vec3& max)
{
  console.XDebug() << "TSubLattice<T>::initNeighborTable(" << min <<  "," << max << ")\n";
  // make size fit range
  double xsize=max.X()-min.X();
  xsize=m_nrange*ceil(xsize/m_nrange);
  double ysize=max.Y()-min.Y();
  ysize=m_nrange*ceil(ysize/m_nrange);
  double zsize=max.Z()-min.Z();
  zsize=m_nrange*ceil(zsize/m_nrange);
  Vec3 grow=Vec3(xsize,ysize,zsize)-(max-min); // size increase
  Vec3 nmin=min-0.5*grow; // distribute symmetrically
  Vec3 nmax=max+0.5*grow;
  console.XDebug() << "range=" << m_nrange << ", new min,max: " << nmin << ", " << nmax << "\n";
  
  // construct particle array
  TML_Comm *ntcomm=new TML_Comm(m_worker_comm);
  m_ppa=new  ParallelParticleArray<T>(ntcomm,m_dims,nmin,nmax,m_nrange,m_alpha);
  //m_ppa=new  ParallelParticleArray<T>(ntcomm,3,nmin,nmax,m_nrange);
  
  makeFields(); // put here to make sure ppa is initialized before makeFields
  
  console.XDebug() << "end TSubLattice<T>::initNeighborTable\n";
}

template <class T>
void TSubLattice<T>::do2dCalculations()
{
  T::setDo2dCalculations(true);
}


/*!
  Initialize particle storage with some circular boundaries. 
  The dimensions are global.
 
  \param min minimum corner
  \param max maximum corner
*/
template <class T>
void TSubLattice<T>::initNeighborTable(const Vec3& min,const Vec3& max,const vector<bool>& circ)
{
  console.XDebug() << "TSubLattice<T>::initNeighborTable(" << min <<  "," << max << ") circ\n";
  double xsize,ysize,zsize;
  // if dimension is circular, check if size fits range, otherwise make it fit
  // x - dim
  if(circ[0])
  {
    xsize=max.X()-min.X();
    if(fabs(xsize/m_nrange-lrint(xsize/m_nrange))>1e-6)
    {
      console.Critical() << "circular x-dimension doesn't fit range !\n";
    }
    //xsize+=2.0*m_nrange; // padding on the circular ends
  }
  else
  {
    xsize=max.X()-min.X();
    xsize=m_nrange*ceil(xsize/m_nrange);
  }
  // y - dim
  if(circ[1])
  {
    ysize=max.Y()-min.Y();
    if(fabs(ysize/m_nrange-lrint(ysize/m_nrange))>1e-6)
    {
      console.Critical() << "circular y-dimension doesn't fit range !\n";
    }
    ysize+=2.0*m_nrange; // padding on the circular ends
  }
  else
  {
    ysize=max.Y()-min.Y();
    ysize=m_nrange*ceil(ysize/m_nrange);
  }
  // z - dim
  if(circ[2])
  {
    zsize=max.Z()-min.Z();
    if(fabs(zsize/m_nrange-lrint(zsize/m_nrange))>1e-6)
    {
      console.Critical() << "circular z-dimension doesn't fit range !\n";
    }
    zsize+=2.0*m_nrange; // padding on the circular ends
  }
  else
  {
    zsize=max.Z()-min.Z();
    zsize=m_nrange*ceil(zsize/m_nrange);
  }
  Vec3 grow=Vec3(xsize,ysize,zsize)-(max-min); // size increase
  Vec3 nmin=min-0.5*grow; // distribute symmetrically
  Vec3 nmax=max+0.5*grow;
  console.XDebug() << "range, new min, max: " << m_nrange << " " << nmin << nmax << "\n";
  // construct particle array
  TML_Comm *ntcomm=new TML_Comm(m_worker_comm);
  m_ppa=new ParallelParticleArray<T>(ntcomm,m_dims,circ,nmin,nmax,m_nrange,m_alpha);
  
  makeFields(); // put here to make sure ppa is initialized before makeFields
  
  console.XDebug() << "end TSubLattice<T>::initNeighborTable (circ)\n";
}

/*!
  Receive particle from a TML Communicator
 
  \param comm the Communicator
  \warning the type of particle is not checked
*/
template <class T>
void TSubLattice<T>::receiveParticles(TML_Comm* comm)
{
  console.XDebug() << "TSubLattice<T>::receiveParticles()\n";

  vector<T> recv_buffer;
  CMPIBarrier barrier(MPI_COMM_WORLD);

  //comm->recv_scatter_packed(recv_buffer,0); // assuming process 0 is root
  comm->recv_broadcast_cont_packed(recv_buffer,0);
  console.XDebug() << "recvd " << recv_buffer.size() << " particles \n";
  m_ppa->insert(recv_buffer);

  barrier.wait("TSubLattice<T>::receiveParticles");

  console.XDebug() << "end TSubLattice<T>::receiveParticles()\n";
}


/*!
  Construct particle which is owned by this SubLattice and add it
 
  \param rad particle radius
  \param mass particle mass
  \param pos particle position
  \param id global particle id
 
*/
template <class T>
void TSubLattice<T>::addOwnedParticleByParam(double rad,double mass,const Vec3& pos,int id)
{
  T* CPo=new T(rad,mass,pos,Vec3(0.0,0.0,0.0),Vec3(0.0,0.0,0.0),id);
  addOwnedParticle(CPo);
}

/*!
  Add particle which is owned by this SubLattice
 
  \param particle pointer to the particle
*/
template <class T>
void TSubLattice<T>::addOwnedParticle(T* particle)
{
  m_ppa->insert(*particle);
}


/*!
  Add wall to the sublattice and setup the WallInteractionGroup
 
  \param wallparam the parameters of the wall
*/
template <class T>
void TSubLattice<T>::addWall(const CEWallIGP* wallparam)
{
  console.XDebug() << "TSubLattice<T>::addWall\n" ;
  
  AWallInteractionGroup<T>* newCEWIG=new CEWallInteractionGroup<T>(&m_tml_worker_comm,wallparam);
  newCEWIG->addToFields(&m_fields);
  newCEWIG->Update(m_ppa);
  m_WIG.push_back(newCEWIG);
}

/*!
  Add bonded wall to the sublattice and setup the WallInteractionGroup

  \param wallparam the parameters of the wall
*/
template <class T>
void TSubLattice<T>::addBondedWall(const CBWallIGP* wallparam)
{
  console.XDebug() << "TSubLattice<T>::addBondedWall\n" ;
  
  AWallInteractionGroup<T>* newCEWIG=new CBWallInteractionGroup<T>(&m_tml_worker_comm,wallparam);
  newCEWIG->addToFields(&m_fields);
  newCEWIG->Update(m_ppa);
  m_WIG.push_back(newCEWIG);
}


/*!
  Add wall with viscous drag.

  \param wallparam  the parameters of the wall
*/
template <class T>
void TSubLattice<T>::addViscWall(const CVWallIGP* wallparam)
{
  console.XDebug() << "TSubLattice<T>::addViscWall\n" ;
        
  AWallInteractionGroup<T>* newCVWIG=new CViscWallIG<T>(&m_tml_worker_comm,wallparam);
  newCVWIG->addToFields(&m_fields);
  newCVWIG->Update(m_ppa);
  m_WIG.push_back(newCVWIG);
}

/*!
  Add bonded wall with per-direction spring constants to the
	sublattice and setup the WallInteractionGroup

  \param wallparam the parameters of the wall
*/
template <class T>
void TSubLattice<T>::addSoftBondedWall(const CSoftBWallIGP* wallparam)
{
  console.XDebug() << "TSubLattice<T>::addSoftBondedWall\n" ;
  AWallInteractionGroup<T>* newCEWIG=new CSoftBWallInteractionGroup<T>(&m_tml_worker_comm,wallparam);
  newCEWIG->addToFields(&m_fields);
  newCEWIG->Update(m_ppa);
  m_WIG.push_back(newCEWIG);
}



/*!
  Add a PairInteractionGroup to the lattice
*/
template <class T>
void TSubLattice<T>::addPairIG()
{
  console.XDebug()  << "TSubLattice<T>::addPairIG()\n";
  CVarMPIBuffer param_buffer(MPI_COMM_WORLD,2000);
  AParallelInteractionStorage* new_pis = NULL;

  // get params
  param_buffer.receiveBroadcast(0);
  string type = param_buffer.pop_string();
  console.XDebug()<< "PIG type: " << type.c_str() << "\n";
  string name = param_buffer.pop_string();
  console.XDebug()<< "PIG name: " << name.c_str() << "\n";

  if(type=="Elastic")
  {
    CElasticIGP eigp;
    eigp.m_k=param_buffer.pop_double();
    new_pis=new ParallelInteractionStorage_NE<T,CElasticInteraction>(m_ppa,eigp);
  } else if (type=="Friction") {
    CFrictionIGP figp;
    figp.k=param_buffer.pop_double();
    figp.mu=param_buffer.pop_double();
    figp.k_s=param_buffer.pop_double();
    figp.dt=param_buffer.pop_double();
    console.XDebug() << "k,mu,k_s,dt: " << figp.k << " , " << figp.mu << " , "
    << figp.k_s << " , " << figp.dt << "\n";
    new_pis=new ParallelInteractionStorage_ED<T,CFrictionInteraction>(m_ppa,figp);
  } else if (type=="AdhesiveFriction") {
    CAdhesiveFrictionIGP figp;
    figp.k=param_buffer.pop_double();
    figp.mu=param_buffer.pop_double();
    figp.k_s=param_buffer.pop_double();
    figp.dt=param_buffer.pop_double();
    figp.r_cut=param_buffer.pop_double();
    console.XDebug()
      << "k,mu,k_s,dt,r_cut: " << figp.k << " , " << figp.mu << " , "
      << figp.k_s << " , " << figp.dt << " " << figp.r_cut << "\n";
    new_pis=new ParallelInteractionStorage_ED<T,CAdhesiveFriction>(m_ppa,figp);
  } else if (type=="FractalFriction") {
    FractalFrictionIGP figp;
    figp.k=param_buffer.pop_double();
    figp.mu_0=param_buffer.pop_double();
    figp.k_s=param_buffer.pop_double();
    figp.dt=param_buffer.pop_double();
    console.XDebug() << "k,mu_0,k_s,dt: " << figp.k << " , " << figp.mu_0 << " , "
                     << figp.k_s << " , " << figp.dt << "\n";
    figp.x0=param_buffer.pop_double();
    figp.y0=param_buffer.pop_double();
    figp.dx=param_buffer.pop_double();
    figp.dy=param_buffer.pop_double();
    figp.nx=param_buffer.pop_int();
    figp.ny=param_buffer.pop_int();
    console.XDebug() 
      <<"x0,y0,dx,dy,nx,ny: "
      << figp.x0 << " , " << figp.y0 << " , "
      << figp.dx << " , " << figp.dy << " ,"
      << figp.nx << " , " << figp.ny << "\n";
    figp.mu = std::auto_ptr<double>(new double[figp.nx*figp.ny]);
  
    for(int i=0;i<figp.nx*figp.ny;i++)
    {
      (figp.mu.get())[i]=param_buffer.pop_double();
      // console.XDebug() << i << " , " << figp.mu[i] << "\n";
    }
    new_pis = new ParallelInteractionStorage_ED<T,CFractalFriction>(m_ppa,figp);
  } else if(type=="VWFriction") {
    VWFrictionIGP figp;
  
    figp.k=param_buffer.pop_double();
    figp.mu=param_buffer.pop_double();
    figp.k_s=param_buffer.pop_double();
    figp.dt=param_buffer.pop_double();
    figp.m_alpha=param_buffer.pop_double();
    console.XDebug()
      << "k,mu,k_s,dt,alpha: " << figp.k << " , " << figp.mu << " , "
      << figp.k_s << " , " << figp.dt << "\n";
    new_pis=new ParallelInteractionStorage_ED<T,CVWFriction>(m_ppa,figp);
  } else {
    cerr 
      << "Unknown interaction group name "
      << type
      << " in TSubLattice<T>::addPairIG()" << endl;
  }

  // add InteractionGroup to map
  m_dpis.insert(make_pair(name,new_pis));
  console.XDebug()  << "end TSubLattice<T>::addPairIG()\n";
}


/*!
  Add a PairInteractionGroup to the lattice
*/
template <class T>
void TSubLattice<T>::addRotPairIG()
{
  console.XDebug()  << "TSubLattice<T>::addRotPairIG()\n";
  CVarMPIBuffer param_buffer(MPI_COMM_WORLD,2000);
  AParallelInteractionStorage* new_pis;

  // get params
  param_buffer.receiveBroadcast(0);
  string type=string(param_buffer.pop_string());
  console.XDebug()<< "PIG type: " << type.c_str() << "\n";
  string name=string(param_buffer.pop_string());
  console.XDebug()<< "PIG name: " << name.c_str() << "\n";

  if(type=="RotElastic"){
    CRotElasticIGP reigp;
    reigp.m_kr=param_buffer.pop_double();
    new_pis=new ParallelInteractionStorage_NE<CRotParticle,CRotElasticInteraction>(m_ppa,reigp);
  } else if (type=="RotFriction"){
    CRotFrictionIGP rfigp;
    rfigp.k=param_buffer.pop_double();
    rfigp.mu_s=param_buffer.pop_double();
    rfigp.mu_d=param_buffer.pop_double();    
    rfigp.k_s=param_buffer.pop_double();
    rfigp.dt=param_buffer.pop_double();
    console.XDebug()
      << "k,mu_d,mu_s, k_s,dt: " << rfigp.k << " , "
      << rfigp.mu_d << " , " << rfigp.mu_s << " , "
      << rfigp.k_s << " , " << rfigp.dt << "\n";

    new_pis=new ParallelInteractionStorage_ED<CRotParticle,CRotFrictionInteraction>(m_ppa,rfigp);
  } else {cout << "try to add a wrong type " << endl;  }

  // add InteractionGroup to map

  m_dpis.insert(make_pair(name,new_pis));
  console.XDebug()  << "end TSubLattice<T>::addRotPairIG()\n";
}

/*!
  Add a Triangle Mesh
*/
template <class T>
void TSubLattice<T>::addTriMesh()
{
  console.XDebug()  << "TSubLattice<T>::addTriMesh()\n";

  // MPI buffers
  CVarMPIBuffer param_buffer(MPI_COMM_WORLD);
  // data buffers
  vector<MeshNodeData> node_recv_buffer;
  vector<MeshTriData> tri_recv_buffer;

  // receive params
  param_buffer.receiveBroadcast(0);

  // extract name 
  string name = param_buffer.pop_string();
  console.XDebug()<< "TriMesh name: " << name.c_str() << "\n";

  // receive nodes
  m_world_comm.recv_broadcast_cont_packed(node_recv_buffer,0);
  console.XDebug() << "recvd " << node_recv_buffer.size() << " nodes \n";

  // receive triangles
  m_world_comm.recv_broadcast_cont_packed(tri_recv_buffer,0);
  console.XDebug() << "recvd " << tri_recv_buffer.size() << " triangles \n";

  // load mesh into new TriMesh
  TriMesh* new_tm=new TriMesh(); 
  new_tm->LoadMesh(node_recv_buffer,tri_recv_buffer);

  m_mesh.insert(make_pair(name,new_tm));  

  console.XDebug()  << "end TSubLattice<T>::addTriMesh()\n";
}

/*!
  Add a TriMesh interaction group
*/
template <class T>
void TSubLattice<T>::addTriMeshIG()
{
  console.XDebug()  << "TSubLattice<T>::addTriMeshIG()\n";

  // MPI buffers
  CVarMPIBuffer param_buffer(MPI_COMM_WORLD);

  // receive params
  param_buffer.receiveBroadcast(0);

  // extract name & type
  string type = param_buffer.pop_string();
  console.XDebug()<< "TriMeshIG type: " << type.c_str() << "\n";
  string name = param_buffer.pop_string();
  console.XDebug()<< "TriMeshIG name: " << name.c_str() << "\n";
  string meshname = param_buffer.pop_string();
  console.XDebug()<< "TriMeshIG mesh name: " << meshname.c_str() << "\n";

  // get pointer to mesh
  TriMesh* tmp=m_mesh[meshname];
  if(tmp==NULL){
    throw runtime_error("unknown mesh name in TSubLattice<T>::addTriMeshIG:" + meshname);
  }
  // switch on type,extract params & construc new TriMeshIG
  if(type=="Elastic")
  {
    AParallelInteractionStorage* new_pis;
    ETriMeshIP tmi;
    tmi.k=param_buffer.pop_double();
    new_pis = new TriMesh_PIS_NE<T,ETriMeshInteraction>(tmp,m_ppa,tmi);
    m_dpis.insert(make_pair(name,new_pis));
  } else { // unknown type-> throw
    throw runtime_error("unknown type in TSubLattice<T>::addTriMeshIG:" + type);
  }

  
  console.XDebug()  << "end TSubLattice<T>::addTriMeshIG()\n";
}

/*!
  Add a bonded TriMesh interaction group
*/
template <class T>
void TSubLattice<T>::addBondedTriMeshIG()
{
  console.XDebug()  << "TSubLattice<T>::addBondedTriMeshIG()\n";

  // MPI buffers
  CVarMPIBuffer param_buffer(MPI_COMM_WORLD);

  // receive params
  BTriMeshIP param;
  param_buffer.receiveBroadcast(0);

  // extract name.meshname & params
  string name = param_buffer.pop_string();
  console.XDebug()<< "BTriMeshIG name: " << name.c_str() << "\n";
  string meshname = param_buffer.pop_string();
  console.XDebug()<< "BTriMeshIG mesh name: " << meshname.c_str() << "\n";
  param.k = param_buffer.pop_double();
  console.XDebug()<< "BTriMeshIG k : " << param.k << "\n";
  param.brk = param_buffer.pop_double();
  console.XDebug()<< "BTriMeshIG r_break: " << param.brk << "\n";
  string buildtype = param_buffer.pop_string();
  console.XDebug()<< "BTriMeshIG build type: " << buildtype.c_str() << "\n";
  
  // get pointer to mesh
  TriMesh* tmp=m_mesh[meshname];
  if(tmp==NULL){
    throw runtime_error("unknown mesh name in TSubLattice<T>::addTriMeshIG:" + meshname);
  }
  
  // setup new interaction storage
  TriMesh_PIS_EB<T,BTriMeshInteraction>* new_pis=new TriMesh_PIS_EB<T,BTriMeshInteraction>(tmp,m_ppa,param);
  // switch on buildtype, extract buildparam & build
  if(buildtype=="BuildByTag"){
    int tag=param_buffer.pop_int();
    int mask=param_buffer.pop_int();
    new_pis->buildFromPPATagged(tag,mask);
    m_bpis.insert(make_pair(name,new_pis));
  } else if(buildtype=="BuildByGap"){
    double max_gap=param_buffer.pop_double();
    new_pis->buildFromPPAByGap(max_gap);
    m_bpis.insert(make_pair(name,new_pis));
  } else { // unknown type-> throw
    throw runtime_error("unknown build type in TSubLattice<T>::addBondedTriMeshIG:" + buildtype);
  }

  console.XDebug()  << "end TSubLattice<T>::addBondedTriMeshIG()\n";
}

/*!
  Add a SingleInteractionGroup to the lattice
 
  \todo currently gravity is hardwired, change this
*/
template <class T>
void TSubLattice<T>::addSingleIG()
{
  console.XDebug()  << "TSubLattice<T>::addSingleIG()\n";
  CVarMPIBuffer param_buffer(MPI_COMM_WORLD);

  // get params
  param_buffer.receiveBroadcast(0);

  string type = param_buffer.pop_string();
  console.XDebug()<< "SIG type: " << type.c_str() << "\n";

  // setup InteractionGroup
  if(type=="Gravity"){
    esys::lsm::BodyForceIGP prms = esys::lsm::BodyForceIGP::extract(&param_buffer);

    // add to map
    m_singleParticleInteractions.insert(
      std::pair<string,AInteractionGroup<T>*>(
        prms.Name(),
        new esys::lsm::BodyForceGroup<T>(prms, *m_ppa)
      )
    );
  }
  else {
    throw std::runtime_error(
      std::string("Trying to setup SIG of unknown type: ")
      +
      type
    );
  }

  console.XDebug()  << "end TSubLattice<T>::addSingleIG()\n";
}


/*!
  Add a DampingGroup to the lattice. Receive the parameters from master.
*/
template <class T>
void TSubLattice<T>::addDamping()
{
  console.XDebug()  << "TSubLattice<T>::addDamping()\n";
  CVarMPIBuffer param_buffer(MPI_COMM_WORLD);

  // CDampingGroup<T> *DG;
  AParallelInteractionStorage* DG;
  // get params
  param_buffer.receiveBroadcast(0);

  string type = param_buffer.pop_string();
  console.XDebug()<< "Interaction Group type: " << type.c_str() << "\n";

  // setup InteractionGroup
  if(type=="Damping")
  {
    CDampingIGP *params=extractDampingIGP(&param_buffer);
    DG=new ParallelInteractionStorage_Single<T,CDamping<T> >(m_ppa,*params);
    //DG=new CDampingGroup<CParticle>();
    //DG->setParam(params);
  }
  /*} else if (type=="RotDamping") {

    CDampingIGP *params=extractDampingIGP(&param_buffer);
    DG=new CDampingGroup<CRotParticle>();
    DG->setParam(params);
  }  */
  else
  {
    console.Error() << "Trying to setup Damping of unknown type: " << type << "\n";
    return;
  }

  // add to map
  m_damping.insert(make_pair("Damping",DG));
  console.XDebug()  << "end TSubLattice<T>::addDamping()\n";
}

/*
template <class T>
void TSubLattice<T>::addRotDamping()
{
  console.XDebug()  << "TSubLattice<T>::addRotDamping()\n";
  CVarMPIBuffer param_buffer(MPI_COMM_WORLD);
                                                                                
  CRotDampingGroup<CRotParticle> *DG;
                                                                                
  // get params
  param_buffer.receiveBroadcast(0);
                                                                                
  string type=string(param_buffer.pop_string());
  console.XDebug()<< "Interaction Group type: " << type.c_str() << "\n";
                                                                                
  // setup InteractionGroup
  if(type=="RotDamping"){
    CRotDampingIGP *params=extractRotDampingIGP(&param_buffer);
    DG=new CRotDampingGroup<CRotParticle>();
    DG->setParam(params);
  } else {
    console.Error() << "Trying to setup Damping of unknown type\n";
    return;
  }
                                                                                
  // add to map
  m_rotdamping.push_back(DG);
  console.XDebug()  << "end TSubLattice<T>::addRotDamping()\n";
}


*/

/*!
  Add bonded interaction group to the lattice. Receive the parameters from master. 
  The bonds are created from the neighbor table.
*/
template <class T>
void TSubLattice<T>::addBondedIG()
{
  console.XDebug()  << "TSubLattice<T>::addBondedIG()\n";
  CVarMPIBuffer param_buffer(MPI_COMM_WORLD);

  // get params
  param_buffer.receiveBroadcast(0);
  int tag=param_buffer.pop_int();
  string name = param_buffer.pop_string();
  double k=param_buffer.pop_double();
  double rbreak=param_buffer.pop_double();

  console.XDebug()
    << "Got BondedIG parameters: " << tag
    << " " << name.c_str() << " "
    << k << " " << rbreak << "\n";
  // setup InteractionGroup
  CBondedIGP param;
  param.k=k;
  param.rbreak=rbreak;
  param.tag = tag;
  ParallelInteractionStorage_E<T,CBondedInteraction> *B_PIS = 
    new ParallelInteractionStorage_EB<T,CBondedInteraction>(m_ppa,param);

  // recv broadcast connection data
  console.XDebug()
    << "rank=" << m_world_comm.rank()
    << "TSubLattice<T>::addBondedIG(): receiving conn_data.\n";

  vector<int> conn_data;
  m_world_comm.recv_broadcast_cont(conn_data,0);
  console.XDebug() 
    << "rank=" << m_world_comm.rank()
    << "TSubLattice<T>::addBondedIG(): conn_data.size()="
    << conn_data.size() << "\n";

  vector<int> vi(2,-1);
  for(int i=0;i<conn_data.size();i+=2)
  {
    vi[0] = (conn_data[i]);
    vi[1] = (conn_data[i+1]);
    B_PIS->tryInsert(vi);
  }

  // add InteractionGroup to map
  m_bpis.insert(make_pair(name,B_PIS));

  console.XDebug()  << "end TSubLattice<T>::addBondedIG()\n";
}

/*!
  Add bonded interaction group to the lattice. Receive the parameters from master.
  The bonds are created from the neighbor table.
*/
template <class T>
void TSubLattice<T>::addRotBondedIG()
{
  console.XDebug()  << "TSubLattice<T>::addBondedIG()\n";
  CVarMPIBuffer param_buffer(MPI_COMM_WORLD);

  // get params
  param_buffer.receiveBroadcast(0);
  int tag=param_buffer.pop_int();
  string name=string(param_buffer.pop_string());
  double kr = param_buffer.pop_double();
  double ks = param_buffer.pop_double();
  double kt = param_buffer.pop_double();
  double kb = param_buffer.pop_double();
  double max_nForce  = param_buffer.pop_double();
  double max_shFroce = param_buffer.pop_double();
  double max_tMoment = param_buffer.pop_double();
  double max_bMoment = param_buffer.pop_double();

  console.XDebug() << "Got RotBondedIG parameters: " << tag << " " << name.c_str() << " " << kr << " " << ks << "\n";
  // setup InteractionGroup
  CRotBondedIGP param;
  param.kr=kr;
  param.ks=ks;
  param.kt=kt;
  param.kb=kb;
  param.max_nForce = max_nForce ;
  param.max_shFroce = max_shFroce ;
  param.max_tMoment = max_tMoment ;
  param.max_bMoment = max_bMoment ;
  ParallelInteractionStorage_E<CRotParticle,CRotBondedInteraction> *B_PIS = 
    new ParallelInteractionStorage_EB<CRotParticle,CRotBondedInteraction>(m_ppa,param);

 // recv broadcast connection data
  console.XDebug() << "rank=" << m_world_comm.rank() << "TSubLattice<T>::addBondedIG(): receiving conn_data.\n";
  vector<int> conn_data;
  m_world_comm.recv_broadcast_cont(conn_data,0);
  console.XDebug() << "rank=" << m_world_comm.rank() << "TSubLattice<T>::addBondedIG(): conn_data.size()=" << conn_data.size() << "\n";
  vector<int> vi(2,-1);
  for(int i=0;i<conn_data.size();i+=2){
    vi[0] = (conn_data[i]);
    vi[1] = (conn_data[i+1]);
    B_PIS->tryInsert(vi);
  }

  // add InteractionGroup to map
  m_bpis.insert(make_pair(name,B_PIS));

  console.XDebug()  << "end TSubLattice<T>::addRotBondedIG()\n";
}

/*!
  Set excluding IG. The names of 2 interaction groups are received from the master. 
  The 1st is set to be excluding the 2nd. If one of the named IGs does not exist, 
  the operation is ignored and a warning emitted. 
*/
template <class T>
void TSubLattice<T>::setExIG()
{
  console.XDebug()  << "TSubLattice<T>::addExIG()\n";
  CVarMPIBuffer pbuffer(MPI_COMM_WORLD);

  // get params
  pbuffer.receiveBroadcast(0);
  string s1 = pbuffer.pop_string();
  string s2 = pbuffer.pop_string();

  //console.XDebug()<< s1.c_str()  << "  " <<  s2.c_str() << "\n";
  if((m_bpis[s1]!=NULL)&&(m_dpis[s2]!=NULL))
  {
    m_dpis[s2]->addExIG(m_bpis[s1]);
  }
  else
  {
    console.Error() << "TSubLattice<T>::setExIG() - nonexisting interaction group \n";
  }
  
  console.XDebug()  << "end TSubLattice<T>::addExIG()\n";
}

/*!
  Exchange positions of shared particles with the other sublattices. 
*/
template <class T>
void TSubLattice<T>::exchangePos()
{
  console.XDebug() << "TSubLattice<T>::exchangePos() \n" ;
  
  m_ppa->exchange(&T::getExchangeValues,&T::setExchangeValues);
  
  console.XDebug() << "end TSubLattice<T>::exchangePos()\n" ;
}

/*!
  Reset the forces on all particles and walls to 0
*/
template <class T>
void TSubLattice<T>::zeroForces()
{
  console.XDebug() << "TSubLattice<T>::zeroForces()\n";

  // particles
  m_ppa->forAllParticles(&T::zeroForce);
  // trimeshes
  for(map<string,TriMesh*>::iterator iter=m_mesh.begin();
      iter!=m_mesh.end();
      iter++){
    (iter->second)->zeroForces();
  }
  
  // walls
  for(typename vector<AWallInteractionGroup<T>*>::iterator iter=m_WIG.begin();iter!=m_WIG.end();iter++)
  {
    (*iter)->zeroForce();
  }
  console.XDebug() << "end TSubLattice<T>::zeroForces() \n";
}

/*!
  Calculate the forces for all interactions contained in the sublattice. 
  Interactions contained in more than one sublattice are calculated in 
  each of them. Slightly more computation, but less communication, i.e. 
  saves one syncronisation point (exchange of forces).
*/
template <class T>
void TSubLattice<T>::calcForces()
{
  console.XDebug() << "TSubLattice<T>::calcForces() \n";
  
  // the walls 
  for(typename vector<AWallInteractionGroup<T>*>::iterator it=m_WIG.begin();it!=m_WIG.end();it++)
  {
    (*it)->calcForces();
  }
  // single particle IGs 
  for(
    typename NameIGroupMap::iterator siter=this->m_singleParticleInteractions.begin();
    siter != m_singleParticleInteractions.end();
    siter++
  )
  {
    (siter->second)->calcForces();
  }
  // dynamically created IGs
  for(typename map<string,AParallelInteractionStorage*>::iterator iter=m_dpis.begin();iter!=m_dpis.end();iter++)
  {
    (iter->second)->calcForces();
  }
  // bonded IGs
  for(typename map<string,AParallelInteractionStorage*>::iterator iter=m_bpis.begin();iter!=m_bpis.end();iter++)
  {
    (iter->second)->calcForces();
  }
  // last, but not least - damping
  for(typename map<string,AParallelInteractionStorage*>::iterator iter=m_damping.begin();iter!=m_damping.end();iter++)
  {
    (iter->second)->calcForces();
  }

  console.XDebug() << "end TSubLattice<T>::calcForces() \n";
}

/*!
  Do the time integration for the particles owned by the SubLattice by 1st order method
 
  \param dt the time step
 */
template <class T>
void TSubLattice<T>::integrate(double dt)
{
  console.XDebug() << "TSubLattice<T>::integrate \n";
  m_ppa->forAllParticles(&T::integrate,dt);
  m_ppa->forAllParticles(&T::rescale) ;        
  console.XDebug() << "end TSubLattice<T>::integrate \n";
}

/*!
  Do one step, i.e. calculate forces, velocities, positions
*/
template <class T>
void TSubLattice<T>::oneStep()
{
  zeroForces();
  calcForces();
  integrate(m_dt);
}

template <class T>
void TSubLattice<T>::rebuildParticleArray()
{
  m_ppa->rebuild();
}

template <class T>
void TSubLattice<T>::rebuildInteractions()
{
  CMPIBarrier barrier(m_worker_comm);
  m_pTimers->start("RebuildInteractions");
  m_pTimers->resume("NeighbourSearch");
  for(typename map<string,AParallelInteractionStorage*>::iterator iter=m_bpis.begin();
          iter!=m_bpis.end();
          iter++)
  {
    console.Debug() << "exchg & rebuild PIS " << iter->first << " at node " << m_rank << "\n";
    (iter->second)->exchange();  
    (iter->second)->rebuild();
  }

  for(typename map<string,AParallelInteractionStorage*>::iterator iter=m_dpis.begin();
      iter!=m_dpis.end();
      iter++)
  {
    console.Debug() << "exchg & rebuild PIS " << iter->first << " at node " << m_rank << "\n";
    (iter->second)->exchange();
    m_pTimers->pause("RebuildInteractions");
    m_pTimers->pause("NeighbourSearch");
    barrier.wait("dpis::exchange");
    m_pTimers->resume("RebuildInteractions");
    m_pTimers->resume("NeighbourSearch");
    (iter->second)->rebuild();
  }
  resetDisplacements();
  m_pTimers->stop("RebuildInteractions");
}

/*!
  Search neighbors
*/
template <class T>
void TSubLattice<T>::searchNeighbors()
{
  console.Debug() << "CSubLattice<T>::searchNeighbors()\n";
  CMPIBarrier barrier(getWorkerComm());
  m_pTimers->start("NeighbourSearch");
  m_pTimers->start("RebuildParticleArray");
  rebuildParticleArray();
  m_pTimers->stop("RebuildParticleArray");
  m_pTimers->pause("NeighbourSearch");
  barrier.wait("PPA rebuild");
  rebuildInteractions();
  m_pTimers->stop("NeighbourSearch");
  console.Debug() << "end CSubLattice<T>::searchNeighbors()\n";
}

/*!
  Update the interaction groups from an existing Neighbortable
 
  \todo check for rebuild Neighbortable
*/
template <class T>
void TSubLattice<T>::updateInteractions()
{
  console.Debug() << "updateInteractions() \n";
  console.Debug() << "m_ppa->getTimeStamp() " << m_ppa->getTimeStamp() << " m_last_ns " << m_last_ns << "\n";
  bool need_update=false;

  for(typename map<string,AParallelInteractionStorage*>::iterator iter=m_bpis.begin();
          iter!=m_bpis.end();
          iter++)
  {
    bool n=(iter->second)->update();
    need_update=need_update || n;
  }
  for(typename map<string,AParallelInteractionStorage*>::iterator iter=m_dpis.begin();
          iter!=m_dpis.end();
          iter++)
  {
    bool n=(iter->second)->update();
    need_update=need_update || n;
  }
  if((m_ppa->getTimeStamp() > m_last_ns) || need_update)
  {
    for(typename vector<AWallInteractionGroup<T>*>::iterator it=m_WIG.begin();
        it!=m_WIG.end();
        it++)
    {
      (*it)->Update(m_ppa);
    }
    for(typename map<string,AParallelInteractionStorage*>::iterator iter=m_damping.begin();
         iter!=m_damping.end();
         iter++){
      (iter->second)->update();
    }
    m_last_ns=m_ppa->getTimeStamp();
  }

  console.Debug() << "end TSubLattice<T>::updateInteractions()\n";
}

/*!
  check if any of the owned particles has moved futher than the search range for the neighbor table
  5 Flops(1 dotproduct)/particle looked at
*/
template <class T>
void TSubLattice<T>::checkNeighbors()
{
  console.Debug() << "TSubLattice<T>::checkNeighbors()\n";
  CMPISGBufferLeaf buffer(MPI_COMM_WORLD,0,8);
  double mdsqr=0; // squared max. displacement
  double alpha=0.5*m_alpha;
  double srsqr=alpha*alpha; // squared search range
  vector<Vec3> displ; // displacements
  
  // get displacement data
  m_ppa->forAllParticlesGet(displ,&T::getDisplacement);
  
  // find maximum particle displacement
  vector<Vec3>::iterator it=displ.begin();
  while((it!=displ.end())&&(mdsqr<srsqr))
  {
    double sqdisp=(*it)*(*it);
    mdsqr = ((mdsqr < sqdisp) ? sqdisp : mdsqr);
    it++;
    //console.XDebug() << "sq. disp: " << sqdisp << "\n";
  }
  console.XDebug() << "max squared displacement " << mdsqr << "alpha^2 = " << srsqr << "\n";
  if (mdsqr>srsqr)
  { // replace with append(bool) -> to be implemented
    buffer.append(int(1));
  }
  else
  {
    buffer.append(int(0));
  }
  buffer.send();
  console.Debug() << "end TSubLattice<T>::checkNeighbors()\n";
}


/*!
  Reset the displacement of all particles
*/
template <class T>
void TSubLattice<T>::resetDisplacements()
{
  console.Debug() << "slave " << m_rank << " resetDisplacements()\n";
  m_ppa->forAllParticles(&T::resetDisplacement);
  console.Debug() << "slave " << m_rank << " end resetDisplacements()\n";
}

/*!
  Move all particles with a given tag to a given position. 
  Parameters (tag,posn) are received from master. 
*/
template <class T>
void TSubLattice<T>::moveParticleTo()
{
  console.Debug() << "TSubLattice<T>::moveParticleTo()\n";
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  
  buffer.receiveBroadcast(0); // get data from master
  int tag=buffer.pop_int();
  Vec3 mv=buffer.pop_vector();
  m_ppa->forParticleTag(tag,(void (T::*)(Vec3))(&T::moveToRel),mv);
  console.Debug() << "end TSubLattice<T>::moveParticleTo()\n";
}

/*!
  Move trimesh node by a given amount.
  Parameters (name,id,displacement) are received from master. 
*/
template <class T>
void TSubLattice<T>::moveSingleNode()
{
  console.Debug() << "TSubLattice<T>::moveSingleNode()\n";
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  
  buffer.receiveBroadcast(0); // get data from master
  string name=string(buffer.pop_string());
  int id=buffer.pop_int();
  Vec3 disp=buffer.pop_vector();

  console.Debug() << "id : " << id << " disp " << disp << "\n";

  (m_mesh[name])->moveNode(id,disp);

  console.Debug() << "end TSubLattice<T>::moveSingleNode()\n";
}

/*!
  Move tagged trimesh nodes by a given amount.
  Parameters (name,tag,displacement) are received from master. 
*/
template <class T>
void TSubLattice<T>::moveTaggedNodes()
{
  console.Error() << "TSubLattice<T>::moveTaggedNodes() NOT IMPLEMENTED\n";
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  
  buffer.receiveBroadcast(0); // get data from master
  string name=string(buffer.pop_string());
  int tag=buffer.pop_int();
  Vec3 disp=buffer.pop_vector();
}

/*!
  Tag particle closest to given position. Params received from master.
*/
template <class T>
void TSubLattice<T>::tagParticleNearestTo()
{
  console.Debug() << "TSubLattice<T>::tagParticleNearestTo()\n";
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  
  buffer.receiveBroadcast(0); // get data from master
  int tag=buffer.pop_int();
  int mask=buffer.pop_int();
  Vec3 pos=buffer.pop_vector();
  
  // warning - this is ugly
  T* part_ptr=m_ppa->getParticlePtrByPosition(pos);
  if(part_ptr!=NULL){
    int old_tag=part_ptr->getTag();
    int new_tag=(old_tag & (~mask)) | (tag & mask);
    part_ptr->setTag(new_tag);

    cout << "pos, realpos: " << pos << " " << part_ptr->getPos() << " old tag, new tag " << old_tag << " " << part_ptr->getTag() << endl;
  }
  console.Debug() << "end TSubLattice<T>::tagParticleNearestTo()\n";  
}

/*!
  Make tagged particles non-dynamic, i.e. give them infinite mass.
  Parameters (tag,posn) are received from master. 
*/
template <class T>
void TSubLattice<T>::setParticleNonDynamic()
{
  console.Debug() << "TSubLattice<T>::setParticleNonDynamic()\n";
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  
  buffer.receiveBroadcast(0); // get data from master
  int tag=buffer.pop_int();
  m_ppa->forParticleTag(tag,(void (T::*)())(&T::setNonDynamic));
  console.Debug() << "end TSubLattice<T>::setParticleNonDynamic()\n";
}

/*!
  Move wall by a given vector. Parameters are received from master. 
*/
template <class T>
void TSubLattice<T>::moveWallBy()
{
  console.XDebug() << "TSubLattice<T>::moveWallBy()\n";
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  
  buffer.receiveBroadcast(0); // get data from master
  int id=buffer.pop_int();
  Vec3 mv=buffer.pop_vector();
  if(id<m_WIG.size())
  {
    (m_WIG[id])->moveWallBy(mv);
  }
}

/*!
  Apply given force to a wall. Parameters are received from master. 
*/
template <class T>
void TSubLattice<T>::applyForceToWall()
{
  console.XDebug() << "TSubLattice<T>::applyForceToWall()\n";
  CVarMPIBuffer buffer(MPI_COMM_WORLD);

  buffer.receiveBroadcast(0); // get data from master
  int id=buffer.pop_int();
  Vec3 f=buffer.pop_vector();
  if(id<m_WIG.size()){
    (m_WIG[id])->applyForce(f);
  }
}

/*!
  Set the velocity of a wall. Only for visc. walls, i.e. m_vel is set, 
  but position is not affected. Parameters are received from master. 
*/
template <class T>
void TSubLattice<T>::setVelocityOfWall()
{
  console.XDebug() << "TSubLattice<T>::setVelocityOfWall()\n";
  CVarMPIBuffer buffer(MPI_COMM_WORLD);

  buffer.receiveBroadcast(0); // get data from master
  int id=buffer.pop_int();
  Vec3 v=buffer.pop_vector();
  if(id<m_WIG.size()){
    (m_WIG[id])->setVelocity(v);
  }
}

/*!
  Set the velocity of a particle. Parameters are received from master. 
*/
template <class T>
void TSubLattice<T>::setParticleVelocity()
{
  console.Debug() << "TSubLattice<T>::setParticleVelocity()\n";
  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  
  buffer.receiveBroadcast(0); // get data from master
  int id=buffer.pop_int();
  Vec3 mv=buffer.pop_vector();
  m_ppa->forParticle(id,(void (T::*)(Vec3))(&T::setVel),mv);
  console.XDebug() << "end TSubLattice<T>::setParticleVelocity()\n";
}

/*!
  Send data of owned particles to the master
*/
template <class T>
void TSubLattice<T>::sendDataToMaster()
{
  console.Debug() << "TSubLattice<T>::sendDataToMaster()\n";
  vector<Vec3> positions;
  vector<double> radii;
  vector<int> ids;
  
  m_ppa->forAllParticlesGet(positions,(Vec3 (T::*)() const)(&T::getPos));
  m_ppa->forAllParticlesGet(radii,(double (T::*)() const)(&T::getRad));
  m_ppa->forAllParticlesGet(ids,(int (T::*)() const)(&T::getID));
  
  m_world_comm.send_gather(positions,0);
  m_world_comm.send_gather(radii,0);
  m_world_comm.send_gather(ids,0);
  
  console.Debug() << "end TSubLattice<T>::sendDataToMaster()\n";
}

/*!
  Send number of owned particles to the master
*/
template <class T>
void TSubLattice<T>::countParticles()
{
  console.Debug()<<"TSubLattice<T>::countParticles()\n";
  CMPIVarSGBufferLeaf buffer(MPI_COMM_WORLD,0);
  //-- pack particles into buffer
  buffer.append(m_ppa->size());
  // send
  buffer.send();
}

/*!
  Print structural information 
*/
template <class T>
void TSubLattice<T>::printStruct()
{
  cout<< "My Rank : " << m_rank << "\n" ;
  if(m_ppa!=NULL)
  {
          cout << *m_ppa << endl;
  }
}

template <class T>
void TSubLattice<T>::printData()
{
  cout << "Data: my rank : " << m_rank << "particles : \n" ;
  m_ppa->forAllParticles((void (T::*)())(&T::print));
}

template <class T>
void TSubLattice<T>::printTimes()
{
  console.Debug() << "time spent calculating force : " << forcetime << " sec\n";
  console.Debug() << "time spent communicating     : " << commtime << " sec\n";
  console.Debug() << "time spent packing           : " << packtime << " sec\n";
  console.Debug() << "time spent unpacking         : " << unpacktime << " sec\n";
}

//======= FIELD FUNCTIONS ==========================

/*!
  Setup fields for the lattice
*/
template <class T>
void TSubLattice<T>::makeFields()
{
  console.XDebug() << "TSubLattice<T>::makeFields() not impl \n";
  m_fields=T::generateFields(m_ppa);
}


/*!
  Get size of a field and send it to the master
*/
template <class T>
void TSubLattice<T>::countFieldData()
{
  console.XDebug() << "TSubLattice<T>::countFieldData()\n";

  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIVarSGBufferLeaf sgbuffer(MPI_COMM_WORLD,0);
  int ndata;

  buffer.receiveBroadcast(0); // get field name
  string fieldname = buffer.pop_string();

  AField* f=m_fields[fieldname];
  if(f!=NULL)
  {
    ndata=f->getSize();
  }
  else
  {
    ndata=0;
  }

  sgbuffer.append(ndata);
  sgbuffer.send();

  console.XDebug() << "end TSubLattice<T>::countFieldData()\n";
}

/*!
  send field data to master
*/
template <class T>
void TSubLattice<T>::collectFieldData()
{
  console.XDebug() << "TSubLattice<T>::collectFieldData()\n";
  
  CVarMPIBuffer buffer(MPI_COMM_WORLD,36000);
  CMPIVarSGBufferLeaf sgbuffer(MPI_COMM_WORLD,0,36000);
  int i;
  
  buffer.receiveBroadcast(0); // get field name
  string fieldname = buffer.pop_string();
  AField* theField = m_fields[fieldname];
  
  console.XDebug() << "TSubLattice<T>::collectFieldData: " <<  " fieldName "  << fieldname  << "\n";
  console.XDebug() << "TSubLattice<T>::collectFieldData: " <<  " theField "  << theField  << "\n";
  if(theField==NULL)
  {
    console.Error() << "unknown field  name :  " << fieldname <<  "\n";
    console.Error() << "known fields  are : \n" ;
    printAllFields();
  }
  else
  {
    // update field values
    theField->updateFieldValues();
    //-- pack data into buffer
    int max_idx=theField->getSize();
    console.XDebug() << "Field size: " << max_idx << "\n";
    sgbuffer.append(max_idx);
    if(theField->isScalar())
    {
      for(i=0;i<max_idx;i++)
      {
        console.XDebug() << "index : " << i << "\n";
        sgbuffer.append((*theField)(i,0));
        sgbuffer.append(theField->idx(i));
      }
    }
    else if (theField->isVector())
    {
      for(i=0;i<max_idx;i++)
      {
        for (int j=0; j < 3; j++)
        {
          console.XDebug()
            << "TSubLattice<T>::collectFieldData: "
            << "(*theField)(" << i << "," << j << ") = " << (*theField)(i,j) << "\n";
          sgbuffer.append((*theField)(i,j));
        }
        sgbuffer.append(theField->idx(i));
      }
    }
  }
  sgbuffer.send();
  console.XDebug() << "end TSubLattice<T>::collectFieldData()\n";
}

/*!
  add scalar per-particle field to saver list
*/
template <class T>
void TSubLattice<T>::addScalarParticleField()
{
  console.Debug() << "TSubLattice<T>::addScalarParticleField\n";
  string fieldname;
  int id,is_tagged;
  
  m_world_comm.recv_broadcast_cont(fieldname,0);
  console.XDebug() << "recvd. fieldname: " << fieldname << "\n";
  m_world_comm.recv_broadcast(id,0);
  console.XDebug() << "recvd. id: " << id << "\n";
  m_world_comm.recv_broadcast(is_tagged,0);
  console.XDebug() << "recvd. is_tagged: " << is_tagged << "\n";
  
  typename T::ScalarFieldFunction rdf=T::getScalarFieldFunction(fieldname);
  ScalarParticleFieldSlave<T> *new_spfs;
  if(is_tagged==0)
  {
    new_spfs=new ScalarParticleFieldSlave<T>(&m_world_comm,m_ppa,rdf);
  }
  else
  {
    int tag,mask;
    m_world_comm.recv_broadcast(tag,0);
    console.XDebug() << "recvd. tag: " << tag << "\n";
    m_world_comm.recv_broadcast(mask,0);
    console.XDebug() << "recvd. mask: " << mask << "\n";
    new_spfs=new ScalarParticleFieldSlaveTagged<T>(&m_world_comm,m_ppa,rdf,tag,mask);
  }
  m_field_slaves.insert(make_pair(id,new_spfs));
  
  console.Debug() << "end TSubLattice<T>::addScalarParticleField\n";
}

/*!
  add vector per-particle field to saver list
*/
template <class T>
void TSubLattice<T>::addVectorParticleField()
{
  console.XDebug() << "TSubLattice<T>::addVectorParticleField\n";
  string fieldname;
  int id,is_tagged;

  m_world_comm.recv_broadcast_cont(fieldname,0);
  console.XDebug() << "recvd. fieldname: " << fieldname << "\n";
  m_world_comm.recv_broadcast(id,0);
  console.XDebug() << "recvd. id: " << id << "\n";
  m_world_comm.recv_broadcast(is_tagged,0);
  console.XDebug() << "recvd. is_tagged: " << is_tagged << "\n";

  typename T::VectorFieldFunction rdf=T::getVectorFieldFunction(fieldname);
  VectorParticleFieldSlave<T> *new_vpfs;
  if(is_tagged==0)
  {
    new_vpfs=new VectorParticleFieldSlave<T>(&m_world_comm,m_ppa,rdf);
  }
  else
  {
    int tag,mask;
    m_world_comm.recv_broadcast(tag,0);
    console.XDebug() << "recvd. tag: " << tag << "\n";
    m_world_comm.recv_broadcast(mask,0);
    console.XDebug() << "recvd. mask: " << mask << "\n";
    new_vpfs=new VectorParticleFieldSlaveTagged<T>(&m_world_comm,m_ppa,rdf,tag,mask);
  }
  m_field_slaves.insert(make_pair(id,new_vpfs));

  console.Debug() << "end TSubLattice<T>::addVectorParticleField\n";
}


/*!
  add scalar per-interaction field to saver list
*/
template <class T>
void TSubLattice<T>::addScalarInteractionField()
{
  console.Debug() << "TSubLattice<T>::addScalarInteractionField\n";
  string fieldname;
  string igname;
  string igtype;
  int id,is_tagged,tag,mask,is_checked;
  
  m_world_comm.recv_broadcast_cont(fieldname,0);
  console.XDebug() << "recvd. fieldname: " << fieldname << "\n";
  m_world_comm.recv_broadcast(id,0);
  console.XDebug() << "recvd. id: " << id << "\n";
  m_world_comm.recv_broadcast_cont(igname,0);
  console.XDebug() << "recvd. interaction group name: " << igname << "\n";
  m_world_comm.recv_broadcast_cont(igtype,0);
  console.XDebug() << "recvd. interaction group name: " << igtype << "\n";
  m_world_comm.recv_broadcast(is_tagged,0);
  console.XDebug() << "recvd. is_tagged: " << is_tagged << "\n";
  
  // get interaction group
  map<string,AParallelInteractionStorage*>::iterator it=m_dpis.find(igname);
  if(is_tagged==1)
  {
          m_world_comm.recv_broadcast(tag,0);
          m_world_comm.recv_broadcast(mask,0);
  }
  m_world_comm.recv_broadcast(is_checked,0);
  console.XDebug() << "recvd. is_checked: " << is_checked << "\n";
  
  if(it!=m_dpis.end())
  {
    console.XDebug() << "found interaction group in dynamic\n";
    AFieldSlave* new_sifs;
    new_sifs=(it->second)->generateNewScalarFieldSlave(&m_world_comm,fieldname,is_checked,is_tagged,tag,mask);
    m_field_slaves.insert(make_pair(id,new_sifs));
  }
  else
  {
    it=m_bpis.find(igname);
    if(it!=m_bpis.end()){
      console.XDebug() << "found interaction group in bonded\n";
      AFieldSlave* new_sifs;
      new_sifs=(it->second)->generateNewScalarFieldSlave(&m_world_comm,fieldname,is_checked,is_tagged,tag,mask);
      m_field_slaves.insert(make_pair(id,new_sifs));
    }
    else // not in dynamic or bonded -> try damping
      {
	//typename map<string,CDampingGroup<T>*>::iterator it2=m_damping.find(igname);
	it=m_damping.find(igname);
	if(it!=m_damping.end()) // found it in damping
	  {
	    AFieldSlave* new_sifs;
	    new_sifs=(it->second)->generateNewScalarFieldSlave(&m_world_comm,fieldname,is_checked,is_tagged,tag,mask);
	    m_field_slaves.insert(make_pair(id,new_sifs));
	  } 
	else // still not found -> unknown name -> error 
	  {
            cerr << "ERROR : unknown interaction group name in addScalarInteractionField " << endl;
	  }
      }
  }
  
  console.Debug() << "end TSubLattice<T>::addScalarInteractionField\n";
}

/*!
  add per-interaction vector field to saver list
*/
template <class T>
void TSubLattice<T>::addVectorInteractionField()
{
  console.Debug() << "TSubLattice<T>::addVectorInteractionField\n";
  string fieldname;
  string igname;
  string igtype;
  int id,is_tagged,tag,mask,is_checked;
  
  m_world_comm.recv_broadcast_cont(fieldname,0);
  console.XDebug() << "recvd. fieldname: " << fieldname << "\n";
  m_world_comm.recv_broadcast(id,0);
  console.XDebug() << "recvd. id: " << id << "\n";
  m_world_comm.recv_broadcast_cont(igname,0);
  console.XDebug() << "recvd. interaction group name: " << igname << "\n";
  m_world_comm.recv_broadcast_cont(igtype,0);
  console.XDebug() << "recvd. interaction group name: " << igtype << "\n";
  m_world_comm.recv_broadcast(is_tagged,0);
  console.XDebug() << "recvd. is_tagged: " << is_tagged << "\n";
  
  // get interaction group
  map<string,AParallelInteractionStorage*>::iterator it=m_dpis.find(igname);
  if(is_tagged==1)
  {
    m_world_comm.recv_broadcast(tag,0);
    m_world_comm.recv_broadcast(mask,0);
  }
  m_world_comm.recv_broadcast(is_checked,0);
  console.XDebug() << "recvd. is_checked: " << is_checked << "\n";
  
  if(it!=m_dpis.end())
  {
    console.XDebug() << "found interaction group in dynamic\n";
    AFieldSlave* new_sifs;
    new_sifs=(it->second)->generateNewVectorFieldSlave(&m_world_comm,fieldname,is_checked,is_tagged,tag,mask);
    m_field_slaves.insert(make_pair(id,new_sifs));
  }
  else
  {
    it=m_bpis.find(igname);
    if(it!=m_bpis.end()){
      console.XDebug() << "found interaction group in bonded\n";
      AFieldSlave* new_sifs;
      new_sifs=(it->second)->generateNewVectorFieldSlave(&m_world_comm,fieldname,is_checked,is_tagged,tag,mask);
      m_field_slaves.insert(make_pair(id,new_sifs));
    }
    else // not in dynamic or bonded -> try damping
      {
	//typename map<string,CDampingGroup<T>*>::iterator it2=m_damping.find(igname);
	it=m_damping.find(igname);
	if(it!=m_damping.end()) // found it in damping
	  {
	    AFieldSlave* new_sifs;
	    new_sifs=(it->second)->generateNewVectorFieldSlave(&m_world_comm,fieldname,is_checked,is_tagged,tag,mask);
	    m_field_slaves.insert(make_pair(id,new_sifs));
	  } 
	else // still not found -> unknown name -> error 
	  {
            cerr << "ERROR : unknown interaction group name in addScalarInteractionField " << endl;
	  }
      }
  }
  
  console.Debug() << "end TSubLattice<T>::addVectorInteractionField\n";
}

/*!
  Add a per-triangle vector field saver. Data received from master.
*/
template <class T>
void TSubLattice<T>::addVectorTriangleField()
{
  console.Debug() << "TSubLattice<T>::addVectorTriangleField()\n";
  string fieldname;
  string meshname;
  int id;

  // receive params
  m_world_comm.recv_broadcast_cont(fieldname,0);
  console.XDebug() << "recvd. fieldname: " << fieldname << "\n";
  m_world_comm.recv_broadcast_cont(meshname,0);
  console.XDebug() << "recvd. meshname: " << meshname << "\n";
  m_world_comm.recv_broadcast(id,0);
  console.XDebug() << "recvd. id: " << id << "\n";

  // get reader function
  Triangle::VectorFieldFunction rdf=Triangle::getVectorFieldFunction(fieldname);
  // check it 
  if(rdf==NULL){
    console.Critical() << "NULL rdf !!!\n";
  }
  VectorTriangleFieldSlave* new_vtfs=new VectorTriangleFieldSlave(&m_world_comm,m_mesh[meshname],rdf);
  m_field_slaves.insert(make_pair(id,new_vtfs));
  console.Debug() << "end TSubLattice<T>::addVectorTriangleField()\n";
}

/*!
  send field data to master
*/
template <class T>
void TSubLattice<T>::sendFieldData()
{
  console.Debug() << "TSubLattice<T>::sendFieldData()\n";
  // receive id's of field to send
  int id;
  m_world_comm.recv_broadcast(id,0);
  console.Debug()  << "received field id " << id << " for data collection\n" ;
  if(m_field_slaves[id] != NULL)
  {
    m_field_slaves[id]->sendData();
  }
  else
  { // can not happen
    cerr << "NULL pointer in m_field_slaves!" << endl;
  }
  // call the sending function of the field
  console.Debug() << "end TSubLattice<T>::sendFieldData()\n";
}

//======= FIELD DEBUGGING FUNCTIONS ================

/*!
  Print all fields available in the SubLattice
  --Use for debugging only--
*/
template <class T>
void TSubLattice<T>::printAllFields()
{
  for(map<string,AField*>::iterator iter=m_fields.begin();iter!=m_fields.end();iter++)
  {
    printField(iter->first);
  }
}

/*!
  Print a field via console.XDebug(). If no field of the given name exists, print an error message. --Use for debugging only--
 
  \param fname the name of the field to be printed
*/
template <class T>
void TSubLattice<T>::printField(const string& fname)
{
  AField* theField=m_fields[fname];
  if(theField!=NULL)
  {
    int max_idx=theField->getSize();
    console.XDebug() << "Field " << fname.c_str() << " contains " << max_idx << "values \n";
    if(theField->isScalar())
    {
      console.XDebug() << "scalar\n";
      for(int i=0;i<max_idx;i++)
      {
        console.XDebug() << (*theField)(i) << "\n";
      }
    }
    if(theField->isVector())
    {
      console.XDebug() << "vector\n";
      for(int i=0;i<max_idx;i++)
      {
        console.XDebug() << "[" << (*theField)(i,0) << " " << (*theField)(i,1) << " " << (*theField)(i,2) << "]\n";
      }
    }
  }
  else
  {
    console.XDebug() << "Error in printField : no field named " << fname.c_str() << "\n";
  }
}

// ---- Checkpointing ----------

template <class T>
void TSubLattice<T>::saveCheckPointData(std::ostream &oStream)
{
  typedef ParallelParticleArray<T> ParticleArray;
  
  ParticleArray &particleArray = dynamic_cast<ParticleArray &>(*m_ppa);
  typename ParticleArray::ParticleIterator particleIt(particleArray.getInnerParticleIterator());
  
  const std::string delim = "\n";
  
  oStream << particleIt.getNumRemaining() << delim;
  while (particleIt.hasNext()) {
    particleIt.next().saveCheckPointData(oStream);
    oStream << delim;
  }

  int numInteractions = 0;
#if 0
  typedef std::map<string,AParallelInteractionStorage*> NameBondedInteractionsMap;
  typedef ParallelInteractionStorage_E<T,CBondedInteraction> BondedInteractions;
  
  typename NameBondedInteractionsMap::iterator it;
  

  for (it = m_bpis.begin(); it != m_bpis.end(); it++) {
    BondedInteractions *pBondedInteractions = dynamic_cast<BondedInteractions *>(it->second);
    if (pBondedInteractions != NULL) {
      numInteractions += pBondedInteractions->getInnerInteractionIterator().getNumRemaining();
    }
  }
#endif

  oStream << numInteractions << delim;
#if 0  
  for (it = m_bpis.begin(); it != m_bpis.end(); it++) {
    BondedInteractions *pBondedInteractions = dynamic_cast<BondedInteractions *>(it->second);
    if (pBondedInteractions != NULL) {
      typename BondedInteractions::InteractionIterator interactionIt = pBondedInteractions->getInnerInteractionIterator();
      while (interactionIt.hasNext()) {
        BondedInteractionCpData(interactionIt.next()).saveCheckPointData(oStream);
        oStream << delim;
      }
    }
    else {
      throw std::runtime_error(
        "TSubLattice<T>::saveCheckPointData - "
        "Could not cast AParallelInteractionStorage to "
        "ParallelInteractionStorage_E<T,CBondedInteraction>" 
      );
    }
  }
#endif
  // dump trimeshdata (if exists)
  oStream << "TMIG " << m_mesh.size() << delim;
  for(typename map<string,TriMesh*>::iterator tm_iter=m_mesh.begin();
      tm_iter!=m_mesh.end();
      tm_iter++){
    oStream << tm_iter->first << delim;
    tm_iter->second->writeCheckPoint(oStream,delim);
  }
}

template <class T>
void TSubLattice<T>::loadCheckPointData(std::istream &iStream)
{
  throw std::runtime_error("TSubLattice<T>::loadCheckPointData not implemented");
}
