
#include "Parallel/SubLatticeControler.h"  

// -- project includes --
#include "Parallel/mpi_tag_defs.h"
#include "Parallel/sublattice_cmd.h"
#include "Parallel/mpicmdbuf.h"
#include "Parallel/mpibarrier.h"
#include "Foundation/console.h"
#include "Model/Particle.h"
#include "Model/RotParticle.h"
#include "Parallel/CheckPointer.h"

CSubLatticeControler::CSubLatticeControler()
  : m_lattice(NULL),
    m_pCheckPointer(NULL),
    m_timingFileName(),
    m_timersPtr(new MpiWTimers())
{
  m_pCheckPointer = new CheckPointer(*this);
  
  m_timersPtr->zeroise("RebuildParticleArray");
  m_timersPtr->zeroise("RebuildInteractions");
  m_timersPtr->zeroise("NeighbourSearch");
  m_timersPtr->zeroise("CheckNeighbours");
  m_timersPtr->zeroise("CheckPoint");
  m_timersPtr->zeroise("BoundaryDataExchange");
  m_timersPtr->zeroise("ForceCalculation");
  m_timersPtr->zeroise("UpdateInteractions");
}

CSubLatticeControler::~CSubLatticeControler()
{
  delete m_pCheckPointer;
  console.Debug() << "slave " << m_global_rank << " finalizing\n";
}

/*!
  Initialize MPI
 */
void CSubLatticeControler::initMPI()
{
  MPI_Comm_size(MPI_COMM_WORLD, &m_global_size);
  MPI_Comm_rank(MPI_COMM_WORLD, &m_global_rank);

  console.Debug() << "Slave initialized at rank " << m_global_rank << "\n" ;

}

/*!
  make a new Lattice of the correct type

  \todo make it throw an exeption if it fails
*/
void CSubLatticeControler::makeLattice()
{
  console.XDebug() << "CSubLatticeControler::makeLattice()\n" ; 

  CVarMPIBuffer param_buffer(MPI_COMM_WORLD);
 
  // get lattice parameters
  param_buffer.receiveBroadcast(0);
  CLatticeParam clp = CLatticeParam::extractLatticeParam(&param_buffer);

  // make the lattice, depending of the type
  if (clp.particle_type()=="Basic"){  
    console.XDebug()<< "making lattice with basic particles\n";
    m_lattice=new TSubLattice<CParticle>(clp,m_global_rank,MPI_COMM_WORLD);
  }else if (clp.particle_type()=="Rot"){ 
    console.XDebug()<< "making lattice with simple rotational particles\n";
    m_lattice=new TSubLattice<CRotParticle>(clp,m_global_rank,MPI_COMM_WORLD);
  }else if (clp.particle_type()=="SPC"){
    console.XDebug()<< "making lattice with spherical patch complex particles\n";
    //m_lattice=new TCPSubLattice<CSPCParticle>(clp,m_global_rank,MPI_COMM_WORLD);
  }else{
    console.Error()<< "ERROR! Trying to make lattice from unknown particles!!\n";
  }
  m_lattice->setTimer(*m_timersPtr);

  console.XDebug() << "end CSubLatticeControler::makeLattice()\n" ; 
}

void CSubLatticeControler::saveCheckPointData(std::ostream &oStream)
{
  if (m_lattice != NULL) {
    m_lattice->saveCheckPointData(oStream);
  }
}

void CSubLatticeControler::loadCheckPointData(std::istream &iStream)
{
  if (m_lattice != NULL) {
    m_lattice->loadCheckPointData(iStream);
  }
}

/*!
  Initialize lattice. All boundaries are assumed to be open, i.e. not circular.
  Recieves all necessary data from Master
*/
void CSubLatticeControler::initLattice()
{
  console.XDebug() << "CSubLatticeControler::initLattice()\n";

  // get corners
  vector<Vec3> bbox;
  m_world_comm.recv_broadcast_cont_packed(bbox,0);
  if(bbox.size()!=2){
    console.Critical()<<"CRITICAL ERROR: bbox.size()!=2\n";
  }
  // init neighbbor table
  m_lattice->initNeighborTable(bbox[0],bbox[1]);

  // send back ppa dims and coords
  vector<int> commdims=m_lattice->getCommDims();
  m_world_comm.send_gather(commdims,0);

  vector<int> commcoords=m_lattice->getCommCoords();
  m_world_comm.send_gather(commcoords,0);

  // receive particles by lattice
  //m_lattice->receiveParticles(&m_world_comm);
  console.XDebug() << "end CSubLatticeControler::initLattice()\n";
}

/*!
  Initialize lattice with at least one circular boundary condition. 
  Recieves all necessary data from Master
*/
void CSubLatticeControler::initLatticeCirc()
{
  console.XDebug() << "CSubLatticeControler::initLatticeCirc()\n";

  // get corners
  vector<Vec3> bbox;
  m_world_comm.recv_broadcast_cont_packed(bbox,0);
  if(bbox.size()!=2){
    console.Critical()<<"CRITICAL ERROR: bbox.size()!=2\n";
  }
  // get circular/non-circular boundary conditions
  vector<int> cv;
  vector<bool> circ;
  m_world_comm.recv_broadcast_cont(cv,0);
  for(vector<int>::iterator iter=cv.begin();
      iter!=cv.end();
      iter++){
    bool c=(*iter==1);
    console.XDebug() << "circ " << c << "\n";
    circ.push_back(c);
  }

  // init neighbbor table
  m_lattice->initNeighborTable(bbox[0],bbox[1],circ);

  // send back ppa dims and coords
  vector<int> commdims=m_lattice->getCommDims();
  m_world_comm.send_gather(commdims,0);

  vector<int> commcoords=m_lattice->getCommCoords();
  m_world_comm.send_gather(commcoords,0);

  // receive particles by lattice
  //m_lattice->receiveParticles(&m_world_comm);
  console.XDebug() << "end CSubLatticeControler::initLattice()\n";
}
 

/*!
  get a wall and add it to the sublattice
*/
void CSubLatticeControler::getWall()
{
  console.XDebug() << "CSubLatticeControler::getWall()\n" ;

  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);
 
  buffer.receiveBroadcast(0);
  CEWallIGP* wigp=extractEWallIGP(&buffer);
  
  m_lattice->addWall(wigp);

  console.XDebug() << "end CSubLatticeControler::getWall()\n" ;
}

/*!
  get a bonded wall and add it to the sublattice
*/
void CSubLatticeControler::getBondedWall()
{
  console.XDebug() << "CSubLatticeControler::getBondedWall()\n" ;

  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  buffer.receiveBroadcast(0);
  CBWallIGP* wigp=extractBWallIGP(&buffer);
  m_lattice->addBondedWall(wigp);

  console.XDebug() << "end CSubLatticeControler::getBondedWall()\n" ;
}

/*!
  get a wall with viscous drag and add it to the sublattice
*/
void CSubLatticeControler::getViscWall()
{
  console.XDebug() << "CSubLatticeControler::getViscWall()\n" ;

  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  buffer.receiveBroadcast(0);
  CVWallIGP* wigp=extractVWallIGP(&buffer);
  m_lattice->addViscWall(wigp);

  console.XDebug() << "end CSubLatticeControler::getViscWall()\n" ;
}

/*!
  get a bonded wall with per-direction spring constants and add it to the sublattice
*/
void CSubLatticeControler::getSoftBondedWall()
{
  console.XDebug() << "CSubLatticeControler::getSoftBondedWall()\n" ;

  CVarMPIBuffer buffer(MPI_COMM_WORLD);
  CMPIBarrier barrier(MPI_COMM_WORLD);

  buffer.receiveBroadcast(0);
  CSoftBWallIGP* wigp=extractSoftBWallIGP(&buffer);
  m_lattice->addBondedWall(wigp);

  console.XDebug() << "end CSubLatticeControler::getSoftBondedWall()\n" ;
}

void CSubLatticeControler::performTiming()
{
  CVarMPIBuffer buffer(MPI_COMM_WORLD);

  // get check-point parameters
  buffer.receiveBroadcast(0);
  std::string fileNamePrefix = buffer.pop_string();
  std::stringstream sStream;
  sStream << fileNamePrefix << MpiInfo().rank() << ".csv";
  setTimingFileName(sStream.str());
}

void CSubLatticeControler::saveTimingData()
{
  if (getTimingFileName() != "") {
    m_timersPtr->appendData(getTimingFileName());
    m_timersPtr->zeroise();
  }
}

void CSubLatticeControler::searchNeighbors()
{
  m_lattice->searchNeighbors();
  return;
}

void CSubLatticeControler::do2dCalculations()
{
  m_lattice->do2dCalculations();
}

/*!
  gets command messages from master and calls the functions of SubLattice
*/
void CSubLatticeControler::run()
{
  CMPILCmdBuffer cmd_buffer(MPI_COMM_WORLD,0);
  int command;
  CMPIBarrier barrier(MPI_COMM_WORLD);

  do{
    command=cmd_buffer.receive();
    console.XDebug() << "rank=" << m_world_comm.rank() << ":Received command " << command << "\n";
    switch(command){
    case CMD_PRINT :  m_lattice->printData(); break;
    case CMD_CALC :
      {
        m_timersPtr->start("ForceCalculation");
        m_lattice->oneStep();
        m_timersPtr->stop("ForceCalculation");
        break;
      }
    case CMD_XCHANGE :
      {
        m_timersPtr->start("BoundaryDataExchange");
        m_lattice->exchangePos();
        m_timersPtr->stop("BoundaryDataExchange");
        break;
      }
    case CMD_NSEARCH : searchNeighbors(); break;
    case CMD_CHECKNEIGHBORS :
      {
        m_timersPtr->start("CheckNeighbours");
        m_lattice->checkNeighbors();
        m_timersPtr->stop("CheckNeighbours");
        break;
      }
    case CMD_UPDATE :
      {
        m_timersPtr->start("UpdateInteractions");
        m_lattice->updateInteractions();
        m_timersPtr->stop("UpdateInteractions");
        break;
      }
    case CMD_PMOVE : m_lattice->moveParticleTo(); break;
    case CMD_PSETND : m_lattice->setParticleNonDynamic(); break;
    case CMD_PVEL : m_lattice->setParticleVelocity(); break;
    case CMD_WMOVE : m_lattice->moveWallBy(); break;
    case CMD_WFORCE : m_lattice->applyForceToWall(); break;
    case CMD_COUNT : m_lattice->countParticles(); break;
    case CMD_COLLECT : m_lattice->sendDataToMaster(); break;
    case CMD_ADDBONDEDIG : m_lattice->addBondedIG(); break;
    case CMD_ADDROTBONDEDIG : m_lattice->addRotBondedIG(); break;
    case CMD_ADDPIG : m_lattice->addPairIG(); break;
    case CMD_ADDROTPIG : m_lattice->addRotPairIG(); break; 
    case CMD_ADDSIG : m_lattice->addSingleIG(); break;
    case CMD_ADDDAMP : m_lattice->addDamping(); break;
    //case CMD_ADDROTDAMP : m_lattice->addRotDamping(); break;
    case CMD_ADDTRIMESHIG : m_lattice->addTriMeshIG(); break;
    case CMD_ADDTRIMESH : m_lattice->addTriMesh(); break;
    case CMD_MAKELATTICE : makeLattice(); break;
    case CMD_INITLATTICE : initLattice(); break;
    case CMD_INITLATTICECIRC : initLatticeCirc(); break;
    case CMD_INITCOMPLEX : m_lattice->initComplex(); break;
    case CMD_GETWALL : getWall(); break;
    case CMD_GETBONDEDWALL : getBondedWall(); break;
    case CMD_GETVISCWALL : getViscWall(); break;
    case CMD_GETSOFTBONDEDWALL : getSoftBondedWall(); break;
    case CMD_PRINTFIELDS : m_lattice->printAllFields(); break;
    case CMD_EXIG : m_lattice->setExIG(); break;
    case CMD_FIELDCOUNT : m_lattice->countFieldData(); break;
    case CMD_FIELDCOLLECT : m_lattice->collectFieldData(); break;  
    case CMD_ADD_SPF : m_lattice->addScalarParticleField(); break;
    case CMD_ADD_VPF : m_lattice->addVectorParticleField(); break;
    case CMD_ADD_SIF : m_lattice->addScalarInteractionField(); break;
    case CMD_ADD_VIF : m_lattice->addVectorInteractionField(); break;
    case CMD_ADD_VTF : m_lattice->addVectorTriangleField(); break;
    case CMD_SEND_FIELDS : m_lattice->sendFieldData(); break;
    case CMD_RECEIVEPARTICLES : m_lattice->receiveParticles(&m_world_comm);break;
    case CMD_SAVECHECKPOINT :
      {
        m_timersPtr->start("CheckPoint");
        m_pCheckPointer->save();
        m_timersPtr->stop("CheckPoint");
        break;
      }
    case CMD_PTAG : m_lattice->tagParticleNearestTo();break;
    case CMD_DO2DCALCULATIONS : do2dCalculations();break;
    case CMD_PERFORMTIMING : performTiming();break;
    case CMD_SAVETIMINGDATA : saveTimingData();break;
    case CMD_MOVENODE : m_lattice->moveSingleNode();break;
    case CMD_MOVETAGGEDNODES : m_lattice->moveTaggedNodes();break;
    case CMD_ADDBONDEDTRIMESHIG : m_lattice->addBondedTriMeshIG(); break;
    case CMD_FINISH : break;
    default: console.Error() << m_global_rank << " got unknown command : " << command << "\n" ;
    }
    
    barrier.wait("SLC past command");
  } while (command!=CMD_FINISH);
#ifdef __TIMING
  m_lattice->printTimes();
#endif //__TIMING
  m_timersPtr->clear();
  barrier.wait("Final");
}

