/*=============================================================================
??
??******************************************************************************
??*                                               *
??* COPYRIGHT ??????? - All Rights Reserved *
??*  *
??* This software is the property of ??????????????. No part of this code   *
??* may be copied in any form or by any means without the expressed written   *
??* consent of ???????????. Copying, use or modification of this software     *
??* by any unauthorised person is illegal unless that    *
??* person has a software license agreement with ?????????????.            *
??*                                                          *
??******************************************************************************
??
*********************************************************************************/

#include "CheckPointLoaderTest.h"
#include "Parallel/CheckPointer.h"
#include "Parallel/CheckPointable.h"
#include "Model/Particle.h"
#include "Model/BondedInteraction.h"
#include "Parallel/IterativeReader.h"
#include "Parallel/CheckPointLoader.h"
#include "Geometry/GeometryInfo.h"
#include "Parallel/CheckPointInfo.h"

#include <iostream>
#include <fstream>

using namespace esys::lsm;

std::ostream &operator<<(std::ostream &oStream, const BondedInteractionCpData &conn)
{
  oStream 
    << "id1 = " << conn.getP1Id() << ", id2 = " << conn.getP2Id() << ", tag = " << conn.getTag();
    
  return oStream;
}

bool operator==(const CParticle &p1, const CParticle &p2)
{
  return 
    (
      (p1.getPos() == p2.getPos())
      &&
      (p1.getID() == p2.getID())
      &&
      (p1.getVel() == p2.getVel())
      &&
      (p1.getTag() == p2.getTag())
      &&
      (p1.getForce() == p2.getForce())
    );
}

bool operator==(const BondedInteractionCpData &c1, const BondedInteractionCpData &c2)
{
  return 
    (
      (c1.getP1Id() == c2.getP1Id())
      &&
      (c1.getP2Id() == c2.getP2Id())
      &&
      (c1.getTag() == c2.getTag())
    );
}

std::ostream &operator<<(std::ostream &oStream, const CheckPointInfo &cpInfo)
{
  cpInfo.write(oStream);
  return oStream;
}

namespace LsmTest
{
  class TestCheckPointer;
  
  class TestLsmData : public CheckPointable
  {
  public:
    typedef std::vector<CParticle>               ParticleVector;
    typedef std::vector<BondedInteractionCpData> ConnectionVector;
    typedef ParticleVector::iterator             particle_iterator;
    typedef ConnectionVector::iterator           connection_iterator;

    TestLsmData(int numParticles = 0)
    {
      const Vec3 zeroVec3;
      for (int i = 0; i < numParticles; i++) {
        Vec3 pos(i, i + i + 1, i + i/2 + 3);
        m_particles.push_back(
          CParticle(
            1.0,
            i*0.1,
            pos, // position
            pos, // old position
            pos, // initial position
            Vec3(i/2, i/3 + i + 1, i/3 + i/2 + 3), // velocity
            Vec3(i, i*2, i*3), // force
            i
          )
       );
       m_particles[i].setTag(i % 5);
      }

      if (m_particles.size() > 0) {
        for (int i = 1; i < m_particles.size(); i++)
        {
          m_connections.push_back(
            BondedInteractionCpData(
              i-1,
              i,
              (i%15 ? 0 : 1)
            )
          );
        }
      }
    }

    bool operator==(const TestLsmData &lsmData) const
    {
      bool equal = 
        (m_particles.size() == lsmData.m_particles.size())
        &&
        (m_connections.size() == lsmData.m_connections.size());
        
      for (int i = 0; equal && (i < m_particles.size()); i++)
      {
        equal = equal && (m_particles[i] == lsmData.m_particles[i]);
      }
      
      for (int i = 0; equal && (i < m_connections.size()); i++)
      {
        equal = equal && (m_connections[i] == lsmData.m_connections[i]);
      }
      
      return equal;
    }

    void write(std::ostream &oStream) const
    {
      oStream
        << "num particles = " << m_particles.size()
        << ", num connections = " << m_connections.size() << "\n";
      for (int i = 0; i < m_particles.size(); i++) {
        oStream << m_particles[i] << "\n";
      }
      for (int i = 0; i < m_connections.size(); i++) {
        oStream << m_connections[i] << "\n";
      }
      
    }
    
    virtual void saveCheckPointData(std::ostream &oStream)
    {
    }

    virtual void loadCheckPointData(std::istream &iStream)
    {
    }
    
    template <class TmplIterator>
    void addIterator(TmplIterator &it)
    {
      while (it.hasNext())
      {
        add(it.next());
      }
    }

    template <class TmplIterator, class TmplParticle>
    void addParticles(TmplIterator &it)
    {
      addIterator(it);
    }

    template <class TmplIterator>
    void addConnections(TmplIterator &it)
    {
      addIterator(it);
    }

    void add(const CParticle &particle)
    {
      m_particles.push_back(particle);
    }
    
    void add(const BondedInteractionCpData &connection)
    {
      m_connections.push_back(connection);
    }

    friend class TestCheckPointer;
  protected:
    ParticleVector   m_particles;
    ConnectionVector m_connections;
  };
  
  std::ostream &operator<<(std::ostream &oStream, const TestLsmData &data)
  {
    data.write(oStream);
    return oStream;
  }

  typedef std::vector<std::string> StringVector;

  class TestCheckPointer
  {
  public:
    TestCheckPointer(TestLsmData &lsmData, int numSlaves, const boost::filesystem::path &dir) : m_data(lsmData)
    {
      for (int i = 1; i <= numSlaves; i++) {
        std::stringstream sStream;
        sStream << "Cp_" << i << ".txt";
        m_fileNames.push_back((dir / sStream.str()).string());
      }
    }

    virtual void save()
    {
      const int numSlaves = m_fileNames.size();
      const int numParticles = m_data.m_particles.size();
      const int numConnections = m_data.m_connections.size();
      const char delim = '\n';
      
      for (int i = 0; i < numSlaves; i++) {
        std::ofstream oStream(m_fileNames[i].c_str());
        
        const int numPartsPerSlave =
          (numParticles % numSlaves)
          ?
          (numParticles/numSlaves) + 1
          :
          (numParticles/numSlaves);
        
        oStream << min((i+1) * numPartsPerSlave, numParticles) - (i * numPartsPerSlave);
        oStream << delim;
        for (int p = i * numPartsPerSlave; (p < ((i+1) * numPartsPerSlave)) && (p < numParticles); p++)
        {
          m_data.m_particles[p].saveCheckPointData(oStream);
          oStream << delim;
        }
        
        const int numConsPerSlave =
          (numConnections % numSlaves)
          ?
          (numConnections/numSlaves) + 1
          :
          (numConnections/numSlaves);
        
        oStream << min((i+1) * numConsPerSlave, numConnections) - (i * numConsPerSlave);
        oStream << delim;
        for (int c = (i * numConsPerSlave); (c < ((i+1) * numConsPerSlave)) && (c < numConnections); c++)
        {
          m_data.m_connections[c].saveCheckPointData(oStream);
          oStream << delim;
        }
        oStream.close();
      }
    }

    const StringVector &getFileNameVector() const
    {
      return m_fileNames;
    }

    private:
      TestLsmData  &m_data;
      StringVector m_fileNames;
  };

  CPPUNIT_TEST_SUITE_REGISTRATION( CheckPointLoaderTest );

  CheckPointLoaderTest::CheckPointLoaderTest()
      : FileTestCase()
  {
  }

  CheckPointLoaderTest::~CheckPointLoaderTest()
  {
  }

  void CheckPointLoaderTest::setUp()
  {
    FileTestCase::setUp();
  }

  void CheckPointLoaderTest::tearDown()
  {
    FileTestCase::tearDown();
  }

  typedef esys::lsm::BoolVector BoolVector;
  void CheckPointLoaderTest::testCheckPointInfo()  
  {
    CheckPointInfo info;
    
    StringVector files;
    files.push_back("a");
    files.push_back("b");
    files.push_back("c");

    info.setLatticeDataFiles(files);
    
    GeometryInfo geoInfo(1.2, Vec3(0.0, 0.1, 0.2), Vec3(1.0, 2.0, 3.0), BoolVector(3, false), false);

    info.setGeometryInfo(geoInfo);
    info.setNumTimeSteps(100);
    info.setTimeStepSize(0.001);
    
    std::stringstream sStream;
    info.write(sStream);
    
    CheckPointInfo newInfo;
    newInfo.read(sStream);
    
    CPPUNIT_ASSERT_EQUAL(info, newInfo);
  }

  void CheckPointLoaderTest::testIt()
  {
    const int numParticles = 1033;
    const int numSlaves    = 8;

    TestLsmData      lsmData(numParticles);
    TestCheckPointer checkPointer(lsmData, numSlaves, getTestDirPath());

    checkPointer.save();

    CheckPointLoader loader(checkPointer.getFileNameVector());
    TestLsmData newLsmData;
    loader.loadInto(newLsmData);

    CPPUNIT_ASSERT_EQUAL(lsmData, newLsmData);
  }

  void CheckPointLoaderTest::testStream()
  {
    std::stringstream sStream;
    sStream << int(5) << char(',') << int(10) << '\n' << int(6) << char(' ') << int(11);
  
    std::istringstream iStream(sStream.str(), std::ios::binary | std::ios::in);

    int i = 0;
    char c = '\0';
    
    iStream >> i;
    CPPUNIT_ASSERT_EQUAL(5, i);
    
    iStream >> c;
    CPPUNIT_ASSERT_EQUAL(',', c);
    
    iStream >> i;
    CPPUNIT_ASSERT_EQUAL(10, i);
    
    iStream >> i;
    CPPUNIT_ASSERT_EQUAL(6, i);
  }
};
