/*=============================================================================
??
??******************************************************************************
??*                                               *
??* COPYRIGHT ??????? - All Rights Reserved *
??*  *
??* This software is the property of ??????????????. No part of this code   *
??* may be copied in any form or by any means without the expressed written   *
??* consent of ???????????. Copying, use or modification of this software     *
??* by any unauthorised person is illegal unless that    *
??* person has a software license agreement with ?????????????.            *
??*                                                          *
??******************************************************************************
??
*********************************************************************************/

#include "ParallelTest/GeometryReaderTest.h"
#include "Parallel/GeometryReader.h"

#include "Foundation/vec3.h"

#include <sstream>

using namespace esys::lsm;

namespace LsmTest
{
  typedef esys::lsm::BoolVector BoolVector;
  typedef esys::lsm::IntVector  IntVector;
  typedef esys::lsm::Vec3Vector Vec3Vector;

  CPPUNIT_TEST_SUITE_REGISTRATION( GeometryReaderTest );

  GeometryReaderTest::GeometryReaderTest()
      : CppUnit::TestCase()
  {
  }

  GeometryReaderTest::~GeometryReaderTest()
  {
  }

  void GeometryReaderTest::setUp()
  {
    CppUnit::TestCase::setUp();
  }

  void GeometryReaderTest::tearDown()
  {
    CppUnit::TestCase::tearDown();
  }

  void GeometryReaderTest::testGeometryInfoEquality()
  {
    GeometryInfo geoInfo1;
    GeometryInfo geoInfo2;

    CPPUNIT_ASSERT(geoInfo1 == geoInfo2);

    geoInfo1.setBBox(Vec3(0.0, 0.0, 1.0), Vec3(0.0, 0.0, 0.0));
    CPPUNIT_ASSERT(!(geoInfo1 == geoInfo2));
    
    geoInfo1 = geoInfo2;
    CPPUNIT_ASSERT(geoInfo1 == geoInfo2);
  }

  void GeometryReaderTest::testGeometryInfoReadWrite()
  {
    BoolVector periodicDims(3, false);
    periodicDims[1] = true;

    GeometryInfo geoInfo1(
        1.2,
        Vec3(1.0, 2.0, 3.0),
        Vec3(10.0, 20.0, 30.0),
        periodicDims,
        false
      );

    std::stringstream sStream;

    geoInfo1.write(sStream);

    GeometryInfo geoInfo2;
    geoInfo2.read(sStream);
    
    CPPUNIT_ASSERT_MESSAGE(sStream.str(), geoInfo1 == geoInfo2);
  }

  void GeometryReaderTest::testParticleDataReadWrite()
  {
    SimpleParticleData pd(0, 0, Vec3(1, 2, 3), 1.0);
    SimpleParticleData pd1(0, 0, Vec3(1, 2, 3), 1.0);
    SimpleParticleData pd2(0, 0, Vec3(1, 2, 3), 1.01);
    SimpleParticleData pd3(7, 5, Vec3(4, 5, 6), 1.0);
    
    CPPUNIT_ASSERT_EQUAL(pd, pd1);
    CPPUNIT_ASSERT(!(pd1 == pd2));
    CPPUNIT_ASSERT(!(pd1 == pd3));

    std::stringstream sStream;
    
    sStream << pd1 << " " << pd2 << " " << pd3;
    
    sStream >> pd;
    CPPUNIT_ASSERT_EQUAL(pd1, pd);
    
    sStream >> pd;
    CPPUNIT_ASSERT_EQUAL(pd2, pd);

    sStream >> pd;
    CPPUNIT_ASSERT_EQUAL(pd3, pd);
  }

  class ParticleGenerator
  {
  public:
    typedef std::vector<SimpleParticleData> ParticleVector;
    typedef ParticleVector::const_iterator  const_iterator;

    ParticleGenerator(int numParticles)
    {
      for (int i = 0; i < numParticles; i++)
      {
        m_particles.push_back(
          SimpleParticleData(
            i,
            i % 8,
            Vec3(i, i + i % 2, i + i % 4),
            1.0 - 0.1*(i % 4)
          )
        );
      }
    }

    int size() const
    {
      return m_particles.size();
    }

    const_iterator begin() const
    {
      return m_particles.begin();
    }

    const_iterator end() const
    {
      return m_particles.end();
    }

    void write(std::ostream &oStream) const
    {
      oStream << "\nBeginParticles\nSimple\n" << size() << "\n";
      for (const_iterator it = begin(); it != end(); it++)
      {
        oStream << *it << "\n";
      }
      oStream << "EndParticles\n";
    }
    
    Vec3 getBBoxMin() const
    {
      Vec3 min = begin()->getPosition();
      for (const_iterator it = begin(); it != end(); it++) {
        min = cmin(min, it->getPosition());
      }
      return min;
    }
    
    Vec3 getBBoxMax() const
    {
      Vec3 max = begin()->getPosition();
      for (const_iterator it = begin(); it != end(); it++) {
        max = cmax(max, it->getPosition());
      }
      return max;
    }

  private:
    ParticleVector m_particles;
  };

  std::ostream &operator<<(std::ostream &oStream, const ParticleGenerator &particleGen)
  {
    particleGen.write(oStream);
    return oStream;
  }

  void GeometryReaderTest::testParticleReader()
  {
    ParticleGenerator particleGen(1);
    std::stringstream sStream;
    
    sStream << particleGen;

    ParticleReader particleReader(sStream, true);
    particleReader.initialise();
    
    CPPUNIT_ASSERT_EQUAL(1, particleReader.getNumElements());
    
    ParticleReader::Iterator &readerIt = particleReader.getIterator();    
    CPPUNIT_ASSERT(readerIt.hasNext());
    CPPUNIT_ASSERT_EQUAL(readerIt.next(), (*particleGen.begin()));
    CPPUNIT_ASSERT(!readerIt.hasNext());
  }

  void GeometryReaderTest::testParticleReaderMultiParticles()
  {
    ParticleGenerator particleGen(512);
    std::stringstream sStream;
    
    sStream << particleGen;

    ParticleReader particleReader(sStream, true);
    ParticleReader::Iterator &readerIt = particleReader.getIterator();

    CPPUNIT_ASSERT_EQUAL(512, particleReader.getNumElements());

    ParticleGenerator::const_iterator genIt = particleGen.begin();
    while (readerIt.hasNext())
    {
      CPPUNIT_ASSERT_EQUAL(readerIt.next(), (*genIt));
      genIt++;
    }
  }

  void GeometryReaderTest::testParticleReaderThrow()
  {
    std::stringstream
      sStream(
        "Simple\n"
        "1\n"
        "0.0 0.1 0.2 1.0 0 0\n"
        "EndParticles"
      );

    ParticleReader particleReader(sStream, true);
    particleReader.initialise();
  }

  class ConnectionGenerator
  {
  public:
    typedef std::vector<SimpleConnectionData> ConnectionVector;
    typedef ConnectionVector::const_iterator  const_iterator;

    ConnectionGenerator(const ParticleGenerator &particleGenerator)
    {
      ParticleGenerator::const_iterator it = particleGenerator.begin();
      SimpleConnectionData::Id prevId = it->getId();
      for (
        it++;
        it != particleGenerator.end();
        it++
      )
      {
        m_connections.push_back(
          SimpleConnectionData(
            prevId,
            it->getId(),
            (it-particleGenerator.begin()) % 5)
        );
        prevId = it->getId();
      }
    }

    int size() const
    {
      return m_connections.size();
    }

    const_iterator begin() const
    {
      return m_connections.begin();
    }

    const_iterator end() const
    {
      return m_connections.end();
    }

    void write(std::ostream &oStream) const
    {
      oStream << "BeginConnect\n" << size() << "\n";
      for (const_iterator it = begin(); it != end(); it++)
      {
        oStream << *it << "\n";
      }
      oStream << "EndConnect\n";
    }

  private:
    ConnectionVector m_connections;
  };
  
  std::ostream &operator<<(std::ostream &oStream, const ConnectionGenerator &conGenerator)
  {
    conGenerator.write(oStream);
    return oStream;
  }

  void GeometryReaderTest::testConnectionReader()
  {
    ParticleGenerator particleGenerator(55);
    ConnectionGenerator conGenerator(particleGenerator);
    
    std::stringstream sStream;
    sStream << "blah\nblah\nblahdie blah\n";
    sStream << conGenerator;
    
    ConnectionReader connectionReader(sStream);
    ConnectionReader::Iterator &readerIt = connectionReader.getIterator();
    
    CPPUNIT_ASSERT_EQUAL(conGenerator.size(), connectionReader.getNumElements());
    CPPUNIT_ASSERT_EQUAL(conGenerator.size(), readerIt.getNumRemaining());
    
    ConnectionGenerator::const_iterator genIt = conGenerator.begin();
    while ((genIt != conGenerator.end()) && readerIt.hasNext()) {
      const SimpleConnectionData &data = readerIt.next();
      CPPUNIT_ASSERT_EQUAL(*genIt, data);
      genIt++;
    }
    
    CPPUNIT_ASSERT_MESSAGE(
      "Number of iterated elements differs from number of generated elements.",
      ((genIt == conGenerator.end()) && (!readerIt.hasNext()))
    );
  }

  void GeometryReaderTest::testConnectionReaderThrow()
  {
    std::stringstream
      sStream(
        "1\n"
        "1 2 0\n"
        "EndConnect"
      );

    ConnectionReader connectionReader(sStream);
    connectionReader.initialise();
  }
  
  void GeometryReaderTest::testGeometryReaderVersion1_1()
  {
    ParticleGenerator particles(11);
    ConnectionGenerator connections(particles);
    
    GeometryInfo geoInfo =
      GeometryInfo(
        1.1,
        particles.getBBoxMin(),
        particles.getBBoxMax(),
        BoolVector(3, false),
        true
      );

    std::stringstream sStream;
    geoInfo.write(sStream);
    particles.write(sStream);
    connections.write(sStream);

    GeometryReader geoReader(sStream);

    GeometryInfo readGeoInfo = geoReader.getGeometryInfo();
    
    CPPUNIT_ASSERT_EQUAL(geoInfo, readGeoInfo);

    GeometryReader::ParticleIterator &particleIt = geoReader.getParticleIterator();

    CPPUNIT_ASSERT_EQUAL(particles.size(), particleIt.getNumRemaining());
    while (particleIt.hasNext()) {
      const SimpleParticleData &particle = particleIt.next();
      CPPUNIT_ASSERT_EQUAL(
        particle.getRadius()*particle.getRadius(),
        particle.getMass()
      );
    }

    GeometryReader::ConnectionIterator &connectionIt = geoReader.getConnectionIterator();
    CPPUNIT_ASSERT_EQUAL(connections.size(), connectionIt.getNumRemaining());
    while (connectionIt.hasNext()) {
      connectionIt.next();
    }
  }
  
  void GeometryReaderTest::testGeometryReaderVersion1_2()
  {
    ParticleGenerator particles(413);
    ConnectionGenerator connections(particles);
    
    BoolVector periodicDims(3, false);
    periodicDims[0] = 1;
    
    GeometryInfo geoInfo =
      GeometryInfo(
        1.2,
        particles.getBBoxMin(),
        particles.getBBoxMax(),
        periodicDims,
        false
      );

    std::stringstream sStream;
    geoInfo.write(sStream);
    particles.write(sStream);
    connections.write(sStream);

    GeometryReader geoReader(sStream);

    GeometryInfo readGeoInfo = geoReader.getGeometryInfo();
    
    CPPUNIT_ASSERT_EQUAL(geoInfo, readGeoInfo);

    GeometryReader::ParticleIterator &particleIt = geoReader.getParticleIterator();

    CPPUNIT_ASSERT_EQUAL(particles.size(), particleIt.getNumRemaining());
    while (particleIt.hasNext()) {
      const SimpleParticleData &particle = particleIt.next();
      CPPUNIT_ASSERT_EQUAL(
        particle.getRadius()*particle.getRadius()*particle.getRadius(),
        particle.getMass()
      );
    }

    GeometryReader::ConnectionIterator &connectionIt = geoReader.getConnectionIterator();
    CPPUNIT_ASSERT_EQUAL(connections.size(), connectionIt.getNumRemaining());
    while (connectionIt.hasNext()) {
      connectionIt.next();
    }
  }
  
}; // namespace LsmTest
