/********************************************************************************/
/* EXPRESSION.CPP	 							*/
/* 										*/
/* Description:	  Parser evaluation functions					*/
/* Author(s):     D. Place (QUAKES)						*/
/* Creation date: March 1999							*/
/*										*/
/********************************************************************************/
#include "stdafx.h"
#include "mtypes.h"
#include "expression.h"
#include "functions.h"
#include "parse.h"
#include "shake_debug.h"
//#include "all.h"

/********************************************************************************/
/* abstract class ExpressionNode	 					*/
/* 										*/
/* Description:									*/
/*   An expression (whatever it is: symbol, numerical expression,		*/ 
/*   function, declaration...)							*/
/*   An expression can be a list of Expressions (childs) where a value		*/
/*   is assigned to it. 							*/
/*   For instance the expression 2+5, will be encoded as an expression '+'	*/
/*   having two childs '2' and '5'						*/
/*										*/
/********************************************************************************/

void ExpressionNode::setloc(yyltype *loc) {
	   first_line = loc->first_line ;
	   first_column = loc->first_column ;
	   curfile = loc->text ;
}

ExpressionNode::ExpressionNode() : List<ExpressionNode>(), OpValue(), ParserStatics(), ParserError() { 
    ExpressionType="ExpressionNode"; 
    first_line = 0 ;
    first_column = 0 ;
}

ExpressionNode::~ExpressionNode() {
    First() ;
    while (!IsEnd()) {
	DB_DELETE(Get(),"?") ;
	delete Get() ;
	Clear() ;
    }
}

/********************************************************************************/
/* class ValueRef			 					*/
/* 										*/
/* Sym: <number>|<string>							*/
/* Description:									*/
/*   its value contain a number or a string , no childs				*/
/*										*/
/********************************************************************************/

ValueRef::ValueRef(const OpValue &Val) : ExpressionNode() {
    *(OpValue *)this = Val ;
    ExpressionType="ValueRef"; 
}   

int ValueRef::Eval() { 
    return true;
}

/********************************************************************************/
/* class SymbolRef			 					*/
/* 										*/
/* Sym: <name>									*/
/* Description:									*/
/*   its value containt the value of the symbol name				*/
/*   the first child contains the symbol name					*/
/*										*/
/********************************************************************************/

SymbolRef::SymbolRef(const OpValue &SymbolName) : ExpressionNode() {
    ValueRef *sn ;
    First() ; 
    sn = new ValueRef(SymbolName) ;
    DB_NEWALLOC(sn,"ValueRef") ;
    *this << sn ;
    ExpressionType="SymbolRef"; 
}

int SymbolRef::Eval() {
    Symbol *Sym ;
    First() ;
    Sym = VT->LookUp(Get()->Content.str) ;	// Lookup in table of variables
    if (!Sym) {
	yyltype loc;
	loc.first_line = first_line ;
	loc.first_column = first_column ;
	loc.text = curfile ;
	Error("Undefined variable",&loc) ;
	return false ;
    }
    *(OpValue *)(this) = *Sym->value ;	// Set the value of expression 
    PtrValue = Sym->value ;
    return true ;
}

int SymbolRef::Declare() {
    Symbol *Sym ;
    First() ;
    Sym = VT->LookUp(Get()->Content.str) ;	// Lookup in table of variables
    if (!Sym) {
	Sym = VT->AddSymbol(Get()->Content.str, ValueType) ;
	PtrValue = Sym->value ;
	return true ;
    } else {
	yyltype loc;
	loc.first_line = first_line ;
	loc.first_column = first_column ;
	loc.text = curfile ;
	Error("Variable already declared",&loc) ;
	return false ;				// Already in global
    }
}

/********************************************************************************/
/* class AffectationInst		 					*/
/* 										*/
/* Sym: <var> = <exp>								*/
/* Description:									*/
/*   its value contains the value of the symbol or expression			*/
/*   The first child is the symbol ref (name of the array)			*/
/*   the second child is the expression						*/
/* 										*/
/********************************************************************************/

AffectationInst::AffectationInst(ExpressionNode *Var, ExpressionNode *exp) {
    First() ;
    *this << Var ;
    *this << exp ;
    ExpressionType="AffectationInst"; 
}

int AffectationInst::Eval() {
    First() ;
    ExpressionNode *Var=Get() ;
    Next() ;
    ExpressionNode *Exp=Get() ;
    if ( Var->Eval() && Exp->Eval() ) {
	*Var->PtrValue = *Exp ;
	*(OpValue *)this = *Exp ;
	return true ;
    } else {
	return false ;
    }
}


/********************************************************************************/
/* class ArrayRef			 					*/
/* 										*/
/* Sym: <name>[dim1][dim2]...[dimN]						*/
/* Description:									*/
/*   its value contains the value of the reference (i.e. a Pointer or a lvalue)	*/
/*   The first child is the symbol ref (name of the array)			*/
/*   the other childs are the indices (leading dim. = second child)		*/
/* 										*/
/********************************************************************************/

ArrayRef::ArrayRef(ExpressionNode *Ptr) : ExpressionNode() {
    First() ;
    *this << Ptr ;
    ExpressionType="ArrayRef"; 
}

int ArrayRef::Eval() {
    First() ;
    ExpressionNode *Ptr=Get() ;
    if (!Ptr->Eval()) return false ; 
    Next() ;
    ExpressionNode *Indice ;
    if (Ptr->ValueType != PTR) {
	    yyltype loc;
	    loc.first_line = first_line ;
	    loc.first_column = first_column ;
	    loc.text = curfile ;
	    Error("Not an array or pointer",&loc) ;
	    return false ;
    }
    PtrValue = Ptr->Content.ptr ;
    while (!IsEnd()) {
	Indice=Get() ;
        if (!Indice->Eval()) return false ;
	if (Indice->ValueType != INT) {
	    yyltype loc;
	    loc.first_line = Indice->first_line ;
	    loc.first_column = Indice->first_column ;
	    loc.text = Indice->curfile ;
	    Error("Indices must be an integer",&loc) ;
	    return false ;
	}
	if (PtrValue->ValueType != PTR) {
	    yyltype loc;
	    loc.first_line = first_line ;
	    loc.first_column = first_column ;
	    loc.text = curfile ;
	    Error("Dimension array mismatch or nor an array",&loc) ;
	    return false ;
	}
	if ((Indice->Content.l <0) || (Indice->Content.l >= PtrValue->Content.ptr->GetSize())) {
	    yyltype loc;
	    loc.first_line = Indice->first_line ;
	    loc.first_column = Indice->first_column ;
	    loc.text = Indice->curfile ;
	    Error("Out of range reference",&loc) ;
	    return false ;
	}
	PtrValue = &(*PtrValue->Content.ptr)[Indice->Content.l] ;
	Next() ;
    }
    *(OpValue *)(this) = *PtrValue ;
    return true ;
}

int ArrayRef::Declare() {
    First() ;
    ExpressionNode *Ptr=Get() ;
    Ptr->ValueType =PTR ;
    if (!Ptr->Declare()) return false ; 
    if (!Ptr->Eval()) return false ;
    Next() ;
    ExpressionNode *Indice ;
    int dim=0 ;
    int Size[10] ;
    while (!IsEnd()) {
	Indice = Get() ;
	if (!Indice->Eval()) return false ;
	if (Indice->ValueType != INT) {
	    yyltype loc;
	    loc.first_line = Indice->first_line ;
	    loc.first_column = Indice->first_column ;
	    loc.text = Indice->curfile ;
	    Error("Indices must be an integer",&loc) ;
	    return false ;
	}
	if (Indice->Content.l <=0) {
	    yyltype loc;
	    loc.first_line = Indice->first_line ;
	    loc.first_column = Indice->first_column ;
	    loc.text = Indice->curfile ;
	    Error("Dimension must be a least 1",&loc) ;
	    return false ;
	}
	Size[dim++] = Indice->GetIntVal() ;
	Next() ;
    }

    return Alloc(*Ptr->PtrValue,1,dim,Size) ;
}

int ArrayRef::Alloc(OpValue &Val, int curdim, int dim, int *Size) {
    if (dim==curdim) {
	Val.ValueType = PTR ;
	Val.Content.ptr = new OpValueArray(ValueType,Size[curdim-1]) ;
        DB_NEWALLOC(Val.Content.ptr,"OpValueArray") ;
    } else {
	int i;
	Val.ValueType = PTR ;
	Val.Content.ptr = new OpValueArray(PTR,Size[curdim-1]) ;
        DB_NEWALLOC(Val.Content.ptr,"OpValueArray") ;
	for (i=0; i< Size[curdim-1]; i++) {
	    Alloc( (*Val.Content.ptr)[i],curdim+1, dim,Size) ;
	}
    }
    return true ;
}

/********************************************************************************/
/* class Declaration			 					*/
/* 										*/
/* Sym: <type> <symbolref>|<arrayref>						*/
/* Description:									*/
/*   its value is a dummy value of same type than the varible being declared	*/
/*   The first child is a symbol ref. or array ref.				*/
/* 										*/
/********************************************************************************/

Declaration::Declaration(OType val_type, ExpressionNode *exp) {
    First() ;
    ValueType = val_type ;
    *this << exp ;
    ExpressionType="Declaration"; 
}

int Declaration::Eval() {
    bool res = true ;
    First() ;
    while ((!IsEnd())&&(res)) {
        Get()->ValueType = ValueType ;
        res = res && Get()->Declare() ;
        Next() ;
    }
    return res ;
}
