/********************************************************************************/
/* EXPRESSION.H		 							*/
/* 										*/
/* Description:	  Parser evaluation functions					*/
/* Author(s):     D. Place (QUAKES)						*/
/* Creation date: March 1999							*/
/*										*/
/********************************************************************************/
#ifndef _EXPRESSION_H_
#define _EXPRESSION_H_

#include "mtypes.h"
#include "globals.h"
#include "mainparser.h"

/*!
  \class ExpressionNode
  \brief Expression node for internation representation


   An expression (whatever it is: symbol, numerical expression,	
   function, declaration...)
   An expression can be a list of Expressions (childs) where a value
   is assigned to it.
   For instance the expression 2+5, will be encoded as an expression '+'
   having two childs '2' and '5'
*/
class ExpressionNode : public List<ExpressionNode>, public OpValue, public ParserStatics, public ParserError {
protected:
    char *ExpressionType ;    //!< used for debugging only 
    static int return_has_been_reached ;
public:
    OpValue *PtrValue ;  //!< pointer to value in the Variable Table for affectation operation
    int first_line, first_column;
    char *curfile ;

    ExpressionNode() ;
    virtual ~ExpressionNode() ;

    void setloc(yyltype *loc) ;
    virtual int Eval() {return false ; } ;
    virtual int Declare() { return false ; } ;
} ;


/*!
  \class ValueRef
  \brief Value Reference (<number>|<string>)
  its value contain a number or a string , no childs
*/
class ValueRef: public ExpressionNode { 
public:
    ValueRef(const OpValue &Val) ;
    virtual ~ValueRef(){};
    virtual int Eval()   ;
} ;

/*!
  \class SymbolRef
  \brief Symbol reference (<name>)
  Its value containt the value of the symbol name, the first child contains the symbol name

*/
class SymbolRef: public ExpressionNode { 
public:
    SymbolRef(const OpValue &SymbolName) ;
    virtual ~SymbolRef(){};
    virtual int Eval()   ;
    virtual int Declare() ;
} ;

/*!
  \class  ArrayRef
  \brief Array reference (<name>[dim1][dim2]...[dimN])
  its value contains the value of the reference (i.e. a Pointer or a lvalue)
  The first child is the symbol ref (name of the array)
  the other childs are the indices (leading dim. = second child)
*/
class ArrayRef: public ExpressionNode {
private:
    int Alloc(OpValue &Val, int curdim, int dim, int *Size) ;
public:
    ArrayRef(ExpressionNode *Ptr) ;

    virtual int Eval() ;
    virtual int Declare() ;
};


/*!
  \class AffectationInst
  \brief Affectation Instruction (<var> = <exp>)
  its value contains the value of the symbol or expression
  The first child is the symbol ref (name of the array)
  the second child is the expression
*/
class AffectationInst: public ExpressionNode {
public:
    AffectationInst(ExpressionNode *Var, ExpressionNode *exp) ;

    virtual int Eval() ;
};

/*!
  \class Declaration
  \brief Declaration Instruction (<type> <symbolref>|<arrayref>)
  its value is a dummy value of same type than the varible being declared
  The first child is a symbol ref. or array ref.
*/
class Declaration: public ExpressionNode {
public:
    Declaration(OType val_type, ExpressionNode *exp) ;

    virtual int Eval() ;
};

#endif


