/********************************************************************************/
/* FUNCTIONS.CPP	 							*/
/* 										*/
/* Description:	  Parser evaluation functions					*/
/* Author(s):     D. Place (QUAKES)						*/
/* Creation date: February 1999							*/
/*										*/
/********************************************************************************/
#include "stdafx.h"
#include "functions.h"
#include "mtypes.h"
#include "parse.h"
#include "shake_debug.h"
#include "console.h"

/********************************************************************************/
/* class PredefFNC, PredefFunctionTable						*/
/* 										*/
/* Description:									*/
/*  Table of all function (user defined and predefined)				*/
/* 										*/
/********************************************************************************/

//PredefFunctionTable m_PF ;

PredefFunctionTable::PredefFunctionTable() : List<PredefFNC>() {
    Init() ;
}

PredefFunctionTable::~PredefFunctionTable() {
    Destroy() ;
}

void PredefFunctionTable::Init() {
    PreMathFn<PMFacos> *f_acos = new PreMathFn<PMFacos> ;
    DB_NEWALLOC(f_acos,"PreMathFn") ;
    AddFunction("acos",f_acos) ;

    PreMathFn<PMFasin> *f_asin = new PreMathFn<PMFasin> ;
    DB_NEWALLOC(f_asin,"PreMathFn") ;
    AddFunction("asin",f_asin) ;

    PreMathFn<PMFatan> *f_atan = new PreMathFn<PMFatan> ;
    DB_NEWALLOC(f_atan,"PreMathFn") ;
    AddFunction("atan",f_atan) ;

    PreMathFn<PMFcos> *f_cos = new PreMathFn<PMFcos> ;
    DB_NEWALLOC(f_cos,"PreMathFn") ;
    AddFunction("cos",f_cos) ;

    PreMathFn<PMFsin> *f_sin = new PreMathFn<PMFsin> ;
    DB_NEWALLOC(f_sin,"PreMathFn") ;
    AddFunction("sin",f_sin) ;

    PreMathFn<PMFtan> *f_tan = new PreMathFn<PMFtan> ;
    DB_NEWALLOC(f_tan,"PreMathFn") ;
    AddFunction("tan",f_tan) ;

    PreMathFn<PMFexp> *f_exp = new PreMathFn<PMFexp> ;
    DB_NEWALLOC(f_exp,"PreMathFn") ;
    AddFunction("exp",f_exp) ;

    PreMathFn<PMFlog> *f_log = new PreMathFn<PMFlog> ;
    DB_NEWALLOC(f_log,"PreMathFn") ;
    AddFunction("log",f_log) ;

    PreMathFn<PMFsqr> *f_sqr = new PreMathFn<PMFsqr> ;
    DB_NEWALLOC(f_sqr,"PreMathFn") ;
    AddFunction("sqrt",f_sqr) ;

    PreFnRan *f_ran = new PreFnRan ;
    DB_NEWALLOC(f_ran,"PreFnRan") ;
    AddFunction("ran",f_ran) ;

    PreFnPrint *f_print = new PreFnPrint ;
    DB_NEWALLOC(f_print,"PreMathFn") ;
    AddFunction("print",f_print) ;

    PreFnFopen *f_fopen = new PreFnFopen ;
    DB_NEWALLOC(f_fopen,"PreFnFopen") ;
    AddFunction("fopen",f_fopen) ;

    PreFnFclose *f_fclose = new PreFnFclose ;
    DB_NEWALLOC(f_fclose,"PreFnFclose") ;
    AddFunction("fclose",f_fclose) ;

    PreFnFprint *f_fprint = new PreFnFprint ;
    DB_NEWALLOC(f_fprint,"PreFnFprint") ;
    AddFunction("fprint",f_fprint) ;

}

void PredefFunctionTable::AddFunction(char *name, PredefinedFunction *Fn) {
    if (LookUp(name)) {
	Fn->Error(name, ": Redefinition of function (last definition ignored).") ;
    } else {
       PredefFNC *newfn = new PredefFNC(name,Fn) ; 
       DB_NEWALLOC(newfn,"PredefFNC") ;
       Append(newfn) ;
    }
}

void PredefFunctionTable::RemoveFunction(char *name) {
    First() ;
    while (!IsEnd()) {
	if (!strcmp(Get()->GetName(),name)) {
	    DB_DELETE(Get(),"?") ;
	    delete Get();
	    Clear() ;
	    return ;
	}
	Next() ;
    }
}

void PredefFunctionTable::Destroy() {
    First() ;
    while (!IsEnd()) {
    	DB_DELETE(Get(),"?") ;
	delete Get();
	Clear() ;
    }
}

PredefinedFunction * PredefFunctionTable::LookUp(const char *s)  {
    PredefFNC *FNC ;
    
    First() ;
    while (!IsEnd()) {
	FNC = Get() ;
//	console.Info() << "checking " << FNC->name << " with " << s << "\n" ; 
	if (!strcmp(FNC->GetName(),s)) {
	    return FNC->GetFn() ;
	}
	Next() ;
    }
    return NULL ;
} 

int PredefinedFunction::CheckArgs(ExpressionNode *args) {
    args->First() ; args->Next() ;
    int nb_args=0 ;
    while (!args->IsEnd()) {
	nb_args ++ ;
	args->Next() ;
    }
    if (argc != UNLIMITED_ARGS) {
	if (nb_args != argc) {
	    yyltype loc ;
	    args->First() ;
	    loc.first_line = args->first_line ;
	    loc.first_column = args->first_column ;
	    loc.text = args->curfile ;
	    Error("Number of arguments mismatch",&loc) ;
	    return false;
	}
    }

    args->First() ; args->Next() ;
    int iarg=0 ;

    while (!args->IsEnd()) {
	if ((argt[iarg] != OpValue::NONE) && (args->Get()->ValueType != argt[iarg])) {
	    yyltype loc;
	    loc.first_line = args->Get()->first_line ;
	    loc.first_column = args->Get()->first_column ;
	    loc.text = args->Get()->curfile ;
	    Error("Argument type mismatch",&loc) ;
	    return false ;
	}
	args->Next() ;
	if (argc != UNLIMITED_ARGS) iarg++ ;
    }
    return true ;
}


PredefinedFunction::PredefinedFunction(int iargc, OpValue::OType itype) : ParserError(),ParserStatics()
{
    if (iargc < 1 ) {
	argt = new OpValue::OType[1] ;
	DB_NEWALLOC(argt,"OpValue::OType") ;
	argt[0] = itype ;
    } else {
	argt = new OpValue::OType[iargc] ;
	DB_NEWALLOC(argt,"OpValue::OType") ;
	int i;
	for (i=0; i< iargc; i++) {
	    argt[i] = itype ;
	}
    }
    argc = iargc ; 
}

PredefinedFunction::PredefinedFunction(int iargc, OpValue::OType *itype) : ParserError(),  ParserStatics(){
    if (iargc < 1 ) {
	argt = new OpValue::OType[1] ;
	DB_NEWALLOC(argt,"OpValue::OType") ;
	argt[0] = itype[0] ;
    } else {
	argt = new OpValue::OType[iargc] ;
	DB_NEWALLOC(argt,"OpValue::OType") ;
	int i;
	for (i=0; i< iargc; i++) {
	    argt[i] = itype[i] ;
	}
    }
    argc = iargc ; 
}

int PredefinedFunction::Eval(ExpressionNode *args) { 
    if (!CheckArgs(args)) return false ;
    return true;
}

PredefinedFunction::~PredefinedFunction() {
    if (argt) {
	DB_DELETE(argt,"?") ;
	delete argt ;
    }
}


/********************************************************************************/
/* class PreFnPrint, PreMathFn							*/
/* 										*/
/* Description:									*/
/* standard predefined functions (print & trig fnc.)				*/
/* 										*/
/********************************************************************************/

PreFnPrint::PreFnPrint() : PredefinedFunction(UNLIMITED_ARGS,OpValue::NONE) {  
}  

OpValue PreFnPrint::fn(ExpressionNode *args) {
//    if (!CheckArgs(args)) return OpValue(0L) ;
    args->First() ;
    args->Next() ;
    console.Message() ;
    while (!args->IsEnd()) {
	ExpressionNode *Ptr = args->Get();
	switch (Ptr->ValueType) {
	case OpValue::INT:
	    console.Message(false) << Ptr->Content.l ;
	    break;
	case OpValue::REAL:
	    console.Message(false) << Ptr->Content.d ;
	    break;
	case OpValue::STR:
	    console.Message(false) << Ptr->Content.str ;
	    break;
	case OpValue::PTR:
	    console.Message(false) << "<!ptr!>" ;
	    break;
	case OpValue::NONE:
	    console.Message(false) << "<!utp!>" ;
	    break;
	default:
	    console.Message(false) << "<!???!>" ;
	    break ;
	}
	args->Next() ;
    }
    console.Message(false) << "\n" ;
    return OpValue(1L) ;
}

OpValue PreFnRan::fn(ExpressionNode *) {
    return OpValue((real)(shake_random())/(real)(shake_RAND_MAX)) ;
}

List<CInternalFiles::CSFile> CInternalFiles::Files  ;

int CInternalFiles::nbFile=0 ;

CInternalFiles::CInternalFiles() {
} 

CInternalFiles::~CInternalFiles() {
}

int CInternalFiles::open(char *filename) {
    FILE *file ;

    nbFile=1 ;

    Files.First() ;
    while (!Files.IsEnd()) {
        if (Files.Get()->handle == nbFile) {
            nbFile++ ;
            Files.First() ;
        } else {
            Files.Next() ;
        }
    }

    file = fopen(filename,"a+") ;
    CSFile *newcsf = new CSFile ;
    newcsf->file = file ;
    newcsf->handle = nbFile ;
    Files.Append(newcsf) ;
    return newcsf->handle ;
} 

void CInternalFiles::close(int Id) {
    FILE *file = getFile(Id) ;
    if (file) {
        fclose(file) ;
        Files.First() ;
        while (!Files.IsEnd()) {
            if (Files.Get()->handle == Id) {
                delete Files.Get() ;
                Files.Clear() ;
                return ;
            }
            Files.Next() ;
        }
    }
}

FILE *CInternalFiles::getFile(int Id) {
    Files.First() ;
    while (!Files.IsEnd()) {
        if (Files.Get()->handle == Id) return Files.Get()->file ;
        Files.Next() ;
    }
    return NULL ;
}

PreFnFopen::PreFnFopen() : CInternalFiles() , PredefinedFunction(1,OpValue::STR) {
}

OpValue PreFnFopen::fn(ExpressionNode *args) {
    args->First() ;
    args->Next() ;
    ExpressionNode *Ptr = args->Get();
    int Id = open(Ptr->Content.str) ;
    return OpValue((long)Id) ;
}

PreFnFclose::PreFnFclose() : CInternalFiles() , PredefinedFunction(1,OpValue::INT) {
}

OpValue PreFnFclose::fn(ExpressionNode *args) {
    args->First() ;
    args->Next() ;
    ExpressionNode *Ptr = args->Get();
    close(Ptr->Content.l) ;
    return OpValue(1L) ;
}

PreFnFprint::PreFnFprint() : CInternalFiles() , PredefinedFunction(UNLIMITED_ARGS,OpValue::NONE) {
}

OpValue PreFnFprint::fn(ExpressionNode *args) {
    args->First() ;
    args->Next() ;
    if (args->IsEnd()) return OpValue(0L) ;
    int FileId = args->Get()->Content.l ;
    FILE *file=getFile(FileId) ;
    if (!file) return OpValue(0L) ;
    args->Next() ;
    while (!args->IsEnd()) {
	ExpressionNode *Ptr = args->Get();
	switch (Ptr->ValueType) {
	case OpValue::INT:
	    fprintf(file,"%d",Ptr->Content.l) ;
	    break;
	case OpValue::REAL:
	    fprintf(file,"%f",Ptr->Content.d) ;
	    break;
	case OpValue::STR:
	    fprintf(file,"%s",Ptr->Content.str) ;
	    break;
	case OpValue::PTR:
	    fprintf(file,"<!ptr!>") ;
	    break;
	case OpValue::NONE:
	    fprintf(file,"<!utp!>") ;
	    break;
	default:
	    fprintf(file,"<!???!>") ;
	    break ;
	}
	args->Next() ;
    }
    fprintf(file,"\n") ;
    return OpValue(1L) ;
}


/********************************************************************************/
/* class UserFunction								*/
/* 										*/
/* Description:									*/
/*   Handles arguments checks and user defined function evalution		*/
/* 										*/
/********************************************************************************/

UserFunction::UserFunction(FunctionDeclaration *iFn, int argc, OpValue::OType itype) 
			    : PredefinedFunction(argc,itype) {
   Fnc = iFn ;
}

UserFunction::~UserFunction() {
    if (Fnc) {
        DB_DELETE(Fnc,"?") ;
	delete Fnc ;
    }
    if (argt) {
        DB_DELETE(argt,"?") ;
	delete argt ;
        argt = NULL ;
    }
}

int UserFunction::Eval(ExpressionNode *args) {
    if (!CheckArgs(args)) return false ;
    return Fnc->Call(args) ;
} 

OpValue UserFunction::fn(ExpressionNode *) {
    OpValue res = *Fnc ;
    return res;
} 

/********************************************************************************/
/* class FunctionCall		 						*/
/* 										*/
/* Description:									*/
/* Its value contain the value of the function					*/
/* The first arg. is the function name						*/
/* The folowwing args are the argumnets transmit to the function		*/
/* 										*/
/********************************************************************************/

FunctionCall::FunctionCall(OpValue &fname) {
    First() ;
    ValueType = NONE ;
    ValueRef *VR = new ValueRef(fname) ;
    DB_NEWALLOC(VR,"ValueRef") ;
    *this << VR ;
    ExpressionType="FunctionCall";
}

int FunctionCall::Eval() {
    First() ;
    PredefinedFunction *ThisFn = PF->LookUp(Get()->Content.str) ;
    if (ThisFn == NULL) {
	if ((first_line!=0) && (first_column!=0)) {
	   yyltype loc ; 
	   loc.first_line = first_line ;
	   loc.first_column = first_column ;
	   loc.text =curfile ;
	   Error("Undefined function",&loc) ;
	}
	return false;
    }
    First() ;
    Next() ;

    // Eval arguments
    while (!IsEnd()) {
	if (!Get()->Eval()) return false ;
	Next() ;
    }

    // call function (ie. ThisFn->Eval)
    if (ThisFn->Eval(this)) { 
	*(OpValue *)(this) = ThisFn->fn(this) ;
	return true ;
    } else {
	return false ;
    }
}

/********************************************************************************/
/* class FunctionDeclaration & FncArgs & ReturnValue				*/
/* 										*/
/* Sym: <type> <symbolref><fncargs><block>					*/
/* Description:									*/
/*   its value is a dummy value of same type than the return value type		*/
/*   The first child is a symbol ref. (the function name)			*/
/*   The second child is the list of argumnets		    			*/
/*   The third child is the function core	        			*/
/* 										*/
/*   FncArgs is a list of Declaration						*/
/* 										*/
/*   The value of ReturnValue set the return value of the function being	*/
/*   evaluated									*/
/* 										*/
/********************************************************************************/

FunctionDeclaration::FunctionDeclaration(OType val_type, 
					 ExpressionNode *Name, 
					 ExpressionNode *Args, 
					 ExpressionNode *Block) {
    First() ;
    ValueType = val_type ;
    *this << Name << Args << Block ;
    ExpressionType="FunctionDeclaration";
}

int FunctionDeclaration::Eval() {
    return Declare() ;
}

int FunctionDeclaration::Declare() {
    First() ;
    ExpressionNode *Name = Get() ; Next() ;
    ExpressionNode *Args = Get() ; Next() ;
    Name->First() ;
    Name = Name->Get() ;
    int nargs=0;
    Args->First() ;
    while (!Args->IsEnd()) {
	Args->Next() ;
	nargs++ ;
    }
    UserFunction *thisFn = new UserFunction(this,nargs) ;
    DB_NEWALLOC(thisFn,"UserFunction") ;
    Args->First() ;
    int i;
    for (i=0; i<nargs; i++) {
	thisFn->ArgType(i) = Args->Get()->ValueType ;
	Args->Next() ;
    }
    PF->AddFunction(Name->Content.str,thisFn) ;

//  Add return Symbol for return value ;
    Symbol *Sym;
    Sym = VT->LookUp("return") ;	// Lookup in table of variables
    if (!Sym) {
	Sym = VT->AddSymbol("return", NONE) ;
    }

    return true ;
}

int FunctionDeclaration::Call(ExpressionNode *Args) {
    First() ;     
    ExpressionNode *Name = Get() ; Next() ;
    ExpressionNode *FnArgs = Get() ; Next() ;
    ExpressionNode *Block = Get() ;      
    Name->First() ;
    Name = Name->Get() ;
    VT->PushStack() ;

    // declare arguments in variable table
    if (!FnArgs->Declare()) {
	VT->PopStack() ;
	return false ; 
    }

    // perform affectation to function arguents
    FnArgs->First() ; Args->First() ; Args->Next() ; 
    while (!FnArgs->IsEnd()) {
	// From the list of arguments get the current declaration (ie. the expression "int a" for instance) 
	// pointer (FnArgs->Get())
	// then get the symbol pointer (FnArgs->Get()->Get()) and set the value in the variable table
	// (ie. *FnArgs->Get()->Get()->PtrValue = *Args->Get() )
	FnArgs->Get()->First() ; 
	*FnArgs->Get()->Get()->PtrValue = *Args->Get() ;
	FnArgs->Next() ; Args->Next() ;
    }
    int res = Block->Eval() ; // Evaluate function
    VT->PopStack() ; // Delete local variables created during Eval
    if (res) {
	Symbol *Sym;
	Sym = VT->LookUp("return") ;	// Lookup in table of variables the return value
	if (Sym) {
	    *((OpValue *)(this)) = *Sym->value ;
	    Sym->value->ValueType = NONE ;
	} else {
	    yyltype loc ; 
	    loc.first_line = first_line ;
	    loc.first_column = first_column ;
	    loc.text =curfile ;
	    Error("All function should return a value",&loc) ;
	    return false;
	}
    }
    return_has_been_reached = false ;
    return res;
}

/********************************************************************************/

FncArgs::FncArgs() { 
    ExpressionType="FncArgs";
}

int FncArgs::Declare() {
    First() ;
    while (!IsEnd()) {
	if (!Get()->Eval()) return false ;
	Next() ;
    }
    return true;
}

/********************************************************************************/

ReturnValue::ReturnValue(ExpressionNode *val) {
    First() ;
    *this << val ;
    ExpressionType="ReturnValue";
}

ReturnValue::ReturnValue() {
}
    
int ReturnValue::Eval() {
    return_has_been_reached = true ;
    First() ;
    if (!Get()->Eval()) return false ;
	Symbol *Sym;
	Sym = VT->LookUp("return") ;	// Lookup in table of variables the return value
	if (Sym) {
	    *Sym->value = *(OpValue *)(Get()) ;
	} else {
	    yyltype loc ; 
	    loc.first_line = first_line ;
	    loc.first_column = first_column ;
	    loc.text =curfile ;
	    Error("unexpected return statement",&loc) ;
	    return false;
	}
    return true;
}
