/********************************************************************************/
/* FUNCTIONS.H		 							*/
/* 										*/
/* Description:	  Parser evaluation functions					*/
/* Author(s):     D. Place (QUAKES)						*/
/* Creation date: February 1999							*/
/*										*/
/********************************************************************************/
#ifndef _FUNCTIONS_H_
#define _FUNCTIONS_H_

#include "yyltype.h"
#include "mtypes.h"
#include "expression.h"
#include "shake_debug.h"
#include "mainparser.h"
#include "t_list.h"

#define UNLIMITED_ARGS	-1

/*!
  \class PredefinedFunction
  \brief Handles arguments checks and function evalution
*/
class PredefinedFunction: public ParserError, public ParserStatics {
protected:
    int argc ;
    OpValue::OType *argt ;
    
    int CheckArgs(ExpressionNode *args) ;
public:
    PredefinedFunction(int argc=0, OpValue::OType itype=OpValue::NONE) ;
    PredefinedFunction(int argc, OpValue::OType *itype) ;
    virtual ~PredefinedFunction() ;
    virtual OpValue fn(ExpressionNode *args) = 0 ;
    // return UNLIMITED_ARGS if unlimited number of args
    int ArgCount() {return argc ; } ;	       
    // return type of argumnent #iargc
    OpValue::OType & ArgType(int iargc) { if (argc==UNLIMITED_ARGS) return argt[0] ; else return argt[iargc]; } ;    
    virtual int Eval(ExpressionNode *args) ;
} ;

class FunctionDeclaration;

/*!
  \class UserFunction
  \brief Handles arguments checks and user defined function evalution
*/
class UserFunction: public PredefinedFunction {
protected:
    FunctionDeclaration *Fnc;
public:
    UserFunction(FunctionDeclaration *iFn, int argc=0, OpValue::OType itype=OpValue::NONE) ;
    ~UserFunction() ;
    OpValue fn(ExpressionNode *args) ;
    int Eval(ExpressionNode *args) ;
} ;

/*!
  \class PreFnPrint
  \brief standard predefined print functions with unlimited number of aguments
*/
class PreFnPrint : public PredefinedFunction {
public:
    PreFnPrint() ; 
    OpValue fn(ExpressionNode *args) ;
} ;

/*!
  \class PreFnPrint
  \brief standard predefined ran functions (return random double)
*/
class PreFnRan : public PredefinedFunction {
public:
    PreFnRan() : PredefinedFunction(0,OpValue::NONE) { } ; 
    OpValue fn(ExpressionNode *args) ;
} ;

/*!
  \class CInternalFiles
  \brief Internal files to open, read, write etc...
*/
class CInternalFiles {
protected:
    struct CSFile {
        FILE *file ;
        int handle ;
    } ;
    static List<CSFile> Files;
    static int nbFile ;
public:
    CInternalFiles() ;
    ~CInternalFiles() ;

    int open(char *file) ;
    FILE *getFile(int Id) ;
    void close(int Id) ;
} ;

class PreFnFopen : public CInternalFiles, public PredefinedFunction {
public:
    PreFnFopen() ; 
    OpValue fn(ExpressionNode *args) ;
} ;

class PreFnFclose : public CInternalFiles, public PredefinedFunction {
public:
    PreFnFclose()  ; 
    OpValue fn(ExpressionNode *args) ;
} ;

class PreFnFprint : public CInternalFiles, public PredefinedFunction {
public:
    PreFnFprint() ; 
    OpValue fn(ExpressionNode *args) ;
} ;


template <class F>
class PreMathFn : public PredefinedFunction {
public:
    PreMathFn() : PredefinedFunction(1,OpValue::REAL) { } ; 
    OpValue fn(ExpressionNode *args) ;
} ;

template <class F>
OpValue PreMathFn<F>::fn(ExpressionNode *args) {
//    if (!CheckArgs(args)) return OpValue(0L) ;
    args->First() ;
    args->Next() ;
    return OpValue(F::calc(args->Get()->GetRealVal())) ;
}

//PredefFunctionTable dummy_instatiation ;

class PMFacos { public: inline static double calc(double d) { return acos(d); } ; } ;
class PMFasin { public: inline static double calc(double d) { return asin(d); } ; } ;
class PMFatan { public: inline static double calc(double d) { return atan(d); } ; } ;
class PMFcos { public: inline static double calc(double d) { return cos(d); } ; } ;
class PMFsin { public: inline static double calc(double d) { return sin(d); } ; } ;
class PMFtan { public: inline static double calc(double d) { return tan(d); } ; } ;
class PMFexp { public: inline static double calc(double d) { return exp(d); } ; } ;
class PMFlog { public: inline static double calc(double d) { return log(d); } ; } ;
class PMFsqr { public: inline static double calc(double d) { return sqrt(d); } ; } ;


/*!
  \class PredefFNC
  \brief Associate Predifine function code with name
*/
class PredefFNC {
protected:
    PredefinedFunction *Fn ; 
    char *name; 
public:
    PredefFNC(char *iname, PredefinedFunction *iFn) { name = iname; Fn = iFn ; } ;
    ~PredefFNC() { 
    	DB_DELETE(Fn,"?") ;
	delete Fn ; 
    } ;
    PredefinedFunction *GetFn() { return Fn ; } ; 
    char *GetName() { return name; } ; 


} ;

/*!
  \class PredefFunctionTable
  \brief Table of all Predifined Function
*/
class PredefFunctionTable : public List<PredefFNC> {
public:
    PredefFunctionTable() ;
    ~PredefFunctionTable() ;
    void AddFunction(char *name, PredefinedFunction *Fn) ;
    PredefinedFunction * LookUp(const char *s)  ; // create and link value from Global Variable
    void RemoveFunction(char *name) ;
    void Destroy() ;
    void Init() ;
} ;


/*!
  \class FunctionDeclaration
  \brief Fruntion declaration (<type> <symbolref><fncargs><block>)

  its value is a dummy value of same type than the return value type		
  The first child is a symbol ref. (the function name)			
  The second child is the list of argumnets		    			
  The third child is the function core	        			
*/
class FunctionDeclaration: public ExpressionNode {
public:
    FunctionDeclaration(OType val_type, ExpressionNode *Name, ExpressionNode *Args, ExpressionNode *Block) ;

    virtual int Eval() ;
    virtual int Declare() ;
    int Call(ExpressionNode *Args) ;
};

/*!
  \class FncArgs
  \brief Argument list for functions (ie a list of Declaration)
*/
class FncArgs: public ExpressionNode {
public:
    FncArgs() ;

    virtual int Declare() ;
};

/*!
  \class ReturnValue
  \brief The value of ReturnValue set the return value of the function being evaluated
*/
class ReturnValue: public ExpressionNode {
public:
    ReturnValue(ExpressionNode *val) ;
    ReturnValue() ;

    virtual int Eval() ;
};

/*!
  \class FunctionCall
  \brief Function call
  Its value contain the value of the function
  The first arg. is the function name
  The folowwing args are the argumnets transmit to the function
*/
class FunctionCall: public ExpressionNode {
public:
    FunctionCall(OpValue &fname) ;

    virtual int Eval() ;
};

/*!
   \class BinaryFunction
   \brief Function with two operandes (sym: <exp> '+|-|*...' <exp>)

   Its value contain the sum or difference or ... of the first and second child
*/
template<class F>
class BinaryFunction : public ExpressionNode {
public:
    BinaryFunction() : ExpressionNode() { ExpressionType="BinaryFunction"; } ;
    BinaryFunction(ExpressionNode *Left, ExpressionNode *Right);
    virtual int Eval() ;
} ;


template<class F>
BinaryFunction<F>::BinaryFunction(ExpressionNode *Left, ExpressionNode *Right): ExpressionNode() {
    First() ;
    *this << Left ;
    *this << Right ;
} 

template<class F>
int BinaryFunction<F>::Eval() {
    ExpressionNode * L , * R ;
    First() ; L = Get() ; Next() ; R = Get() ;
    L->Eval() ; R->Eval() ;
    if (L->IsANumber() && R->IsANumber()) {
	ValueType = MAX(L->ValueType,R->ValueType) ;
	switch (ValueType) {
	case INT:
	    *(OpValue *)(this) = F::calc(L->GetIntVal(),R->GetIntVal()) ;
	    return true;
	case REAL:
	    *(OpValue *)(this) = F::calc(L->GetRealVal(),R->GetRealVal()) ;
	    return true; ;
	default:
	    break ;
	}
    }
    ValueType = NONE ;
    return 0 ;
} 

class AddOp {
public:
    static long calc(long a, long b) { return a+b; } ;
    static double calc(double a, double b) { return a+b; } ;
};
class MulOp {
public:
    static long calc(long a, long b) { return a*b; } ;
    static double calc(double a, double b) { return a*b; } ;
};
class DivOp {
public:
    static long calc(long a, long b) { return a/b; } ;
    static double calc(double a, double b) { return a/b; } ;
};
class SubOp {
public:
    static long calc(long a, long b) { return a-b; } ;
    static double calc(double a, double b) { return a-b; } ;
};
class PowOp {
public:
    static long calc(long a, long b) { return (long)(pow(double(a),double(b))); } ;
    static double calc(double a, double b) { return pow(a,b); } ;
};


/*!
   \class BooleanFunction
   \brief Function with two operandesBoolean functions (sym: <exp> '==|&&|<...' <exp>)

   Its value contain the result of the operation between the first and second child
*/
template<class F>
class BooleanFunction : public ExpressionNode {
public:
    BooleanFunction() : ExpressionNode() { ExpressionType="BooleanFunction"; } ;
    BooleanFunction(ExpressionNode *Left, ExpressionNode *Right);
    virtual int Eval() ;
} ;


template<class F>
BooleanFunction<F>::BooleanFunction(ExpressionNode *Left, ExpressionNode *Right): ExpressionNode() {
    First() ;
    *this << Left ;
    *this << Right ;
} 

template<class F>
int BooleanFunction<F>::Eval() {
    ExpressionNode * L , * R ;
    First() ; L = Get() ; Next() ; R = Get() ;
    L->Eval() ; R->Eval() ;
    ValueType = INT ;
    switch (MAX(L->ValueType,R->ValueType)) {
    case STR:
	if (L->ValueType == R->ValueType) {
	    *(OpValue *)(this) = (long)(!strcmp(L->Content.str, R->Content.str)) ;
	}
	return true ;
    case PTR:
    case INT:
	*(OpValue *)(this) = F::calc(L->GetIntVal(),R->GetIntVal()) ;
	return true;
    case REAL:
	*(OpValue *)(this) = F::calc(L->GetRealVal(),R->GetRealVal()) ;
	return true; ;
    default:
	break ;
    }
    ValueType = NONE ;
    return 0 ;
} 

class EQOp {
public:
    static long calc(long a, long b) { return a == b; } ;
    static long calc(double a, double b) { return a == b; } ;
};
class GEOp {
public:
    static long calc(long a, long b) { return a >= b; } ;
    static long calc(double a, double b) { return a >= b; } ;
};
class LEOp {
public:
    static long calc(long a, long b) { return a <= b; } ;
    static long calc(double a, double b) { return a <= b; } ;
};
class GTOp {
public:
    static long calc(long a, long b) { return a > b; } ;
    static long calc(double a, double b) { return a > b; } ;
};
class LTOp {
public:
    static long calc(long a, long b) { return a < b; } ;
    static long calc(double a, double b) { return a < b; } ;
};
class OROp {
public:
    static long calc(long a, long b) { return a || b; } ;
    static long calc(double a, double b) { return a || b; } ;
};
class ANDOp {
public:
    static long calc(long a, long b) { return a && b; } ;
    static long calc(double a, double b) { return a && b; } ;
};
class NOTOp {
public:
    static long calc(long , long b) { return ! b; } ;
    static long calc(double , double b) { return ! b; } ;
};

#endif




