/********************************************************************************/
/* MTYPES.CPP		 							*/
/* 										*/
/* Description:	  Parser evaluation functions					*/
/* Author(s):     D. Place (QUAKES)						*/
/* Creation date: February 1999							*/
/*										*/
/********************************************************************************/
#include "stdafx.h"
#include "mtypes.h"
#include "parse.h"
#include "functions.h"
#include "shake_debug.h"
#include <string.h>

int ExpressionNode::return_has_been_reached = false ;

/********************************************************************************/
/* base class Symbol	 							*/
/* 										*/
/* Description:									*/
/*   A symbol is a variable name or function name that contains a value		*/
/*										*/
/********************************************************************************/

Symbol::Symbol()  {
    name = NULL ;
    value = NULL ;
    deleteValue = false ;
}

Symbol::Symbol(const char *s, OpValue::OType iType)  {
    value = new OpValue(iType) ;
    DB_NEWALLOC(value,"OpValue") ;
    deleteValue = true ;
    name = NULL ;
    SetName(s) ;
}

Symbol::~Symbol() {
    if (name) {
        DB_DELETE(name,"?") ;
	delete name ;
    }
    if (deleteValue) {
        DB_DELETE(value,"?") ;
	delete value ;
    }
}

void Symbol::SetName(const char *s) {
    if (name) {
        DB_DELETE(name,"?") ;
	delete name ;
    }
    name = new char[strlen(s)+1] ;
    DB_NEWALLOC(name,"char") ;
    strcpy(name, s) ;
}

OpValue & Symbol::operator = (OpValue &s)  {
    if (!value) {
	deleteValue = true ; 
	value = new OpValue(s.ValueType) ;
	DB_NEWALLOC(value,"OpValue") ;
    }
    (*value) = s ;
    return *value ; 
}
/********************************************************************************/
/* class StackList	 							*/
/* 										*/
/* Description:									*/
/*   A list of symbols which can be (not compulsory) used as a stack		*/
/*   PushStack placed a marker and PopStack erase objects inserted in the	*/
/*   since the last call of PushStack						*/
/*   This is use for declaration of local variable than needs to be delete	*/ 
/*   when exiting the function							*/
/*										*/
/********************************************************************************/

StackList::StackList() : List<Symbol>() { 
    StackPos = 0 ; 
    Stack[StackPos] = NULL ; StackPos++ ; 
    Stack[StackPos] = NULL ;
} 

int StackList::PushStack() { 
    if (StackPos >= STACKSIZE-1) return SL_STACKOVERFLOW ;
    Stack[StackPos] = End ; StackPos++ ; 
    Stack[StackPos] = NULL ; 
    return SL_NOERROR ;
} 

int StackList::PopStack() { 
    if (StackPos <= 0) return SL_STACKUNDERFLOW ;
    StackPos-- ;
    Current = Stack[StackPos] ;
    Stack[StackPos] = NULL ; 
    if (Current == NULL) { return SL_NOERROR ; }
    Next() ;
    while (!IsEnd()) Clear() ;
    return SL_NOERROR ;
} 

void StackList::Destroy() { 
    StackPos = 0 ;
    Stack[StackPos] = NULL ; StackPos++ ; 
    Stack[StackPos] = NULL ;
    List<Symbol>::Destroy() ;
} 

void StackList::BeforeFirstLocal() { Current = Stack[StackPos-1] ; }

int StackList::IsEndOfStack(int i) { 
    if (Stack[i]==NULL) return (Current==NULL) ;
    if (Stack[i]->Next==NULL) return (Current==NULL) ;
    return Current==Stack[i]->Next ; 
}


/********************************************************************************/
/* class VaribaleTable	 							*/
/* 										*/
/* Description:									*/
/*   The list of all varibales and functions declared in the script		*/
/*										*/
/********************************************************************************/

VariableTable::VariableTable() : StackList() {
    Init() ;
}

void VariableTable::Init() {
}

Symbol * VariableTable::AddSymbol(char *name, OpValue::OType iType) {
    Symbol *Sym=new Symbol(name,iType) ;
    DB_NEWALLOC(Sym,"Symbol") ;
    Last() ;
    *this << Sym ;
    return Sym ;
}

Symbol * VariableTable::AddGlobalSymbol(char *name, OpValue::OType iType) {
    Symbol *Sym=new Symbol(name,iType) ;
    DB_NEWALLOC(Sym,"Symbol") ;
    InsertAtStart(Sym) ;
    return Sym ;
}

void VariableTable::DelSymbol(char *name) {
    First() ;
    while (!IsEnd()) {
      if (!strcmp(Get()->name,name)) {
        DB_DELETE(Get(),"?") ;
	delete Get() ;
	Clear() ;
      } else {
	Next() ;
      }
    }
}

void VariableTable::Destroy() {
    First() ;
    while (!IsEnd()) {
        DB_DELETE(Get(),"?") ;
	delete Get() ;
	Next() ;
    }
    StackList::Destroy() ;
}

Symbol * VariableTable::LookUp(const char *s)  {
    Symbol *Sym ;
    
    First() ;
    while (!IsEndOfStack(1)) {
	Sym = Get() ;
	if (!strcmp(Sym->name,s)) {
	    return Sym ;
	}
	Next() ;
    }
    BeforeFirstLocal() ; Next() ;
    while (!IsEnd()) {
	Sym = Get() ;
	if (!strcmp(Sym->name,s)) {
	    return Sym ;
	}
	Next() ;
    }
    return NULL ;
}

int VariableTable::PushStack()  {
    return StackList::PushStack() ;
}

int VariableTable::PopStack() {
    BeforeFirstLocal() ;
    Next() ;
    while (!IsEnd()) {
        DB_DELETE(Get(),"?") ;
	delete Get() ;
	Next() ;
    }
    return StackList::PopStack() ;
}

/********************************************************************************/
/* base class OpValue & OpValueArray						*/
/* 										*/
/* Description:									*/
/*   Untyped description of a value						*/
/*   OpValue is a value of type INT, REAL, String, Pointer to an array		*/
/*   or Untyped									*/
/*										*/
/********************************************************************************/

OpValue::OpValue(OType iType)  {
    ValueType = iType ;
    PleaseDelete = false ;
    switch (ValueType) {
    case INT:
	Content.l = 0L ; 
	break;
    case REAL:
	Content.d = 0.0 ; 
	break;
    case STR:
	Content.str = NULL ;
	break;
    case PTR:
	Content.ptr = NULL ;
	break;
    default:
	break;
    } 
}

OpValue::OpValue(long l)  {
    ValueType = INT ;
    Content.l = l ; 
    PleaseDelete = false ;
}

OpValue::OpValue(double d)  {
    ValueType = REAL ;
    Content.d = d ; 
    PleaseDelete = false ;
}

OpValue::OpValue(OpValueArray * ptr)  {
    ValueType = PTR ;
    Content.ptr = ptr ; 
    PleaseDelete = false ;
}

OpValue::OpValue(char *str, int deleteit)  {
    ValueType = STR ;
    if (deleteit!=0) {
	Content.str = str ;
    } else {
        Content.str = new char[strlen(str)+1] ;
	DB_NEWALLOC(Content.str,"char") ;
	strcpy(Content.str,str) ;
    }
    PleaseDelete = true ;
}

OpValue::~OpValue() {
  if ((PleaseDelete)&&(Content.str)&&(ValueType==STR)) {
    DB_DELETE(Content.str,"?") ;
    delete [] Content.str ;
  }
}

int OpValue::Allocate(OpValue::OType ot, int iDim, int NDim, int *Dims) {
  int res = true ;
  if (iDim == NDim) {
    ValueType = PTR ;
    Content.ptr = new OpValueArray(ot, Dims[iDim-1]) ;
    DB_NEWALLOC(Content.ptr,"OpValueArray") ;
  } else {
    int i ;
    ValueType = PTR ;    
    Content.ptr = new OpValueArray(OpValue::PTR, Dims[iDim-1]) ;
    DB_NEWALLOC(Content.ptr,"OpValueArray") ;
    for (i=0; i< Dims[iDim-1]; i++) {
      res = res && (*Content.ptr)[i].Allocate(ot,iDim+1,NDim,Dims);
    }
  }	
  return res;
}	

int OpValue::IsANumber()  {
    return (ValueType == INT)||(ValueType==REAL) ;
}

OpValue & OpValue::operator = (long l)  {
   *this = OpValue(l) ;
   return *this ;
}

OpValue & OpValue::operator = (double d)  {
   *this = OpValue(d) ;
   return *this ;
}

OpValue & OpValue::operator = (char * str) {
   *this = OpValue(str) ;
   return *this ;
}

OpValue & OpValue::operator = (OpValueArray * ptr) {
   *this = OpValue(ptr) ;
   return *this ;
}

OpValue & OpValue::operator = (const OpValue &ov) {
    if (ov.ValueType == ValueType) {
	if ((PleaseDelete)&&(Content.str)&&(ValueType==STR)) {
            DB_DELETE(Content.str,"?") ;
	    delete Content.str ;
	}
	ValueType = NONE ;
    }
    if (ValueType == NONE) {
	ValueType = ov.ValueType ;
	PleaseDelete = false ;
	switch (ValueType) {
	case INT:
	    Content.l = ov.Content.l ; 
	    break;
	case REAL:
	    Content.d = ov.Content.d ; 
	    break;
	case STR:
	    if (ov.Content.str) {
		Content.str = new char[strlen(ov.Content.str)+1] ;
		DB_NEWALLOC(Content.str,"char") ;
		strcpy(Content.str,ov.Content.str) ;
		PleaseDelete = true ;
	    }
	    break;
	case PTR:
	    Content.ptr = ov.Content.ptr ;
	    break;
	default:
	    break;
	} 
    } else { // some cast is necessary...
	switch (ov.ValueType) {
	case INT:
	    switch (ValueType) {
	    case REAL:
		Content.d = (double) ov.Content.l ;
		break;
	    default:
		// invalid cast...
		break;
	    }
	    break;
	case REAL:
	    switch (ValueType) {
	    case INT:
		Content.l = (long) ov.Content.d ;
		break;
	    default:
		// invalid cast...
		break;
	    }
	    break;
	case STR:
	case PTR:
	default:
	    // invalid cast...
	    break;
	}
	return *this ;
    }
    return *this ;
} 

double OpValue::GetRealVal() {
    switch (ValueType) {
    case INT:
	return (double)(Content.l) ;
    case REAL:
	return Content.d ;
    default:
	return 0.0;
    }
}

long OpValue::GetIntVal() {
    switch (ValueType) {
    case INT:
	return Content.l ;
    case REAL:
	return (long)(Content.d) ;
    case STR:
	return (long)(Content.str) ;
    case PTR:
	return (long)(Content.ptr) ;
    default:
	return 0L;
    }
}

int OpValue::GetValueFrom(void *ptr,int ind) {
    int res = true ;
    switch (ValueType) {
    case INT:
	*this = (long)(((integer *)(ptr))[ind]) ;
	break ;
    case REAL:
	*this = ((real *)(ptr))[ind] ;
	break ;
    case STR:
	*this = ((char**)(ptr))[ind] ;
	break ;
    case PTR:
	res = Content.ptr->GetValueFrom(ptr,ind) ;
	break ;
    default:
	return false;
    }
    return res ;
}

int OpValue::CopyValueTo(void *ptr,int ind) {
    int res = true ;
    switch (ValueType) {
    case INT:
	((integer *)(ptr))[ind] = Content.l;
	break ;
    case REAL:
	((real *)(ptr))[ind] = Content.d;
	break ;
    case STR:
	strcpy(((char**)(ptr))[ind],Content.str) ;
	break ;
    case PTR:
	res = Content.ptr->CopyValueTo(ptr,ind) ;
	break ;
    default:
	return false;
    }
    return res ;
}


int OpValueArray::GetValueFrom(void *ptr,int ind) {
    int res = true ;
    int i;
    for (i=0; i<Size; i++)
	res = res & values[i].GetValueFrom(ptr,ind*Size+i) ;
    return res ;
}

int OpValueArray::CopyValueTo(void *ptr,int ind) {
    int res = true ;
    int i;
    for (i=0; i<Size; i++)
	res = res & values[i].CopyValueTo(ptr,ind*Size+i) ;
    return res ;
}


OpValueArray::OpValueArray() {
    ValueType = PTR ; // the type of Array is PTR...
    Content.ptr = this ;
    values = NULL ;
    Size = 0 ;
}

OpValueArray::OpValueArray(OType ot, int iSize) {
    int i;

    ValueType = PTR ;
    Content.ptr = this ;
    Size = iSize ;
    values = new OpValue[Size] ;
    DB_NEWALLOC(values,"OpValue") ;
    
    for (i=0; i<Size; i++)
	values[i].ValueType = (ot) ;
}

OpValueArray::~OpValueArray() {
    if (values) {
        DB_DELETE(values,"?") ;
	delete [] values ;
    }
    Size = 0 ;
}

OpValue & OpValueArray::operator [] (int ind) {
    return values[ind] ;
}




