/********************************************************************************/
/* MTYPES.H		 							*/
/* 										*/
/* Description:	  Parser evaluation functions					*/
/* Author(s):     D. Place (QUAKES)						*/
/* Creation date: February 1999							*/
/*										*/
/********************************************************************************/
#ifndef _MTYPES_H_
#define _MTYPES_H_
#include "t_list.h"
#include "yyltype.h"
#include "shake_math.h"
#include <math.h>

//#define YYSTYPE Function
//struct yyltype ;
class Symbol ;
class OpValueArray ;

/*!
\class OpValue
\brief Untyped description of a value

OpValue is a value of type INT, REAL, String, Pointer to an array or Untyped
*/
class OpValue {
protected:
    int PleaseDelete ;
public:
    union {
	long l ;
	double d ;
	char *str ;
	OpValueArray *ptr ; //!< A value can be a pointer to a 1D array ;
	Symbol *sym ;
    } Content ;
    enum OType { INT, REAL, STR, PTR, NONE };
    OType ValueType ;

    OpValue(OType iType=NONE)  ;
    OpValue(long l)  ;
    OpValue(double d)  ;
    OpValue(OpValueArray *ptr)  ;
    OpValue(char *str, int deleteit=0)  ;

    double GetRealVal() ;
    long GetIntVal() ;

    virtual ~OpValue() ;

    int IsANumber()  ;

    OpValue & operator = (long l)  ;
    OpValue & operator = (double d)  ;
    OpValue & operator = (char * str) ;
    OpValue & operator = (OpValueArray * ptr) ;
    OpValue & operator = (const OpValue & ov) ;

//    OpValue & operator = (symbol & sym) ;

    int Allocate(OType ot, int iDim, int NDim, int *Dims) ;
    virtual int GetValueFrom(void *ptr,int ind=0) ;
    virtual int CopyValueTo(void *ptr,int ind=0) ;
} ;

/*!
  \class OpValueArray
  \brief Untyped description of an array
*/
class OpValueArray: public  OpValue {
protected:
    int Size;
    OpValue *values ; // Note that OpValue can embed an Array for multidimensional arrays
public:

    OpValueArray() ;
    OpValueArray(OType ot, int iSize) ;

    virtual ~OpValueArray();

    inline int GetSize() { return Size; } ;
    OpValue & operator [] (int ind) ;
    int GetValueFrom(void *ptr,int ind=0) ;
    int CopyValueTo(void *ptr,int ind=0) ;
} ;

/*! 
  \class Symbol
  \brief A symbol is a variable name or function name that contains a value
*/
class Symbol {
private:
    int deleteValue ;
public:
    char *name ;
    OpValue *value ;
    Symbol()  ; 
    Symbol(OpValue::OType iType)  ; 
    Symbol(const char *s, OpValue::OType iType=OpValue::NONE)  ; 
    ~Symbol() ;

    void SetName(const char *s);
    OpValue & operator = (OpValue &s)  ; // set the value of the symbol 
} ;

#define STACKSIZE 1000

/*! 
  \class StackList
  \brief List of symbols

  A list of symbols which can be (not compulsory) used as a stack
  PushStack placed a marker and PopStack erase objects inserted in the
  since the last call of PushStack
  This is use for declaration of local variable than needs to be delete
  when exiting the function
*/
class StackList : public List<Symbol> {
private:
    Node<Symbol> *Stack[STACKSIZE] ;
    int StackPos ;
public:
    enum ESLError {SL_NOERROR=0, SL_STACKOVERFLOW, SL_STACKUNDERFLOW } ;

    StackList() ;
    virtual ~StackList(){};
    virtual int PushStack();
    virtual int PopStack() ;
    void BeforeFirstLocal() ; // go to the element in front of the first local 
    int IsEndOfStack(int i) ;  // last element of the stack #i
    void Destroy() ;
};

/*!
  \class VaribaleTable
  \brief The list of all varibales and functions declared in the script
*/
class VariableTable : public StackList {

public:
    VariableTable() ;
    virtual ~VariableTable(){};
    Symbol * AddSymbol(char *name, OpValue::OType iType) ;
    Symbol * AddGlobalSymbol(char *name, OpValue::OType iType) ;
    void DelSymbol(char *name) ;
    void Destroy() ;
    void Init() ;
    Symbol * LookUp(const char *s)  ; // create and link value from Global Variable

    virtual int PushStack()  ; // Start local declaration
    virtual int PopStack() ; // delete last local declaration
} ;


#endif

/********************************************************************************/
/* 								END OF MTYPES.H	*/
/********************************************************************************/


