#ifndef __PARALLEL_INTERACTION_STORAGE_H
#define __PARALLEL_INTERACTION_STORAGE_H

//--- STL includes ---
#include <list>
#include <vector>

using std::list;
using std::vector;
using std::pair;

//--- Project includes ---
#include "Foundation/vec3.h"
#include "tml/comm/comm.h"

// forward declared class to avoid circular include
class AFieldSlave; 
class AParallelParticleArray;

/*!
  \class AParallelInteractionStorage
  \brief abstract base class for parallel interaction storage array
*/
class AParallelInteractionStorage
{
 protected:
  AParallelParticleArray* m_ppa;

 public:
  AParallelInteractionStorage(AParallelParticleArray* ppa){m_ppa=ppa;};
  virtual ~AParallelInteractionStorage(){};

  virtual void exchange()=0;
  virtual void rebuild()=0;
  virtual bool update()=0;
  //  virtual void tryInsert(const vector<int>&)=0;
  virtual bool isIn(const vector<int>&)=0;
  virtual void calcForces()=0;
  virtual void addExIG(AParallelInteractionStorage*){}; // do nothing
  virtual AFieldSlave* generateNewScalarFieldSlave(TML_Comm*,const string&,int,int,int,int)=0;
  virtual AFieldSlave* generateNewVectorFieldSlave(TML_Comm*,const string&,int,int,int,int)=0;
};



/*!
  \class TParallelInteractionStorage
  \brief templated abstract base class for parallel interaction storage array.
  Adds the vector of interactions and access functions to AParallelInteractionStorage
*/
template <typename I>
class TParallelInteractionStorage : public AParallelInteractionStorage
{
 public:
  typedef I TParallelInteractionStorage::interaction_type;

 protected:
  list<I> m_interactions;

 public:
  TParallelInteractionStorage(AParallelParticleArray* ppa):AParallelInteractionStorage(ppa){};
  virtual ~TParallelInteractionStorage(){};

  class InteractionIterator {
  public:
    typedef I Interaction;
    typedef typename list<I>::iterator Iterator;

    InteractionIterator(Iterator begin, Iterator end, AParallelParticleArray* ppa);

    bool hasNext();

    Interaction &next();

    int getNumRemaining();
    
  protected:
    bool isInner(const Iterator &it);

  private:
    int                    m_numRemaining;
    Iterator               m_it;
    Iterator               m_end;
    AParallelParticleArray *m_ppa;
  };

  InteractionIterator getInnerInteractionIterator();

  //!< access functions
  template <typename P> vector<pair<Vec3,P> > forAllInnerInteractionsGetWithPos(P (I::*rdf)() const);
  template <typename P> vector<pair<pair<Vec3,Vec3>,P> > forAllInnerInteractionsGetWithPos2(P (I::*rdf)() const);
  template <typename P> void forAllInnerInteractionsGet(P&,typename P::value_type (I::*rdf)() const);

  //!< access functions with tags
  template <typename P> vector<pair<Vec3,P> > forAllTaggedInnerInteractionsGetWithPos(P (I::*rdf)() const,int,int);
  template <typename P> void forAllTaggedInnerInteractionsGet(P&,typename P::value_type (I::*rdf)() const,int,int);

  //!< generate FieldSlave of correct type
  virtual AFieldSlave* generateNewScalarFieldSlave(TML_Comm*,const string&,int,int,int,int); // const ?
  virtual AFieldSlave* generateNewVectorFieldSlave(TML_Comm*,const string&,int,int,int,int);
};

#include "pis/pi_storage.hpp"

#endif //__PARALLEL_INTERACTION_STORAGE_H
