
#include "tml/comm/comm.h"

//--- I/O ---
#include <iostream>
using std::cout;
using std::endl;
using std::flush;

/*!
  Default constructor for TML_comm. Sets the MPI communicator
  to MPI_COMM_NULL;
*/
TML_Comm::TML_Comm()  
{
  m_comm=MPI_COMM_NULL;
}

/*!
  construct TML_Comm from MPI communicator

  \param comm the MPI communicator
*/
TML_Comm::TML_Comm(MPI_Comm comm)
{
  m_comm=comm;
}

TML_Comm& TML_Comm::operator=(const TML_Comm& rhs)
{
  MPI_Comm_dup(rhs.m_comm,&m_comm);

  return *this;
}

int TML_Comm::rank()  const
{
  int rank;
  
  if(m_comm!=MPI_COMM_NULL){
    MPI_Comm_rank(m_comm,&rank);
  } else {
    rank=MPI_UNDEFINED;
  }

  return rank;
}

int TML_Comm::size()
{
  int size;

  if(m_comm!=MPI_COMM_NULL){
    MPI_Comm_size(m_comm,&size);
  } else {
    size=MPI_UNDEFINED;
  }

  return size;
}

/*!
  Construct a new communicator containing the processes
  which are given as input

  \param ids the ranks of the processes which form the new communicator in the current communicator
  \todo error handling
*/
TML_Comm TML_Comm::include(const vector<int>& ids)
{
  TML_Comm newcomm;
  MPI_Group grp,ngrp;

  // extract group
  MPI_Comm_group(m_comm,&grp);
  // vector->array
  int nids=ids.size();
  int *ranks=new int[nids];
  for(int i=0;i<nids;i++){
    ranks[i]=ids[i];
  }
  // make new group
  int err=MPI_Group_incl(grp,nids,ranks,&ngrp);
  if(err!=MPI_SUCCESS){
    cout << "Error in TML_Comm::include group construction, rank " 
	 << rank() << " error " << err << endl << flush;
  }
  int gsize,grnk;
  MPI_Group_size(ngrp,&gsize);
  MPI_Group_rank(ngrp,&grnk);
  delete ranks;
  // construct new MPI communicator
  err=MPI_Comm_create(m_comm,ngrp,&(newcomm.m_comm));

  if(err!=MPI_SUCCESS){
    cout << "Error in TML_Comm::include communicator construction, rank " 
	 << rank() << " error " << err << endl << flush;
  }

  return newcomm;
}

/*!
  Construct a new communicator containing the processes from the current communicator 
  except to ones which are given as input

  \param ids the ranks of the processes which are excluded from  the new communicator 
  \todo error handling
*/
TML_Comm TML_Comm::exclude(const vector<int>& ids)
{
  TML_Comm newcomm;
  MPI_Group grp,ngrp;

  // extract group
  MPI_Comm_group(m_comm,&grp);
  // vector->array
  int nids=ids.size();
  int *ranks=new int[nids];
  for(int i=0;i<nids;i++){
    ranks[i]=ids[i];
  }
  // make new group
  MPI_Group_excl(grp,nids,ranks,&ngrp);
  delete ranks;
  // construct new MPI communicator
  MPI_Comm_create(m_comm,ngrp,&(newcomm.m_comm));

  return newcomm;
}

/*!
  Wait on a barrier. Wrapper for MPI_Barrier.
*/
void TML_Comm::barrier()
{
  MPI_Barrier(m_comm);
}
